/******************************************************************************
 * This file is part of dOpenCL.
 * 
 * dOpenCL is an implementation of the OpenCL application programming
 * interface for distributed systems. See <http://dopencl.uni-muenster.de/>
 * for more information.
 * 
 * Developed by: Research Group Parallel and Distributed Systems
 *               Department of Mathematics and Computer Science
 *               University of Muenster, Germany
 *               <http://pvs.uni-muenster.de/>
 * 
 * Copyright (C) 2013  Philipp Kegel <philipp.kegel@uni-muenster.de>
 *
 * dOpenCL is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * dOpenCL is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with dOpenCL. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Permission to use dOpenCL for scientific, non-commercial work is
 * granted under the terms of the dOpenCL Academic License provided
 * appropriate credit is given. See the dOpenCL Academic License for
 * more details.
 * 
 * You should have received a copy of the dOpenCL Academic License
 * along with dOpenCL. If not, see <http://dopencl.uni-muenster.de/>.
 ******************************************************************************/

/**
 * @file 	CommandQueue.cpp
 *
 * @date    2012-08-12
 * @author  Philipp Kegel
 */

#include <cstddef>
#include <ostream>
#include <vector>

#include <boost/shared_ptr.hpp>

#define __CL_ENABLE_EXCEPTIONS
#include <CL/cl.hpp>

#include <dcl/DCLTypes.h>
#include <dcl/Event.h>

#include <dcl/util/Logger.h>

#include "CommandQueue.h"
#include "Context.h"
#include "Event.h"
#include "Kernel.h"
#include "Memory.h"

#include "command/Command.h"
#include "command/ReadWriteCommand.h"
#include "command/SetCompleteCommand.h"

namespace dcld {

/**
 * @brief Callback used to execute enqueued commands.
 */
static void executeCommand(cl_event event, cl_int execution_status, void *user_data) {
    dcld::command::Command *command = static_cast<command::Command *>(user_data);
    assert(command != 0);
    command->execute(execution_status);
    delete command;
}

#ifdef PROFILE
/**
 * @brief Callback used to log event profiling info.
 */
static void logEventProfilingInfo(cl_event object_, cl_int execution_status, void *user_data) {
    std::string *str = static_cast<std::string *>(user_data);
    cl::Event event;
    cl_int err;
    cl_ulong queued, submit, start, end;

    /* assign event to C++ wrapper */
    event() = object_;
    err = ::clRetainEvent(object_);
    assert(err == CL_SUCCESS);

    /* query profiling info */
    try {
        event.getProfilingInfo(CL_PROFILING_COMMAND_QUEUED, &queued);
        event.getProfilingInfo(CL_PROFILING_COMMAND_SUBMIT, &submit);
        event.getProfilingInfo(CL_PROFILING_COMMAND_START, &start);
        event.getProfilingInfo(CL_PROFILING_COMMAND_END, &end);

        dcl::util::Logger << dcl::util::Debug
                << "Command completed (" << (str ? *str : "") << ")\n"
                << "\tqueued =" << queued << '\n'
                << "\tsubmit =" << submit << '\n'
                << "\tstart  =" << start << '\n'
                << "\tend    =" << end << '\n'
                << "\tdurance=" << (static_cast<double>(end - start) / 1000000000.0) << " sec"
                << std::endl;
    } catch (const cl::Error& err) {
        dcl::util::Logger << dcl::util::Error
                << "OpenCL error (ID=" << err.err() << "): " << err.what()
                << std::endl;
    }

    delete str;
}
#endif

/******************************************************************************/

CommandQueue::CommandQueue(const boost::shared_ptr<Context>& context,
        const cl::Device& device, cl_command_queue_properties properties) :
    _commandQueue(*context, device, properties), _context(context) { }

CommandQueue::~CommandQueue() { }

/**
 * @brief Returns the native command queue.
 */
CommandQueue::operator cl::CommandQueue() const {
    return _commandQueue;
}

/*
 * INFO: Synchronization
 *
 * Synchronization always is based on events in a wait list which act as
 * synchronization points (the only exception is the enqueueBarrier command).
 * For events associated with a command on the same compute node and for user
 * events, this requires no special measures; the events are just passed to the
 * implementation. But for events which are associated with a command on another
 * compute node (instances of dcld::EventListener), a synchronization across
 * compute nodes is required, if a memory object is associated with such an
 * event.
 * For each associated memory object, a data transfer from the memory object's
 * owner (the compute node that executes the command associated with the event)
 * is enqueued to this command queue. A native event is created for each data
 * transfer.
 * While the associated memory objects are removed from the event, the native
 * events associated with the data transfers are added to it. Thus,
 * synchronization is only enqueued once to the command queue that first uses
 * this event for synchronization.
 * The native events that have been added (associated) with the event during
 * synchronization are added to the native event wait list, such that the OpenCL
 * implementation synchronizes the memory objects. This is particularly
 * important if the event is used by another command queue on the same compute
 * node: while the data has already been transferred to the compute node, it is
 * only visible to the device that performed the synchronization. Thus, all
 * other devices have to synchronize in order to obtain the data.
 */
void CommandQueue::synchronize(
        const std::vector<boost::shared_ptr<dcl::Event> >& eventWaitList,
        VECTOR_CLASS<cl::Event>& nativeEventWaitList) {
    bool hasEventListeners = false;

    nativeEventWaitList.clear();

    if (eventWaitList.empty()) return;

    dcl::util::Logger << dcl::util::Debug
            << "Synchronizing event wait list with " << eventWaitList.size()
            << " event(s)" << std::endl;

    for (std::vector<boost::shared_ptr<dcl::Event> >::const_iterator it =
            eventWaitList.begin(); it != eventWaitList.end(); ++it) {
        boost::shared_ptr<EventListener> eventListener =
                boost::dynamic_pointer_cast<EventListener>(*it);
        if (eventListener) {
            /* event is an event listener */
            VECTOR_CLASS<cl::Event> synchronizeEvents;
            eventListener->synchronize(_commandQueue, synchronizeEvents);
            /* FIXME Synchronize memory objects associated with multiple event in wait list only once
             * Different event listeners may be associated with the same memory object
             * However, a memory object must only be synchronized once.
             * Synchronizing with multiple events associated with the same memory
             * object may be considered undefined behavior. */
            nativeEventWaitList.insert(nativeEventWaitList.end(),
                    synchronizeEvents.begin(), synchronizeEvents.end());

            hasEventListeners = true;
        } else {
            boost::shared_ptr<Event> eventImpl = boost::dynamic_pointer_cast<Event>(*it);
            if (!eventImpl) throw cl::Error(CL_INVALID_EVENT_WAIT_LIST);
            nativeEventWaitList.push_back(*eventImpl);
        }
    }

    if (hasEventListeners) {
        /* Flush the command queue to ensure instant execution of the acquire
         * operation */
        flush();
    }
}

void CommandQueue::flush() {
    _commandQueue.flush();
}

/* clFinish is a synchronization point */
void CommandQueue::finish() {
    _commandQueue.finish();

    /* TODO Implement synchronization for clFinish
     * It is not clear from the specification what this means; maybe it acts as
     * an active release, i.e., the memory objects modified by this command queue
     * are replicated to all other compute nodes. */
}

void CommandQueue::enqueueCopyBuffer(
        const boost::shared_ptr<dcl::Buffer>& src,
        const boost::shared_ptr<dcl::Buffer>& dst,
        size_t srcOffset,
        size_t dstOffset,
        size_t size,
        const std::vector<boost::shared_ptr<dcl::Event> > * eventWaitList,
        dcl::dcl_object_id commandId,
        boost::shared_ptr<dcl::Event> *event) {
    boost::shared_ptr<Buffer> srcImpl = boost::dynamic_pointer_cast<Buffer>(src);
    boost::shared_ptr<Buffer> dstImpl = boost::dynamic_pointer_cast<Buffer>(dst);
    VECTOR_CLASS<cl::Event> nativeEventWaitList;
    cl::Event copyBuffer;

    /* Obtain wait list of native events */
    if (eventWaitList) {
        synchronize(*eventWaitList, nativeEventWaitList);
    }

    /* Enqueue copy buffer
     * Only create event if requested by caller */
    _commandQueue.enqueueCopyBuffer(
            *srcImpl, *dstImpl,
            srcOffset, dstOffset, size,
            &nativeEventWaitList, event ? &copyBuffer : 0);
#ifdef FORCE_FLUSH
    _commandQueue.flush();
#endif

    if (event) { // an event should be associated with this command
        try {
            /* Wrap event */
            event->reset(new SimpleEvent(commandId, _context, dstImpl, copyBuffer));
        } catch (const std::bad_alloc&) {
            throw cl::Error(CL_OUT_OF_RESOURCES);
        }
    }
}

void CommandQueue::enqueueReadBuffer(
        const boost::shared_ptr<dcl::Buffer>& buffer,
        bool blockingRead,
        size_t offset,
        size_t size,
        const std::vector<boost::shared_ptr<dcl::Event> > *eventWaitList,
        dcl::dcl_object_id commandId,
        boost::shared_ptr<dcl::Event> *event) {
    boost::shared_ptr<Buffer> bufferImpl = boost::dynamic_pointer_cast<Buffer>(buffer);
    void *ptr;
    VECTOR_CLASS<cl::Event> nativeEventWaitList;
    cl::Event mapData, unmapData;
    cl::UserEvent dataSending(*_context);

    if (!bufferImpl) throw cl::Error(CL_INVALID_MEM_OBJECT);

    /* Obtain wait list of native events */
    if (eventWaitList) {
        synchronize(*eventWaitList, nativeEventWaitList);
    }

    /* Enqueue map buffer (implicit download) */
    ptr = _commandQueue.enqueueMapBuffer(
            *bufferImpl,
            CL_FALSE,    // non-blocking map
            CL_MAP_READ, // map for reading
            offset, size,
            &nativeEventWaitList, &mapData);
#ifdef PROFILE
    mapData.setCallback(CL_COMPLETE, &logEventProfilingInfo, new std::string("map buffer for reading"));
#endif
    /* Enqueue unmap buffer
     * Only create a native event, if an event should be associated with this
     * read buffer command */
    nativeEventWaitList.assign(1, dataSending);
    _commandQueue.enqueueUnmapMemObject(
            *bufferImpl,
            ptr,
            &nativeEventWaitList, event ? &unmapData : 0);
#ifdef FORCE_FLUSH
    _commandQueue.flush();
#else
    if (blockingRead) {
        _commandQueue.flush();
    }
#endif

    try {
        /* Schedule data sending
         * A 'command submitted' message will be sent to the host in order to
         * start data receipt. */
        mapData.setCallback(CL_COMPLETE, &executeCommand,
                new command::ReadBufferCommand(_context->host(), commandId, size, ptr, dataSending));
        /* The read buffer command is finished on the host such that no 'command
         * complete' message must be sent by the compute node. */

        if (event) { // an event should be associated with this command
            /* Wrap event */
            /* WARNING: No callback must be registered for any native event of
             * the ReadMemoryEvent object, that access the object. As the host
             * finishes the read buffer command, the application may delete the
             * ReadMemoryEvent object, while or even *before* the callbacks are
             * processed. Thus, a callback that accesses the ReadMemoryEvent
             * object may raise a SIGSEGV. */
            event->reset(new ReadMemoryEvent(commandId, _context, mapData, unmapData));
        }
    } catch (const std::bad_alloc&) {
        throw cl::Error(CL_OUT_OF_RESOURCES);
    }
}

void CommandQueue::enqueueWriteBuffer(
        const boost::shared_ptr<dcl::Buffer>& buffer,
        bool blockingWrite,
        size_t offset,
        size_t size,
        const std::vector<boost::shared_ptr<dcl::Event> > *eventWaitList,
        dcl::dcl_object_id commandId,
        boost::shared_ptr<dcl::Event> *event) {
    boost::shared_ptr<Buffer> bufferImpl = boost::dynamic_pointer_cast<Buffer>(buffer);
    void *ptr;
    VECTOR_CLASS<cl::Event> nativeEventWaitList;
    cl::Event mapData, unmapData;
    cl::UserEvent dataReceipt(*_context);

    if (!bufferImpl) throw cl::Error(CL_INVALID_MEM_OBJECT);

    /* Obtain wait list of native events */
    if (eventWaitList) {
        synchronize(*eventWaitList, nativeEventWaitList);
    }

    /* Enqueue map buffer */
    /* !!! WARNING !!!
     * NVIDIA only: the reference count of events in wait list is *not*
     * increased by clEnqueueMapBuffer. This may be a bug, if the event is not
     * retained by other means than its reference count.
     * !!!!!!!!!!!!!!! */
    ptr = _commandQueue.enqueueMapBuffer(
            *bufferImpl,
            CL_FALSE,     // non-blocking map
            CL_MAP_WRITE, // map for writing
            offset, size,
            &nativeEventWaitList, &mapData);
    /* Enqueue unmap buffer (implicit upload) */
    nativeEventWaitList.assign(1, dataReceipt);
    /* !!! WARNING !!!
     * NVIDIA only: the reference count of dataReceipt is *not* increased by
     * clEnqueueUnmapMemObject. This may be a bug, if the event is not retained
     * by other means than its reference count.
     * !!!!!!!!!!!!!!! */
//  std::cout << "!!! before unmap dataReceipt refCount=" << dataReceipt.getInfo<CL_EVENT_REFERENCE_COUNT>() << std::endl;
//  ::clRetainEvent(dataReceipt()); // tentative fix for NVIDIA
    _commandQueue.enqueueUnmapMemObject(
            *bufferImpl,
            ptr,
            &nativeEventWaitList, &unmapData);
//  std::cout << "!!! after unmap dataReceipt refCount=" << dataReceipt.getInfo<CL_EVENT_REFERENCE_COUNT>() << std::endl;
#ifdef PROFILE
    unmapData.setCallback(CL_COMPLETE, &logEventProfilingInfo, new std::string("unmap buffer after writing"));
#endif
#ifdef FORCE_FLUSH
    _commandQueue.flush();
#else
    if (blockingWrite) {
        _commandQueue.flush();
    }
#endif

    try {
        /* TODO Sketch of a possible solution for a command-based implementation
         * Complex commands, i.e., commands which include interaction with the
         * host, should be implemented by a class and must be independent of an
         * event. */
#if 0
        /* The write buffer command notifies the host about its submission and
         * completion, regardless an event is associated with it or not. */
        boost::shared_ptr<command::WriteBuffer> writeBuffer(
                boost::make_shared<command::WriteBufferCommand>(_context->host(),
                        commandId, size, ptr, mapBuffer, dataReceipt, unmapData));

        if (event) { // an event should be associated with this command
            /* An event wrapper retains its associated command for further
             * commands from the host. Moreover, it broadcast command execution
             * status changes to other compute nodes in the command's context,
             * while the command only communicates with the host. */
            event->reset(new WriteMemoryEvent(commandId, _context, bufferImpl, writeBuffer));
        }
#else
        /* Schedule data receipt
         * A 'command submitted' message will be sent to the host. */
        mapData.setCallback(CL_COMPLETE, &executeCommand,
                new command::WriteBufferCommand(_context->host(), commandId, size, ptr, dataReceipt));
        /* Schedule completion message for host
         * A 'command complete' message is sent to the host. It is required for
         * a blocking write and to properly delete the associated command in the
         * ICD. In particular, this message must also be sent if no event is
         * associated with this command. */
        unmapData.setCallback(CL_COMPLETE, &executeCommand,
                new command::SetCompleteCommand(_context->host(), commandId, cl::UserEvent(*_context)));

        if (event) { // an event should be associated with this command
            /* Wrap event */
            /* This event must only broadcast its status on other compute nodes
             * but not to the host, as a 'command complete' message is already
             * sent to the host by the callback that has been set for the native
             * event unmapData. */
            /* FIXME Avoid race condition in write buffer command
             * Callbacks are registered for unmapData to 1) notify the host
             * about command completion, and 2) to synchronize the memory object
             * associated with WriteMemoryEvent.
             * As the execution order of callback is unspecified, the host may
             * be notified about command completion (callback 1) before
             * callback 2 is executed. If the application and the network
             * respond quickly to callback 1 in order to delete the
             * WriteMemoryEvent object, it may be deleted *before* callback 2 is
             * processed. In this case callback 2 tries to access the deleted
             * WriteMemoryEvent object, such that a SIGSEGV will be raised. */
            event->reset(new WriteMemoryEvent(commandId, _context, bufferImpl, mapData, unmapData));
        }
#endif
    } catch (const std::bad_alloc&) {
        throw cl::Error(CL_OUT_OF_RESOURCES);
    }
}

void CommandQueue::enqueueMapBuffer(
        const boost::shared_ptr<dcl::Buffer>& buffer,
        bool blockingMap,
        cl_map_flags mapFlags,
        size_t offset,
        size_t size,
        const std::vector<boost::shared_ptr<dcl::Event> > *eventWaitList,
        dcl::dcl_object_id commandId,
        boost::shared_ptr<dcl::Event> *event) {
    boost::shared_ptr<Buffer> bufferImpl = boost::dynamic_pointer_cast<Buffer>(buffer);
    VECTOR_CLASS<cl::Event> nativeEventWaitList;

    if (!bufferImpl) throw cl::Error(CL_INVALID_MEM_OBJECT);

    /* Obtain wait list of native events */
    if (eventWaitList) {
        synchronize(*eventWaitList, nativeEventWaitList);
    }

    /* TODO Implement CommandQueue::enqueueMapBuffer */
    assert(!"CommandQueue::enqueueMapBuffer not implemented");
}

void CommandQueue::enqueueUnmapMemObject(
        const boost::shared_ptr<dcl::Memory>& memory,
        const std::vector<boost::shared_ptr<dcl::Event> > *eventWaitList,
        dcl::dcl_object_id commandId,
        boost::shared_ptr<dcl::Event> *event) {
    /* TODO Implement CommandQueue::enqueueUnmapMemObject */
    assert(!"CommandQueue::enqueueUnmapMemObject not implemented");
}

void CommandQueue::enqueueNDRangeKernel(
        const boost::shared_ptr<dcl::Kernel>& kernel,
        const cl::NDRange& offset,
        const cl::NDRange& global,
        const cl::NDRange& local,
        const std::vector<boost::shared_ptr<dcl::Event> > *eventWaitList,
        dcl::dcl_object_id commandId,
        boost::shared_ptr<dcl::Event> *event) {
    boost::shared_ptr<Kernel> kernelImpl = boost::dynamic_pointer_cast<Kernel>(kernel);
    VECTOR_CLASS<cl::Event> nativeEventWaitList;
    cl::Event ndRangeKernel;

    if (!kernelImpl) throw cl::Error(CL_INVALID_KERNEL);

    /* Obtain wait list of native events */
    if (eventWaitList) {
        synchronize(*eventWaitList, nativeEventWaitList);
    }

    /* Enqueue ND range kernel
     * Only create event if requested by caller */
    _commandQueue.enqueueNDRangeKernel(
            *kernelImpl,
            offset, global, local,
            &nativeEventWaitList, event ? &ndRangeKernel : 0);
#ifdef FORCE_FLUSH
    _commandQueue.flush();
#endif

    if (event) { // an event should be associated with this command
        try {
            /* Wrap event */
            event->reset(new SimpleEvent(commandId, _context,
                    kernelImpl->writeMemoryObjects(), ndRangeKernel));
        } catch (const std::bad_alloc&) {
            throw cl::Error(CL_OUT_OF_RESOURCES);
        }
    }
}

/*
 * The semantics of enqueueMarker is unclear:
 * What is clEnqueueMarkerWithWaitList good for, if no event is returned?
 */
void CommandQueue::enqueueMarker(
        const std::vector<boost::shared_ptr<dcl::Event> > *eventWaitList,
        dcl::dcl_object_id commandId, boost::shared_ptr<dcl::Event> *event) {
    VECTOR_CLASS<cl::Event> nativeEventWaitList;
    cl::Event marker;

    /* Obtain wait list of native events */
    if (eventWaitList) {
        synchronize(*eventWaitList, nativeEventWaitList);
    }

#if defined(CL_VERSION_1_2)
    /* Use native enqueueMarkerWithWaitList */
    _commandQueue.enqueueMarkerWithWaitList(
            &nativeEventWaitList, event ? &marker : 0);
#else // #if defined(CL_VERSION_1_2)
    /* Use pre-OpenCL 1.2 APIs to implement CommandQueue::enqueueMarker with wait list */
    /* clEnqueueMarker (deprecated) does not support an event list. */
    if (eventWaitList) {
        assert(!"clEnqueueMarkerWithWaitList not implemented");
#if 0
        /* TODO Use count-down latch to implement clEnqueueMarkerWithWaitList
         * Each event in wait list decrements the latch once. When the latch is
         * decremented to 0 (triggered), it sets the marker to complete. The
         * event which triggered the latch also deletes it.
         * To ensure in-order execution (for in-order command-queues), a marker
         * is enqueued and added to the event wait list. */
        /* WARNING: The marker event is implemented by a user event. As such it
         * cannot be used in the event list of clEnqueueWaitForEvents. Hence,
         * clEnqueueMarkerWithWaitList cannot be fully implemented using the
         * OpenCL 1.1 APIs. */
        cl::UserEvent complete;

        if ((_commandQueue.getInfo<CL_QUEUE_PROPERTIES>() & CL_QUEUE_OUT_OF_ORDER_EXEC_MODE_ENABLE)) {
            cl::Event execution;
            _commandQueue.enqueueMarker(&execution);
            nativeEventWaitList.push_back(execution);
        }

        CountDownLatch *cdl = new CountDownLatch(nativeEventWaitList.size(),
                boost::bind(&cl::UserEvent::setStatus, complete, CL_COMPLETE));
        for (VECTOR_CLASS<cl::Event>::const_iterator i = nativeEventWaitList.begin();
                i != nativeEventWaitList.end(); ++i) {
            i->setCallback(CL_COMPLETE, &countDown, cdl);
        }

        marker = complete; // downcast marker to cl::Event
#endif
    } else {
        _commandQueue.enqueueMarker(&marker);
    }
#endif // #if defined(CL_VERSION_1_2)

    if (event) { // an event should be associated with this command
        try {
            /* Wrap event */
            event->reset(new SimpleEvent(commandId, _context, marker));
        } catch (const std::bad_alloc&) {
            throw cl::Error(CL_OUT_OF_RESOURCES);
        }
    }
}

#if defined(CL_USE_DEPRECATED_OPENCL_1_1_APIS) || (defined(CL_VERSION_1_1) && !defined(CL_VERSION_1_2))
void CommandQueue::enqueueWaitForEvents(
        const std::vector<boost::shared_ptr<dcl::Event> >& eventList) {
    VECTOR_CLASS<cl::Event> nativeEventList;

	assert(!eventList.empty()); // event list must no be empty

    dcl::util::Logger << dcl::util::Debug
            << "Synchronizing event list with " << eventList.size() << " event(s)"
            << std::endl;

    /* Obtain native event list
     * Unlike other enqueued commands, wait for events throws CL_INVALID_EVENT
     * rather than CL_INVALID_EVENT_WAIT_LIST if the event list contains an
     * invalid event. */
    for (std::vector<boost::shared_ptr<dcl::Event> >::const_iterator it = eventList.begin();
            it != eventList.end(); ++it) {
        boost::shared_ptr<EventListener> eventListener = boost::dynamic_pointer_cast<EventListener>(*it);
        if (eventListener) {
            /* event is an event listener */
            VECTOR_CLASS<cl::Event> synchronizeEvents;
            eventListener->synchronize(_commandQueue, synchronizeEvents);
            nativeEventList.insert(nativeEventList.end(),
                    synchronizeEvents.begin(), synchronizeEvents.end());
        } else {
            /* TODO Create Event::synchronize method */
            boost::shared_ptr<Event> eventImpl = boost::dynamic_pointer_cast<Event>(*it);
            if (!eventImpl) throw cl::Error(CL_INVALID_EVENT);
            nativeEventList.push_back(*eventImpl);
        }
    }

    _commandQueue.enqueueWaitForEvents(nativeEventList);
}
#endif // #if defined(CL_USE_DEPRECATED_OPENCL_1_1_APIS)

/* enqueueBarrier is a synchronization point */
void CommandQueue::enqueueBarrier(
        const std::vector<boost::shared_ptr<dcl::Event> > *eventWaitList,
        dcl::dcl_object_id commandId, boost::shared_ptr<dcl::Event> *event) {
    VECTOR_CLASS<cl::Event> nativeEventWaitList;
    cl::Event barrier;

    /* Obtain wait list of native events */
    if (eventWaitList) {
	    synchronize(*eventWaitList, nativeEventWaitList);
	}

#if defined(CL_VERSION_1_2)
    /* Use native enqueueBarrierWithWaitList */
    _commandQueue.enqueueBarrierWithWaitList(
            &nativeEventWaitList, event ? &barrier : 0);
#else // #if defined(CL_VERSION_1_2)
    /* Use pre-OpenCL 1.2 APIs to implement CommandQueue::enqueueBarrier with wait list */
    /* clEnqueueBarrier (deprecated) does not support an event wait list nor
     * does is return an event. */
    if (eventWaitList) {
        assert(!"clEnqueueBarrierWithWaitList not implemented");
#if 0
        try {
            /* WARNING: According to the OpenCL 1.1 specification the event list
             * of clEnqueueWaitForEvents must not contain user events. Moreover,
             * unlike clEnqueueBarrier, clEnqueueWaitForEvents is not explicitly
             * defined as a synchronization point. Hence,
             * clEnqueueBarrierWithEventList cannot be fully implemented using
             * the OpenCL 1.1 APIs. */
            _commandQueue.enqueueWaitForEvents(nativeEventWaitList);
        } catch (const cl::Error& err) {
            /* cl::CommandQueue::enqueueWaitForEvents throws CL_INVALID_EVENT if
             * an event in the wait list is not a valid event, but
             * enqueueBarrier (with wait list) must throw
             * CL_INVALID_EVENT_WAIT_LIST in this case. */
            if (err.err() == CL_INVALID_EVENT) {
                throw cl::Error(CL_INVALID_EVENT_WAIT_LIST);
            } else {
                throw err;
            }
        }
#endif
    } else {
        _commandQueue.enqueueBarrier();
    }
    if (event) {
        _commandQueue.enqueueMarker(&barrier);
        /* Ensure that the marker has finished execution before any future
         * command is executed. This is implied for in-order command-queues but
         * requires additional synchronization using clEnqueueWaitForEvents for
         * out-of-order command-queues. */
        _commandQueue.enqueueWaitForEvents(VECTOR_CLASS<cl::Event>(1, barrier));
        /* TODO Lock command-queue when enqueuing clEnqueueBarrierWithWaitList
         * No other command must be enqueued concurrently, such that
         * clEnqeueueWaitForEvents is executed right after clEnqueueMarker. */
    }
#endif // #if defined(CL_VERSION_1_2)

    /* TODO Implement synchronization for clEnqueueBarrierWithWaitList
     * It is not clear from the specification what this means; maybe it acts as
     * an active release, i.e. changes to the memory objects by this
     * command-queue are replicated to all other compute nodes. */

    if (event) { // an event should be associated with this command
        /* We assume that an event returned by clEnqueueBarrierWithWaitList
         * cannot be used to synchronize changes to a memory object across
         * command queues, as the specification explicitly stipulates to use
         * the event for the command that modified the memory objects for
         * synchronization (see Appendix A.1). */
        try {
            /* Wrap event */
            event->reset(new SimpleEvent(commandId, _context, barrier));
        } catch (const std::bad_alloc&) {
            throw cl::Error(CL_OUT_OF_RESOURCES);
        }
    }
}

const boost::shared_ptr<Context>& CommandQueue::context() const {
    return _context;
}

} /* namespace dcld */
