/******************************************************************************
 * This file is part of dOpenCL.
 * 
 * dOpenCL is an implementation of the OpenCL application programming
 * interface for distributed systems. See <http://dopencl.uni-muenster.de/>
 * for more information.
 * 
 * Developed by: Research Group Parallel and Distributed Systems
 *               Department of Mathematics and Computer Science
 *               University of Muenster, Germany
 *               <http://pvs.uni-muenster.de/>
 * 
 * Copyright (C) 2013  Philipp Kegel <philipp.kegel@uni-muenster.de>
 *
 * dOpenCL is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * dOpenCL is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with dOpenCL. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Permission to use dOpenCL for scientific, non-commercial work is
 * granted under the terms of the dOpenCL Academic License provided
 * appropriate credit is given. See the dOpenCL Academic License for
 * more details.
 * 
 * You should have received a copy of the dOpenCL Academic License
 * along with dOpenCL. If not, see <http://dopencl.uni-muenster.de/>.
 ******************************************************************************/

/*
 * Context.cpp
 *
 *  Created on: 06.03.2012
 *      Author: Philipp Kegel
 */

#include <cassert>
#include <cstddef>
#include <ostream>
#include <string>
#include <vector>

#define __CL_ENABLE_EXCEPTIONS
#include <CL/cl.hpp>

#include <dcl/ComputeNode.h>
#include <dcl/DCLTypes.h>
#include <dcl/Host.h>
#include <dcl/Remote.h>

#include <dcl/comm/ContextMessage.h>

#include <dcl/util/Logger.h>

#include "Context.h"

namespace dcld {

/**
 * @brief Callback for completion of program build
 */
static void onContextError(const char *errinfo, const void *private_info,
		size_t cb, void *user_data) {
	dcld::Context *context = static_cast<dcld::Context *>(user_data);
	assert(context != 0);
	context->onError(std::string(errinfo), private_info, cb);
}

/**
 * @brief Callback for printf
 */
static void onContextPrintf(cl_context object, cl_uint printf_data_len,
        char *printf_data_ptr, void *user_data) {
    dcld::Context *context = static_cast<dcld::Context *>(user_data);
    assert(context != 0);
    context->onPrintf(std::string(printf_data_ptr,
            printf_data_len == 0 ? std::string::npos : printf_data_len));
}

/******************************************************************************/

Context::Context(dcl::dcl_object_id contextId, dcl::Host& host,
		const std::vector<dcl::ComputeNode *>& computeNodes,
		const cl::Platform& platform,
		const std::vector<cl::Device>& devices) :
	dcl::Remote(contextId), _host(host), _computeNodes(computeNodes) {
    cl_context_properties properties[3];

//	if (computeNodes.empty()) { throw cl::Error(CL_INVALID_VALUE); }
	if (devices.empty()) { throw cl::Error(CL_INVALID_VALUE); }

	/* TODO Remove self from list of compute nodes */

	/*
	 * Initialize context properties
	 */
	properties[0] = CL_CONTEXT_PLATFORM;
	properties[1] = (cl_context_properties) platform();
	properties[2] = 0; /* end of list */

	_context = cl::Context(devices, properties, &onContextError, this);
	_ioCommandQueue = cl::CommandQueue(_context, devices.front());

	/* TODO Register printf callback if platform supports OpenCL 1.2 */
	/* OpenCL 1.2 not supported by C++ binding */
//	clSetPrintfCallback(_context(), &onContextPrintf, this);
}

Context::~Context() { }

Context::operator cl::Context() const {
	return _context;
}

dcl::Host& Context::host() const {
	return _host;
}

const cl::CommandQueue& Context::ioCommandQueue() const {
    return _ioCommandQueue;
}

const std::vector<dcl::ComputeNode *>& Context::computeNodes() const {
	return _computeNodes;
}

void Context::onError(const std::string& errorInfo, const void *private_info,
		size_t cb) {
    dcl::comm::ContextErrorMessage message(_id, errorInfo,
            dcl::Binary(cb, private_info));
    _host.sendMessage(message);

    dcl::util::Logger << dcl::util::Debug
            << "Context error (ID=" << _id
            << ", info=" << errorInfo
            << ')' << std::endl;
}

void Context::onPrintf(const std::string& printfData) {
    dcl::comm::ContextPrintfMessage message(_id, printfData);
    _host.sendMessage(message);

    dcl::util::Logger << dcl::util::Debug
            << "Context printf (ID=" << _id
            << ", data=" << printfData
            << ')' << std::endl;
}

} /* namespace dcld */
