/******************************************************************************
 * This file is part of dOpenCL.
 * 
 * dOpenCL is an implementation of the OpenCL application programming
 * interface for distributed systems. See <http://dopencl.uni-muenster.de/>
 * for more information.
 * 
 * Developed by: Research Group Parallel and Distributed Systems
 *               Department of Mathematics and Computer Science
 *               University of Muenster, Germany
 *               <http://pvs.uni-muenster.de/>
 * 
 * Copyright (C) 2013  Philipp Kegel <philipp.kegel@uni-muenster.de>
 *
 * dOpenCL is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * dOpenCL is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with dOpenCL. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Permission to use dOpenCL for scientific, non-commercial work is
 * granted under the terms of the dOpenCL Academic License provided
 * appropriate credit is given. See the dOpenCL Academic License for
 * more details.
 * 
 * You should have received a copy of the dOpenCL Academic License
 * along with dOpenCL. If not, see <http://dopencl.uni-muenster.de/>.
 ******************************************************************************/

/*
 * DeviceManager.cpp
 *
 *  Created on: 03.02.2011
 *      Author: Philipp Kegel
 */

#include <ostream>
#include <sstream>
#include <stdexcept>
#include <string>
#include <utility>
#include <vector>

#define __CL_ENABLE_EXCEPTIONS
#include <CL/cl.hpp>

#include <dcl/DCLTypes.h>

#include <dcl/util/Logger.h>

#include "DeviceManager.h"

namespace dcld {

/**
 * @brief Extracts version information from an OpenCL version string.
 * The OpenCL version string is given as
 * OpenCL<space><miajor.major version<space><platform-specific information>
 *
 * @param[in]  version  OpenCL version string
 * @param[out] major    major version number
 * @param[out] minor    minor version number
 * @param[out] info     platform-specific information
 */
static void getOpenCLVersion(
		const std::string& version,
		unsigned int& major,
		unsigned int& minor,
		std::string& info) {
	std::istringstream iss(version);
	std::string token;

	/*
	 * Validate prefix
	 */
	iss >> token; // read first token
	if (token.compare("OpenCL")) {
		/* version string does not start with 'OpenCL' */
		throw std::invalid_argument("Invalid OpenCL version string");
	}

	/*
	 * Extract version numbers
	 */
	if (iss.eof() || iss.get() != ' ') { // skip space
		throw std::invalid_argument("No version number found");
	}
	major = 0;
	minor = 0;
	iss >> major; // read major version number
	iss.ignore(); // skip period
	iss >> minor; // read minor version number

	if (iss.good()) {
		/*
		 * Extract platform-specific information
		 */
		if (iss.get() != ' ') { // skip space
			throw std::invalid_argument("Invalid version number");
		}
		std::getline(iss, info); // read platform-specific information (all remaining characters)
	}
}

/******************************************************************************/

DeviceManager::DeviceManager(const std::string *platformName) {
	VECTOR_CLASS<cl::Platform> platforms;
	VECTOR_CLASS<cl::Platform>::const_iterator platform;

	/* The number of platform may be zero without throwing an error.
	 * If an ICD loader is used, CL_PLATFORM_NOT_FOUND_KHR will be thrown. */
	cl::Platform::get(&platforms);

	/*
	 * Select platform
	 */
	platform = platforms.begin();
	while (platform != platforms.end()) {
		std::string version;
		unsigned int major, minor;
		std::string info;

		/*
		 * Obtain platform version
		 * dOpenCL daemon requires OpenCL version 1.1
		 */
		platform->getInfo(CL_PLATFORM_VERSION, &version);
		getOpenCLVersion(version, major, minor, info);

		if (platformName) {
			std::string name;

			/*
			 * select platform by name
			 */
			platform->getInfo(CL_PLATFORM_NAME, &name);
			if ((name.find(*platformName) != std::string::npos)) {
				if (major < 1 || (major == 1 && minor < 1)) {
					dcl::util::Logger << dcl::util::Warning
							<< "Platform '" << name << "' (version "
							<< version << ") does not support OpenCL 1.1 or higher."
							<< std::endl;
					platform = platforms.end();
				}
				break;
			}
		} else {
			/*
			 * select first appropriate platform
			 */
			if ((major == 1 && minor >= 1) || major > 1) {
				break;
			}
		}

		++platform;
	}

	if (platform == platforms.end()) {
		if (!platforms.empty()) {
			dcl::util::Logger << dcl::util::Error
					<< "No OpenCL 1.1 compliant platform found." << std::endl;
		}
		throw cl::Error(CL_PLATFORM_NOT_FOUND_KHR);
	}

	_platform = *platform;

	/*
	 * Initialize device list
	 */
	_platform.getDevices(CL_DEVICE_TYPE_ALL, &_devices);

	dcl::util::Logger << dcl::util::Info
			<< "Using platform '" << _platform.getInfo<CL_PLATFORM_NAME>() << "'\n"
			<< "\tfound " << _devices.size() << " device(s):\n";
	for (VECTOR_CLASS<cl::Device>::const_iterator i = _devices.begin();
			i != _devices.end(); ++i) {
		dcl::util::Logger << dcl::util::Info
			<< "\t\t" << i->getInfo<CL_DEVICE_NAME>() << '\n';
	}
	dcl::util::Logger.flush();
}

void DeviceManager::getDeviceIDs(
		cl_device_type deviceTypes,
		std::vector<dcl::dcl_object_id>& deviceIds) const {
	deviceIds.clear();
	for (dcl::dcl_object_id id = 0; id < _devices.size(); ++id) {
		cl_device_type deviceType;

		_devices[id].getInfo(CL_DEVICE_TYPE, &deviceType);
		if (deviceType == (deviceTypes & deviceType)) {
			/* Add device ID to list */
			deviceIds.push_back(id);
		}
	}
}

const cl::Platform& DeviceManager::getPlatform() const {
    return _platform;
}

const cl::Device& DeviceManager::getDevice(
		dcl::dcl_object_id deviceId) const {
	try {
		return _devices.at(deviceId);
	} catch (std::out_of_range&) {
		throw cl::Error(CL_INVALID_DEVICE);
	}
}

void DeviceManager::getDevices(
		const std::vector<dcl::dcl_object_id>& deviceIDs,
		std::vector<cl::Device>& devices) const {
	devices.clear();
	for (std::vector<dcl::dcl_object_id>::const_iterator i = deviceIDs.begin();
	        i != deviceIDs.end(); ++i) {
		devices.push_back(getDevice(*i));
	}
}

} /* namespace dcld */
