/******************************************************************************
 * This file is part of dOpenCL.
 * 
 * dOpenCL is an implementation of the OpenCL application programming
 * interface for distributed systems. See <http://dopencl.uni-muenster.de/>
 * for more information.
 * 
 * Developed by: Research Group Parallel and Distributed Systems
 *               Department of Mathematics and Computer Science
 *               University of Muenster, Germany
 *               <http://pvs.uni-muenster.de/>
 * 
 * Copyright (C) 2013  Philipp Kegel <philipp.kegel@uni-muenster.de>
 *
 * dOpenCL is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * dOpenCL is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with dOpenCL. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Permission to use dOpenCL for scientific, non-commercial work is
 * granted under the terms of the dOpenCL Academic License provided
 * appropriate credit is given. See the dOpenCL Academic License for
 * more details.
 * 
 * You should have received a copy of the dOpenCL Academic License
 * along with dOpenCL. If not, see <http://dopencl.uni-muenster.de/>.
 ******************************************************************************/

/*
 * DeviceManager.h
 *
 *  Created on: 03.02.2011
 *      Author: Philipp Kegel
 */

#ifndef DEVICEMANAGER_H_
#define DEVICEMANAGER_H_

#include <string>
#include <vector>

#define __CL_ENABLE_EXCEPTIONS
#include <CL/cl.hpp>

#include <dcl/DCLTypes.h>

namespace dcld {

/**
 * @brief A device manager lists all devices provided by a specified OpenCL platform and assigns IDs to these devices.
 */
class DeviceManager {
public:
	/**
	 * @brief Creates a device manager.
	 *
	 * @param[in]  platformName name of the platform which devices should be managed
	 *             If platformName is NULL, the first platform available will be used.
	 */
	DeviceManager(
			const std::string * /* platformName */);

	/**
	 * @brief Creates a list of device IDs.
	 *
	 * @param device_types	types of devices, which IDs should be returned
	 * @param device_ids	the device IDs
	 */
	void getDeviceIDs(
			cl_device_type,						/* device_types */
			std::vector<dcl::dcl_object_id>&	/* device_ids */) const;

    const cl::Platform& getPlatform() const;

	/**
	 * @brief Returns the OpenCL device that corresponds to a given ID.
	 *
	 * @param id	device ID
	 * @return the device
	 */
	const cl::Device& getDevice(
			dcl::dcl_object_id	/* device_id */) const;

	/**
	 * @brief Returns OpenCL devices that correspond to a given list of IDs.
	 *
	 * @param deviceIDs			device IDs
	 * @param devices			a vector of OpenCL devices found
	 */
	void getDevices(
			const std::vector<dcl::dcl_object_id>&,	/* deviceIDs */
			VECTOR_CLASS<cl::Device>& 				/* devices */) const;

private:
	cl::Platform _platform; /**< selected platform; default is first platform */
	VECTOR_CLASS<cl::Device> _devices; /**< device list */
};

} /* namespace dcld */

#endif /* DEVICEMANAGER_H_ */
