/******************************************************************************
 * This file is part of dOpenCL.
 * 
 * dOpenCL is an implementation of the OpenCL application programming
 * interface for distributed systems. See <http://dopencl.uni-muenster.de/>
 * for more information.
 * 
 * Developed by: Research Group Parallel and Distributed Systems
 *               Department of Mathematics and Computer Science
 *               University of Muenster, Germany
 *               <http://pvs.uni-muenster.de/>
 * 
 * Copyright (C) 2013  Philipp Kegel <philipp.kegel@uni-muenster.de>
 *
 * dOpenCL is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * dOpenCL is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with dOpenCL. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Permission to use dOpenCL for scientific, non-commercial work is
 * granted under the terms of the dOpenCL Academic License provided
 * appropriate credit is given. See the dOpenCL Academic License for
 * more details.
 * 
 * You should have received a copy of the dOpenCL Academic License
 * along with dOpenCL. If not, see <http://dopencl.uni-muenster.de/>.
 ******************************************************************************/

/**
 * @file Kernel.cpp
 *
 * @date 09.12.2011
 * @author Philipp Kegel
 */

#include <cstddef>
#include <cstdlib>
#include <set>
#include <vector>

#include <boost/shared_ptr.hpp>

#define __CL_ENABLE_EXCEPTIONS
#include <CL/cl.hpp>

#include <dcl/Memory.h>

#include "Kernel.h"
#include "Memory.h"
#include "Program.h"

namespace dcld {

Kernel::Kernel(const Program& program, const char *name) :
        _kernel(program, name) {
    _writeMemoryObjects.resize(_kernel.getInfo<CL_KERNEL_NUM_ARGS>());
}

Kernel::Kernel(const cl::Kernel& kernel) : _kernel(kernel) { }

Kernel::~Kernel() { }

Kernel::operator cl::Kernel() const {
	return _kernel;
}

void Kernel::getInfo(
		cl_kernel_info param_name,
		void **param_value_ret,
		size_t *param_value_size_ret) const {
	cl_int errcode;
	size_t param_value_size;

	/* Obtain kernel info using OpenCL C API to avoid unnecessary type
	 * conversions. */
	errcode = ::clGetKernelInfo(_kernel(), param_name, 0, 0, &param_value_size);
	if (errcode == CL_SUCCESS) {
		void *param_value = (void *) malloc(param_value_size);

		if (param_value) {
			errcode = ::clGetKernelInfo(_kernel(),
					param_name,
					param_value_size,
					param_value,
					0);
			if (errcode == CL_SUCCESS) {
				*param_value_size_ret = param_value_size;
				*param_value_ret = param_value;
			} else {
				/* clean up */
				free(param_value);
			}
		} else {
			throw cl::Error(CL_OUT_OF_HOST_MEMORY);
		}
	} else {
		throw cl::Error(errcode);
	}
}

void Kernel::getWorkGroupInfo(
        const cl::Device& device,
        cl_kernel_work_group_info param_name,
        void **param_value_ret,
        size_t *param_value_size_ret) const {
    cl_int errcode;
    size_t param_value_size;

    /* Obtain kernel work group info using OpenCL C API to avoid unnecessary
     * type conversions. */
    errcode = ::clGetKernelWorkGroupInfo(
            _kernel(), device(), param_name, 0, 0, &param_value_size);
    if (errcode == CL_SUCCESS) {
        void *param_value = (void *) malloc(param_value_size);

        if (param_value) {
            errcode = ::clGetKernelWorkGroupInfo(
                    _kernel(),
                    device(),
                    param_name,
                    param_value_size,
                    param_value,
                    0);
            if (errcode == CL_SUCCESS) {
                *param_value_size_ret = param_value_size;
                *param_value_ret = param_value;
            } else {
                /* clean up */
                free(param_value);
            }
        } else {
            throw cl::Error(CL_OUT_OF_HOST_MEMORY);
        }
    } else {
        throw cl::Error(errcode);
    }
}

void Kernel::setArg(cl_uint index,
        const boost::shared_ptr<dcl::Memory>& memory) {
    const boost::shared_ptr<Memory>& memoryImpl = boost::dynamic_pointer_cast<Memory>(memory);
    if (memoryImpl) {
        _kernel.setArg(index, memoryImpl->operator cl::Memory());

        if (memoryImpl->isOutput()) { // memory object will be modified by kernel
            /* If a writable (CL_MEM_WRITE_ONLY, CL_MEM_READ_WRITE)
             * memory object is set as kernel argument, it is assumed
             * that the kernel will modify this memory object. */
            assert(index < _writeMemoryObjects.size());
            if (_writeMemoryObjects.size() <= index) {
                _writeMemoryObjects.resize(index + 1);
            }
            _writeMemoryObjects[index] = memoryImpl;
        }
    } else {
        throw cl::Error(CL_INVALID_MEM_OBJECT);
    }
}

void Kernel::setArg(cl_uint index, const cl::Sampler& sampler) {
	_kernel.setArg(index, sampler);
}

void Kernel::setArg(cl_uint index, size_t size, const void *argPtr) {
	/* Const-ness has to be stripped from the argument value, as the OpenCL C++
	 * binding requires a non-const argument value. Though, this argument is
	 * directly passed to the C API which argument is const. */
	_kernel.setArg(index, size, const_cast<void *>(argPtr));
}

std::vector<boost::shared_ptr<Memory> > Kernel::writeMemoryObjects() {
    std::set<boost::shared_ptr<Memory> > writeMemoryObjects;

    /* copy memory objects from argument list to set to remove duplicates */
    for (std::vector<boost::shared_ptr<Memory> >::const_iterator it = _writeMemoryObjects.begin();
            it != _writeMemoryObjects.end(); ++it) {
        /* ignore empty (NULL) entries */
        if (*it) writeMemoryObjects.insert(*it);
    }

    /* return value optimization should kick-in here to avoid copying */
    return std::vector<boost::shared_ptr<Memory> >(writeMemoryObjects.begin(),
            writeMemoryObjects.end());
}

} /* namespace dcld */
