/******************************************************************************
 * This file is part of dOpenCL.
 * 
 * dOpenCL is an implementation of the OpenCL application programming
 * interface for distributed systems. See <http://dopencl.uni-muenster.de/>
 * for more information.
 * 
 * Developed by: Research Group Parallel and Distributed Systems
 *               Department of Mathematics and Computer Science
 *               University of Muenster, Germany
 *               <http://pvs.uni-muenster.de/>
 * 
 * Copyright (C) 2013  Philipp Kegel <philipp.kegel@uni-muenster.de>
 *
 * dOpenCL is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * dOpenCL is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with dOpenCL. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Permission to use dOpenCL for scientific, non-commercial work is
 * granted under the terms of the dOpenCL Academic License provided
 * appropriate credit is given. See the dOpenCL Academic License for
 * more details.
 * 
 * You should have received a copy of the dOpenCL Academic License
 * along with dOpenCL. If not, see <http://dopencl.uni-muenster.de/>.
 ******************************************************************************/

/**
 * @file Kernel.h
 *
 * @date 09.12.2011
 * @author Philipp Kegel
 */

#ifndef KERNEL_H_
#define KERNEL_H_

#include <cstddef>
#include <vector>

#include <boost/shared_ptr.hpp>

#define __CL_ENABLE_EXCEPTIONS
#include <CL/cl.hpp>

#include <dcl/Kernel.h>
#include <dcl/Memory.h>

namespace dcld {

/* forward declarations */
class Memory;
class Program;

/******************************************************************************/

/**
 * @brief A decorator for a native kernel.
 *
 * This wrapper is required to store memory objects which are passed to a
 * kernel as arguments.
 */
class Kernel:
        public dcl::Kernel { // implement
public:
    Kernel(
            const Program& program,
            const char *   name);
    Kernel(
            const cl::Kernel& kernel);
    virtual ~Kernel();

    /**
     * @brief Returns the wrapped OpenCL kernel.
     */
    operator cl::Kernel() const;

    void getInfo(
            cl_kernel_info param_name,
            void **        param_value_ret,
            size_t *       param_value_size_ret) const;
    void getWorkGroupInfo(
            const cl::Device&         device,
            cl_kernel_work_group_info param_name,
            void **                   param_value_ret,
            size_t *                  param_value_size_ret) const;

    void setArg(
            cl_uint                               index,
            const boost::shared_ptr<dcl::Memory>& memoryObject);
    void setArg(
            cl_uint            index,
            const cl::Sampler& sampler);
    void setArg(
            cl_uint     index,
            size_t      size,
            const void *argPtr = 0);

    /**
     * @brief Returns the memory objects (possibly) written to with this kernel
     *
     * @return a list of memory objects
     */
    std::vector<boost::shared_ptr<Memory> > writeMemoryObjects();

private:
    /* Kernels must not be copied */
    Kernel(const Kernel&);
    Kernel& operator=(const Kernel&);

    cl::Kernel _kernel; /**< Native kernel */

    /**
     * @brief Memory objects used by this kernel
     */
    std::vector<boost::shared_ptr<Memory> > _writeMemoryObjects;
};

} /* namespace dcld */

#endif /* KERNEL_H_ */
