/******************************************************************************
 * This file is part of dOpenCL.
 * 
 * dOpenCL is an implementation of the OpenCL application programming
 * interface for distributed systems. See <http://dopencl.uni-muenster.de/>
 * for more information.
 * 
 * Developed by: Research Group Parallel and Distributed Systems
 *               Department of Mathematics and Computer Science
 *               University of Muenster, Germany
 *               <http://pvs.uni-muenster.de/>
 * 
 * Copyright (C) 2013  Philipp Kegel <philipp.kegel@uni-muenster.de>
 *
 * dOpenCL is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * dOpenCL is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with dOpenCL. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Permission to use dOpenCL for scientific, non-commercial work is
 * granted under the terms of the dOpenCL Academic License provided
 * appropriate credit is given. See the dOpenCL Academic License for
 * more details.
 * 
 * You should have received a copy of the dOpenCL Academic License
 * along with dOpenCL. If not, see <http://dopencl.uni-muenster.de/>.
 ******************************************************************************/

/**
 * @file    Memory.cpp
 *
 * @date    01.09.2011
 * @author  Philipp Kegel
 */

#include <cstddef>
#include <cstdlib>
#include <ostream>

#include <boost/bind.hpp>
#include <boost/shared_ptr.hpp>

#define __CL_ENABLE_EXCEPTIONS
#include <CL/cl.hpp>
#include <CL/cl_wwu_dcl.h>

#include <dcl/DataTransfer.h>
#include <dcl/DCLException.h>
#include <dcl/DCLTypes.h>
#include <dcl/Remote.h>

#include <dcl/util/Logger.h>

#include "Context.h"
#include "Memory.h"

namespace dcld {

struct ExecData {
    dcl::Process *process;
    size_t size;
    void *ptr;
    cl::UserEvent event;
};

static void execAcquire(cl_event event, cl_int execution_status, void *user_data) {
    ExecData *syncData = static_cast<ExecData *>(user_data);

    assert(execution_status == CL_COMPLETE || execution_status < 0);
    assert(syncData != NULL);

    if (execution_status == CL_COMPLETE) {
        dcl::util::Logger << dcl::util::Debug
                << "(SYN) Acquiring memory object data from process '"
                << syncData->process->url() << '\''
                << std::endl;

        try {
            boost::shared_ptr<dcl::DataTransfer> recv =
                    syncData->process->receiveData(syncData->size, syncData->ptr);
            recv->setCallback(
                    boost::bind(&cl::UserEvent::setStatus, syncData->event, _1));
        } catch (const dcl::IOException& e) {
            dcl::util::Logger << dcl::util::Error
                    << "Data receipt failed: " << e.what() << std::endl;
            syncData->event.setStatus(CL_IO_ERROR_WWU);
        }
    } else {
        dcl::util::Logger << dcl::util::Error
                << "(SYN) Acquiring memory object data failed"
                << std::endl;

        syncData->event.setStatus(execution_status);
    }

    delete syncData;
}

static void execRelease(cl_event event, cl_int execution_status, void *user_data) {
    ExecData *syncData = static_cast<ExecData *>(user_data);

    assert(execution_status == CL_COMPLETE || execution_status < 0);
    assert(syncData != NULL);

    if (execution_status == CL_COMPLETE) {
        dcl::util::Logger << dcl::util::Debug
                << "(SYN) Releasing memory object data to process '"
                << syncData->process->url() << '\''
                << std::endl;

        try {
            boost::shared_ptr<dcl::DataTransfer> send =
                    syncData->process->sendData(syncData->size, syncData->ptr);
            send->setCallback(
                    boost::bind(&cl::UserEvent::setStatus, syncData->event, _1));
        } catch (const dcl::IOException& e) {
            dcl::util::Logger << dcl::util::Error
                    << "Data sending failed: " << e.what() << std::endl;
            syncData->event.setStatus(CL_IO_ERROR_WWU);
        }
    } else {
        dcl::util::Logger << dcl::util::Error
                << "(SYN) Releasing memory object data failed"
                << std::endl;

        syncData->event.setStatus(execution_status);
    }

    delete syncData;
}

/******************************************************************************
 * Memory object
 ******************************************************************************/

Memory::Memory(
        dcl::dcl_object_id memId,
        const boost::shared_ptr<Context>& context) :
        dcl::Remote(memId), _context(context) {
}

Memory::~Memory() { }

size_t Memory::size() const {
    return (operator cl::Memory()).getInfo<CL_MEM_SIZE>();
}

bool Memory::isInput() const{
	return (static_cast<cl::Memory>(*this).getInfo<CL_MEM_FLAGS>() &
	        (CL_MEM_READ_ONLY | CL_MEM_READ_WRITE));
}

bool Memory::isOutput() const {
	return (static_cast<cl::Memory>(*this).getInfo<CL_MEM_FLAGS>() &
	        (CL_MEM_WRITE_ONLY | CL_MEM_READ_WRITE));
}

/******************************************************************************
 * Buffer
 ******************************************************************************/

Buffer::Buffer(dcl::dcl_object_id memId,
        const boost::shared_ptr<Context>& context, cl_mem_flags flags,
		size_t size) :
	dcld::Memory(memId, context)
{
    cl_mem_flags rwFlags = flags &
            (CL_MEM_READ_WRITE | CL_MEM_READ_ONLY | CL_MEM_WRITE_ONLY);
    cl_mem_flags hostPtrFlags = flags &
            (CL_MEM_COPY_HOST_PTR | CL_MEM_USE_HOST_PTR);
    cl_mem_flags allocHostPtr = flags &
            CL_MEM_ALLOC_HOST_PTR;

    /*
     * Always let the OpenCL implementation allocate the memory on the compute
     * node to ensure optimal performance.
     * If CL_MEM_USE_HOST_PTR is specified, the compute node should try to use
     * pinned memory to ensure optimal performance for frequent data transfers.
     */

    if (hostPtrFlags) {
        /* receive buffer data from host */
        /* FIXME Use map/unmap to avoid using malloc/free *or* only copy data on host */
        void *host_ptr = (void *) malloc(size);
        if (!host_ptr) throw cl::Error(CL_OUT_OF_RESOURCES);
        _context->host().receiveData(size, host_ptr)->wait();

        _buffer = cl::Buffer(*_context, rwFlags | CL_MEM_COPY_HOST_PTR | allocHostPtr, size, host_ptr);

        free(host_ptr);
    } else {
        /* create uninitialized buffer */
        _buffer = cl::Buffer(*_context, rwFlags | allocHostPtr, size);
    }
}

Buffer::~Buffer() { }

Buffer::operator cl::Memory() const {
    return _buffer;
}

Buffer::operator cl::Buffer() const {
	return _buffer;
}

void Buffer::acquire(
        dcl::Process& process,
        const cl::CommandQueue& commandQueue,
        const cl::Event& releaseEvent,
        cl::Event *acquireEvent) {
    cl::Event mapEvent;
    cl::UserEvent dataReceipt(*_context);

    dcl::util::Logger << dcl::util::Debug
            << "(SYN) Acquiring buffer from process '" << process.url() << "' (ID="
            << remoteId() << ')'
            << std::endl;

    /* map buffer to host memory when releaseEvent is complete */
    VECTOR_CLASS<cl::Event> mapWaitList(1, releaseEvent);
    void *ptr = commandQueue.enqueueMapBuffer(
            _buffer,
            CL_FALSE,
            CL_MAP_WRITE,
            0, size(),
            &mapWaitList, &mapEvent);

    ExecData *syncData = new ExecData;
    syncData->process = &process;
    syncData->size    = size();
    syncData->ptr     = ptr;
    syncData->event   = dataReceipt;

    /* receive buffer data when mapping is complete */
    mapEvent.setCallback(CL_COMPLETE, &execAcquire, syncData);

    /* WARNING: do not use cb_data after this point, as the callback of mapEvent
     *          deletes dataTransfer concurrently */

    /* unmap buffer when acquire operation is complete */
    VECTOR_CLASS<cl::Event> unmapWaitList(1, dataReceipt);
    commandQueue.enqueueUnmapMemObject(
            _buffer,
            ptr,
            &unmapWaitList, acquireEvent);
}

void Buffer::release(
        dcl::Process& process,
        const cl::CommandQueue& commandQueue,
        const cl::Event& releaseEvent) const {
    cl::Event mapEvent;
    cl::UserEvent dataSending(*_context);

    dcl::util::Logger << dcl::util::Debug
            << "(SYN) Releasing buffer to process '" << process.url() << "' (ID="
            << remoteId() << ')'
            << std::endl;

    /* map buffer when releaseEvent is complete */
    VECTOR_CLASS<cl::Event> mapWaitList(1, releaseEvent);
    void *ptr = commandQueue.enqueueMapBuffer(
            _buffer,
            CL_FALSE,
            CL_MAP_READ,
            0, size(),
            &mapWaitList, &mapEvent
    );

    ExecData *syncData = new ExecData;
    syncData->process = &process;
    syncData->size    = size();
    syncData->ptr     = ptr;
    syncData->event   = dataSending;

    /* send buffer data when mapping is complete */
    mapEvent.setCallback(CL_COMPLETE, &execRelease, syncData);

    /* WARNING: do not use cb_data after this point, as the callback of mapEvent
     *          deletes dataTransfer concurrently */

    /* unmap buffer when acquire operation is complete */
    VECTOR_CLASS<cl::Event> unmapWaitList(1, dataSending);
    commandQueue.enqueueUnmapMemObject(
            _buffer,
            ptr,
            &unmapWaitList, NULL);
}

} /* namespace dcld */
