/******************************************************************************
 * This file is part of dOpenCL.
 * 
 * dOpenCL is an implementation of the OpenCL application programming
 * interface for distributed systems. See <http://dopencl.uni-muenster.de/>
 * for more information.
 * 
 * Developed by: Research Group Parallel and Distributed Systems
 *               Department of Mathematics and Computer Science
 *               University of Muenster, Germany
 *               <http://pvs.uni-muenster.de/>
 * 
 * Copyright (C) 2013  Philipp Kegel <philipp.kegel@uni-muenster.de>
 *
 * dOpenCL is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * dOpenCL is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with dOpenCL. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Permission to use dOpenCL for scientific, non-commercial work is
 * granted under the terms of the dOpenCL Academic License provided
 * appropriate credit is given. See the dOpenCL Academic License for
 * more details.
 * 
 * You should have received a copy of the dOpenCL Academic License
 * along with dOpenCL. If not, see <http://dopencl.uni-muenster.de/>.
 ******************************************************************************/

/**
 * @file Program.cpp
 *
 * @date    2012-07-30
 * @author  Philipp Kegel
 */

#include <cstddef>
#include <cstdlib>
#include <ostream>
#include <vector>

#include <boost/make_shared.hpp>
#include <boost/shared_ptr.hpp>

#define __CL_ENABLE_EXCEPTIONS
#include <CL/cl.hpp>

#include <dcl/DataTransfer.h>
#include <dcl/DCLTypes.h>
#include <dcl/Kernel.h>
#include <dcl/ProgramBuildListener.h>
#include <dcl/Remote.h>

#include <dcl/comm/ProgramMessage.h>

#include <dcl/util/Logger.h>

#include "Context.h"
#include "Program.h"
#include "Kernel.h"

namespace dcld {

namespace detail {

class ProgramBuild: public dcl::Remote {
public:
    /* FIXME Pass program build listener proxy (dcl::ProgramBuildListener) ctor
     * rather than program build ID. A program build listener proxy hides the ID
     * and details of communication with the remote program build listener which
     * is notified about the completion of this program build. */
    ProgramBuild(
            dcl::dcl_object_id programBuildId,
            Program& program,
            const VECTOR_CLASS<cl::Device>& devices) :
        dcl::Remote(programBuildId), _program(program), _devices(devices) { }
    virtual ~ProgramBuild() { }

    void onComplete() {
        std::vector<dcl::dcl_object_id> deviceIds;
        std::vector<cl_build_status> buildStatus;

        /*
         * Query program build status
         */
        deviceIds.reserve(_devices.size());
        buildStatus.reserve(_devices.size());
        for (VECTOR_CLASS<cl::Device>::const_iterator i = _devices.begin();
                i != _devices.end(); ++i) {
            cl::Device device = *i;

            /* TODO Get device ID */
            deviceIds.push_back(0);

            /* Get build status for device */
            buildStatus.push_back(
                    static_cast<cl::Program>(_program).getBuildInfo<CL_PROGRAM_BUILD_STATUS>(device));
        }

        {
            dcl::comm::ProgramBuildMessage message(_id, deviceIds, buildStatus);

            _program.context()->host().sendMessage(message);
            /* TODO Handle errors */

            dcl::util::Logger << dcl::util::Debug
                    << "Sent update of program build status (ID=" << _id
                    << ')' << std::endl;
        }
    }

private:
    Program& _program;
    VECTOR_CLASS<cl::Device> _devices;
};

} /* namespace detail */

/******************************************************************************/

/**
 * @brief Callback for completion of program build
 */
static void onProgramBuildComplete(cl_program object_, void *user_data) {
    dcld::detail::ProgramBuild *programBuild = static_cast<dcld::detail::ProgramBuild *>(user_data);
    assert(programBuild != 0);
    programBuild->onComplete();

    /* clean up */
    delete programBuild;
}

/******************************************************************************/

Program::Program(const boost::shared_ptr<Context>& context, size_t length) :
    _context(context)
{
    char *string;
    cl::Program::Sources sources;

    /* Receive program sources */
    try {
        string = new char[length];
        _context->host().receiveData(length, string)->wait(); // blocking receive
    } catch (const std::bad_alloc&) {
        throw cl::Error(CL_OUT_OF_RESOURCES);
    }

    /* Create native program */
    sources.push_back(std::make_pair(string, length));
    _program = cl::Program(*_context, sources);

    /* Clean up */
    delete string;
}

Program::Program(
        const boost::shared_ptr<Context>& context,
        const VECTOR_CLASS<cl::Device>& devices,
        const std::vector<size_t>& lengths,
        VECTOR_CLASS<cl_int> *binary_status) :
    _context(context)
{
    char **strings;
    cl::Program::Binaries binaries;

    /* Receive program binaries */
    try {
        strings = new char *[lengths.size()];

        for (unsigned int i = 0; i < lengths.size(); ++i) {
            void *binary;
            size_t length = lengths[i];

            strings[i] = new char[length];
            binary = static_cast<void *>(strings[i]);
            _context->host().receiveData(length, binary)->wait(); // blocking receive

            binaries.push_back(std::make_pair(binary, length));
        }
    } catch (const std::bad_alloc&) {
        throw cl::Error(CL_OUT_OF_RESOURCES);
    }

    /* Create native program */
    _program = cl::Program(*context, devices, binaries, binary_status);

    /* Clean up */
    binaries.clear();
    for (unsigned int i = 0; i < lengths.size(); ++i) {
        delete[] strings[i];
    }
    delete strings;
}

Program::~Program() { }

Program::operator cl::Program() const {
    return _program;
}

#if 0
void Program::build(
        const std::vector<cl::Device>& devices,
        const char *options,
        const boost::shared_ptr<dcl::ProgramBuildListener>& programBuildListener) {
    /* Start asynchronous program build */
    _program.build(devices, options, &onProgramBuildComplete,
            new detail::ProgramBuild(*this, devices, programBuildListener));
}
#endif
void Program::build(
        const std::vector<cl::Device>& devices,
        const char *options,
        dcl::dcl_object_id programBuildId) {
    /* TODO Start asynchronous program build */
//    _program.build(devices, options, &onProgramBuildComplete,
//            new detail::ProgramBuild(programBuildId, *this, devices));
    _program.build(devices, options);
}

void Program::createKernels(
        std::vector<boost::shared_ptr<dcl::Kernel> >& kernels) {
    VECTOR_CLASS<cl::Kernel> nativeKernels;

    /* Create kernels */
    _program.createKernels(&nativeKernels);

    /* Wrap kernels */
    kernels.clear();
    kernels.reserve(nativeKernels.size());
    for (VECTOR_CLASS<cl::Kernel>::const_iterator i = nativeKernels.begin();
            i != nativeKernels.end(); ++i) {
        kernels.push_back(boost::make_shared<Kernel>(*i));
    }
}

const boost::shared_ptr<Context>& Program::context() const {
    return _context;
}

} /* namespace dcld */
