/******************************************************************************
 * This file is part of dOpenCL.
 * 
 * dOpenCL is an implementation of the OpenCL application programming
 * interface for distributed systems. See <http://dopencl.uni-muenster.de/>
 * for more information.
 * 
 * Developed by: Research Group Parallel and Distributed Systems
 *               Department of Mathematics and Computer Science
 *               University of Muenster, Germany
 *               <http://pvs.uni-muenster.de/>
 * 
 * Copyright (C) 2013  Philipp Kegel <philipp.kegel@uni-muenster.de>
 *
 * dOpenCL is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * dOpenCL is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with dOpenCL. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Permission to use dOpenCL for scientific, non-commercial work is
 * granted under the terms of the dOpenCL Academic License provided
 * appropriate credit is given. See the dOpenCL Academic License for
 * more details.
 * 
 * You should have received a copy of the dOpenCL Academic License
 * along with dOpenCL. If not, see <http://dopencl.uni-muenster.de/>.
 ******************************************************************************/

/*
 * Session.cpp
 *
 *  Created on: 03.02.2011
 *      Author: Philipp Kegel
 */

#include <cassert>
#include <cstddef>
#include <cstdlib>
#include <cstring>
#include <ostream>
#include <map>
#include <stdexcept>
#include <utility>
#include <vector>

#include <boost/shared_ptr.hpp>

#define __CL_ENABLE_EXCEPTIONS
#include <CL/cl.hpp>

#include <dcl/Binary.h>
#include <dcl/CLObjectRegistry.h>
#include <dcl/CommandQueue.h>
#include <dcl/ComputeNode.h>
#include <dcl/Context.h>
#include <dcl/DCLException.h>
#include <dcl/DCLTypes.h>
#include <dcl/Event.h>
#include <dcl/Host.h>
#include <dcl/Kernel.h>
#include <dcl/Memory.h>
#include <dcl/Program.h>

#include <dcl/util/Logger.h>

#include "CommandQueue.h"
#include "Context.h"
#include "DeviceManager.h"
#include "Event.h"
#include "Kernel.h"
#include "Memory.h"
#include "Program.h"
#include "Session.h"

namespace dcld {

Session::Session(
        DeviceManager& deviceManager,
        dcl::Host& host) :
    _deviceManager(deviceManager), _host(host) {
    _host.setCLCommandListener(this);
}

Session::~Session() {
	_host.setCLCommandListener();
}

/******************************************************************************
 * Command listener APIs
 ******************************************************************************/

/* Device APIs */
void Session::getDeviceIDs(
        cl_device_type device_type,
        std::vector<dcl::dcl_object_id>& device_ids) const {
	_deviceManager.getDeviceIDs(device_type, device_ids);
}

void Session::getDeviceInfo(
		dcl::dcl_object_id device_id,
        cl_device_info param_name,
        void **param_value_ret,
        size_t *param_value_size_ret) const {
	cl_int err;
	size_t param_value_size;
	cl::Device device = _deviceManager.getDevice(device_id);

	/* Obtain device info using OpenCL C API to avoid unnecessary type
	 * conversions. */
	err = ::clGetDeviceInfo(device(), param_name, 0, 0, &param_value_size);
	if (err == CL_SUCCESS) {
		void *param_value = (void *) malloc(param_value_size);
		if (param_value) {
			err = ::clGetDeviceInfo(device(), param_name, param_value_size,
					param_value, 0);
			if (err == CL_SUCCESS) {
				*param_value_size_ret = param_value_size;
				*param_value_ret = param_value;
			} else {
				/* clean up */
				free(param_value);
				throw cl::Error(err);
			}
		} else {
			throw cl::Error(CL_OUT_OF_HOST_MEMORY);
		}
	} else {
		throw cl::Error(err);
	}
}

/* Context APIs */
dcl::Context * Session::createContext(
		dcl::dcl_object_id context_id,
		const std::vector<dcl::ComputeNode *>& computeNodes,
        const std::vector<dcl::dcl_object_id>& device_ids) {
	std::vector<cl::Device> devices;

    /* Lookup devices */
    _deviceManager.getDevices(device_ids, devices);

    /* Create context and add to list */
    Context *context = new Context(context_id, _host, computeNodes,
            _deviceManager.getPlatform(), devices);
    /* TODO Verify insert operation */
    _contexts.insert(std::make_pair(context_id,
            boost::shared_ptr<Context>(context)));

    dcl::util::Logger << dcl::util::Info
            << "Context created (ID=" << context_id << ')' << std::endl;

    return context;
}

void Session::releaseContext(
		dcl::dcl_object_id context_id) {
	/* Delete context from list */
	if (_contexts.erase(context_id) != 1) {
		throw cl::Error(CL_INVALID_CONTEXT);
	}

    dcl::util::Logger << dcl::util::Info
            << "Context released (ID=" << context_id << ')' << std::endl;
}

/* Command Queue APIs */
dcl::CommandQueue * Session::createCommandQueue(
		dcl::dcl_object_id context_id,
		dcl::dcl_object_id device_id,
		dcl::dcl_object_id command_queue_id,
        cl_command_queue_properties properties) {
    /* Create command queue and add to list */
    dcl::CommandQueue *commandQueue = new CommandQueue(getContext(context_id),
            _deviceManager.getDevice(device_id), properties);
    /* TODO Verify insert operation */
    _command_queues.insert(std::make_pair(command_queue_id,
                    boost::shared_ptr<dcl::CommandQueue>(commandQueue)));

	dcl::util::Logger << dcl::util::Info
			<< "Command queue created (ID=" << command_queue_id << ')' << std::endl;

	return commandQueue;
}

void Session::releaseCommandQueue(
		dcl::dcl_object_id command_queue_id) {
	/* Delete command queue from list */
	if (_command_queues.erase(command_queue_id) != 1) {
		throw cl::Error(CL_INVALID_COMMAND_QUEUE);
	}

    dcl::util::Logger << dcl::util::Info
            << "Command queue released (ID=" << command_queue_id << ')' << std::endl;
}

dcl::Buffer * Session::createBuffer(
		dcl::dcl_object_id context_id,
        dcl::dcl_object_id buffer_id,
        cl_mem_flags flags,
        size_t size) {
	/* Create buffer and add to list */
    dcl::Buffer *buffer = new Buffer(buffer_id, getContext(context_id), flags, size);
    /* TODO Verify insert operation */
	_buffers.insert(std::make_pair(buffer_id,
	        boost::shared_ptr<dcl::Buffer>(buffer)));

	dcl::util::Logger << dcl::util::Info
			<< "Buffer created (ID=" << buffer_id << ')' << std::endl;

	return buffer;
}

void Session::releaseMemObject(
		dcl::dcl_object_id mem_object_id) {
    /* Delete buffer or image from list */
	if (_buffers.erase(mem_object_id) != 1) {
		throw cl::Error(CL_INVALID_MEM_OBJECT);
	}

    dcl::util::Logger << dcl::util::Info
            << "Memory object released (ID=" << mem_object_id << ')'
            << std::endl;
}

dcl::Program * Session::createProgram(
		dcl::dcl_object_id context_id,
        dcl::dcl_object_id program_id,
		size_t length) {
	/* Create program and add to list */
    dcl::Program *program = new Program(getContext(context_id), length);
    /* TODO Verify insert operation */
	_programs.insert(std::make_pair(program_id,
	        boost::shared_ptr<dcl::Program>(program)));

	dcl::util::Logger << dcl::util::Info
			<< "Program created from source (ID=" << program_id << ')'
			<< std::endl;

	return program;
}

dcl::Program * Session::createProgram(
		dcl::dcl_object_id context_id,
        dcl::dcl_object_id program_id,
        const std::vector<dcl::dcl_object_id>& device_id_list,
		const std::vector<size_t>& lengths,
		std::vector<cl_int> *binary_status) {
	VECTOR_CLASS<cl::Device> devices;

	/* Lookup devices */
	_deviceManager.getDevices(device_id_list, devices);

	/* Create program and add to list */
    dcl::Program *program = new Program(getContext(context_id), devices,
            lengths, binary_status);
    /* TODO Verify insert operation */
    _programs.insert(std::make_pair(program_id,
            boost::shared_ptr<dcl::Program>(program)));

	dcl::util::Logger << dcl::util::Info
			<< "Program created from binaries (ID=" << program_id << ')'
			<< std::endl;

	return program;
}

void Session::releaseProgram(
		dcl::dcl_object_id program_id) {
	if (_programs.erase(program_id) != 1) {
		throw cl::Error(CL_INVALID_PROGRAM);
	}

    dcl::util::Logger << dcl::util::Info
            << "Program released (ID=" << program_id << ')' << std::endl;
}

void Session::buildProgram(
		dcl::dcl_object_id program_id,
		const std::vector<dcl::dcl_object_id> *device_id_list,
		const char *options,
		dcl::dcl_object_id program_build_id) {
	std::vector<cl::Device> devices;

	if (device_id_list != 0) {
		_deviceManager.getDevices(*device_id_list, devices); // lookup devices
	}

#if 0
    /* TODO Generate proxy for ProgramBuildListener from program build ID */
    boost::shared_ptr<dcl::ProgramBuildListener> programBuildListener(
            boost::make_shared<ProgramBuildListener>(program_id, getProgram(program_id)));
	getProgram(program_id)->build(devices, options, programBuildListener);
#endif
    getProgram(program_id)->build(devices, options, program_build_id);

	dcl::util::Logger << dcl::util::Info
			<< "Program build submitted (program ID=" << program_id
			<< ", build ID=" << program_build_id
			<< ')' << std::endl;
}

dcl::Kernel * Session::createKernel(
		dcl::dcl_object_id program_id,
        dcl::dcl_object_id kernel_id,
        const char *name) {
    boost::shared_ptr<Program> programImpl = boost::dynamic_pointer_cast<Program>(getProgram(program_id));

    if (!programImpl) throw cl::Error(CL_INVALID_PROGRAM);

	/* Create kernel and add to list */
    Kernel *kernel = new Kernel(*programImpl, name);
    /* TODO Verify insert operation */
	_kernels.insert(std::make_pair(kernel_id,
	        boost::shared_ptr<dcl::Kernel>(kernel)));

	dcl::util::Logger << dcl::util::Info
			<< "Kernel created (ID=" << kernel_id
			<< ", name=" << name
			<< ')' << std::endl;

	return kernel;
}

void Session::createKernelsInProgram(
		dcl::dcl_object_id program_id,
		const std::vector<dcl::dcl_object_id>& kernel_ids) {
    std::vector<boost::shared_ptr<dcl::Kernel> > kernels;

	/* Create kernels */
	getProgram(program_id)->createKernels(kernels);
	if (kernel_ids.size() != kernels.size()) throw cl::Error(CL_INVALID_VALUE);

	/* Add kernels to list */
	for (unsigned int i = 0; i < kernels.size(); ++i) {
	    /* TODO Verify insert operation */
		_kernels.insert(std::make_pair(kernel_ids[i], kernels[i]));
	}

    dcl::util::Logger << dcl::util::Info
            << "Kernels in program created (program ID=" << program_id
            << ", #kernels=" << kernels.size()
            << ')' << std::endl;
}

void Session::releaseKernel(
		dcl::dcl_object_id kernel_id) {
	if (_kernels.erase(kernel_id) != 1) {
		throw cl::Error(CL_INVALID_KERNEL);
	}

    dcl::util::Logger << dcl::util::Info
            << "Kernel released (ID=" << kernel_id << ')' << std::endl;
}

void Session::setKernelArg(
		dcl::dcl_object_id kernel_id,
        cl_uint arg_index,
        dcl::dcl_kernel_arg_type arg_type,
        size_t arg_size,
        const void *arg_value) {
	boost::shared_ptr<dcl::Kernel> kernel = getKernel(kernel_id);

	if (!kernel) {
	    throw cl::Error(CL_INVALID_KERNEL);
	}

	/* Lookup memory object or sampler by ID */
	switch (arg_type) {
	case dcl::MEMORY:
		if (arg_value == 0) {
			kernel->setArg(arg_index, arg_size);
		} else {
			kernel->setArg(arg_index, getMemory(
					*reinterpret_cast<const dcl::dcl_object_id *>(arg_value)));
		}
		break;
	case dcl::SAMPLER:
		kernel->setArg(arg_index, getSampler(
				*reinterpret_cast<const dcl::dcl_object_id *> (arg_value)));
		break;
	default:
		kernel->setArg(arg_index, arg_size, arg_value);
		/* no break */
	}

	dcl::util::Logger << dcl::util::Info
			<< "Kernel argument set (ID=" << kernel_id << ')' << std::endl;
}

void Session::getKernelInfo(
		dcl::dcl_object_id kernel_id,
        cl_kernel_info param_name,
        void **param_value_ret,
        size_t *param_value_size_ret) const {
	getKernel(kernel_id)->getInfo(param_name, param_value_ret,
					param_value_size_ret);

    dcl::util::Logger << dcl::util::Info
            << "Got kernel info (ID=" << kernel_id << ')' << std::endl;
}

void Session::getKernelWorkGroupInfo(
        dcl::dcl_object_id kernel_id,
        dcl::dcl_object_id device_id,
        cl_kernel_work_group_info param_name,
        void **param_value_ret,
        size_t *param_value_size_ret) const {
    getKernel(kernel_id)->getWorkGroupInfo(_deviceManager.getDevice(device_id),
            param_name, param_value_ret, param_value_size_ret);

    dcl::util::Logger << dcl::util::Info
            << "Got kernel work group info (kernel ID=" << kernel_id
            << ", device ID=" << device_id
            << ')' << std::endl;
}

dcl::Event * Session::createEvent(
        dcl::dcl_object_id context_id,
        dcl::dcl_object_id command_id,
        const std::vector<dcl::dcl_object_id>& memoryObjectIds) {
    std::vector<boost::shared_ptr<Memory> > memoryObjects;

    /* Lookup memory objects */
    memoryObjects.reserve(memoryObjectIds.size());
    for (std::vector<dcl::dcl_object_id>::const_iterator i = memoryObjectIds.begin();
            i != memoryObjectIds.end(); ++i) {
        boost::shared_ptr<Memory> memoryImpl = boost::dynamic_pointer_cast<Memory>(getMemory(*i));
        if (!memoryImpl) throw cl::Error(CL_INVALID_MEM_OBJECT);
        memoryObjects.push_back(memoryImpl);
    }

    /* Create event
     * Technically, events can only be created by enqueuing a command. However,
     * since the command has been enqueued on another compute node, a user
     * event is created as a substitute for that event. The status of the
     * substitute event is controlled by messages about execution status changes
     * of its associated command. */
    dcl::Event *event = new EventListener(command_id, getContext(context_id), memoryObjects);

    /* Add event to list */
    bool inserted = _events.insert(std::make_pair(command_id,
            boost::shared_ptr<dcl::Event>(event))).second;
    assert(inserted); // insert must not fail

    dcl::util::Logger << dcl::util::Info
            << "Event created (ID=" << command_id << ')' << std::endl;

    return event;
}

dcl::UserEvent * Session::createUserEvent(
		dcl::dcl_object_id context_id,
        dcl::dcl_object_id event_id) {
    dcl::UserEvent *event = new UserEvent(getContext(context_id));

	/* Add user event to list */
    bool inserted = _userEvents.insert(std::make_pair(event_id,
            boost::shared_ptr<dcl::UserEvent>(event))).second;
    assert(inserted); // insert must not fail

	dcl::util::Logger << dcl::util::Info
			<< "User event created (ID=" << event_id << ')' << std::endl;

	return event;
}

void Session::releaseEvent(
		dcl::dcl_object_id event_id) {
    if (_events.erase(event_id) != 1
            && _userEvents.erase(event_id) != 1) {
        throw cl::Error(CL_INVALID_EVENT);
    }

	dcl::util::Logger << dcl::util::Info
			<< "Event released (ID=" << event_id << ')' << std::endl;
}

void Session::setUserEventStatus(
		dcl::dcl_object_id event_id,
        cl_int status) {
	getUserEvent(event_id)->setStatus(status);

    dcl::util::Logger << dcl::util::Info
            << "Set user event status (ID=" << event_id
            << ", status="
            << ')' << std::endl;
}

void Session::getEventProfilingInfo(
		dcl::dcl_object_id event_id,
		cl_profiling_info param_name,
		cl_ulong& param_value) {
	getEvent(event_id)->getProfilingInfo(param_name, param_value);
}

void Session::flush(dcl::dcl_object_id command_queue_id) {
	getCommandQueue(command_queue_id)->flush();

	dcl::util::Logger << dcl::util::Info
			<< "Flushed command queue (ID=" << command_queue_id << ')'
			<< std::endl;
}

void Session::finish(dcl::dcl_object_id command_queue_id) {
    /* TODO Finish command queue asynchronously */
	getCommandQueue(command_queue_id)->finish();

	dcl::util::Logger << dcl::util::Info
			<< "Finished command queue (ID=" << command_queue_id << ')'
			<< std::endl;
}

void Session::enqueueCopyBuffer(
		dcl::dcl_object_id command_queue_id,
		dcl::dcl_object_id command_id,
		dcl::dcl_object_id src_buffer_id,
		dcl::dcl_object_id dst_buffer_id,
		size_t src_offset,
		size_t dst_offset,
		size_t size,
		const std::vector<dcl::dcl_object_id> *event_id_wait_list,
		bool event) {
    std::vector<boost::shared_ptr<dcl::Event> > eventWaitList;
    boost::shared_ptr<dcl::Event> copyBuffer;

    /* Lookup events of event wait list */
    if (event_id_wait_list) {
        getEventWaitList(*event_id_wait_list, eventWaitList);
    }

    getCommandQueue(command_queue_id)->enqueueCopyBuffer(
            getBuffer(src_buffer_id),
            getBuffer(dst_buffer_id),
            src_offset, dst_offset, size,
            &eventWaitList, command_id, event ? &copyBuffer : 0);

	if (event) { // an event should be associated with this command
        /* Add event to list */
        bool inserted = _events.insert(std::make_pair(command_id, copyBuffer)).second;
        assert(inserted); // insert must not fail
	}

	dcl::util::Logger << dcl::util::Info
			<< "Enqueued copy buffer (command queue ID=" << command_queue_id
			<< ", src buffer ID=" << src_buffer_id
			<< ", dst buffer ID=" << dst_buffer_id
			<< ", command ID=" << command_id
			<< ')' << std::endl;
}

void Session::enqueueReadBuffer(
		dcl::dcl_object_id command_queue_id,
		dcl::dcl_object_id command_id,
		dcl::dcl_object_id buffer_id,
		bool blocking_read,
		size_t offset,
		size_t size,
		const std::vector<dcl::dcl_object_id> *event_id_wait_list,
		bool event) {
    std::vector<boost::shared_ptr<dcl::Event> > eventWaitList;
    boost::shared_ptr<dcl::Event> readBuffer;

    /* Lookup events of event wait list */
    if (event_id_wait_list) {
        getEventWaitList(*event_id_wait_list, eventWaitList);
    }

    getCommandQueue(command_queue_id)->enqueueReadBuffer(
            getBuffer(buffer_id),
            blocking_read,
            offset, size,
            &eventWaitList, command_id, event ? &readBuffer : 0);

    if (event) { // an event should be associated with this command
        /* Add event to list */
        bool inserted = _events.insert(std::make_pair(command_id, readBuffer)).second;
        assert(inserted); // insert must not fail
    }

	dcl::util::Logger << dcl::util::Info
			<< "Enqueued data download from buffer (command queue ID="
			<< command_queue_id << ", buffer ID=" << buffer_id
			<< ", command ID=" << command_id
			<< ')' << std::endl;
}

void Session::enqueueWriteBuffer(
		dcl::dcl_object_id command_queue_id,
		dcl::dcl_object_id command_id,
		dcl::dcl_object_id buffer_id,
		bool blocking_write,
		size_t offset,
		size_t size,
		const std::vector<dcl::dcl_object_id> *event_id_wait_list,
		bool event) {
    std::vector<boost::shared_ptr<dcl::Event> > eventWaitList;
    boost::shared_ptr<dcl::Event> writeBuffer;

    /* Lookup events of event wait list */
    if (event_id_wait_list) {
        getEventWaitList(*event_id_wait_list, eventWaitList);
    }

    getCommandQueue(command_queue_id)->enqueueWriteBuffer(
            getBuffer(buffer_id),
            blocking_write,
            offset, size,
            &eventWaitList, command_id, event ? &writeBuffer : 0);

    if (event) { // an event should be associated with this command
        /* Add event to list */
        bool inserted = _events.insert(std::make_pair(command_id, writeBuffer)).second;
        assert(inserted); // insert must not fail
    }

	dcl::util::Logger << dcl::util::Info
			<< "Enqueued data upload to buffer (command queue ID="
			<< command_queue_id << ", buffer ID=" << buffer_id
			<< ", command ID=" << command_id
			<< ')' << std::endl;
}

void Session::enqueueMapBuffer(
        dcl::dcl_object_id command_queue_id,
        dcl::dcl_object_id command_id,
        dcl::dcl_object_id buffer_id,
        bool blocking_map,
        cl_map_flags map_flags,
        size_t offset,
        size_t size,
        const std::vector<dcl::dcl_object_id> *event_id_wait_list,
        bool event) {
    std::vector<boost::shared_ptr<dcl::Event> > eventWaitList;
    boost::shared_ptr<dcl::Event> mapBuffer;

    /* Lookup events of event wait list */
    if (event_id_wait_list) {
        getEventWaitList(*event_id_wait_list, eventWaitList);
    }

    getCommandQueue(command_queue_id)->enqueueMapBuffer(
            getBuffer(buffer_id),
            blocking_map, map_flags, offset, size,
            &eventWaitList, command_id, event ? &mapBuffer : 0);

    if (event) { // an event should be associated with this command
        /* Add event to list */
        bool inserted = _events.insert(std::make_pair(command_id, mapBuffer)).second;
        assert(inserted); // insert must not fail
    }

    dcl::util::Logger << dcl::util::Info
            << "Enqueued map buffer (command queue ID=" << command_queue_id
            << ", buffer ID=" << buffer_id
            << ", command ID=" << command_id
            << ')' << std::endl;
}

void Session::enqueueUnmapMemObject(
        dcl::dcl_object_id command_queue_id,
        dcl::dcl_object_id command_id,
        dcl::dcl_object_id mem_object_id,
        const std::vector<dcl::dcl_object_id> *event_id_wait_list,
        bool event) {
    std::vector<boost::shared_ptr<dcl::Event> > eventWaitList;
    boost::shared_ptr<dcl::Event> unmapMemObject;

    /* Lookup events of event wait list */
    if (event_id_wait_list) {
        getEventWaitList(*event_id_wait_list, eventWaitList);
    }

    getCommandQueue(command_queue_id)->enqueueUnmapMemObject(
            getMemory(mem_object_id),
            &eventWaitList, command_id, event ? &unmapMemObject : 0);

    if (event) { // an event should be associated with this command
        /* Add event to list */
        bool inserted = _events.insert(std::make_pair(command_id, unmapMemObject)).second;
        assert(inserted); // insert must not fail
    }

    dcl::util::Logger << dcl::util::Info
            << "Enqueued unmap memory object (command queue ID=" << command_queue_id
            << ", memory object ID=" << mem_object_id
            << ", command ID=" << command_id
            << ')' << std::endl;
}

void Session::enqueueNDRangeKernel(
		dcl::dcl_object_id command_queue_id,
		dcl::dcl_object_id command_id,
		dcl::dcl_object_id kernel_id,
		const cl::NDRange& offset,
		const cl::NDRange& global,
		const cl::NDRange& local,
		const std::vector<dcl::dcl_object_id> *event_id_wait_list,
		bool event) {
    std::vector<boost::shared_ptr<dcl::Event> > eventWaitList;
    boost::shared_ptr<dcl::Event> ndRangeKernel;

    /* Lookup events of event wait list */
	if (event_id_wait_list) {
		getEventWaitList(*event_id_wait_list, eventWaitList);
	}

	getCommandQueue(command_queue_id)->enqueueNDRangeKernel(
	        getKernel(kernel_id),
			offset, global, local,
			&eventWaitList, command_id, event ? &ndRangeKernel : 0);

	if (event) { // an event should be associated with this command
        /* Add event to list */
        bool inserted = _events.insert(std::make_pair(command_id, ndRangeKernel)).second;
        assert(inserted); // insert must not fail
	}

	dcl::util::Logger << dcl::util::Info
			<< "Enqueued ND range kernel (command queue ID=" << command_queue_id
			<< ", kernel ID=" << kernel_id
			<< ", command ID=" << command_id
			<< ')' << std::endl;
}

void Session::enqueueMarker(
		dcl::dcl_object_id command_queue_id,
		dcl::dcl_object_id command_id,
        const std::vector<dcl::dcl_object_id> *event_id_wait_list,
        bool event) {
    std::vector<boost::shared_ptr<dcl::Event> > eventWaitList;
    boost::shared_ptr<dcl::Event> marker;

    /* Lookup events of event wait list */
    if (event_id_wait_list) {
        getEventWaitList(*event_id_wait_list, eventWaitList);
    }

    getCommandQueue(command_queue_id)->enqueueMarker(
            event_id_wait_list ? &eventWaitList : 0, command_id,
            event ? &marker : 0);

    if (event) { // an event should be associated with this command
        /* Add event to list */
        bool inserted = _events.insert(std::make_pair(command_id, marker)).second;
        assert(inserted); // insert must not fail
    }

	dcl::util::Logger << dcl::util::Info
			<< "Enqueued marker (command queue ID=" << command_queue_id
			<< ", command ID=" << command_id
			<< ')' << std::endl;
}

#if defined(CL_USE_DEPRECATED_OPENCL_1_1_APIS) || (defined(CL_VERSION_1_1) && !defined(CL_VERSION_1_2))
void Session::enqueueWaitForEvents(
		dcl::dcl_object_id command_queue_id,
		const std::vector<dcl::dcl_object_id>& event_id_list) {
	std::vector<boost::shared_ptr<dcl::Event> > eventList;

    /* Lookup events of event list
     * Unlike other enqueued commands, wait for events throws CL_INVALID_EVENT
     * rather than CL_INVALID_EVENT_WAIT_LIST if the event list contains an
     * invalid event. Moreover, the event list must not contain user events */
	eventList.reserve(event_id_list.size());
    for (std::vector<dcl::dcl_object_id>::const_iterator i = event_id_list.begin();
            i != event_id_list.end(); ++i) {
        eventList.push_back(getEvent(*i));
    }

    getCommandQueue(command_queue_id)->enqueueWaitForEvents(eventList);

	dcl::util::Logger << dcl::util::Info
			<< "Enqueued wait for events (command queue ID=" << command_queue_id
			<< ')' << std::endl;
}
#endif // #if defined(CL_USE_DEPRECATED_OPENCL_1_1_APIS)

void Session::enqueueBarrier(
        dcl::dcl_object_id command_queue_id,
        dcl::dcl_object_id command_id,
        const std::vector<dcl::dcl_object_id> *event_id_wait_list,
        bool event) {
    std::vector<boost::shared_ptr<dcl::Event> > eventWaitList;
    boost::shared_ptr<dcl::Event> barrier;

    /* Lookup events of event wait list */
    if (event_id_wait_list) {
        getEventWaitList(*event_id_wait_list, eventWaitList);
    }

    getCommandQueue(command_queue_id)->enqueueBarrier(
            event_id_wait_list ? &eventWaitList : 0, command_id,
            event ? &barrier : 0);

    if (event) { // an event should be associated with this command
        /* Add event to list */
        bool inserted = _events.insert(std::make_pair(command_id, barrier)).second;
        assert(inserted); // insert must not fail
    }

    dcl::util::Logger << dcl::util::Info
            << "Enqueued barrier (command queue ID=" << command_queue_id
            << ", command ID=" << command_id
            << ')' << std::endl;
}

void Session::enqueueBroadcastBuffer(
        const std::vector<dcl::dcl_object_id>& command_queue_ids,
        dcl::dcl_object_id command_id,
        dcl::dcl_object_id srcId,
        const std::vector<dcl::dcl_object_id>& dstIds,
        size_t srcOffset,
        const std::vector<size_t> dstOffsets,
        size_t cb,
        const std::vector<dcl::dcl_object_id> *event_id_wait_list,
        bool event) {
    /* TODO Implement Session::enqueueBroadcastBuffer */
    assert(!"Session::enqueueBroadcastBuffer not implemented");
}

void Session::enqueueReduceBuffer(
        dcl::dcl_object_id command_queue_id,
        dcl::dcl_object_id command_id,
        const std::vector<dcl::dcl_object_id>& srcIds,
        dcl::dcl_object_id dstId,
        dcl::dcl_object_id kernel_id,
        const cl::NDRange& offset,
        const cl::NDRange& global,
        const cl::NDRange& local,
        const std::vector<dcl::dcl_object_id> *event_id_wait_list,
        bool event) {
    /* TODO Implement Session::enqueueReduceBuffer */
    assert(!"Session::enqueueReduceBuffer not implemented");
}

/******************************************************************************/

boost::shared_ptr<Context> Session::getContext(dcl::dcl_object_id id) const {
	std::map<dcl::dcl_object_id, boost::shared_ptr<Context> >::const_iterator i;

	i = _contexts.find(id);
	if (i == _contexts.end()) { /* context not found */
		throw cl::Error(CL_INVALID_CONTEXT);
	} else {
		return i->second;
	}
}

boost::shared_ptr<dcl::CommandQueue> Session::getCommandQueue(
        dcl::dcl_object_id id) const {
	std::map<dcl::dcl_object_id, boost::shared_ptr<dcl::CommandQueue> >::const_iterator i;

	i = _command_queues.find(id);
	if (i == _command_queues.end()) { /* command queue not found */
		throw cl::Error(CL_INVALID_COMMAND_QUEUE);
	}

	return i->second;
}

boost::shared_ptr<dcl::Memory> Session::getMemory(dcl::dcl_object_id id) const {
	std::map<dcl::dcl_object_id, boost::shared_ptr<dcl::Buffer> >::const_iterator i;

	i = _buffers.find(id);
	if (i == _buffers.end()) { // buffer not found
	    /* TODO Lookup image */

		throw cl::Error(CL_INVALID_MEM_OBJECT);
	}

	return i->second;
}

boost::shared_ptr<dcl::Buffer> Session::getBuffer(dcl::dcl_object_id id) const {
    std::map<dcl::dcl_object_id, boost::shared_ptr<dcl::Buffer> >::const_iterator i;

    i = _buffers.find(id);
    if (i == _buffers.end()) { // buffer not found
        throw cl::Error(CL_INVALID_MEM_OBJECT);
    }

    return i->second;
}

cl::Sampler Session::getSampler(dcl::dcl_object_id id) const {
	std::map<dcl::dcl_object_id, cl::Sampler>::const_iterator i;

	i = _samplers.find(id);
	if (i == _samplers.end()) { /* sampler not found */
		throw cl::Error(CL_INVALID_SAMPLER);
	}

	return i->second;
}

boost::shared_ptr<dcl::Program> Session::getProgram(dcl::dcl_object_id id) const {
	std::map<dcl::dcl_object_id, boost::shared_ptr<dcl::Program> >::const_iterator i;

	i = _programs.find(id);
	if (i == _programs.end()) { /* program not found */
		throw cl::Error(CL_INVALID_PROGRAM);
	}

	return i->second;
}

boost::shared_ptr<dcl::Kernel> Session::getKernel(dcl::dcl_object_id id) const {
	std::map<dcl::dcl_object_id, boost::shared_ptr<dcl::Kernel> >::const_iterator i;

	i = _kernels.find(id);
	if (i == _kernels.end()) { /* kernel not found */
		throw cl::Error(CL_INVALID_KERNEL);
	}

	return i->second;
}

void Session::getEventWaitList(
        const std::vector<dcl::dcl_object_id>& eventIdWaitList,
        std::vector<boost::shared_ptr<dcl::Event> >& eventWaitList) const {
    eventWaitList.clear();

    for (std::vector<dcl::dcl_object_id>::const_iterator id = eventIdWaitList.begin();
            id != eventIdWaitList.end(); ++id) {
        std::map<dcl::dcl_object_id, boost::shared_ptr<dcl::Event> >::const_iterator i;

        i = _events.find(*id);
        if (i == _events.end()) { // event not found
            std::map<dcl::dcl_object_id, boost::shared_ptr<dcl::UserEvent> >::const_iterator j;

            j = _userEvents.find(*id);
            if (j == _userEvents.end()) { // user event not found
                throw cl::Error(CL_INVALID_EVENT_WAIT_LIST);
            }

            eventWaitList.push_back(j->second);
        } else {
            eventWaitList.push_back(i->second);
        }
    }
}

boost::shared_ptr<dcl::Event> Session::getEvent(dcl::dcl_object_id id) const {
	std::map<dcl::dcl_object_id, boost::shared_ptr<dcl::Event> >::const_iterator i;

	i = _events.find(id);
    if (i == _events.end()) { // event not found
        throw cl::Error(CL_INVALID_EVENT);
    }

	return i->second;
}

boost::shared_ptr<dcl::UserEvent> Session::getUserEvent(dcl::dcl_object_id id) const {
	std::map<dcl::dcl_object_id, boost::shared_ptr<dcl::UserEvent> >::const_iterator i;

	i = _userEvents.find(id);
	if (i == _userEvents.end()) { // user event not found
		throw cl::Error(CL_INVALID_EVENT);
	}

	return i->second;
}

} /* namespace dcld */
