/******************************************************************************
 * This file is part of dOpenCL.
 * 
 * dOpenCL is an implementation of the OpenCL application programming
 * interface for distributed systems. See <http://dopencl.uni-muenster.de/>
 * for more information.
 * 
 * Developed by: Research Group Parallel and Distributed Systems
 *               Department of Mathematics and Computer Science
 *               University of Muenster, Germany
 *               <http://pvs.uni-muenster.de/>
 * 
 * Copyright (C) 2013  Philipp Kegel <philipp.kegel@uni-muenster.de>
 *
 * dOpenCL is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * dOpenCL is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with dOpenCL. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Permission to use dOpenCL for scientific, non-commercial work is
 * granted under the terms of the dOpenCL Academic License provided
 * appropriate credit is given. See the dOpenCL Academic License for
 * more details.
 * 
 * You should have received a copy of the dOpenCL Academic License
 * along with dOpenCL. If not, see <http://dopencl.uni-muenster.de/>.
 ******************************************************************************/

/*
 * Session.h
 *
 *  Created on: 03.02.2011
 *      Author: Philipp Kegel
 */

#ifndef SESSION_H_
#define SESSION_H_

#include <cstddef>
#include <map>
#include <vector>

#include <boost/shared_ptr.hpp>

#define __CL_ENABLE_EXCEPTIONS
#include <CL/cl.hpp>

#include <dcl/CLCommandListener.h>
#include <dcl/CommandQueue.h>
#include <dcl/ComputeNode.h>
#include <dcl/Context.h>
#include <dcl/DCLTypes.h>
#include <dcl/Event.h>
#include <dcl/Host.h>
#include <dcl/Kernel.h>
#include <dcl/Memory.h>
#include <dcl/Program.h>

namespace dcld {

/* forward declarations */
class Context;
class DeviceManager;

/******************************************************************************/

/**
 * The Session class saves the state of a connected host, i.e., it manages all
 * OpenCL objects that have been created on a compute node to implement a
 * corresponding OpenCL object on the host.
 *
 * All commands issued by the host are forwarded to a session.
 * The session resolves the object IDs contained in these commands into a
 * OpenCL objects to execute the command on these objects.
 */
/* TODO Make Session a factory class for dOpenCL objects.
 * A session should only provide constructor and destructor methods for dOpenCL
 * objects and should also keep ownership of these objects.
 * However, lookup facilities should be moved to CLObjectRegistry.
 */
class Session:
    public dcl::CLCommandListener {
public:
    Session(
            DeviceManager&                      /* deviceManager */,
            dcl::Host&                          /* host */);
    virtual ~Session();

	/* Command listener APIs */
	void getDeviceIDs(
			cl_device_type						/* device_type */,
			std::vector<dcl::dcl_object_id>&	/* device_ids */) const;
	void getDeviceInfo(
			dcl::dcl_object_id	/* device_id */,
			cl_device_info	 	/* param_name */,
			void **				/* param_value */,
			size_t *			/* param_value_size */) const;

	dcl::Context * createContext(
	        dcl::dcl_object_id                     /* id */,
	        const std::vector<dcl::ComputeNode *>& /* computeNodes */,
	        const std::vector<dcl::dcl_object_id>& /* device_ids */);
	void releaseContext(
			dcl::dcl_object_id	/* context_id */);

	dcl::CommandQueue * createCommandQueue(
			dcl::dcl_object_id			/* context_id */,
			dcl::dcl_object_id			/* device_id */,
			dcl::dcl_object_id			/* command_queue_id */,
			cl_command_queue_properties	/* properties */);
	void releaseCommandQueue(
			dcl::dcl_object_id	/* command_queue_id */);

	dcl::Buffer * createBuffer(
			dcl::dcl_object_id	/* context_id */,
			dcl::dcl_object_id	/* buffer_id */,
			cl_mem_flags		/* flags */,
			size_t				/* size */);
	void releaseMemObject(
			dcl::dcl_object_id	/* mem_object_id */);

	dcl::Program * createProgram(
			dcl::dcl_object_id	/* context_id */,
			dcl::dcl_object_id	/* program_id */,
			size_t              /* length */);
	dcl::Program * createProgram(
			dcl::dcl_object_id						/* context_id */,
			dcl::dcl_object_id						/* program_id */,
			const std::vector<dcl::dcl_object_id>&	/* device_id_list */,
			const std::vector<size_t>&				/* lengths */,
			std::vector<cl_int> *					/* binary_status */);
	void releaseProgram(
			dcl::dcl_object_id	/* program_id */);
	void buildProgram(
			dcl::dcl_object_id						/* program_id */,
			const std::vector<dcl::dcl_object_id> *	/* device_id_list */,
			const char *							/* options */,
            dcl::dcl_object_id                      /* program_build_id */);

	dcl::Kernel * createKernel(
			dcl::dcl_object_id	/* program_id */,
			dcl::dcl_object_id	/* kernel_id */,
			const char *		/* name */);
	void createKernelsInProgram(
			dcl::dcl_object_id						/* program_id */,
			const std::vector<dcl::dcl_object_id>&	/* kernel_ids */);
	void releaseKernel(
			dcl::dcl_object_id	/* kernel_id */);
	void setKernelArg(
			dcl::dcl_object_id          /* kernel_id */,
			cl_uint                     /* arg_index */,
			dcl::dcl_kernel_arg_type    /* arg_type */,
			size_t                      /* arg_size */,
			const void *                /* arg_value */);

	void getKernelInfo(
			dcl::dcl_object_id  /* kernel_id */,
			cl_kernel_info      /* param_name */,
			void **             /* param_value */,
			size_t *            /* param_value_size */) const;
    void getKernelWorkGroupInfo(
            dcl::dcl_object_id          /* device_id */,
            dcl::dcl_object_id          /* kernel_id */,
            cl_kernel_work_group_info   /* param_name */,
            void **                     /* param_value */,
            size_t *                    /* param_value_size */) const;

    dcl::Event * createEvent(
            dcl::dcl_object_id                      /* context_id */,
            dcl::dcl_object_id                      /* command_id */,
            const std::vector<dcl::dcl_object_id>&  /* memoryObjectIds */);
	dcl::UserEvent * createUserEvent(
			dcl::dcl_object_id  /* context_id */,
			dcl::dcl_object_id  /* event_id */);
	void releaseEvent(
			dcl::dcl_object_id  /* event_id */);
	void setUserEventStatus(
			dcl::dcl_object_id  /* event_id */,
			cl_int              /* status */);
	void getEventProfilingInfo(
			dcl::dcl_object_id  /* event */,
			cl_profiling_info   /* param_name */,
			cl_ulong&           /* param_value */);

	void flush(
			dcl::dcl_object_id  /* command_queue_id */);
	void finish(
			dcl::dcl_object_id  /* command_queue_id */);

	void enqueueCopyBuffer(
			dcl::dcl_object_id                          /* command_queue_id */,
			dcl::dcl_object_id                          /* command_id */,
			dcl::dcl_object_id                          /* src_buffer_id */,
			dcl::dcl_object_id                          /* dst_buffer_id */,
			size_t                                      /* src_offset */,
			size_t                                      /* dst_offset */,
			size_t                                      /* size */,
			const std::vector<dcl::dcl_object_id> *     /* event_id_wait_list */,
			bool                                        /* event */);
	void enqueueReadBuffer(
			dcl::dcl_object_id                          /* command_queue_id */,
			dcl::dcl_object_id                          /* command_id */,
			dcl::dcl_object_id                          /* buffer_id */,
			bool                                        /* blocking_read */,
			size_t                                      /* offset */,
			size_t                                      /* size */,
			const std::vector<dcl::dcl_object_id> *     /* event_id_wait_list */,
			bool                                        /* event */);
	void enqueueWriteBuffer(
			dcl::dcl_object_id                          /* command_queue_id */,
			dcl::dcl_object_id                          /* command_id */,
			dcl::dcl_object_id                          /* buffer_id */,
			bool                                        /* blocking_write */,
			size_t                                      /* offset */,
			size_t                                      /* size */,
			const std::vector<dcl::dcl_object_id> *     /* event_id_wait_list */,
			bool                                        /* event */);
    void enqueueMapBuffer(
            dcl::dcl_object_id                          /* command_queue_id */,
            dcl::dcl_object_id                          /* command_id */,
            dcl::dcl_object_id                          /* buffer_id */,
            bool                                        /* blocking_map */,
            cl_map_flags                                /* map_flags */,
            size_t                                      /* offset */,
            size_t                                      /* size */,
            const std::vector<dcl::dcl_object_id> *     /* event_id_wait_list */,
            bool                                        /* event */);
    void enqueueUnmapMemObject(
            dcl::dcl_object_id                          /* command_queue_id */,
            dcl::dcl_object_id                          /* command_id */,
            dcl::dcl_object_id                          /* buffer_id */,
            const std::vector<dcl::dcl_object_id> *     /* event_id_wait_list */,
            bool                                        /* event */);
	void enqueueNDRangeKernel(
			dcl::dcl_object_id                          /* command_queue_id */,
			dcl::dcl_object_id                          /* command_id */,
			dcl::dcl_object_id                          /* kernel_id */,
			const cl::NDRange&                          /* offset */,
			const cl::NDRange&                          /* global */,
			const cl::NDRange&                          /* local */,
			const std::vector<dcl::dcl_object_id> *     /* event_id_wait_list */,
			bool                                        /* event */);

	void enqueueMarker(
			dcl::dcl_object_id                      /* command_queue_id */,
			dcl::dcl_object_id                      /* command_id */,
            const std::vector<dcl::dcl_object_id> * /* event_id_wait_list */,
            bool                                    /* event */);
#if defined(CL_USE_DEPRECATED_OPENCL_1_1_APIS) || (defined(CL_VERSION_1_1) && !defined(CL_VERSION_1_2))
	void enqueueWaitForEvents(
			dcl::dcl_object_id                      /* command_queue_id */,
			const std::vector<dcl::dcl_object_id>&  /* event_id_wait_list */);
#endif // #if defined(CL_USE_DEPRECATED_OPENCL_1_1_APIS)
	void enqueueBarrier(
            dcl::dcl_object_id                      /* command_queue_id */,
            dcl::dcl_object_id                      /* command_id */,
            const std::vector<dcl::dcl_object_id> * /* event_id_wait_list */,
            bool                                    /* event */);

    void enqueueBroadcastBuffer(
            const std::vector<dcl::dcl_object_id>&  /* command_queue_ids */,
            dcl::dcl_object_id                      /* command_id */,
            dcl::dcl_object_id                      /* srcId */,
            const std::vector<dcl::dcl_object_id>&  /* dstIds */,
            size_t                                  /* srcOffset */,
            const std::vector<size_t>               /* dstOffsets */,
            size_t                                  /* cb */,
            const std::vector<dcl::dcl_object_id> * /* event_id_wait_list */,
            bool                                    /* event */);
    void enqueueReduceBuffer(
            dcl::dcl_object_id                      /* command_queue_id */,
            dcl::dcl_object_id                      /* command_id */,
            const std::vector<dcl::dcl_object_id>&  /* srcIds */,
            dcl::dcl_object_id                      /* dstId */,
            dcl::dcl_object_id                      /* kernel_id */,
            const cl::NDRange&                      /* offset */,
            const cl::NDRange&                      /* global */,
            const cl::NDRange&                      /* local */,
            const std::vector<dcl::dcl_object_id> * /* event_id_wait_list */,
            bool                                    /* event */);

    /*
	 * Object lookup methods
	 */
    /* TODO Move object lookup to CLObjectRegistry of session's host
     * The session should only maintain sets of objects, rather than maps to
     * keep these objects alive. */
	boost::shared_ptr<Context> getContext(dcl::dcl_object_id) const;
	boost::shared_ptr<dcl::CommandQueue> getCommandQueue(dcl::dcl_object_id) const;
    boost::shared_ptr<dcl::Memory> getMemory(dcl::dcl_object_id) const;
	boost::shared_ptr<dcl::Buffer> getBuffer(dcl::dcl_object_id) const;
	cl::Sampler getSampler(dcl::dcl_object_id) const;
	boost::shared_ptr<dcl::Program> getProgram(dcl::dcl_object_id) const;
	boost::shared_ptr<dcl::Kernel> getKernel(dcl::dcl_object_id) const;
	void getEventWaitList(
	        const std::vector<dcl::dcl_object_id>&         /* eventIdWaitList */,
	        std::vector<boost::shared_ptr<dcl::Event> >&   /* eventWaitList */) const;
    boost::shared_ptr<dcl::Event> getEvent(dcl::dcl_object_id) const;
	boost::shared_ptr<dcl::UserEvent> getUserEvent(dcl::dcl_object_id) const;

private:
	/* Session must not be copied or assigned */
	Session(const Session&);
	Session& operator=(const Session&);

    /* TODO Remove device manager from Session */
	DeviceManager& _deviceManager;
	dcl::Host& _host;

	/** Context list */
	std::map<dcl::dcl_object_id, boost::shared_ptr<Context> > _contexts;
	/** Buffer list */
	std::map<dcl::dcl_object_id, boost::shared_ptr<dcl::Buffer> > _buffers;
	/** Sampler list */
	std::map<dcl::dcl_object_id, cl::Sampler> _samplers;
	/** Command queue list */
	std::map<dcl::dcl_object_id, boost::shared_ptr<dcl::CommandQueue> > _command_queues;
	/** Program list */
	std::map<dcl::dcl_object_id, boost::shared_ptr<dcl::Program> > _programs;
	/** Kernel list */
	std::map<dcl::dcl_object_id, boost::shared_ptr<dcl::Kernel> > _kernels;
	/** Event lists */
	std::map<dcl::dcl_object_id, boost::shared_ptr<dcl::Event> > _events;
	/** User event list */
	std::map<dcl::dcl_object_id, boost::shared_ptr<dcl::UserEvent> > _userEvents;
};

} /* namespace dcld */

#endif /* SESSION_H_ */
