/******************************************************************************
 * This file is part of dOpenCL.
 * 
 * dOpenCL is an implementation of the OpenCL application programming
 * interface for distributed systems. See <http://dopencl.uni-muenster.de/>
 * for more information.
 * 
 * Developed by: Research Group Parallel and Distributed Systems
 *               Department of Mathematics and Computer Science
 *               University of Muenster, Germany
 *               <http://pvs.uni-muenster.de/>
 * 
 * Copyright (C) 2013  Philipp Kegel <philipp.kegel@uni-muenster.de>
 *
 * dOpenCL is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * dOpenCL is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with dOpenCL. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Permission to use dOpenCL for scientific, non-commercial work is
 * granted under the terms of the dOpenCL Academic License provided
 * appropriate credit is given. See the dOpenCL Academic License for
 * more details.
 * 
 * You should have received a copy of the dOpenCL Academic License
 * along with dOpenCL. If not, see <http://dopencl.uni-muenster.de/>.
 ******************************************************************************/

/*
 * dOpenCLd.cpp
 *
 *  Created on: 16.01.2011
 *      Author: Philipp Kegel
 *
 * Implementation of dOpenCL daemon
 */

#include <ostream>
#include <stdexcept>
#include <string>
#include <map>
#include <utility>

#include <boost/thread/condition_variable.hpp>
#include <boost/thread/locks.hpp>
#include <boost/thread/mutex.hpp>

#include <dcl/ComputeNode.h>
#include <dcl/ConnectionManager.h>
#include <dcl/Host.h>

#include <dcl/util/Logger.h>

#include "Session.h"
#include "DeviceManager.h"
#include "dOpenCLd.h"

namespace dcld {

dOpenCLd::dOpenCLd(const std::string& url, const std::string *platform) :
	_connectionManager(dcl::ComputeNodeConnectionManager::create(url)),
	_deviceManager(platform) {
	_connectionManager->addConnectionListener(*this);
}

dOpenCLd::~dOpenCLd() {
	_connectionManager->removeConnectionListener(*this);
	delete _connectionManager;
	/* TODO Delete sessions
	 * Use map of shared pointers */
}

void dOpenCLd::run() {
	boost::lock_guard<boost::mutex> lock(_interruptMutex);

	_connectionManager->start();

    /* TODO Process connections in run method rather than in the callback methods */

	/* Suspend the calling (main) thread to prevent the daemon from exiting */
	_interrupt = false;
	while (!_interrupt) _interrupted.wait(_interruptMutex);

	_connectionManager->stop();

	dcl::util::Logger << dcl::util::Info
	        << "Shutting down dOpenCL daemon ..." << std::endl;
}

void dOpenCLd::terminate() {
	boost::lock_guard<boost::mutex> lock(_interruptMutex);
	/* Resume the main thread which will terminate the daemon */
	_interrupt = true;
	_interrupted.notify_all();
}

/******************************************************************************
 * dOpenCL connection listener API
 ******************************************************************************/

bool dOpenCLd::connected(dcl::Host& host) {
	boost::lock_guard<boost::mutex> lock(_sessionsMutex);
	std::map<const dcl::Host *, Session *>::iterator i = _sessions.find(&host);

	if (i == _sessions.end()) {
		Session *session = new Session(_deviceManager, host); // create new session

		/*
		 * Add session to list
		 */
		bool inserted = _sessions.insert(std::make_pair(&host, session)).second;
		assert(inserted);

		dcl::util::Logger << dcl::util::Info
				<< "Session created (host='" << host.url() << "')" << std::endl;
	} else {
		/* TODO Handle reestablished connection */
	}

	return true;
}

void dOpenCLd::disconnected(dcl::Host& host) {
	boost::lock_guard<boost::mutex> lock(_sessionsMutex);
	std::map<const dcl::Host *, Session *>::iterator i = _sessions.find(&host);

	/* TODO Retain session for future reconnect
	 * Do not drop a session if its process has lost connection. Rather preserve
	 * the session to allow the process to reconnect and continue the session.
	 * Only drop sessions on request, or after a certain timeout has expired. */

	if (i != _sessions.end()) {
		/*
		 * Remove session from list
		 */
		delete i->second;
		_sessions.erase(i);

		dcl::util::Logger << dcl::util::Info
				<< "Session destroyed (host='" << host.url() << "')" << std::endl;
	}

	/* TODO Delete host when disconnected
	 * The ownership of the host must be clarified, as the host may be used by
	 * multiple connection listeners.
	 * Possible solution: return bool (see connected) to indicate if host should
	 * be deleted. If all listeners agree on deleting the host, it is deleted
	 * by the communication manager. */
}

bool dOpenCLd::connected(dcl::ComputeNode& computeNode) {
	/* TODO Implement dOpenCLd::connected for compute nodes */
	return false;
}

void dOpenCLd::disconnected(dcl::ComputeNode& computeNode) {
	/* TODO Implement dOpenCLd::disconnected for compute nodes */
}

} /* namespace dcld */
