/******************************************************************************
 * This file is part of dOpenCL.
 * 
 * dOpenCL is an implementation of the OpenCL application programming
 * interface for distributed systems. See <http://dopencl.uni-muenster.de/>
 * for more information.
 * 
 * Developed by: Research Group Parallel and Distributed Systems
 *               Department of Mathematics and Computer Science
 *               University of Muenster, Germany
 *               <http://pvs.uni-muenster.de/>
 * 
 * Copyright (C) 2013  Philipp Kegel <philipp.kegel@uni-muenster.de>
 *
 * dOpenCL is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * dOpenCL is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with dOpenCL. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Permission to use dOpenCL for scientific, non-commercial work is
 * granted under the terms of the dOpenCL Academic License provided
 * appropriate credit is given. See the dOpenCL Academic License for
 * more details.
 * 
 * You should have received a copy of the dOpenCL Academic License
 * along with dOpenCL. If not, see <http://dopencl.uni-muenster.de/>.
 ******************************************************************************/

/*!
 * \file    CLCommandListener.h
 *
 * \date    2011-08-07
 * \author  Philipp Kegel
 */

#ifndef CLCOMMANDLISTENER_H_
#define CLCOMMANDLISTENER_H_

#include <cstddef>
#include <vector>

/* TODO Remove OpenCL C++ binding dependency */
#include <CL/cl.hpp>

#include "DCLTypes.h"

namespace dcl {

/* forward declaration */
class Buffer;
class CommandQueue;
class ComputeNode;
class Context;
class Event;
class Kernel;
class Program;
class UserEvent;

/******************************************************************************/

/* TODO Split CLCommandListener into one interface class per OpenCL object */
/*!
 * \brief A interface for listening for incoming OpenCL commands from the dOpenCL ICD
 */
class CLCommandListener {
public:
	virtual ~CLCommandListener() { }

	virtual void getDeviceIDs(
            cl_device_type              /* device_type */,
            std::vector<dcl_object_id>& /* device_ids */) const = 0;
	/* TODO Adopt C++ API of Device::getInfo from OpenCL C++ binding */
	virtual void getDeviceInfo(
			dcl_object_id   /* device_id */,
            cl_device_info  /* param_name */,
            void **         /* param_value */,
            size_t *        /* param_value_size */) const = 0;

	virtual Context * createContext(
	        dcl_object_id                      /* context_id */,
	        const std::vector<ComputeNode *>&  /* computeNodes */,
	        const std::vector<dcl_object_id>&  /* device_ids */) = 0;
	virtual void releaseContext(
			dcl_object_id	/* context_id */) = 0;

    virtual CommandQueue * createCommandQueue(
    		dcl_object_id               /* context_id */,
			dcl_object_id               /* device_id */,
	        dcl_object_id               /* command_queue_id */,
	        cl_command_queue_properties /* properties */) = 0;
	virtual void releaseCommandQueue(
			dcl_object_id	/* command_queue_id */) = 0;

	virtual Buffer * createBuffer(
			dcl_object_id	/* context_id */,
            dcl_object_id	/* buffer_id */,
            cl_mem_flags	/* flags */,
            size_t          /* size */) = 0;
	virtual void releaseMemObject(
			dcl_object_id	/* mem_object_id */) = 0;

	virtual Program * createProgram(
			dcl_object_id	/* context_id */,
            dcl_object_id	/* program_id */,
			size_t			/* lengths */) = 0;
	virtual Program * createProgram(
			dcl_object_id						/* context_id */,
            dcl_object_id						/* program_id */,
            const std::vector<dcl_object_id>&	/* device_id_list */,
			const std::vector<size_t>&			/* lengths */,
			std::vector<cl_int> *				/* binary_status */) = 0;
	virtual void releaseProgram(
			dcl_object_id	/* program_id */) = 0;
	virtual void buildProgram(
			dcl_object_id						/* program_id */,
			const std::vector<dcl_object_id> *	/* device_id_list */,
			const char *						/* options */,
            dcl_object_id                       /* program_build_id */) = 0;

    virtual Kernel * createKernel(
    		dcl_object_id	/* program_id */,
            dcl_object_id	/* kernel_id */,
            const char *	/* name */) = 0;
    virtual void createKernelsInProgram(
    		dcl_object_id						/* program_id */,
    		const std::vector<dcl_object_id>&	/* kernel_ids */) = 0;
    virtual void releaseKernel(
    		dcl_object_id	/* kernel_id */) = 0;
	virtual void getKernelInfo(
			dcl_object_id   /* kernel_id */,
			cl_kernel_info  /* param_name */,
			void **         /* param_value */,
			size_t *        /* param_value_size */) const = 0;
    virtual void getKernelWorkGroupInfo(
            dcl_object_id               /* device_id */,
            dcl_object_id               /* kernel_id */,
            cl_kernel_work_group_info   /* param_name */,
            void **                     /* param_value */,
            size_t *                    /* param_value_size */) const = 0;
	/* TODO Adopt C++ API of Kernel::setArg from OpenCL C++ binding */
    virtual void setKernelArg(
    		dcl_object_id     	    /* kernel_id */,
            cl_uint                 /* arg_index */,
            dcl_kernel_arg_type     /* arg_type */,
            size_t                  /* arg_size */,
            const void *            /* arg_value */) = 0;

	virtual Event * createEvent(
            dcl_object_id                       /* context_id */,
            dcl_object_id                       /* event_id */,
            const std::vector<dcl_object_id>&   /* memObjectIds */) = 0;
	virtual UserEvent * createUserEvent(
			dcl_object_id   /* context_id */,
            dcl_object_id   /* event_id */) = 0;
	virtual void releaseEvent(
			dcl_object_id   /* event_id */) = 0;
	virtual void setUserEventStatus(
			dcl_object_id   /* event_id */,
            cl_int          /* status */) = 0;
	virtual void getEventProfilingInfo(
			dcl_object_id       /* event */,
			cl_profiling_info   /* param_name */,
			cl_ulong&           /* param_value */) = 0;

	virtual void flush(
			dcl_object_id	/* command_queue_id */) = 0;
	virtual void finish(
			dcl_object_id	/* command_queue_id */) = 0;

	virtual void enqueueCopyBuffer(
			dcl_object_id						/* command_queue_id */,
			dcl_object_id						/* command_id */,
			dcl_object_id						/* src_buffer_id */,
			dcl_object_id						/* dst_buffer_id */,
			size_t								/* src_offset */,
			size_t								/* dst_offset */,
			size_t								/* size */,
			const std::vector<dcl_object_id> *	/* event_id_wait_list */,
			bool								/* event */) = 0;
	virtual void enqueueReadBuffer(
			dcl_object_id						/* command_queue_id */,
			dcl_object_id						/* command_id */,
			dcl_object_id						/* buffer_id */,
			bool                                /* blocking_read */,
			size_t								/* offset */,
			size_t								/* size */,
			const std::vector<dcl_object_id> *	/* event_id_wait_list */,
			bool								/* event */) = 0;
	virtual void enqueueWriteBuffer(
			dcl_object_id						/* command_queue_id */,
			dcl_object_id						/* command_id */,
			dcl_object_id						/* buffer_id */,
			bool                                /* blocking_write */,
			size_t								/* offset */,
			size_t								/* size */,
			const std::vector<dcl_object_id> *	/* event_id_wait_list */,
			bool								/* event */) = 0;
    virtual void enqueueMapBuffer(
            dcl_object_id                       /* command_queue_id */,
            dcl_object_id                       /* command_id */,
            dcl_object_id                       /* buffer_id */,
            bool                                /* blocking_map */,
            cl_map_flags                        /* map_flags */,
            size_t                              /* offset */,
            size_t                              /* size */,
            const std::vector<dcl_object_id> *  /* event_id_wait_list */,
            bool                                /* event */) = 0;
    virtual void enqueueUnmapMemObject(
            dcl_object_id                       /* command_queue_id */,
            dcl_object_id                       /* command_id */,
            dcl_object_id                       /* buffer_id */,
            const std::vector<dcl_object_id> *  /* event_id_wait_list */,
            bool                                /* event */) = 0;
    virtual void enqueueNDRangeKernel(
			dcl_object_id						/* command_queue_id */,
			dcl_object_id						/* command_id */,
			dcl_object_id						/* kernel_id */,
			const cl::NDRange&					/* offset */,
			const cl::NDRange&					/* global */,
			const cl::NDRange&					/* local */,
			const std::vector<dcl_object_id> *	/* event_id_wait_list */,
			bool								/* event */) = 0;
	virtual void enqueueMarker(
			dcl_object_id						/* command_queue_id */,
			dcl_object_id						/* command_id */,
            const std::vector<dcl_object_id> *  /* event_id_wait_list */,
            bool                                /* event */) = 0;
#if defined(CL_USE_DEPRECATED_OPENCL_1_1_APIS) || (defined(CL_VERSION_1_1) && !defined(CL_VERSION_1_2))
	virtual void enqueueWaitForEvents(
			dcl_object_id						/* command_queue_id */,
			const std::vector<dcl_object_id>&	/* event_id_list */) = 0;
#endif // #if defined(CL_USE_DEPRECATED_OPENCL_1_1_APIS)
	virtual void enqueueBarrier(
            dcl_object_id                       /* command_queue_id */,
            dcl_object_id                       /* command_id */,
            const std::vector<dcl_object_id> *  /* event_id_wait_list */,
            bool                                /* event */) = 0;

    virtual void enqueueBroadcastBuffer(
            const std::vector<dcl_object_id>&   /* command_queue_ids */,
            dcl_object_id                       /* command_id */,
            dcl_object_id                       /* srcId */,
            const std::vector<dcl_object_id>&   /* dstIds */,
            size_t                              /* srcOffset */,
            const std::vector<size_t>           /* dstOffsets */,
            size_t                              /* cb */,
            const std::vector<dcl_object_id> *  /* event_id_wait_list */,
            bool                                /* event */) = 0;
    virtual void enqueueReduceBuffer(
            dcl_object_id                       /* command_queue_id */,
            dcl_object_id                       /* command_id */,
            const std::vector<dcl_object_id>&   /* srcIds */,
            dcl_object_id                       /* dstId */,
            dcl_object_id                       /* kernel_id */,
            const cl::NDRange&                  /* offset */,
            const cl::NDRange&                  /* global */,
            const cl::NDRange&                  /* local */,
            const std::vector<dcl_object_id> *  /* event_id_wait_list */,
            bool                                /* event */) = 0;
};

} /* namespace dcl */

#endif /* CLCOMMANDLISTENER_H_ */
