/******************************************************************************
 * This file is part of dOpenCL.
 * 
 * dOpenCL is an implementation of the OpenCL application programming
 * interface for distributed systems. See <http://dopencl.uni-muenster.de/>
 * for more information.
 * 
 * Developed by: Research Group Parallel and Distributed Systems
 *               Department of Mathematics and Computer Science
 *               University of Muenster, Germany
 *               <http://pvs.uni-muenster.de/>
 * 
 * Copyright (C) 2013  Philipp Kegel <philipp.kegel@uni-muenster.de>
 *
 * dOpenCL is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * dOpenCL is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with dOpenCL. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Permission to use dOpenCL for scientific, non-commercial work is
 * granted under the terms of the dOpenCL Academic License provided
 * appropriate credit is given. See the dOpenCL Academic License for
 * more details.
 * 
 * You should have received a copy of the dOpenCL Academic License
 * along with dOpenCL. If not, see <http://dopencl.uni-muenster.de/>.
 ******************************************************************************/

/**
 * @file    ComputeNode.h
 *
 * @date    30.10.2011
 * @author  Philipp Kegel
 *
 * C++ API declarations for dOpenCL communication layer
 */

#ifndef DCL_COMPUTENODE_H_
#define DCL_COMPUTENODE_H_

#include <vector>

/* TODO Remove GCF headers from ComputeNode interface
 * These file introduces a GCF dependency */
#include <gcf/GCFTypedefs.h>
#include <gcf/Serializable.h>

#include "Process.h"

/* TODO Remove message classes from ComputeNode interface
 * These classes introduce a GCF dependency */
#include "comm/Request.h"
#include "comm/Response.h"

namespace dcl {

/* forward declaration */
class Device;

/******************************************************************************/

class ComputeNode :
public virtual Process { // extends Process
public:
    virtual ~ComputeNode() { }

    /**
     * @brief Obtains a list of devices that are hosted by the compute node.
     *
     * @param[out]  devices a list of devices associated with this compute node
     */
    virtual void getDevices(
            std::vector<Device *>&  /* devices */) = 0;

    /**
     * @brief Sends a request message to this compute node.
     *
     * @param[in]  request  the request to send
     */
    virtual void sendRequest(
            dcl::comm::Request&         /* request */) const = 0;

    /**
     * @brief Waits for this compute node's response to the specified request.
     *
     * This methods throws a dcl::ProtocolException, if the response does not
     * have the expected type.
     *
     * @param[in]  request      the sent request
     * @param[in]  responseType the expected type of response
     * @return the received response
     */
    virtual dcl::comm::Response * awaitResponse(
            const dcl::comm::Request&   /* request */,
            GCFTypeID_t                 /* responseType */) = 0;
    /**
     * @brief Waits for this compute node's response to the specified request.
     *
     * The expected response is Response::TYPE, i.e., a simple response which
     * only contains an error code.
     *
     * @param[in]  request  the sent request
     */
    virtual void awaitResponse(
            const dcl::comm::Request&   /* request */) = 0;

    /**
     * @brief Execute a command on this compute node.
     *
     * This methods throws a dcl::ProtocolException, if the response does not
     * have the expected type.
     *
     * @param[in]  request      the request to send
     * @param[in]  responseType the expected type of response
     * @return the received response
     */
    virtual dcl::comm::Response * executeCommand(
            dcl::comm::Request&         /* request */,
            GCFTypeID_t                 /* responseType */) = 0;

    /**
     * @brief Execute a command on this compute node.
     *
     * The expected response for this command is Response::TYPE, i.e., a simple
     * response which only contains an error code.
     *
     * @param[in]  request  the request to send
     */
    virtual void executeCommand(
            dcl::comm::Request&         /* request */) = 0;
};

/******************************************************************************/

/**
 * The following function are free functions to avoid hiding of virtual methods.
 * E.g., if sendMessage becomes a static member function of ComputeNode, it
 * hides the virtual sendMessage which ComputeNode inherits from Process.
 */

/* TODO Move ComputeNode::sendMessage to ComputeNodeImpl::sendMessage */
/**
 * @brief Sends a message to multiple compute nodes.
 *
 * @param[in]  computeNodes a list of compute nodes to send a message to
 * @param[in]  message      the message to send
 */
void sendMessage(
        const std::vector<ComputeNode *>&,
        gcf::Serializable&  /* message */);

/* TODO Discard ComputeNode::sendRequest
 * Use ComputeNodeImpl::sendRequest instead */
/**
 * @brief Sends a request message to multiple compute nodes.
 *
 * @param[in]  computeNodes a list of compute nodes to send a request to
 * @param[in]  request      the request to send
 */
void sendRequest(
        const std::vector<ComputeNode *>&,
        dcl::comm::Request&);

/* TODO Discard ComputeNode::executeCommand
 * Use ComputeNodeImpl::executeCommand instead */
/**
 * @brief Execute a command on multiple compute nodes.
 *
 * @param[in]  computeNodes a list of compute nodes to execute a command on
 * @param[in]  request      the request to send
 * @param[in]  responseType the expected type of responses
 * @param[out] responses    the received responses
 */
void executeCommand(
        const std::vector<ComputeNode *>&,
        dcl::comm::Request&,
        GCFTypeID_t responseType = dcl::comm::Response::TYPE,
        std::vector<dcl::comm::Response *> *responses = 0);

} /* namespace dcl */

#endif /* DCL_COMPUTENODE_H_ */
