/******************************************************************************
 * This file is part of dOpenCL.
 * 
 * dOpenCL is an implementation of the OpenCL application programming
 * interface for distributed systems. See <http://dopencl.uni-muenster.de/>
 * for more information.
 * 
 * Developed by: Research Group Parallel and Distributed Systems
 *               Department of Mathematics and Computer Science
 *               University of Muenster, Germany
 *               <http://pvs.uni-muenster.de/>
 * 
 * Copyright (C) 2013  Philipp Kegel <philipp.kegel@uni-muenster.de>
 *
 * dOpenCL is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * dOpenCL is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with dOpenCL. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Permission to use dOpenCL for scientific, non-commercial work is
 * granted under the terms of the dOpenCL Academic License provided
 * appropriate credit is given. See the dOpenCL Academic License for
 * more details.
 * 
 * You should have received a copy of the dOpenCL Academic License
 * along with dOpenCL. If not, see <http://dopencl.uni-muenster.de/>.
 ******************************************************************************/

/*
 * EnqueueCollective.h
 *
 *  Created on: 02.07.2012
 *      Author: Philipp Kegel
 */

#ifndef ENQUEUECOLLECTIVE_H_
#define ENQUEUECOLLECTIVE_H_

#include <cstddef>
#include <vector>

#include <gcf/GCFTypedefs.h>

#include <dcl/DCLTypes.h>

#include "Request.h"

namespace dcl {

namespace comm {

class EnqueueBroadcastBuffer : public Request {
public:
	EnqueueBroadcastBuffer(
			const std::vector<dcl_object_id>&  /* commandQueueIds */,
			dcl_object_id                      /* commandId */,
			dcl_object_id                      /* srcBufferId */,
			const std::vector<dcl_object_id>&  /* dstBufferIds */,
			size_t                             /* srcOffset */,
			const std::vector<size_t>&         /* dstOffsets */,
			size_t                             /* cb */,
			const std::vector<dcl_object_id> *eventIdWaitList = 0,
			bool event = false);
	EnqueueBroadcastBuffer(const EnqueueBroadcastBuffer&);

	const std::vector<dcl_object_id>& commandQueueIds() const;
    dcl_object_id commandId() const;
    dcl_object_id srcBufferId() const;
    const std::vector<dcl_object_id>& dstBufferIds() const;
    size_t srcOffset() const;
    void dstOffsets(
    		std::vector<size_t>& /* dstOffsets */) const;
    size_t cb() const;
    const std::vector<dcl_object_id>& eventIdWaitList(void) const;
    bool event() const;

    DECLARE_SERIALIZABLE_PUBLIC(EnqueueBroadcastBuffer, 100 + ENQUEUE_BROADCAST_BUFFER)

private:
    std::vector<dcl_object_id> _commandQueueIds;
    dcl_object_id _commandId;
    dcl_object_id _srcBufferId;
    std::vector<dcl_object_id> _dstBufferIds;
    gcf_uint32 _srcOffset;
    std::vector<gcf_uint32> _dstOffsets;
    gcf_uint32 _cb;
    std::vector<dcl_object_id> _eventIdWaitList;
    gcf_bool _event;

	DECLARE_SERIALIZABLE_EVENT_PRIVATE(EnqueueBroadcastBuffer)
};

/******************************************************************************/

class EnqueueReduceBuffer : public Request {
public:
	EnqueueReduceBuffer(
			dcl_object_id                      /* commandQueueId */,
			dcl_object_id                      /* commandId */,
			const std::vector<dcl_object_id>&  /* srcIds */,
			dcl_object_id                      /* dstId */,
			dcl_object_id                      /* kernelId */,
			const std::vector<size_t>&         /* offset */,
			const std::vector<size_t>&         /* global */,
			const std::vector<size_t>&         /* local */,
			const std::vector<dcl_object_id> *eventIdWaitList = 0,
			bool event = false);
	EnqueueReduceBuffer(const EnqueueReduceBuffer&);

	dcl_object_id commandQueueId() const;
	dcl_object_id commandId() const;
	const std::vector<dcl_object_id>& srcIds() const;
	dcl_object_id dstId() const;
	dcl_object_id kernelId() const;

	void offset(std::vector<size_t>&) const;
	void global(std::vector<size_t>&) const;
	void local(std::vector<size_t>&) const;

    const std::vector<dcl_object_id>& eventIdWaitList(void) const;
	bool event() const;

	DECLARE_SERIALIZABLE_PUBLIC(EnqueueReduceBuffer, 100 + ENQUEUE_REDUCE_BUFFER)

private:
	dcl_object_id _commandQueueId;
	dcl_object_id _commandId;
	std::vector<dcl_object_id> _srcIds;
	dcl_object_id _dstId;
	dcl_object_id _kernelId;
	std::vector<gcf_uint32> _offset;
	std::vector<gcf_uint32> _global;
	std::vector<gcf_uint32> _local;
    std::vector<dcl_object_id> _eventIdWaitList;
	gcf_bool _event;

	DECLARE_SERIALIZABLE_EVENT_PRIVATE(EnqueueReduceBuffer)
};

} /* namespace comm */

} /* namespace dcl */

#endif /* ENQUEUECOLLECTIVE_H_ */
