/******************************************************************************
 * This file is part of dOpenCL.
 * 
 * dOpenCL is an implementation of the OpenCL application programming
 * interface for distributed systems. See <http://dopencl.uni-muenster.de/>
 * for more information.
 * 
 * Developed by: Research Group Parallel and Distributed Systems
 *               Department of Mathematics and Computer Science
 *               University of Muenster, Germany
 *               <http://pvs.uni-muenster.de/>
 * 
 * Copyright (C) 2013  Philipp Kegel <philipp.kegel@uni-muenster.de>
 *
 * dOpenCL is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * dOpenCL is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with dOpenCL. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Permission to use dOpenCL for scientific, non-commercial work is
 * granted under the terms of the dOpenCL Academic License provided
 * appropriate credit is given. See the dOpenCL Academic License for
 * more details.
 * 
 * You should have received a copy of the dOpenCL Academic License
 * along with dOpenCL. If not, see <http://dopencl.uni-muenster.de/>.
 ******************************************************************************/

/*!
 * \file EventCommand.h
 *
 * Request/response messages for event objects.
 *
 * \date 2011-07-11
 * \author Philipp Kegel
 */

#ifndef EVENT_COMMAND_H_
#define EVENT_COMMAND_H_

#include <vector>

#ifdef __APPLE__
#include <OpenCL/cl.h>
#else
#include <CL/cl.h>
#endif

#include <gcf/GCFTypedefs.h>

#include <dcl/DCLTypes.h>

#include "Request.h"
#include "Response.h"

namespace dcl {

namespace comm {

/*!
 * \brief A request message for creating an event.
 *
 * \date    2012-07-17
 * \author  Philipp Kegel
 */
class CreateEvent: public Request {
public:
	CreateEvent(
			dcl_object_id                     contextId,
			dcl_object_id                     eventId,
			const std::vector<dcl_object_id>& memObjectIds);
	CreateEvent(
	        const CreateEvent& rhs);
	virtual ~CreateEvent();

	dcl_object_id contextId() const;
	dcl_object_id eventId() const;
	const std::vector<dcl_object_id>& memObjectIds() const;

	DECLARE_SERIALIZABLE_PUBLIC(CreateEvent, 100 + CREATE_EVENT)

private:
	dcl_object_id _contextId;
	dcl_object_id _eventId;
	std::vector<dcl_object_id> _memObjectIds;

	DECLARE_SERIALIZABLE_EVENT_PRIVATE(CreateEvent)
};

/******************************************************************************/

/*!
 * \brief A request message for creating a user event.
 */
class CreateUserEvent: public Request {
public:
	CreateUserEvent(
			dcl_object_id contextId,
			dcl_object_id eventId);
	CreateUserEvent(
	        const CreateUserEvent& rhs);
	virtual ~CreateUserEvent();

	dcl_object_id contextId() const;
	dcl_object_id eventId() const;

	DECLARE_SERIALIZABLE_PUBLIC(CreateUserEvent, 100 + CREATE_USER_EVENT)

private:
	dcl_object_id _contextId;
	dcl_object_id _eventId;

	DECLARE_SERIALIZABLE_EVENT_PRIVATE(CreateUserEvent)
};

/******************************************************************************/

/*!
 * \brief A request message for deleting an event.
 *
 * \date    Jul 19, 2011
 * \author  Tunc Taylan Turunc
 */
class DeleteEvent: public Request {
public:
	DeleteEvent(
	        dcl_object_id eventId);
	DeleteEvent(
	        const DeleteEvent& rhs);
	virtual ~DeleteEvent();

	dcl_object_id eventId() const;

	DECLARE_SERIALIZABLE_PUBLIC(DeleteEvent, 100 + RELEASE_EVENT)

private:
	dcl_object_id _eventId;

	DECLARE_SERIALIZABLE_EVENT_PRIVATE(DeleteEvent)
};

/******************************************************************************/

/*!
 * \brief A request message for setting the status of a user event.
 */
class SetUserEventStatusRequest: public Request {
public:
	SetUserEventStatusRequest(
			dcl_object_id eventID,
			cl_int        status);
	SetUserEventStatusRequest(
	        const SetUserEventStatusRequest& rhs);
	virtual ~SetUserEventStatusRequest();

	dcl_object_id eventId() const;
	cl_int status() const;

	DECLARE_SERIALIZABLE_PUBLIC(SetUserEventStatusRequest, 100 + SET_USER_EVENT_STATUS)

private:
	dcl_object_id _eventId;
	gcf_uint32 _status;

	DECLARE_SERIALIZABLE_EVENT_PRIVATE(SetUserEventStatusRequest)
};

/******************************************************************************/

/*!
 * \brief A request message for obtaining event profiling info.
 */
class GetEventProfilingInfo: public Request {
public:
	GetEventProfilingInfo(
			dcl_object_id eventID);
	GetEventProfilingInfo(
	        const GetEventProfilingInfo& rhs);
	virtual ~GetEventProfilingInfo();

	dcl_object_id eventId() const;

	DECLARE_SERIALIZABLE_PUBLIC(GetEventProfilingInfo, 100 + GET_EVENT_PROFILING_INFO)

private:
	dcl_object_id _eventId;

	DECLARE_SERIALIZABLE_EVENT_PRIVATE(GetEventProfilingInfo)
};

/******************************************************************************/

/*!
 * \brief A response message containing event profiling info.
 */
class EventProfilingInfo: public Response {
public:
	EventProfilingInfo(
			const Request& request,
			cl_ulong       received,
			cl_ulong       queued,
			cl_ulong       submit,
			cl_ulong       start,
			cl_ulong       end);
	EventProfilingInfo(
	        const EventProfilingInfo& rhs);
	virtual ~EventProfilingInfo();

    cl_ulong received() const;
	cl_ulong queued() const;
	cl_ulong submit() const;
	cl_ulong start() const;
	cl_ulong end() const;

	DECLARE_SERIALIZABLE_PUBLIC(EventProfilingInfo, 200 + Request::GET_EVENT_PROFILING_INFO)

private:
    gcf_uint64 _received;
	gcf_uint64 _queued;
	gcf_uint64 _submit;
	gcf_uint64 _start;
	gcf_uint64 _end;

	DECLARE_SERIALIZABLE_EVENT_PRIVATE(EventProfilingInfo)
};

} /* namespace comm */

} /* namespace dcl */

#endif /* EVENT_COMMAND_H_ */
