/******************************************************************************
 * This file is part of dOpenCL.
 * 
 * dOpenCL is an implementation of the OpenCL application programming
 * interface for distributed systems. See <http://dopencl.uni-muenster.de/>
 * for more information.
 * 
 * Developed by: Research Group Parallel and Distributed Systems
 *               Department of Mathematics and Computer Science
 *               University of Muenster, Germany
 *               <http://pvs.uni-muenster.de/>
 * 
 * Copyright (C) 2013  Philipp Kegel <philipp.kegel@uni-muenster.de>
 *
 * dOpenCL is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * dOpenCL is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with dOpenCL. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Permission to use dOpenCL for scientific, non-commercial work is
 * granted under the terms of the dOpenCL Academic License provided
 * appropriate credit is given. See the dOpenCL Academic License for
 * more details.
 * 
 * You should have received a copy of the dOpenCL Academic License
 * along with dOpenCL. If not, see <http://dopencl.uni-muenster.de/>.
 ******************************************************************************/

/**
 * @file    ProgramCommand.h
 *
 * @date    23.12.2011
 * @author  Philipp Kegel
 */

#ifndef PROGRAM_COMMAND_H_
#define PROGRAM_COMMAND_H_

#include <cstddef>
#include <vector>

#include <gcf/GCFTypedefs.h>

#include <dcl/DCLTypes.h>

#include "Request.h"

namespace dcl {

namespace comm {

class CreateProgramWithBinary: public Request {
public:
	CreateProgramWithBinary(
			const dcl_object_id programId,
			const dcl_object_id contextId,
			const std::vector<dcl_object_id>& deviceIds,
			const std::vector<size_t>& lengths);
	CreateProgramWithBinary(const CreateProgramWithBinary&);
	virtual ~CreateProgramWithBinary();

	dcl_object_id programId() const;
	dcl_object_id contextId() const;
	void getDeviceIds(std::vector<dcl_object_id>& /* deviceIds */) const;
	void getLengths(std::vector<size_t>& lengths) const;

	DECLARE_SERIALIZABLE_PUBLIC(CreateProgramWithBinary,
			100 + CREATE_PROGRAM_WITH_BINARY);

private:
	dcl_object_id _programId;
	dcl_object_id _contextId;
	std::vector<dcl_object_id> _deviceIds;
	std::vector<gcf_uint32> _lengths;

	DECLARE_SERIALIZABLE_EVENT_PRIVATE(CreateProgramWithBinary)
};

/******************************************************************************/

/**
 * @brief Query a program's build log.
 * Other build info is provided locally.
 */
class GetProgramBuildLog: public Request {
public:
	/**
	 * @brief
	 *
	 * @param programId
	 * @param deviceId
	 * @param size      maximum size of the build log to return
	 * An error will be returned by the compute node if the build log is longer this size.
	 * A value of 0 indicates that no build log should be returned.
	 */
	GetProgramBuildLog(
			const dcl_object_id /* programId */,
			const dcl_object_id /* deviceId */,
			size_t /* size */);
	GetProgramBuildLog(const GetProgramBuildLog&);
	virtual ~GetProgramBuildLog();

	dcl_object_id programId(void) const;
	dcl_object_id deviceId(void) const;
	size_t size(void) const;

	DECLARE_SERIALIZABLE_PUBLIC(GetProgramBuildLog,
			100 + GET_PROGRAM_BUILD_LOG);

private:
	dcl_object_id _programId;
	dcl_object_id _deviceId;
	gcf_uint32 _size;

	DECLARE_SERIALIZABLE_EVENT_PRIVATE(GetProgramBuildLog)
};

/******************************************************************************/

/**
 * @brief A program's build log.
 */
class ProgramBuildLog: public Response {
public:
	/**
	 * @brief
	 *
	 * @param size      total size of the build log
	 */
	ProgramBuildLog(
			size_t /* size */);
	ProgramBuildLog(const ProgramBuildLog&);
	virtual ~ProgramBuildLog();

	size_t size(void) const;

	DECLARE_SERIALIZABLE_PUBLIC(ProgramBuildLog,
			200 + Request::GET_PROGRAM_BUILD_LOG);

private:
	gcf_uint32 _size;

	DECLARE_SERIALIZABLE_EVENT_PRIVATE(ProgramBuildLog)
};

} /* namespace comm */

} /* namespace dcl */

#endif /* PROGRAM_COMMAND_H_ */
