/******************************************************************************
 * This file is part of dOpenCL.
 * 
 * dOpenCL is an implementation of the OpenCL application programming
 * interface for distributed systems. See <http://dopencl.uni-muenster.de/>
 * for more information.
 * 
 * Developed by: Research Group Parallel and Distributed Systems
 *               Department of Mathematics and Computer Science
 *               University of Muenster, Germany
 *               <http://pvs.uni-muenster.de/>
 * 
 * Copyright (C) 2013  Philipp Kegel <philipp.kegel@uni-muenster.de>
 *
 * dOpenCL is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * dOpenCL is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with dOpenCL. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Permission to use dOpenCL for scientific, non-commercial work is
 * granted under the terms of the dOpenCL Academic License provided
 * appropriate credit is given. See the dOpenCL Academic License for
 * more details.
 * 
 * You should have received a copy of the dOpenCL Academic License
 * along with dOpenCL. If not, see <http://dopencl.uni-muenster.de/>.
 ******************************************************************************/

/**
 * @file    Stopwatch.h
 *
 * @date    2010-02-23
 * @author  Michel Steuwer <michel.steuwer@uni-muenster.de>
 * @author  Philipp Kegel
 */

#ifndef STOPWATCH_H_
#define STOPWATCH_H_

#define _POSIX_C_SOURCE 199309L

#include <cstring>
#include <ctime>
#include <iostream>
#include <iomanip>
#include <sstream>

#ifndef _POSIX_TIMERS
#error No POSIX timers available
#endif
#ifndef _POSIX_MONOTONIC_CLOCK
#warning No monotonic clock available
#endif

namespace dcl {

namespace util {

/**
 * @brief Simple watch for time measurements.
 *
 * This class provides simple time measurements capabilities.
 * Only works on POSIX-compatible systems, since clock_gettime is used.
 * Link with rt on Linux.
 * After creation start should be called, then stop.
 * The time can be accessed by the getTime function or printed directly with the printTime function.
 */
class Stopwatch {
public:
    /**
     * Default constructor
     */
    Stopwatch(): start_time(), diff_time(0.0f), running(false) { }

    /**
     * Default destructor
     */
    virtual ~Stopwatch() { }

    /**
     * Start the clock
     */
    inline void start() {
        {
            diff_time = 0.0f;
        #ifdef _POSIX_MONOTONIC_CLOCK
            clock_gettime(CLOCK_MONOTONIC, &start_time);
        #else
            clock_gettime(CLOCK_REALTIME, &start_time);
        #endif
            running = true;
        }
    }

    /**
     * Stop the clock
     */
    inline void stop() {
        diff_time = getDiffTime();
        running = false;
    }

    /**
     * Returns the measured time as a float.
     * Should only be called after stop was called.
     * Otherwise it returns 0.0f.
     *
     * @return Returns the time difference measured between the last start and stop call.
     *         If no stop call was made 0.0f is returned.
     */
    inline float getTime() const {
        return diff_time;
    }

    /**
     * Prints the measured time to the standard output.
     *
     * @param[in]  message  An optional message printed to identify the measurement on the console.
     */
    inline void printTime(const char* message = "") const
    {
        std::ostringstream ostr;
        ostr.precision(10);
        if (strcmp(message, "") == 0) {
            ostr << "*** Time: ";
        } else {
            ostr << "*** " <<  message << ": ";
        }
        ostr <<  diff_time << " msecs.";
        std::cout.width(80);
        std::cout.fill('*');
        std::cout << std::left << ostr.str() << "\n";
    }

private:
    /**
     * Private copy constructor. To prevent C++ form building one.
     */
    Stopwatch(const Stopwatch& rhs);

    /**
     * Private assignment operator. To prevent C++ from building one.
     */
    Stopwatch& operator=(const Stopwatch& rhs);

    /**
     * Private function to determine the time passed since the last call of start.
     *
     * @return The elapsed time since the last call of start
     */
    inline float getDiffTime() const {
        struct timespec stop_time;

    #ifdef _POSIX_MONOTONIC_CLOCK
        clock_gettime(CLOCK_MONOTONIC, &stop_time);
    #else
        clock_gettime(CLOCK_REALTIME, &stop_time);
    #endif

        return (float) (1000.0      * (stop_time.tv_sec - start_time.tv_sec)
                     + (   0.000001 * (stop_time.tv_nsec -start_time.tv_nsec)));
    }

    struct timespec start_time; /**< Here is the time stamp stored when start is called */
    float diff_time; /**< Here the time difference between a start and stop call is saved. Returned by getTime */
    bool running; /**< Indicates whether the clock is running. That is start is called, but stop not yet */
};

} /* namespace util */

} /* namespace dcl */

#endif /* STOPWATCH_H_ */
