/******************************************************************************
 * This file is part of dOpenCL.
 * 
 * dOpenCL is an implementation of the OpenCL application programming
 * interface for distributed systems. See <http://dopencl.uni-muenster.de/>
 * for more information.
 * 
 * Developed by: Research Group Parallel and Distributed Systems
 *               Department of Mathematics and Computer Science
 *               University of Muenster, Germany
 *               <http://pvs.uni-muenster.de/>
 * 
 * Copyright (C) 2013  Philipp Kegel <philipp.kegel@uni-muenster.de>
 *
 * dOpenCL is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * dOpenCL is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with dOpenCL. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Permission to use dOpenCL for scientific, non-commercial work is
 * granted under the terms of the dOpenCL Academic License provided
 * appropriate credit is given. See the dOpenCL Academic License for
 * more details.
 * 
 * You should have received a copy of the dOpenCL Academic License
 * along with dOpenCL. If not, see <http://dopencl.uni-muenster.de/>.
 ******************************************************************************/

/**
 * @file    ComputeNode.cpp
 *
 * @date    01.11.2011
 * @author  Philipp Kegel
 */

#include <algorithm>
#include <vector>

#include <boost/bind.hpp>

#include <gcf/CommunicationManager.h>
#include <gcf/GCFTypedefs.h>
#include <gcf/Serializable.h>

#include <dcl/ComputeNode.h>
#include <dcl/Process.h>

#include <dcl/comm/Request.h>
#include <dcl/comm/Response.h>

namespace dcl {

void sendMessage(
        const std::vector<ComputeNode *>& computeNodes,
        gcf::Serializable& message) {
    std::for_each(computeNodes.begin(), computeNodes.end(),
            boost::bind(&ComputeNode::sendMessage, _1, boost::ref(message)));
}

void sendRequest(
		const std::vector<ComputeNode *>& computeNodes,
		dcl::comm::Request& request) {
    std::for_each(computeNodes.begin(), computeNodes.end(),
            boost::bind(&ComputeNode::sendRequest, _1, boost::ref(request)));
}

void executeCommand(
		const std::vector<ComputeNode *>& computeNodes,
		dcl::comm::Request& request,
		GCFTypeID_t responseType,
		std::vector<dcl::comm::Response *> *responses) {
	/*
	 * Send request to all compute nodes, such that the command is executed
	 * simultaneously
	 */
	sendRequest(computeNodes, request);

	/*
	 * Await responses from all compute nodes
	 */
	if (responses) {
		responses->clear();
		responses->reserve(computeNodes.size());
	}

	for (std::vector<ComputeNode *>::const_iterator i = computeNodes.begin();
			i != computeNodes.end(); ++i) {
		dcl::comm::Response *response = (*i)->awaitResponse(request, responseType);
		/* TODO Receive message from *all* compute nodes, i.e., do not stop receipt on first failure */

		if (responses) {
			responses->push_back(response);
		} else {
			/* discard response */
			delete response;
		}
	}
}

} /* namespace dcl */
