/******************************************************************************
 * This file is part of dOpenCL.
 * 
 * dOpenCL is an implementation of the OpenCL application programming
 * interface for distributed systems. See <http://dopencl.uni-muenster.de/>
 * for more information.
 * 
 * Developed by: Research Group Parallel and Distributed Systems
 *               Department of Mathematics and Computer Science
 *               University of Muenster, Germany
 *               <http://pvs.uni-muenster.de/>
 * 
 * Copyright (C) 2013  Philipp Kegel <philipp.kegel@uni-muenster.de>
 *
 * dOpenCL is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * dOpenCL is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with dOpenCL. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Permission to use dOpenCL for scientific, non-commercial work is
 * granted under the terms of the dOpenCL Academic License provided
 * appropriate credit is given. See the dOpenCL Academic License for
 * more details.
 * 
 * You should have received a copy of the dOpenCL Academic License
 * along with dOpenCL. If not, see <http://dopencl.uni-muenster.de/>.
 ******************************************************************************/

/**
 * @file ComputeNodeConnectionManagerImpl.cpp
 *
 * @date 26.10.2011
 * @author Philipp Kegel
 */

#include <cassert>
#include <ostream>
#include <set>
#include <string>
#include <utility>
#include <vector>

#include <boost/thread/condition_variable.hpp>
#include <boost/thread/locks.hpp>
#include <boost/thread/mutex.hpp>

#include <gcf/CommunicationManager.h>
#include <gcf/CommunicationManagerProperties.h>
#include <gcf/ConnectionListener.h>
#include <gcf/Process.h>
#include <gcf/ProcessLookup.h>
#include <gcf/Properties.h>
#include <gcf/ReplicateMappings.h>
#include <gcf/Serializable.h>
#include <gcf/SerializationModule.h>

#include <dcl/ComputeNode.h>
#include <dcl/ConnectionListener.h>
#include <dcl/DCLException.h>
#include <dcl/DCLTypes.h>

#include <dcl/util/Logger.h>

#include "ComputeNodeImpl.h"
#include "ComputeNodeConnectionManagerImpl.h"
#include "HostImpl.h"
#include "ProcessImpl.h"

#include "comm/CLRequestDispatcher.h"
#include "comm/ProcessMessage.h"

namespace dclgcf {

/******************************************************************************
 *
 * Compute node connection manager implementation
 *
 ******************************************************************************/

ComputeNodeConnectionManagerImpl::ComputeNodeConnectionManagerImpl(
		const gcf::CommunicationManagerProperties& commProps) :
		ConnectionManagerImpl(commProps) {
	dcl::util::Logger << dcl::util::Info
			<< "Bound to " << commProps.hostname << ':' << commProps.port
			<< std::endl;

	_clRequestDispatcher = new comm::CLRequestDispatcher(*this);
}

ComputeNodeConnectionManagerImpl::~ComputeNodeConnectionManagerImpl() {
	delete _clRequestDispatcher;
}

void ComputeNodeConnectionManagerImpl::start() {
	_communicationManager->addMessageListener(*_clRequestDispatcher,
	        comm::CLRequestDispatcher::callbackPolicy());

	ConnectionManagerImpl::start();
}

void ComputeNodeConnectionManagerImpl::stop() {
	_communicationManager->removeMessageListener(*_clRequestDispatcher);

	ConnectionManagerImpl::stop();
}

void ComputeNodeConnectionManagerImpl::createComputeNodes(
        const std::vector<std::string>& urls,
        std::vector<ComputeNodeImpl *>& computeNodes) {
    std::set<const gcf::Process *> processes;

    /*
     * Create process list
     */
    for (std::vector<std::string>::const_iterator url = urls.begin();
            url != urls.end(); ++url) {
        std::string host;
        gcf_uint16 port = DEFAULT_COMPUTE_NODE_PORT;

        resolveURL(*url, host, port);
        assert(!host.empty());

        /* WARNING: Broken GCF API design
         * For the following line to work correctly,
         * USE_PROCESSLOOKUP_SINGLETON must be (un)defined, if it has been
         * (un)defined during the compilation of GCF */
        gcf::Process& process = _communicationManager->getProcessLookup().createProcess(host, port);
        bool inserted = processes.insert(&process).second;
        assert(inserted); // URL must only be specified once
    }

    createComputeNodes(processes, computeNodes);
}

void ComputeNodeConnectionManagerImpl::createComputeNodes(
        const std::set<const gcf::Process *>& processes,
        std::vector<ComputeNodeImpl *>& computeNodes) {
    boost::lock_guard<boost::recursive_mutex> lock(_connectionsMutex);

    computeNodes.clear();
    computeNodes.reserve(processes.size());

    /*
     * (Lookup or) create compute nodes and add to list
     */
    for (std::set<const gcf::Process *>::const_iterator process = processes.begin();
            process != processes.end(); ++process) {
        std::map<const gcf::Process *, ComputeNodeImpl *>::const_iterator i;

        assert(*process != 0); // process must not be NULL
        /* WARNING: The equals operator can throw an assertion (see implementation
         * of gcf::Process), if the local and the remote process have not been
         * connected yet. Therefore, the local process must be properly initialized. */
        if (_communicationManager->getOwnProcess().equals(*process)) {
            /* TODO Return reference to local compute node instance */
            continue;
        }

        i = _computeNodes.find(*process);
        if (i == _computeNodes.end()) {
            ComputeNodeImpl *computeNode = new ComputeNodeImpl(
                    *_communicationManager, **process);

            /* Process will not have a valid process ID before it is connected.
             * Hence the process ID cannot be used to identify a process. */
            _computeNodes.insert(std::make_pair(*process, computeNode));
            computeNodes.push_back(computeNode);
        } else {
            computeNodes.push_back(i->second);
        }
    }
}

void ComputeNodeConnectionManagerImpl::destroyComputeNode(
        ComputeNodeImpl *computeNode) {
    if (computeNode) {
        const gcf::Process *process = &computeNode->process();

        computeNode->disconnect();

        {
            boost::lock_guard<boost::recursive_mutex> lock(_connectionsMutex);
            _computeNodes.erase(process);
        }

        delete computeNode;
        _communicationManager->getProcessLookup().freeProcess(process);
    } else {
        throw dcl::InvalidArgument(DCL_INVALID_NODE);
    }
}

void ComputeNodeConnectionManagerImpl::connectComputeNodes(
        const std::vector<ComputeNodeImpl *>& computeNodes) {
    /* Compute node that will connect to this process */
    std::set<ComputeNodeImpl *> activeComputeNodes;
    /* Compute nodes that this process will connect to */
    std::set<ComputeNodeImpl *> passiveComputeNodes;

    /* A collective node-to-node connection operation:
     * A connection is be initiated by exactly one of the involved compute
     * nodes. Therefore, compute nodes must agree on which ones actively connect
     * while others passively accept connections.
     * For this purpose, an order of compute nodes is derived from the order of
     * GCF processes. Based on this order, computes nodes either have a 'lower'
     * or 'higher' rank than others.
     * A compute node only actively connects to other compute nodes that have a
     * lower rank, and passively wait for connections from compute nodes with
     * higher ranks. */

    /*
     * Determine active and passive compute nodes
     */
    for (std::vector<ComputeNodeImpl *>::const_iterator computeNode = computeNodes.begin();
            computeNode != computeNodes.end(); ++computeNode) {
        assert(*computeNode != 0); // compute node must not be NULL
        const gcf::Process& process = (*computeNode)->process();

        /* WARNING: The equals operator can throw an assertion (see implementation
         * of gcf::Process), if the local and the remote process have not been
         * connected yet. Therefore, the local process must be properly initialized. */
        if (_communicationManager->getOwnProcess().equals(&process)) {
            /* a compute node cannot connect to itself */
            continue;
        }

        /* WARNING Do not use operator< to compare processes */
        if (process.getProcessId() < _communicationManager->getOwnProcess().getProcessId()) {
            /* actively connect to passive compute nodes */
            passiveComputeNodes.insert(*computeNode);
        } else {
            /* await connection from active compute nodes */
            activeComputeNodes.insert(*computeNode);
        }
    }

    /* Connect to passive compute nodes (parallelized operation) */
    ComputeNodeImpl::connect(
            std::vector<ComputeNodeImpl *>(passiveComputeNodes.begin(), passiveComputeNodes.end()),
            ProcessImpl::COMPUTE_NODE);

    /* Await connection of active compute nodes */
    ComputeNodeImpl::awaitConnection(
            std::vector<ComputeNodeImpl *>(activeComputeNodes.begin(), activeComputeNodes.end()));
}

bool ComputeNodeConnectionManagerImpl::addConnectionListener(
		dcl::ConnectionListener& listener) {
	boost::lock_guard<boost::mutex> lock(_connectionListenersMutex);
	return (_connectionListeners.insert(&listener)).second;
}

bool ComputeNodeConnectionManagerImpl::removeConnectionListener(
		dcl::ConnectionListener& listener) {
	boost::lock_guard<boost::mutex> lock(_connectionListenersMutex);
	return (_connectionListeners.erase(&listener) == 1);
}

ProcessImpl * ComputeNodeConnectionManagerImpl::getProcess(
        const gcf::Process& process) const {
    boost::lock_guard<boost::recursive_mutex> lock(_connectionsMutex);
    std::map<const gcf::Process *, HostImpl *>::const_iterator i;
    std::map<const gcf::Process *, ComputeNodeImpl *>::const_iterator j;

    /* Do not use ComputeNodeConnectionManagerImpl::getHost and
     * ConnectionManagerImpl::getComputeNode to search for node in order to
     * avoid multiple locking operations on _connectionsMutex */

    /* search host list */
    i = _hosts.find(&process);
    if (i != _hosts.end()) return i->second;

    /* search compute node list */
    j = _computeNodes.find(&process);
    if (j != _computeNodes.end()) return j->second;

    return 0;
}

HostImpl * ComputeNodeConnectionManagerImpl::getHost(
        const gcf::Process& process) const {
    boost::lock_guard<boost::recursive_mutex> lock(_connectionsMutex);
    std::map<const gcf::Process *, HostImpl *>::const_iterator i;

    /* search host list */
    i = _hosts.find(&process);
    if (i != _hosts.end()) return i->second;

    return 0;
}

void ComputeNodeConnectionManagerImpl::hostConnectionEstablished(
		const gcf::Process& process) {
	HostImpl *host;
	bool accepted = false;

    /* TODO Process must not be associated with a compute node yet */
	assert(getHost(process) == 0); // host must not have been created yet
    host = new HostImpl(*_communicationManager, process, _clRequestDispatcher);

	/*
	 * Notify connection listeners
	 */
	{
		boost::lock_guard<boost::mutex> lock(_connectionListenersMutex);
		for (std::set<dcl::ConnectionListener *>::const_iterator i = _connectionListeners.begin();
		        i != _connectionListeners.end(); ++i) {
		    /* WARNING: do not change expression order, otherwise short-cut
		     * evaluation might skip calling the listener */
			accepted = (*i)->connected(*host) || accepted;
		}
	}

	if (accepted) {
        /* add host to list */
	    {
            boost::lock_guard<boost::recursive_mutex> lock(_connectionsMutex);
            bool inserted = _hosts.insert(std::make_pair(&process, host)).second;
            assert(inserted && "Could not add host to list"); // assert insertion of host into list
	    }

        /* send connection approval */
        dcl::comm::ConnectionApprovedMessage message;
        _communicationManager->send(process, message);

        dcl::util::Logger << dcl::util::Info
                << "Accepted connection from host " << host->url() << std::endl;
	} else {
	    /* TODO Send rejection message if incoming connection has been rejected */
	    delete host;

        dcl::util::Logger << dcl::util::Warning
                << "Rejecting connection from host " << host->url() << std::endl;
	}
}

void ComputeNodeConnectionManagerImpl::computeNodeConnectionEstablished(
		const gcf::Process& process) {
	ComputeNodeImpl *computeNode;
    bool accepted = false;

    /* TODO Process must not be associated with a host yet */
    computeNode = getComputeNode(process);
    if (computeNode) {
        /* update compute node's connection status */
        computeNode->setConnectionStatus(ProcessImpl::CONNECTED);
    } else {
        computeNode = new ComputeNodeImpl(*_communicationManager, process);
    }

	/*
	 * Notify connection listeners
	 */
	{
		boost::lock_guard<boost::mutex> lock(_connectionListenersMutex);
		for (std::set<dcl::ConnectionListener *>::const_iterator i = _connectionListeners.begin();
		        i != _connectionListeners.end(); ++i) {
            /* WARNING: do not change expression order, otherwise short-cut
             * evaluation might skip calling the listener */
			accepted = (*i)->connected(*computeNode) || accepted;
		}
	}

    if (accepted) {
        /* add compute node to list;
         * insertion may fail if compute node is re-connected */
        {
            boost::lock_guard<boost::recursive_mutex> lock(_connectionsMutex);
            _computeNodes.insert(std::make_pair(&process, computeNode));
        }

        /* send connection approval */
        dcl::comm::ConnectionApprovedMessage message;
        _communicationManager->send(process, message);

        dcl::util::Logger << dcl::util::Info
                << "Accepted connection from compute node " << computeNode->url() << std::endl;
    } else {
        /* remove compute node from list */
        {
            boost::lock_guard<boost::recursive_mutex> lock(_connectionsMutex);
            _computeNodes.erase(&process);
        }

        /* TODO Send rejection message if incoming connection has been rejected */

        delete computeNode;

        dcl::util::Logger << dcl::util::Warning
                << "Rejected connection from compute node " << computeNode->url() << std::endl;
    }
}

bool ComputeNodeConnectionManagerImpl::hostConnectionLost(
        const gcf::Process& process) {
    HostImpl *host = getHost(process);
    if (host) {
        dcl::util::Logger << dcl::util::Warning
                << "Lost connection to host '" << host->url() << '\''
                << std::endl;

        host->setConnectionStatus(HostImpl::DISCONNECTED); // update host connection status
        /* FIXME Enable GCF DataStreamProcessor to detect closed streams
         * Deleting the data stream is a work-around to ensure that a closed
         * stream is removed from the GCF DataStreamProcessor. Otherwise, the
         * stream will make the DataStreamProcessor to spin in its data receipt
         * procedure, as a closed socket is indicated as incoming data. */
        host->setDataStream(0);

        /*
         * Notify connection listeners
         */
        {
            boost::lock_guard<boost::mutex> lock(_connectionListenersMutex);
            std::set<dcl::ConnectionListener *>::const_iterator i;

            for (i = _connectionListeners.begin();
                    i != _connectionListeners.end(); ++i) {
                (*i)->disconnected(*host);
            }
        }
    }

    return host != 0;
}

bool ComputeNodeConnectionManagerImpl::computeNodeConnectionLost(
        const gcf::Process& process) {
    ComputeNodeImpl *computeNode = getComputeNode(process);
    if (computeNode) {
        dcl::util::Logger << dcl::util::Warning
                << "Lost connection to compute node '" << computeNode->url() << '\''
                << std::endl;

        /* TODO Attempt to re-established failed connection transparently */

        computeNode->setConnectionStatus(ComputeNodeImpl::DISCONNECTED);
        /* FIXME Enable GCF DataStreamProcessor to detect closed streams
         * Deleting the data stream is a work-around to ensure that a closed
         * stream is removed from the GCF DataStreamProcessor. Otherwise, the
         * stream will make the DataStreamProcessor to spin in its data receipt
         * procedure, as a closed socket is indicated as incoming data. */
        computeNode->setDataStream(0);

        /*
         * Notify connection listeners
         */
        {
            boost::lock_guard<boost::mutex> lock(_connectionListenersMutex);
            for (std::set<dcl::ConnectionListener *>::const_iterator i =
                    _connectionListeners.begin();
                    i != _connectionListeners.end(); ++i) {
                (*i)->disconnected(*computeNode);
            }
        }
    }

    return computeNode != 0;
}

} /* namespace dclgcf */
