/******************************************************************************
 * This file is part of dOpenCL.
 * 
 * dOpenCL is an implementation of the OpenCL application programming
 * interface for distributed systems. See <http://dopencl.uni-muenster.de/>
 * for more information.
 * 
 * Developed by: Research Group Parallel and Distributed Systems
 *               Department of Mathematics and Computer Science
 *               University of Muenster, Germany
 *               <http://pvs.uni-muenster.de/>
 * 
 * Copyright (C) 2013  Philipp Kegel <philipp.kegel@uni-muenster.de>
 *
 * dOpenCL is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * dOpenCL is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with dOpenCL. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Permission to use dOpenCL for scientific, non-commercial work is
 * granted under the terms of the dOpenCL Academic License provided
 * appropriate credit is given. See the dOpenCL Academic License for
 * more details.
 * 
 * You should have received a copy of the dOpenCL Academic License
 * along with dOpenCL. If not, see <http://dopencl.uni-muenster.de/>.
 ******************************************************************************/

/**
 * @file ComputeNodeConnectionManagerImpl.h
 *
 * @date 26.10.2011
 * @author Philipp Kegel
 */

#ifndef COMPUTENODECONNECTIONMANAGERIMPL_H_
#define COMPUTENODECONNECTIONMANAGERIMPL_H_

#include <map>
#include <set>
#include <string>
#include <utility>
#include <vector>

#include <boost/thread/mutex.hpp>

#include <gcf/CommunicationManager.h>
#include <gcf/CommunicationManagerProperties.h>
#include <gcf/ConnectionListener.h>
#include <gcf/MessageListener.h>
#include <gcf/Process.h>
#include <gcf/ReplicateMappings.h>
#include <gcf/Serializable.h>

#include <dcl/ConnectionListener.h>
#include <dcl/ConnectionManager.h>

#include "ConnectionManagerImpl.h"

namespace dclgcf {

/* forward declarations */
class ComputeNodeImpl;
class HostImpl;
class ProcessImpl;

namespace comm {

class CLRequestDispatcher;

} /* namespace comm */

/******************************************************************************/

class ComputeNodeConnectionManagerImpl:
        public ConnectionManagerImpl,              // extend ConnectionManagerImpl
        public dcl::ComputeNodeConnectionManager { // implement ComputeNodeConnectionManager
public:
    ComputeNodeConnectionManagerImpl(
            const gcf::CommunicationManagerProperties&  /* properties */);
    virtual ~ComputeNodeConnectionManagerImpl();

    void start();
    void stop();

    void createComputeNodes(
            const std::vector<std::string>& /* urls */,
            std::vector<ComputeNodeImpl *>& /* computeNodes */);
    /**
     * @brief Creates compute nodes from the given GCF processes.
     *
     * WARNING: Compute nodes returned by this method may not be connected such
     * that subsequent communication with these compute nodes will fail.
     */
    void createComputeNodes(
            const std::set<const gcf::Process *>&   /* processes */,
            std::vector<ComputeNodeImpl *>&         /* computeNodes */);
    void destroyComputeNode(
            ComputeNodeImpl *  /* computeNode */);

    void connectComputeNodes(
            const std::vector<ComputeNodeImpl *>&   /* computeNodes */);

    bool addConnectionListener(dcl::ConnectionListener&);
    bool removeConnectionListener(dcl::ConnectionListener&);

    /**
     * @brief Returns the dOpenCL process that is associated with a specified GCF process.
     *
     * This method overrides ConnectionManagerImpl::getProcess in order to also
     * search the list of hosts.
     */
    ProcessImpl * getProcess(
            const gcf::Process& /* process */) const;

    HostImpl * getHost(
            const gcf::Process& /* process */) const;

protected:
    void hostConnectionEstablished(const gcf::Process&);
    void computeNodeConnectionEstablished(const gcf::Process&);

    bool hostConnectionLost(const gcf::Process&);
    bool computeNodeConnectionLost(const gcf::Process&);

private:
    /* Connection manager must not be copied */
    ComputeNodeConnectionManagerImpl(const ComputeNodeConnectionManagerImpl&);
    ComputeNodeConnectionManagerImpl& operator=(const ComputeNodeConnectionManagerImpl&);

    comm::CLRequestDispatcher *_clRequestDispatcher;

    std::set<dcl::ConnectionListener *> _connectionListeners;
    boost::mutex _connectionListenersMutex;

    std::map<const gcf::Process *, HostImpl *> _hosts;
};

} /* namespace dclgcf */

#endif /* COMPUTENODECONNECTIONMANAGERIMPL_H_ */
