/******************************************************************************
 * This file is part of dOpenCL.
 * 
 * dOpenCL is an implementation of the OpenCL application programming
 * interface for distributed systems. See <http://dopencl.uni-muenster.de/>
 * for more information.
 * 
 * Developed by: Research Group Parallel and Distributed Systems
 *               Department of Mathematics and Computer Science
 *               University of Muenster, Germany
 *               <http://pvs.uni-muenster.de/>
 * 
 * Copyright (C) 2013  Philipp Kegel <philipp.kegel@uni-muenster.de>
 *
 * dOpenCL is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * dOpenCL is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with dOpenCL. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Permission to use dOpenCL for scientific, non-commercial work is
 * granted under the terms of the dOpenCL Academic License provided
 * appropriate credit is given. See the dOpenCL Academic License for
 * more details.
 * 
 * You should have received a copy of the dOpenCL Academic License
 * along with dOpenCL. If not, see <http://dopencl.uni-muenster.de/>.
 ******************************************************************************/

/*!
 * \file ComputeNodeImpl.cpp
 *
 * \date 2011-10-30
 * \author Philipp Kegel
 */

#include <algorithm>
#include <cassert>
#include <ostream>
#include <vector>

#include <boost/bind.hpp>
#include <boost/shared_ptr.hpp>

#include <boost/thread/condition_variable.hpp>
#include <boost/thread/locks.hpp>
#include <boost/thread/recursive_mutex.hpp>
#include <boost/thread/thread_time.hpp>

#ifdef __APPLE__
#include <OpenCL/cl.h>
#else
#include <CL/cl.h>
#endif

#include <gcf/CommunicationManager.h>
#include <gcf/GCFError.h>
#include <gcf/GCFTypedefs.h>
#include <gcf/Process.h>

#include <gcf/link/SocketOptions.h>

#include <gcf/transport/IODataStream.h>

#include <dcl/CLError.h>
#include <dcl/DataTransfer.h>
#include <dcl/DCLException.h>
#include <dcl/DCLTypes.h>
#include <dcl/Device.h>

#include <dcl/comm/ErrorResponse.h>
#include <dcl/comm/Request.h>
#include <dcl/comm/Response.h>

#include <dcl/util/Logger.h>
//#include <dcl/util/Stopwatch.h>

#include "ComputeNodeImpl.h"
#include "ConnectionManagerImpl.h"
#include "DeviceImpl.h"

#include "comm/DeviceIDsResponse.h"
#include "comm/GetDeviceIDs.h"
#include "comm/ProcessMessage.h"

#if 0
#include <dcl/comm/CreateContext.h>
#include <dcl/comm/CreateCommandQueue.h>
#include <dcl/comm/CreateBuffer.h>
#include <dcl/comm/CreateProgramWithSource.h>
#endif

namespace dclgcf {

void ComputeNodeImpl::updateDevices(
        const std::vector<ComputeNodeImpl *>& computeNodes) {
	std::vector<ComputeNodeImpl *> connectedComputeNodes(computeNodes);
//    dcl::util::Stopwatch timer;

	/*
	 * Send request to all compute nodes, such that the command is executed
	 * simultaneously
	 */
//    timer.start();
	dcl::comm::GetDeviceIDs request((cl_device_type) CL_DEVICE_TYPE_ALL);
	for (std::vector<ComputeNodeImpl *>::iterator i = connectedComputeNodes.begin();
			i != connectedComputeNodes.end();) {
		try {
			(*i)->sendRequest(request);
			++i;
		} catch (const dcl::DCLException& err) {
			i = connectedComputeNodes.erase(i);
			dcl::util::Logger << dcl::util::Warning << err.what() << std::endl;
		}
	}
//    timer.stop();
//    timer.printTime("send requests");
//    std::cout << "#connected nodes " << connectedComputeNodes.size() << std::endl;

	/*
	 * Await responses from all compute nodes
	 */
//    timer.start();
	for (std::vector<ComputeNodeImpl *>::const_iterator i = connectedComputeNodes.begin();
			i != connectedComputeNodes.end(); ++i) {
		ComputeNodeImpl *computeNode = *i;

		try {
			boost::shared_ptr<dcl::comm::DeviceIDsResponse> response(
					static_cast<dcl::comm::DeviceIDsResponse *>(
							computeNode->awaitResponse(request, dcl::comm::DeviceIDsResponse::TYPE)));
			assert(response != 0); // response must not be NULL
		    dcl::util::Logger << dcl::util::Info
		            << "Found " << response->deviceIds().size()
		            << " devices on compute node " << computeNode->url() << std::endl;

			computeNode->updateDevices(response->deviceIds());
		} catch (const dcl::DCLException& err) {
			dcl::util::Logger << dcl::util::Error << err.what() << std::endl;
		}
	}
//    timer.stop();
//    timer.printTime("process responses");
}

void ComputeNodeImpl::connect(
        const std::vector<ComputeNodeImpl *>& computeNodes,
        ProcessType localProcessType) {
	std::vector<ComputeNodeImpl *> connectingComputeNodes(computeNodes);
	std::vector<ComputeNodeImpl *>::iterator i;
//    dcl::util::Stopwatch timer;

//    std::cout << "#nodes " << computeNodes.size() << std::endl;
//    timer.start();
	i = connectingComputeNodes.begin();
	while (i != connectingComputeNodes.end()) {
	    /* skip connected compute nodes */
	    if ((*i)->isConnected()) {
            i = connectingComputeNodes.erase(i);
            continue;
	    }

		try {
			(*i)->connectProcess(localProcessType,
			        ConnectionManagerImpl::DEFAULT_CONNECTION_TIMEOUT);
			++i;
		} catch (const dcl::DCLException& err) {
			i = connectingComputeNodes.erase(i);
			dcl::util::Logger << dcl::util::Warning << err.what() << std::endl;
		}
	}
//    timer.stop();
//    timer.printTime("connect processes");
//    std::cout << "#nodes connecting process " << connectingComputeNodes.size() << std::endl;

	/*
	 * The data streams are created after all process connections have been
	 * established. Thus, the compute nodes can initialize their data stream
	 * listeners concurrently.
	 */

	boost::system_time deadline = boost::get_system_time() +
	        boost::posix_time::milliseconds(
	                ConnectionManagerImpl::DEFAULT_CONNECTION_TIMEOUT);

//    timer.start();
	i = connectingComputeNodes.begin();
	while (i != connectingComputeNodes.end()) {
		try {
			(*i)->createDataStream(deadline);
			++i;
		} catch (const dcl::DCLException& err) {
			i = connectingComputeNodes.erase(i);
			dcl::util::Logger << dcl::util::Error << err.what() << std::endl;
		}
	}
//    timer.stop();
//    timer.printTime("create streams");
//    std::cout << "#nodes connecting stream " << connectingComputeNodes.size() << std::endl;

	switch (localProcessType) {
	case HOST:
        /* Prefetch device IDs */
//        timer.start();
	    updateDevices(connectingComputeNodes);
//        timer.stop();
//        timer.printTime("update devices");
	    break;
	default:
	    ;
	    /* no break */
	}

	/* TODO Handle connection error
	 * Devices of compute nodes whose connections failed should become unavailable. */
}

void ComputeNodeImpl::awaitConnection(
        const std::vector<ComputeNodeImpl *>& computeNodes) {
    boost::system_time deadline = boost::get_system_time() +
            boost::posix_time::milliseconds(
                    ConnectionManagerImpl::DEFAULT_CONNECTION_TIMEOUT);
    std::for_each(computeNodes.begin(), computeNodes.end(),
            boost::bind(&ComputeNodeImpl::awaitConnectionStatus<boost::system_time>,
                    _1, ProcessImpl::CONNECTED, deadline));
}

ComputeNodeImpl::ComputeNodeImpl(
        gcf::CommunicationManager& communicationManager,
        const gcf::Process& process) :
    ProcessImpl(communicationManager, process), _devices(0)
{
    dcl::util::Logger << dcl::util::Debug
            << "Created compute node " << url() << std::endl;
}

ComputeNodeImpl::~ComputeNodeImpl() {
    /* clear device list */
    if (_devices) {
        for (std::vector<DeviceImpl *>::iterator i = _devices->begin();
                i != _devices->end(); ++i) {
            delete *i;
        }

        delete _devices;
    }
}

void ComputeNodeImpl::connectProcess(ProcessType localProcessType,
        gcf_uint32 timeout) {
	gcf::SocketOptions socketOptions;

	/* Disable Nagle's algorithm
     * Due to the RPC-style protocol of dOpenCL, short messages usually wait for
     * a response before the next message is send. Hence, waiting for another
     * short message to merge with the first one is pointless.
     * In particular, command status messages suffer from the high latency
     * introduced by Nagle's algorithm. */
    socketOptions.set("SOL_TCP", "TCP_NODELAY");
    /* CommunicationManager::setOptions set the options locally and also sends a
     * message to the remote process in order to let that process set the same
     * options. However, this does not work reliably (at least for TCP_NODELAY).
     * Moreover, the options are sent twice (in CommunicationManager::setOptions
     * and ConnectionManager::setOptions); maybe this is a GCF bug */

	try {
		/*
		 * connect process
		 */
		{
			boost::lock_guard<boost::recursive_mutex> lock(_connectionStatusMutex);
			bool connected = _communicationManager.preconnectProcess(_process,
			        timeout, socketOptions);
			if (!connected) { /* timeout reached or failure */
				throw dcl::ConnectionException(
						"Could not connect to compute node at " + url());
			}
			_connectionStatus = CONNECTING;
			_connectionStatusChanged.notify_all();
		}

	    dcl::util::Logger << dcl::util::Info
	            << "Connecting to compute node " << url() << std::endl;

		/* Inform remote process of the type of dOpenCL process that has
		 * established the incoming connection */
		switch (localProcessType) {
		case HOST:
		{
			dcl::comm::HostIdentificationMessage message;
			_communicationManager.send(_process, message);
			break;
		}
		case COMPUTE_NODE:
		{
			dcl::comm::ComputeNodeIdentificationMessage message;
			_communicationManager.send(_process, message);
			break;
		}
		default:
			/* TODO Throw exception
			 * Processes other than hosts or compute nodes must not create
			 * compute node connections */
			;
			/* no break */
		}
	} catch (const gcf::GCFError& err) {
		throw dcl::ConnectionException(err.message);
	}
}

void ComputeNodeImpl::updateDevices() {
	dcl::comm::GetDeviceIDs request((cl_device_type) CL_DEVICE_TYPE_ALL);
	boost::shared_ptr<dcl::comm::DeviceIDsResponse> response(
			static_cast<dcl::comm::DeviceIDsResponse *>(
					executeCommand(request, dcl::comm::DeviceIDsResponse::TYPE)));
	assert(response != 0); // response must not be NULL
	dcl::util::Logger << dcl::util::Info
			<< "Found " << response->deviceIds().size()
			<< " devices on compute node " << url() << std::endl;

    updateDevices(response->deviceIds());
}

void ComputeNodeImpl::updateDevices(
		const std::vector<dcl::dcl_object_id>& deviceIds) {
	boost::lock_guard<boost::recursive_mutex> lock(_devicesMutex);

	/* FIXME Do not clear but update the device list
	 * Add new devices; mark as valid
	 * Mark existing devices valid or invalid */
	if (_devices) {
		_devices->clear();
	} else {
		_devices = new std::vector<DeviceImpl *>();
		_devices->reserve(deviceIds.size());
	}

	for (std::vector<dcl::dcl_object_id>::const_iterator i = deviceIds.begin();
			i != deviceIds.end(); ++i) {
		_devices->push_back(new DeviceImpl((*i), *this));
		/* TODO Register devices */
	}
}

void ComputeNodeImpl::connect(ProcessType localProcessType) {
    boost::system_time deadline = boost::get_system_time() +
            boost::posix_time::milliseconds(
                    ConnectionManagerImpl::DEFAULT_CONNECTION_TIMEOUT);
    connectProcess(localProcessType, ConnectionManagerImpl::DEFAULT_CONNECTION_TIMEOUT);
    createDataStream(deadline);
}

void ComputeNodeImpl::getDevices(std::vector<dcl::Device *>& devices) {
	boost::lock_guard<boost::recursive_mutex> lock(_devicesMutex);

	/* Device list may be uninitialized, if the compute node's device is queried
	 * for the first time. */
	if (!_devices) {
		updateDevices();
	}
	
	assert(_devices != 0);
	/* Must not return a reference to the device list as it may change
	 * asynchronously */
	devices.assign(_devices->begin(), _devices->end());
}

void ComputeNodeImpl::sendRequest(dcl::comm::Request& request) const {
    /* Do not use sendMessage which should be used for dOpenCL messages only
     * Thus, sending messages and sending requests can be properly
     * distinguished. */
//	sendMessage(request);
    /* gcf::CommunicationManager::send never fails */
    if (!_communicationManager.isConnected(_process)) {
        throw dcl::IOException("No connection to compute node at " + url());
    }
    _communicationManager.send(_process, request);
}

dcl::comm::Response * ComputeNodeImpl::awaitResponse(
		const dcl::comm::Request& request, GCFTypeID_t responseType) {
	dcl::comm::Response *response;

	try {
#ifndef NDEBUG
        /* Wait for responses indefinitely during debug. */
		response = _responseBuffer.get(request);
#else
        /* FIXME Allow for blocking operation (e.g. finish) to wait indefinitely
         * Blocking operations may require more time than the default response
         * timeout. Hence no IO exception must be reported.
         * ALTERNATIVELY implement blocking operations asynchronously. */
        response = _responseBuffer.get(request, boost::posix_time::milliseconds(
                ConnectionManagerImpl::DEFAULT_CONNECTION_TIMEOUT));
#endif
	} catch (const boost::thread_interrupted&) {
		/* ignore interrupt; treat as timeout */
		response = 0;
	}
	if (!response) {
		throw dcl::IOException("No response from compute node at " + url());
	}

	/* Intercept error message */
	if (response->getType() == dcl::comm::ErrorResponse::TYPE) {
		dcl::CLError err(response->errcode());
		delete response;
		throw err;
	}
	/* Intercept illegal response (protocol error) */
	if (response->getType() != responseType) {
		delete response;
		throw dcl::ProtocolException("Illegal response from compute node at " + url());
	}
	return response;
}

void ComputeNodeImpl::awaitResponse(const dcl::comm::Request& request) {
	delete awaitResponse(request, dcl::comm::Response::TYPE);
}

dcl::comm::Response * ComputeNodeImpl::executeCommand(
		dcl::comm::Request& request, GCFTypeID_t responseType) {
//#ifndef NDEBUG
#if 0
    dcl::comm::Response *response;

    sendRequest(request);
    dcl::util::Logger << dcl::util::Debug
            << "\tsent request (request ID=" << request.id
            << ", type=" << request.getType() << ')'
            << std::endl;

    response = awaitResponse(request, responseType);
    dcl::util::Logger << dcl::util::Debug
            << "\treceived response (request ID=" << response->requestId
            << ", type=" << response->getType() << ')'
            << std::endl;

    return response;
#else
	sendRequest(request);
	return awaitResponse(request, responseType);
#endif
}

void ComputeNodeImpl::executeCommand(dcl::comm::Request& request) {
	delete executeCommand(request, dcl::comm::Response::TYPE);
}

dcl::comm::ResponseBuffer& ComputeNodeImpl::responseBuffer() {
	return _responseBuffer;
}

#if 0
template<typename TReq, typename TRes>
TRes * ComputeNodeImpl::awaitTypedResponse(const TReq& request) {
	dcl::comm::Response *response = awaitResponse(request);
	/* Intercept error message */
	if (response->getType() == dcl::comm::ErrorResponse::TYPE) {
		delete response;
		throw dcl::CLError(response->errcode());
	}
	/* Intercept illegal response (protocol error) */
	if (response->getType() != TRes::TYPE) {
		delete response;
		throw dcl::ProtocolException();
	}
	return static_cast<TRes *>(response);
}

template<typename TReq>
dcl::comm::Response * ComputeNodeImpl::awaitTypedResponse(const TReq& request) {
	delete awaitResponse<TReq, dcl::comm::Response>(request);
	return 0;
}

/* explicit instantiation */
template dcl::comm::DeviceIDsResponse * ComputeNodeImpl::awaitTypedResponse(const dcl::comm::GetDeviceIDs&);
template dcl::comm::Response *          ComputeNodeImpl::awaitTypedResponse(const dcl::comm::CreateContext&);
template dcl::comm::Response *          ComputeNodeImpl::awaitTypedResponse(const dcl::comm::CreateCommandQueue&);
template dcl::comm::Response *          ComputeNodeImpl::awaitTypedResponse(const dcl::comm::CreateBuffer&);
template dcl::comm::Response *          ComputeNodeImpl::awaitTypedResponse(const dcl::comm::CreateProgramWithSource&);
#endif

/******************************************************************************/

void sendRequest(
        const std::vector<ComputeNodeImpl *>& computeNodes,
        dcl::comm::Request& request) {
    /* TODO Use communication group to broadcast request */
    std::for_each(computeNodes.begin(), computeNodes.end(),
            boost::bind(&ComputeNodeImpl::sendRequest, _1, boost::ref(request)));
}

void executeCommand(
        const std::vector<ComputeNodeImpl *>& computeNodes,
        dcl::comm::Request& request,
        GCFTypeID_t responseType,
        std::vector<dcl::comm::Response *> *responses) {
    /*
     * Send request to all compute nodes, such that the command is executed
     * simultaneously
     */
    sendRequest(computeNodes, request);

    /*
     * Await responses from all compute nodes
     */
    if (responses) {
        responses->clear();
        responses->reserve(computeNodes.size());
    }

    for (std::vector<ComputeNodeImpl *>::const_iterator i = computeNodes.begin();
            i != computeNodes.end(); ++i) {
        dcl::comm::Response *response = (*i)->awaitResponse(request, responseType);
        /* TODO Receive message from *all* compute nodes, i.e., do not stop receipt on first failure */

        if (responses) {
            responses->push_back(response);
        } else {
            /* discard response */
            delete response;
        }
    }
}

} /* namespace dclgcf */
