/******************************************************************************
 * This file is part of dOpenCL.
 * 
 * dOpenCL is an implementation of the OpenCL application programming
 * interface for distributed systems. See <http://dopencl.uni-muenster.de/>
 * for more information.
 * 
 * Developed by: Research Group Parallel and Distributed Systems
 *               Department of Mathematics and Computer Science
 *               University of Muenster, Germany
 *               <http://pvs.uni-muenster.de/>
 * 
 * Copyright (C) 2013  Philipp Kegel <philipp.kegel@uni-muenster.de>
 *
 * dOpenCL is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * dOpenCL is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with dOpenCL. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Permission to use dOpenCL for scientific, non-commercial work is
 * granted under the terms of the dOpenCL Academic License provided
 * appropriate credit is given. See the dOpenCL Academic License for
 * more details.
 * 
 * You should have received a copy of the dOpenCL Academic License
 * along with dOpenCL. If not, see <http://dopencl.uni-muenster.de/>.
 ******************************************************************************/

/*!
 * \file ComputeNodeImpl.h
 *
 * \date 2011-10-30
 * \author Philipp Kegel
 */

#ifndef COMPUTENODEIMPL_H_
#define COMPUTENODEIMPL_H_

#include <vector>

#include <boost/thread/recursive_mutex.hpp>

#include <gcf/CommunicationManager.h>
#include <gcf/GCFTypedefs.h>
#include <gcf/Process.h>
#include <gcf/StreamCommunication.h>

#include <dcl/ComputeNode.h>
#include <dcl/DCLException.h>
#include <dcl/Device.h>

#include <dcl/comm/Request.h>
#include <dcl/comm/Response.h>

#include "ProcessImpl.h"

#include "comm/ResponseBuffer.h"

namespace dclgcf {

/* forward declarations */
class DeviceImpl;

/******************************************************************************/

class ComputeNodeImpl:
public dcl::ComputeNode, // implements ComputeNode
public ProcessImpl {     // extends ProcessImpl
public:
    /*!
     * \brief Connects to multiple compute nodes concurrently.
     *
     * This method updates the compute nodes' device lists.
     * This method works on a best-effort base: failed connection attempts are
     * ignored.
     *
     * \param[in]  computeNodes     the compute nodes to connect to
     * \param[in]  localProcessType type of the local process that is establishing the connection
     */
    static void connect(
            const std::vector<ComputeNodeImpl *>& computeNodes,
            ProcessType                           localProcessType);

    /*!
     * \brief Awaits for multiple compute nodes to connect.
     *
     * \param[in]  computeNodes the compute nodes which connection to wait for
     */
    static void awaitConnection(
            const std::vector<ComputeNodeImpl *>& computeNodes);


    /*!
     * \brief Creates a compute node instance.
     *
     * \param[in]  communicationManager an associated GCF communication manager
     * \param[in]  process              an associated GCF process
     */
    ComputeNodeImpl(
            gcf::CommunicationManager& communicationManager,
            const gcf::Process&        process);
    virtual ~ComputeNodeImpl();

    /*!
     * \brief Connects to the compute node
     *
     * A connection is established in four steps:
     * 1. connect GCF process
     * 2. await session creation on compute node
     * 3. create GCF I/O data stream
     * 4. update device list
     *
     * \param[in]  localProcessType type of the local process that is establishing the connection
     */
    /* TODO Allow for asynchronous connect using a callback */
    void connect(
            ProcessType localProcessType);

    void getDevices(
            std::vector<dcl::Device *>& devices);

    void sendRequest(
            dcl::comm::Request& request) const;

    dcl::comm::Response * awaitResponse(
            const dcl::comm::Request& request,
            GCFTypeID_t               responseType);
    void awaitResponse(
            const dcl::comm::Request& request);

    dcl::comm::Response * executeCommand(
            dcl::comm::Request& request,
            GCFTypeID_t         responseType);
    void executeCommand(dcl::comm::Request&);

    dcl::comm::ResponseBuffer& responseBuffer();

#if 0
    template<typename TReq, typename TRes>
    TRes * awaitTypedResponse(
            const TReq& request);
    template<typename TReq>
    dcl::comm::Response * awaitTypedResponse(const TReq& request);

    template<typename TReq, typename TRes>
    TRes * executeCommand(
            TReq& request) {
        sendRequest(request);
        return awaitTypedResponse<TReq, TRes>(request);
    }

protected:
    virtual dcl::comm::Response * awaitResponse(
            const dcl::comm::Request& request) = 0;
#endif

private:
    /*!
     * \brief Updates the device lists of multiple compute nodes.
     *
     * The device IDs are queried from multiple remote compute nodes.
     * This method is more efficient than @c updateDevices.
     *
     * \param[in]  computeNode  the compute node to query the device IDs from
     */
    static void updateDevices(
            const std::vector<ComputeNodeImpl *>& computeNodes);


    ComputeNodeImpl(
            const ComputeNodeImpl& computeNode);
    ComputeNodeImpl& operator=(
            const ComputeNodeImpl& computeNode);

    /*!
     * \brief Connects to the compute node's GCF process.
     *
     * \param[in]  localProcessType type of the local process that is establishing the connection
     * \param[in]  timeout          maximum time for connection process
     */
    void connectProcess(
            ProcessType localProcessType,
            gcf_uint32  timeout);
    /*!
     * \brief Initializes the compute node's GCF I/O data stream.
     *
     * \param[in]  timeout  a timeout; may be a duration or time point
     */
    template<typename timeout_type>
    void createDataStream(
            const timeout_type& timeout) {
        /*
         * Await connection approval
         * The peer's listening socket for data streams must have been initialized
         * before a data stream can be created.
         */
        bool connected = awaitConnectionStatus(CONNECTED, timeout);
        if (!connected) { /* session creation timed out */
            throw dcl::ConnectionException("Could not connect to compute node at " + url());
        }

        /*
         * create I/O data stream
         */
        {
            boost::lock_guard<boost::recursive_mutex> lock(_connectionStatusMutex);
            gcf::StreamCommunication& sc = _communicationManager.getStreamCommunication();
            _dataStream = &sc.createIODataStream(_process);
        }
    }

    /*!
     * \brief Updates the compute nodes device list.
     * The device IDs are queried from the remote compute node.
     */
    void updateDevices();
    /*!
     * \brief Updates the compute node's device list using a list of device IDs.
     *
     * \param[in]  deviceIds    the device IDs that should be used for the update
     */
    void updateDevices(
            const std::vector<dcl::dcl_object_id>& deviceIds);


    dcl::comm::ResponseBuffer _responseBuffer; //!< Response buffer

    std::vector<DeviceImpl *> *_devices; //!< Device list
    /*!
     * \brief A mutex associated with this compute node's devices list.
     *
     * This mutex must be recursive as getDeviceIDs locks it and will attempt to
     * lock it again when the device IDs have to be queried in updateDevices.
     */
    boost::recursive_mutex _devicesMutex;
};

/******************************************************************************/

/*!
 * \brief Sends a request to multiple compute nodes.
 *
 * \param[in]  computeNodes a list of compute nodes to send a request to
 * \param[in]  request      the request to send
 */
void sendRequest(
        const std::vector<ComputeNodeImpl *>& computeNodes,
        dcl::comm::Request&                   request);

/*!
 * \brief Executes a command on multiple compute nodes.
 *
 * \param[in]  computeNodes a list of compute node to execute the command on
 * \param[in]  request      the request for a given command
 * \param[in]  responseType the expected type of responses
 * \param[in]  responses    the responses returned by the compute nodes
 */
void executeCommand(
        const std::vector<ComputeNodeImpl *>& computeNodes,
        dcl::comm::Request&                   request,
        GCFTypeID_t                           responseType = dcl::comm::Response::TYPE,
        std::vector<dcl::comm::Response *> *  responses = 0);

} /* namespace dclgcf */

#endif /* COMPUTENODEIMPL_H_ */
