/******************************************************************************
 * This file is part of dOpenCL.
 * 
 * dOpenCL is an implementation of the OpenCL application programming
 * interface for distributed systems. See <http://dopencl.uni-muenster.de/>
 * for more information.
 * 
 * Developed by: Research Group Parallel and Distributed Systems
 *               Department of Mathematics and Computer Science
 *               University of Muenster, Germany
 *               <http://pvs.uni-muenster.de/>
 * 
 * Copyright (C) 2013  Philipp Kegel <philipp.kegel@uni-muenster.de>
 *
 * dOpenCL is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * dOpenCL is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with dOpenCL. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Permission to use dOpenCL for scientific, non-commercial work is
 * granted under the terms of the dOpenCL Academic License provided
 * appropriate credit is given. See the dOpenCL Academic License for
 * more details.
 * 
 * You should have received a copy of the dOpenCL Academic License
 * along with dOpenCL. If not, see <http://dopencl.uni-muenster.de/>.
 ******************************************************************************/

/**
 * @file ConnectionManager.cpp
 *
 * @date 26.10.2011
 * @author Philipp Kegel
 */

#include <cstdlib>
#include <iostream>
#include <fstream>
#include <sstream>
#include <string>

#include <gcf/CommunicationManagerProperties.h>
#include <gcf/GCFTypedefs.h>

#include <rtfutil/Logger.h>

#include <dcl/ConnectionManager.h>
#include <dcl/DCLException.h>

#include <dcl/util/Logger.h>

#include "ComputeNodeConnectionManagerImpl.h"
#include "HostConnectionManagerImpl.h"

namespace dcl {

static dcl::util::Severity::Type getSeverity() {
    const char *loglevel = getenv("DCL_LOG_LEVEL");

    if (loglevel) {
        if (strcmp(loglevel, "ERROR") == 0) {
            return dcl::util::Severity::Error;
        } else if (strcmp(loglevel, "WARNING") == 0) {
            return dcl::util::Severity::Warning;
        } else if (strcmp(loglevel, "INFO") == 0) {
            return dcl::util::Severity::Info;
        } else if (strcmp(loglevel, "DEBUG") == 0) {
            return dcl::util::Severity::Debug;
        } else if (strcmp(loglevel, "VERBOSE") == 0) {
            return dcl::util::Severity::Verbose;
        }
    }

    /* use default log level */
#ifndef NDEBUG
    return dcl::util::Severity::Debug;
#else
    return dcl::util::Severity::Info;
#endif
}

static rtfutil::Logger::LoggingLevel getLoggingLevel() {
    const char *loglevel = getenv("DCL_LOG_LEVEL");

    if (loglevel) {
        if (strcmp(loglevel, "ERROR") == 0) {
            return rtfutil::Logger::Error;
        } else if (strcmp(loglevel, "WARNING") == 0) {
            return rtfutil::Logger::Warning;
        } else if (strcmp(loglevel, "INFO") == 0) {
            return rtfutil::Logger::Info;
        } else if (strcmp(loglevel, "DEBUG") == 0) {
            return rtfutil::Logger::Debug;
        } else if (strcmp(loglevel, "VERBOSE") == 0) {
            return rtfutil::Logger::VerboseDebug;
        }
    }

    /* use default log level */
#ifndef NDEBUG
    return rtfutil::Logger::Debug;
#else
    return rtfutil::Logger::Info;
#endif
}

/******************************************************************************/

HostConnectionManager * HostConnectionManager::create() {
    /* setup dOpenCL logger */
    static std::ofstream dclLogFile("dcl_host.log");
    dcl::util::Logger.setOutput(dclLogFile);
    dcl::util::Logger.setLoggingLevel(getSeverity());
    dcl::util::Logger.setDefaultSeverity(dcl::util::Severity::Info);

    /* setup RTF logger */
    static std::ofstream rtfLogFile("gcf_host.log");
    rtfutil::Logger::output(rtfLogFile);
    rtfutil::Logger::verbose(getLoggingLevel());

    return new dclgcf::HostConnectionManagerImpl();
}

/******************************************************************************/

ComputeNodeConnectionManager * ComputeNodeConnectionManager::create(
        const std::string& url) {
    std::string host;
    gcf_uint16 port = dclgcf::ConnectionManagerImpl::DEFAULT_COMPUTE_NODE_PORT;
    gcf::SocketOptions socketOptions;
    std::string logFileName;

    dclgcf::ConnectionManagerImpl::resolveURL(url, host, port);

    if (host.empty()) {
        throw dcl::InvalidArgument(DCL_INVALID_NODE, "Missing host name");
    }

    /* Generate name of dOpenCL log file */
    {
        std::stringstream ss;

        /* TODO Log to system default location of files
         * Log to /var/log/dcld */
        ss << "dcl_" << host << ".log";
        ss >> logFileName;
    }

    /* setup dOpenCL logger */
    static std::ofstream dclLogFile(logFileName.c_str());
    dcl::util::Logger.setOutput(dclLogFile);
    dcl::util::Logger.setLoggingLevel(getSeverity());
    dcl::util::Logger.setDefaultSeverity(dcl::util::Severity::Info);

    /* Generate name of RTF log file */
    {
        std::stringstream ss;

        ss << "gcf_" << host << ".log";
        ss >> logFileName;
    }

    /* setup RTF logger */
    static std::ofstream rtfLogFile(logFileName.c_str());
    rtfutil::Logger::output(rtfLogFile);
    rtfutil::Logger::verbose(getLoggingLevel());

    /* Disable Nagle's algorithm on listening socket
     * Due to the RPC-style protocol of dOpenCL, short messages usually wait for
     * a response before the next message is send. Hence, waiting for another
     * short message to merge with the first one is pointless.
     * In particular, command status messages suffer from the high latency
     * introduced by Nagle's algorithm. */
    socketOptions.set("SOL_TCP", "TCP_NODELAY");

    /* Configure listening socket: the socket options are also set for incoming
     * connections. This seems to be more reliable that setting the options
     * using CommunicationManager::setOption which sends a message for setting
     * these options after the connection has been established. */
    gcf::CommunicationManagerProperties commProps(
            host, port,
            false /* disable UDP */,
            socketOptions);

    return new dclgcf::ComputeNodeConnectionManagerImpl(commProps);
}

} /* namespace dcl */
