/******************************************************************************
 * This file is part of dOpenCL.
 * 
 * dOpenCL is an implementation of the OpenCL application programming
 * interface for distributed systems. See <http://dopencl.uni-muenster.de/>
 * for more information.
 * 
 * Developed by: Research Group Parallel and Distributed Systems
 *               Department of Mathematics and Computer Science
 *               University of Muenster, Germany
 *               <http://pvs.uni-muenster.de/>
 * 
 * Copyright (C) 2013  Philipp Kegel <philipp.kegel@uni-muenster.de>
 *
 * dOpenCL is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * dOpenCL is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with dOpenCL. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Permission to use dOpenCL for scientific, non-commercial work is
 * granted under the terms of the dOpenCL Academic License provided
 * appropriate credit is given. See the dOpenCL Academic License for
 * more details.
 * 
 * You should have received a copy of the dOpenCL Academic License
 * along with dOpenCL. If not, see <http://dopencl.uni-muenster.de/>.
 ******************************************************************************/

/**
 * @file ConnectionManagerImpl.cpp
 *
 * @date 26.10.2011
 * @author Philipp Kegel
 */

#include <cassert>
#include <ostream>
#include <set>
#include <string>
#include <utility>
#include <vector>

#include <boost/shared_ptr.hpp>

#include <boost/thread/locks.hpp>
#include <boost/thread/mutex.hpp>
#include <boost/thread/recursive_mutex.hpp>

#include <boost/algorithm/string/trim.hpp>

#include <gcf/CallbackPolicy.h>
#include <gcf/CommunicationManager.h>
#include <gcf/CommunicationManagerProperties.h>
#include <gcf/ConnectionListener.h>
#include <gcf/GCFTypedefs.h>
#include <gcf/Process.h>
#include <gcf/Properties.h>
#include <gcf/ReplicateMappings.h>
#include <gcf/Serializable.h>
#include <gcf/SerializationModule.h>
#include <gcf/StreamCommunication.h>

#include <gcf/link/SocketOptions.h>

#include <gcf/transport/DataStream.h>
#include <gcf/transport/IODataStream.h>

#include <dcl/CLObjectRegistry.h>
#include <dcl/ConnectionListener.h>
#include <dcl/DCLException.h>
#include <dcl/DCLTypes.h>

#include <dcl/comm/BuildProgram.h>
#include <dcl/comm/CreateCommandQueue.h>
#include <dcl/comm/CreateBuffer.h>
#include <dcl/comm/CreateContext.h>
#include <dcl/comm/CreateKernel.h>
#include <dcl/comm/CreateKernelsInProgram.h>
#include <dcl/comm/CreateProgramWithSource.h>
#include <dcl/comm/ContextMessage.h>
#include <dcl/comm/DeleteMemory.h>
#include <dcl/comm/DeleteCommandQueue.h>
#include <dcl/comm/DeleteContext.h>
#include <dcl/comm/DeleteKernel.h>
#include <dcl/comm/DeleteProgram.h>
#include <dcl/comm/EnqueueCopy.h>
#include <dcl/comm/EnqueueMapMemObject.h>
#include <dcl/comm/EnqueueNDRangeKernel.h>
#include <dcl/comm/EnqueueReadBuffer.h>
#include <dcl/comm/EnqueueWaitForEvents.h>
#include <dcl/comm/EnqueueWriteBuffer.h>
#include <dcl/comm/ErrorResponse.h>
#include <dcl/comm/EventCommand.h>
#include <dcl/comm/EventSynchronizationMessage.h>
#include <dcl/comm/CommandMessage.h>
#include <dcl/comm/FlushFinishRequest.h>
#include <dcl/comm/GetKernelInfo.h>
#include <dcl/comm/InfoResponse.h>
#include <dcl/comm/ProgramMessage.h>
#include <dcl/comm/Request.h>
#include <dcl/comm/Response.h>
#include <dcl/comm/SetKernelArg.h>

#include <dcl/util/Logger.h>

#include "ComputeNodeImpl.h"
#include "ConnectionManagerImpl.h"
#include "HostImpl.h"
#include "ProcessImpl.h"

#include "comm/CLMessageDispatcher.h"
#include "comm/DeviceIDsResponse.h"
#include "comm/DeviceInfosResponse.h"
#include "comm/GetDeviceIDs.h"
#include "comm/GetDeviceInfos.h"
#include "comm/ProcessMessage.h"

namespace dclgcf {

/******************************************************************************
 *
 * Connection manager base implementation
 *
 ******************************************************************************/

void ConnectionManagerImpl::resolveURL(
		const std::string& url,
		std::string& hostName,
		gcf_uint16& port) {
	size_t delim = url.find(":");

	hostName = url.substr(0, delim);
	boost::trim(hostName); // remove leading and trailing white spaces
	if (delim != std::string::npos) {
		std::istringstream stream(url.substr(delim));
		gcf_uint16 defaultPort = port;

		stream.ignore(); // skip delimiter
		if ((stream >> port).fail()) {
			port = defaultPort;
		}
	}
}

ConnectionManagerImpl::ConnectionManagerImpl(
		const gcf::CommunicationManagerProperties& commProps) {
	try {
		gcf::Properties props;

		props.setProperty(&commProps);
		_communicationManager = new gcf::CommunicationManager(props);
	} catch (const gcf::GCFError& err) {
		throw dcl::ConnectionException(err.message);
	}

	/*
	 * Register all dOpenCL message types
	 * Thus, invalid messages can also be de-/serialized by processes (host,
	 * compute node, resource manager) such that they can report a protocol
	 * error to the application.
	 */
	gcf::SerializationModule& sm = _communicationManager->getSerializationModule();

	/*
	 * Register CL command requests
	 */
	dcl::comm::Request::registerSerializable(sm);
	dcl::comm::BuildProgram::registerSerializable(sm);
	dcl::comm::CreateBuffer::registerSerializable(sm);
	dcl::comm::CreateCommandQueue::registerSerializable(sm);
	dcl::comm::CreateContext::registerSerializable(sm);
	dcl::comm::CreateEvent::registerSerializable(sm);
	dcl::comm::CreateKernel::registerSerializable(sm);
	dcl::comm::CreateKernelsInProgram::registerSerializable(sm);
	dcl::comm::CreateProgramWithSource::registerSerializable(sm);
	dcl::comm::CreateUserEvent::registerSerializable(sm);
	dcl::comm::DeleteMemory::registerSerializable(sm);
	dcl::comm::DeleteContext::registerSerializable(sm);
	dcl::comm::DeleteCommandQueue::registerSerializable(sm);
	dcl::comm::DeleteEvent::registerSerializable(sm);
	dcl::comm::DeleteKernel::registerSerializable(sm);
	dcl::comm::DeleteProgram::registerSerializable(sm);
	dcl::comm::EnqueueBarrier::registerSerializable(sm);
	dcl::comm::EnqueueCopyBuffer::registerSerializable(sm);
	dcl::comm::EnqueueMarker::registerSerializable(sm);
    dcl::comm::EnqueueMapBuffer::registerSerializable(sm);
	dcl::comm::EnqueueNDRangeKernel::registerSerializable(sm);
	dcl::comm::EnqueueReadBuffer::registerSerializable(sm);
    dcl::comm::EnqueueUnmapMemObject::registerSerializable(sm);
	dcl::comm::EnqueueWaitForEvents::registerSerializable(sm);
	dcl::comm::EnqueueWriteBuffer::registerSerializable(sm);
	dcl::comm::FinishRequest::registerSerializable(sm);
	dcl::comm::FlushRequest::registerSerializable(sm);
	dcl::comm::GetDeviceIDs::registerSerializable(sm);
	dcl::comm::GetDeviceInfos::registerSerializable(sm);
	dcl::comm::GetEventProfilingInfo::registerSerializable(sm);
	dcl::comm::GetKernelInfo::registerSerializable(sm);
    dcl::comm::GetKernelWorkGroupInfo::registerSerializable(sm);
	dcl::comm::SetKernelArg::registerSerializable(sm);
	dcl::comm::SetKernelArgBinary::registerSerializable(sm);
	dcl::comm::SetKernelArgMemObject::registerSerializable(sm);
	dcl::comm::SetUserEventStatusRequest::registerSerializable(sm);

	/*
	 * Register CL command responses
	 */
	dcl::comm::Response::registerSerializable(sm);
	dcl::comm::DeviceIDsResponse::registerSerializable(sm);
	dcl::comm::DeviceInfosResponse::registerSerializable(sm);
	dcl::comm::ErrorResponse::registerSerializable(sm);
	dcl::comm::EventProfilingInfo::registerSerializable(sm);
	dcl::comm::InfoResponse::registerSerializable(sm);

	/*
	 * Register CL notifications
	 */
	dcl::comm::CommandExecutionStatusChangedMessage::registerSerializable(sm);
	dcl::comm::ContextErrorMessage::registerSerializable(sm);
    dcl::comm::ContextPrintfMessage::registerSerializable(sm);
	dcl::comm::ProgramBuildMessage::registerSerializable(sm);
	dcl::comm::EventSynchronizationMessage::registerSerializable(sm);

	/*
	 * Register process status messages
	 */
	dcl::comm::HostIdentificationMessage::registerSerializable(sm);
	dcl::comm::ComputeNodeIdentificationMessage::registerSerializable(sm);
	dcl::comm::ConnectionApprovedMessage::registerSerializable(sm);
}

ConnectionManagerImpl::~ConnectionManagerImpl() {
	delete _communicationManager;
}

void ConnectionManagerImpl::start() {
    std::vector<GCFTypeID_t> processMessageTypes;

    /* create list of process status message types*/
    processMessageTypes.push_back(dcl::comm::HostIdentificationMessage::TYPE);
    processMessageTypes.push_back(dcl::comm::ComputeNodeIdentificationMessage::TYPE);
    processMessageTypes.push_back(dcl::comm::ConnectionApprovedMessage::TYPE);

	_communicationManager->addMessageListener(*this,
	        gcf::CallbackPolicy(processMessageTypes, false));

	/* Set buffered to false to immediately register new sessions */
    _communicationManager->addConnectionListener(*this, false);
    _communicationManager->getStreamCommunication().addDataStreamListener(*this);

	_communicationManager->activateCommunication();
}

void ConnectionManagerImpl::stop() {
    _communicationManager->getStreamCommunication().removeDataStreamListener(*this);
	_communicationManager->removeConnectionListener(*this);

	_communicationManager->removeMessageListener(*this);

	_communicationManager->shutdownCommunication();
}

gcf::CommunicationManager& ConnectionManagerImpl::communicationManager() const {
    return *_communicationManager;
}

dcl::CLObjectRegistry& ConnectionManagerImpl::objectRegistry() {
    return _objectRegistry;
}

ProcessImpl * ConnectionManagerImpl::getProcess(
        const gcf::Process& process) const {
    return getComputeNode(process);
}

ComputeNodeImpl * ConnectionManagerImpl::getComputeNode(
        const gcf::Process& process) const {
    boost::lock_guard<boost::recursive_mutex> lock(_connectionsMutex);
    std::map<const gcf::Process *, ComputeNodeImpl *>::const_iterator i;

    /* search compute node list */
    i = _computeNodes.find(&process);
    if (i != _computeNodes.end()) return i->second;

    return 0;
}

#if 0
/*
 * Assume that all compute nodes have been introduced statically by a configuration file.
 */
void ConnectionManagerImpl::getComputeNodes(
        const std::vector<dcl_object_id>& computeNodeIds,
        std::vector<ComputeNodeImpl *>& computeNodes) {
    boost::lock_guard<boost::mutex> lock(_connectionsMutex);

    for (std::vector<dcl_object_id>::const_iterator id = computeNodeIds.begin();
            id != computeNodeIds.end(); ++id) {
        std::map<dcl_object_id, ComputeNodeImpl *>::const_iterator i;

        i = _computeNodes.find(*id);
        if (i == _computeNodes.end()) { // compute node not found
            throw dcl::ConnectionException("Invalid compute node ID");
        }
        computeNodes.push_back(i->second);
    }
}
#endif

/*
 * GCF connection listener API
 */

bool ConnectionManagerImpl::approveConnection(
		const std::string& subjectName) {
	/* This method is not only called for incoming process connections but also
	 * for outgoing connections. */
	return true;
}

void ConnectionManagerImpl::connectionEstablished(
		const gcf::Process& process, const std::string& subjectName) {
	/* TODO Distinguish incoming (host only) and outgoing (host, compute node, device manager) connections
	 * This method is not only called for incoming process connections but also
	 * for outgoing connections. */

	/* TODO Add incoming connections to list of pending connections
	 * Connections will not be used until an identification has been send by
	 * the remote process. To avoid idle connections, pending connections should
	 * be closed after a certain timeout. */
}

void ConnectionManagerImpl::connectionLost(const gcf::Process& process) {
	if (hostConnectionLost(process)) return;
	if (computeNodeConnectionLost(process)) return;

	dcl::util::Logger << dcl::util::Warning <<
			"Lost connection to unknown process" << std::endl;
}

gcf::ReplicateMappings ConnectionManagerImpl::incomingConnectionTransfer(
		const gcf::Process& host,
		const gcf::Process& computeNode,
		gcf::Serializable *transferMessage,
		std::vector<gcf::Serializable *>& connectedEntities,
		std::vector<gcf::Serializable *>& newEntities) {
	/* TODO Handle incoming connection transfers */
	return gcf::ReplicateMappings();
}

/*
 * GCF data stream listener API
 */

bool ConnectionManagerImpl::incomingDataStream(
        gcf::IODataStream &dataStream) {
    const gcf::Process *process = dataStream.getRemoteProcess();

    if (process) {
        /* The connections list must be locked, such that a node is not
         * destroyed while its associated data stream is changed */
        boost::lock_guard<boost::recursive_mutex> lock(_connectionsMutex);
        /* lookup process' associated dOpenCL node */
        ProcessImpl *node = getProcess(*process);

        if (node) {
            node->setDataStream(&dataStream);
            /* TODO Log node type ('host' or 'compute node') */
            dcl::util::Logger << dcl::util::Info
                    << "Incoming data stream from node '" << node->url() << '\''
                    << std::endl;

            return true;
        } else {
            dcl::util::Logger << dcl::util::Warning
                    << "Incoming data stream from unknown node" << std::endl;
        }
    } else {
        dcl::util::Logger << dcl::util::Error
                << "No process for data stream" << std::endl;
    }

    return false;
}

/* FIXME Ensure that closed data streams are detected
 * This method is not called when a data stream is closed by the remote endpoint
 * while no data transfer is pending (see implementation in
 * gcf::DataStreamProcessor). Hence, ProcessImpl::setDataStream(0) is called
 * when the process connection is lost. */
void ConnectionManagerImpl::closedDataStream(
        gcf::DataStream& dataStream) {
    const gcf::Process *process = dataStream.getRemoteProcess();

    if (process) {
        /* The connections list must be locked, such that a node is not
         * destroyed while its associated data stream is changed */
        boost::lock_guard<boost::recursive_mutex> lock(_connectionsMutex);
        /* lookup process' associated dOpenCL node */
        ProcessImpl *node = getProcess(*process);

        if (node) { // node found
            node->setDataStream(0);
            /* TODO Log node type ('host' or 'compute node') */
            dcl::util::Logger << dcl::util::Info
                    << "Closed data stream from node '" << node->url() << '\''
                    << std::endl;
        } else {
            dcl::util::Logger << dcl::util::Warning
                    << "Closed data stream from unknown node" << std::endl;
        }
    } else {
        dcl::util::Logger << dcl::util::Error
                << "No process for data stream" << std::endl;
    }
}

/*
 * GCF message listener API
 */

void ConnectionManagerImpl::dispatch(
		gcf::Serializable& message, const gcf::Process& process) {
	switch (message.getType()) {

	/*
	 * Process identification message
	 */
	case dcl::comm::HostIdentificationMessage::TYPE: // process is host
		hostConnectionEstablished(process);
		break;
	case dcl::comm::ComputeNodeIdentificationMessage::TYPE: // process is compute node
		computeNodeConnectionEstablished(process);
		break;

	/*
	 * Process status messages
	 */
	case dcl::comm::ConnectionApprovedMessage::TYPE:
		connectionApproved(process);
		break;

	default: /* invalid message */
		dcl::util::Logger << dcl::util::Error
				<< "Received message of unknown type" << std::endl;
		break;
	}
}

void ConnectionManagerImpl::hostConnectionEstablished(const gcf::Process&) {
	/* ignore process */
	dcl::util::Logger << dcl::util::Warning
			<< "Ignoring incoming host connection" << std::endl;
}

void ConnectionManagerImpl::computeNodeConnectionEstablished(const gcf::Process&) {
	/* ignore process */
	dcl::util::Logger << dcl::util::Warning
			<< "Ignoring incoming compute node connection" << std::endl;
}

bool ConnectionManagerImpl::hostConnectionLost(const gcf::Process&) {
	return false;
}

bool ConnectionManagerImpl::computeNodeConnectionLost(const gcf::Process&) {
	return false;
}

void ConnectionManagerImpl::connectionApproved(
        const gcf::Process& process) const {
    ComputeNodeImpl *computeNode = getComputeNode(process);
    if (computeNode) {
        computeNode->setConnectionStatus(ComputeNodeImpl::CONNECTED); // update compute node connection status
    } else {
        dcl::util::Logger << dcl::util::Error
                << "Received connection approval from unknown process" << std::endl;
    }
}

} /* namespace dclgcf */
