/******************************************************************************
 * This file is part of dOpenCL.
 * 
 * dOpenCL is an implementation of the OpenCL application programming
 * interface for distributed systems. See <http://dopencl.uni-muenster.de/>
 * for more information.
 * 
 * Developed by: Research Group Parallel and Distributed Systems
 *               Department of Mathematics and Computer Science
 *               University of Muenster, Germany
 *               <http://pvs.uni-muenster.de/>
 * 
 * Copyright (C) 2013  Philipp Kegel <philipp.kegel@uni-muenster.de>
 *
 * dOpenCL is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * dOpenCL is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with dOpenCL. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Permission to use dOpenCL for scientific, non-commercial work is
 * granted under the terms of the dOpenCL Academic License provided
 * appropriate credit is given. See the dOpenCL Academic License for
 * more details.
 * 
 * You should have received a copy of the dOpenCL Academic License
 * along with dOpenCL. If not, see <http://dopencl.uni-muenster.de/>.
 ******************************************************************************/

/*
 * DataTransferImpl.cpp
 *
 *  Created on: 17.12.2011
 *      Author: Philipp Kegel
 */

#include <cassert>
#include <cstddef>
#include <ostream>
#include <vector>

#include <boost/function.hpp>
#include <boost/shared_ptr.hpp>

#include <boost/thread/locks.hpp>
#include <boost/thread/mutex.hpp>
#include <boost/thread/xtime.hpp>

#ifdef __APPLE__
#include <OpenCL/cl.h>
#include <OpenCL/cl_wwu_dcl.h>
#else
#include <CL/cl.h>
#include <CL/cl_wwu_dcl.h>
#endif

#include <gcf/transport/IncomingDataStream.h>
#include <gcf/transport/OutgoingDataStream.h>

#include <dcl/DataTransfer.h>
#include <dcl/DCLException.h>

#include <dcl/util/Logger.h>

#include "DataTransferImpl.h"

namespace dclgcf {

/******************************************************************************
 * Data transfer
 ******************************************************************************/

DataTransferImpl::DataTransferImpl() {
}

void DataTransferImpl::setCallback(
        const boost::function<void (cl_int)>& notify) {
    boost::lock_guard<boost::mutex> lock(_callbacksMutex);

    if (isComplete()) { // data transfer is already finished
        /* The data dispatcher has already completed the data transfer but it is
         * undefined whether the callback list has already been processed or not.
         * Hence, if a new callback is added to the callback list it might
         * eventually be called or not.
         * Adding the new callback to the callback list and calling it might
         * result in a redundant execution of the callback.
         * Therefore, call the new callback immediately without adding it to the
         * callback list to ensure that it is only called once. */
        notify(status());
    } else {
        /* Since we obtained the lock for the callback list while the data
         * transfer was still in progress we can be sure that the data
         * dispatcher will not process the callback list before we added the new
         * callback. */
        _callbacks.push_back(notify);
    }
}

void DataTransferImpl::abort() {
    /* TODO Implement DataTransferImpl::abort */
}

void DataTransferImpl::triggerCallbacks() {
    boost::lock_guard<boost::mutex> lock(_callbacksMutex);
    for(std::vector<boost::function<void (cl_int)> >::const_iterator i = _callbacks.begin();
            i != _callbacks.end(); ++i) {
        (*i)(status());
    }
}

/******************************************************************************
 * Data receipt
 ******************************************************************************/

DataReceipt::DataReceipt(
		const boost::shared_ptr<const gcf::IncomingDataStream::Status>& status) :
	_status(status) { }

unsigned long DataReceipt::submit() const {
	boost::xtime time = _status->submit();
	return time.sec * 1000000000 + time.nsec;
}

unsigned long DataReceipt::start() const {
	boost::xtime time = _status->start();
	return time.sec * 1000000000 + time.nsec;
}

unsigned long DataReceipt::end() const {
	boost::xtime time = _status->end();
	return time.sec * 1000000000 + time.nsec;
}

bool DataReceipt::isComplete() const {
	return (_status->hasFinished() || _status->hasFailed());
}

void DataReceipt::wait() const {
	if (!_status->waitToLength(_status->getLength()) /* wait for end of receipt */
			|| _status->hasFailed()) {
		throw dcl::IOException("Data receipt failed");
	}
}

void DataReceipt::execute() {
	try {
		double latency;
		double durance;
		double bandwidth;

		wait();

		latency = static_cast<double>(start() - submit()) / 1000000.0;
		durance = static_cast<double>(end()   - start() ) / 1000000000.0;
		bandwidth = (_status->getReceivedLength() / static_cast<double>(1024 * 1024)) / durance;

		dcl::util::Logger << dcl::util::Debug
				<< "Received " << _status->getReceivedLength() << " bytes\n"
				<< "\tlatency: " << latency << " ms, bandwidth: " << bandwidth << " MB/s" << std::endl;
	} catch (const dcl::IOException& err) {
		dcl::util::Logger << dcl::util::Error << err.what() << std::endl;
	}

	triggerCallbacks();
}

cl_int DataReceipt::status() const {
	if (_status->hasFailed()) return CL_IO_ERROR_WWU;
	if (_status->hasFinished()) return CL_COMPLETE;
	return CL_RUNNING;
}

/******************************************************************************
 * Data sending
 ******************************************************************************/

DataSending::DataSending(
		const boost::shared_ptr<const gcf::OutgoingDataStream::Status>& status) :
	_status(status) { }

unsigned long DataSending::submit() const {
	boost::xtime time = _status->submit();
	return time.sec * 1000000000 + time.nsec;
}

unsigned long DataSending::start() const {
	boost::xtime time = _status->start();
	return time.sec * 1000000000 + time.nsec;
}

unsigned long DataSending::end() const {
	boost::xtime time = _status->end();
	return time.sec * 1000000000 + time.nsec;
}

bool DataSending::isComplete() const {
	return (_status->hasFinished() || _status->hasFailed());
}

void DataSending::wait() const {
	if (!_status->waitToLength(_status->getLength()) /* wait for end of receipt */
			|| _status->hasFailed()) {
		throw dcl::IOException("Data sending failed");
	}
}

void DataSending::execute() {
	try {
		double latency;
		double durance;
		double bandwidth;

		wait();

		latency = static_cast<double>(start() - submit()) / 1000000.0;
		durance = static_cast<double>(end()   - start() ) / 1000000000.0;
		bandwidth = (_status->getSentLength() / static_cast<double>(1024 * 1024)) / durance;

		dcl::util::Logger << dcl::util::Debug
				<< "Sent " << _status->getSentLength() << " bytes\n"
				<< "\tlatency: " << latency << " ms, bandwidth: " << bandwidth << " MB/s" << std::endl;
	} catch (const dcl::IOException& err) {
		dcl::util::Logger << dcl::util::Error << err.what() << std::endl;
	}

	triggerCallbacks();
}

cl_int DataSending::status() const {
	if (_status->hasFailed()) return CL_IO_ERROR_WWU;
	if (_status->hasFinished()) return CL_COMPLETE;
	return CL_RUNNING;
}

} /* namespace dclgcf */
