/******************************************************************************
 * This file is part of dOpenCL.
 * 
 * dOpenCL is an implementation of the OpenCL application programming
 * interface for distributed systems. See <http://dopencl.uni-muenster.de/>
 * for more information.
 * 
 * Developed by: Research Group Parallel and Distributed Systems
 *               Department of Mathematics and Computer Science
 *               University of Muenster, Germany
 *               <http://pvs.uni-muenster.de/>
 * 
 * Copyright (C) 2013  Philipp Kegel <philipp.kegel@uni-muenster.de>
 *
 * dOpenCL is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * dOpenCL is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with dOpenCL. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Permission to use dOpenCL for scientific, non-commercial work is
 * granted under the terms of the dOpenCL Academic License provided
 * appropriate credit is given. See the dOpenCL Academic License for
 * more details.
 * 
 * You should have received a copy of the dOpenCL Academic License
 * along with dOpenCL. If not, see <http://dopencl.uni-muenster.de/>.
 ******************************************************************************/

/**
 * @file    ProcessImpl.cpp
 *
 * @date    18.03.2012
 * @author  Philipp Kegel
 */

#include <algorithm>
#include <cstddef>
#include <string>
#include <sstream>

#include <boost/shared_ptr.hpp>

#include <boost/thread/condition_variable.hpp>
#include <boost/thread/locks.hpp>
#include <boost/thread/recursive_mutex.hpp>

#include <gcf/CommunicationManager.h>
#include <gcf/Process.h>
#include <gcf/StreamCommunication.h>

#include <gcf/GCFError.h>

#include <gcf/transport/IODataStream.h>

#include <dcl/DataTransfer.h>
#include <dcl/DCLException.h>
#include <dcl/DCLTypes.h>
#include <dcl/Remote.h>

#include "DataTransferImpl.h"
#include "ProcessImpl.h"

#include "comm/DataDispatcher.h"

namespace dclgcf {

ProcessImpl::ProcessImpl(
        gcf::CommunicationManager& communicationManager,
        const gcf::Process& process) :
        dcl::Remote(dcl::Remote::generateId()), _communicationManager(communicationManager),
        _process(process), _dataStream(0),
        /* The process' connection status is derived from the GCF process'
         * connection status. If the GCF process is connected, this process will
         * only be created after its type has been determined, such that this
         * process can be considered connected. */
        /* TODO Set connection status to DISCONNECTED or PROCESS_CONNECTED */
        _connectionStatus(communicationManager.isConnected(process) ? CONNECTED : DISCONNECTED) {
}

ProcessImpl::~ProcessImpl() {
}

void ProcessImpl::disconnect() {
    gcf::DataStream *dataStream;

    {
        boost::lock_guard<boost::recursive_mutex> lock(_connectionStatusMutex);
        /* detach data stream from process */
        dataStream = _dataStream;
        _dataStream = 0;
        /* disconnect process */
        _communicationManager.disconnectProcess(_process);
        _connectionStatus = DISCONNECTED;
        _connectionStatusChanged.notify_all();
    }

    /* destroy data stream */
    if (dataStream) {
        /* FIXME GCF data streams should be usable as a black box, without bothering about locking mechanisms */
        /* Avoid possible deadlock by unlocking ProcessImpl *before* destroying
         * the data stream:
         * If a process loses connection ConnectionManager::connectionLost and
         * ComputeNodeConnectionManager::closedDataStream are called
         * simultaneously. While connectionLost calls Process::disconnect
         * holding a lock on Process and trying to acquire a lock on
         * DataStreamProcessor for destroying the data stream, closedDataStream
         * holds a lock on DataStreamProcessor and tries to acquire a lock
         * on Process for calling calls Process::setDataStream. */
        _communicationManager.getStreamCommunication().destroyDataStream(*dataStream);
    }
}

bool ProcessImpl::isConnected() {
    boost::lock_guard<boost::recursive_mutex> lock(_connectionStatusMutex);
    return _connectionStatus == CONNECTED;
}

const std::string& ProcessImpl::url(void) const {
    if (_url.empty()) {
        std::stringstream ss;

        /* create URL string once */
        if (!(ss << _process.getHostname() << ':' << _process.getPort()) ||
                !(ss >> _url) ||
                !((ss >> std::ws).eof())) // skip whitespace and check if EOF flag has been set
            _url.clear();
    }

    return _url;
}

void ProcessImpl::sendMessage(gcf::Serializable& message) const {
    /* gcf::CommunicationManager::send never fails */
    if (!_communicationManager.isConnected(_process)) {
        throw dcl::IOException("No connection to process at " + url());
    }
    _communicationManager.send(_process, message);
}

boost::shared_ptr<dcl::DataTransfer> ProcessImpl::sendData(size_t size,
        const void *ptr) {
//    boost::lock_guard<boost::recursive_mutex> lock(_connectionStatusMutex);
//    if (!_dataStream) throw dcl::IOException("No data stream to process at " + url());
//    return boost::static_pointer_cast<dcl::DataTransfer>(
//            _dataDispatcher.sendData(*_dataStream, size, ptr));
    return boost::static_pointer_cast<dcl::DataTransfer>(
            _dataDispatcher.sendData(dataStream(), size, ptr));
}

boost::shared_ptr<dcl::DataTransfer> ProcessImpl::receiveData(size_t size,
        void *ptr) {
//    boost::lock_guard<boost::recursive_mutex> lock(_connectionStatusMutex);
//    if (!_dataStream) throw dcl::IOException("No data stream to process at " + url());
//    return boost::static_pointer_cast<dcl::DataTransfer>(
//            _dataDispatcher.receiveData(*_dataStream, size, ptr));
    return boost::static_pointer_cast<dcl::DataTransfer>(
            _dataDispatcher.receiveData(dataStream(), size, ptr));
}

const gcf::Process& ProcessImpl::process() const {
    return _process;
}

gcf::IODataStream& ProcessImpl::dataStream() {
    boost::lock_guard<boost::recursive_mutex> lock(_connectionStatusMutex);
    while (!_dataStream) {
        /* TODO Use timed wait when waiting for data stream;
         * throw dcl::ConnectionException after timeout.
         * This method is deadlock-prone. However, without it (see sendData and
         * receiveData) the data stream sometimes is not ready when sending or
         * receiving data, in particular when using a large number (>=32) of
         * compute nodes. */
        _dataStreamReady.wait(_connectionStatusMutex);
    }
    return *_dataStream;
}

void ProcessImpl::setConnectionStatus(ProcessConnectionStatus status) {
    boost::lock_guard<boost::recursive_mutex> lock(_connectionStatusMutex);
    _connectionStatus = status;
    _connectionStatusChanged.notify_all();
}

void ProcessImpl::setDataStream(gcf::IODataStream *dataStream) {
    {
        boost::lock_guard<boost::recursive_mutex> lock(_connectionStatusMutex);
        std::swap(_dataStream, dataStream);
        if (_dataStream) _dataStreamReady.notify_all();
    }

    /* destroy old data stream */
    if (dataStream) {
        _communicationManager.getStreamCommunication().destroyDataStream(*dataStream);
    }
}

} /* namespace dclgcf */
