/******************************************************************************
 * This file is part of dOpenCL.
 * 
 * dOpenCL is an implementation of the OpenCL application programming
 * interface for distributed systems. See <http://dopencl.uni-muenster.de/>
 * for more information.
 * 
 * Developed by: Research Group Parallel and Distributed Systems
 *               Department of Mathematics and Computer Science
 *               University of Muenster, Germany
 *               <http://pvs.uni-muenster.de/>
 * 
 * Copyright (C) 2013  Philipp Kegel <philipp.kegel@uni-muenster.de>
 *
 * dOpenCL is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * dOpenCL is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with dOpenCL. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Permission to use dOpenCL for scientific, non-commercial work is
 * granted under the terms of the dOpenCL Academic License provided
 * appropriate credit is given. See the dOpenCL Academic License for
 * more details.
 * 
 * You should have received a copy of the dOpenCL Academic License
 * along with dOpenCL. If not, see <http://dopencl.uni-muenster.de/>.
 ******************************************************************************/

/*
 * @file CLMessageDispatcher.cpp
 *
 * @date 2011-12-16
 * @author Philipp Kegel
 * @author Sebastian Pribnow
 */

#include <map>
#include <ostream>
#include <string>
#include <utility>
#include <vector>

#include <boost/thread/thread.hpp>

#ifdef __APPLE__
#include <OpenCL/cl.h>
#else
#include <CL/cl.h>
#endif

#include <gcf/CallbackPolicy.h>
#include <gcf/GCFTypedefs.h>

#include <dcl/Binary.h>
#include <dcl/BlockingQueue.h>
#include <dcl/CLObjectRegistry.h>
#include <dcl/ContextListener.h>
#include <dcl/DCLTypes.h>
#include <dcl/Device.h>
#include <dcl/EventListener.h>
#include <dcl/CommandListener.h>
#include <dcl/Process.h>
#include <dcl/ProgramBuildListener.h>

#include <dcl/comm/ContextMessage.h>
#include <dcl/comm/CommandMessage.h>
#include <dcl/comm/EventSynchronizationMessage.h>
#include <dcl/comm/ProgramMessage.h>

#include <dcl/util/Logger.h>

#include "../ComputeNodeImpl.h"
#include "../ConnectionManagerImpl.h"
#include "../HostImpl.h"

#include "CLMessageDispatcher.h"

namespace dclgcf {

namespace comm {

CLMessageDispatcher::MessageTypes::MessageTypes() {
    /* add context message types */
    push_back(dcl::comm::ContextErrorMessage::TYPE);
    push_back(dcl::comm::ContextPrintfMessage::TYPE);
    /* add event message types */
    push_back(dcl::comm::CommandExecutionStatusChangedMessage::TYPE);
    /* add program build message types */
    push_back(dcl::comm::ProgramBuildMessage::TYPE);
    /* add memory consistency message types */
    push_back(dcl::comm::EventSynchronizationMessage::TYPE);
}

/******************************************************************************/

const gcf::CallbackPolicy& CLMessageDispatcher::callbackPolicy() {
    static gcf::CallbackPolicy policy(
            MessageTypes(),
            false /* do not exclude but include specified types */);
    return policy;
}

CLMessageDispatcher::CLMessageDispatcher(dcl::CLObjectRegistry& objectRegistry) :
        _objectRegistry(objectRegistry) { }

void CLMessageDispatcher::contextError(dcl::dcl_object_id contextId,
        const std::string& errorInfo, const dcl::Binary& privateInfo) const {
    dcl::ContextListener *contextListener = _objectRegistry.lookup<dcl::ContextListener>(contextId);
    if (contextListener) {
        contextListener->onError(errorInfo, privateInfo.value(), privateInfo.size());
    } else {
        dcl::util::Logger << dcl::util::Error
                << "Context listener not found (ID=" << contextId
                << ')' << std::endl;
    }
}

void CLMessageDispatcher::contextPrintf(
        dcl::dcl_object_id contextId,
        const std::string& printfData) const {
    dcl::ContextListener *contextListener = _objectRegistry.lookup<dcl::ContextListener>(contextId);
    if (contextListener) {
        contextListener->onPrintf(printfData);
    } else {
        dcl::util::Logger << dcl::util::Error
                << "Context listener not found (ID=" << contextId
                << ')' << std::endl;
    }
}

void CLMessageDispatcher::executionStatusChanged(dcl::dcl_object_id commandId,
        cl_int executionStatus) const {
    dcl::CommandListener *commandListener = _objectRegistry.lookup<dcl::CommandListener>(commandId);
    if (commandListener) {
        commandListener->onExecutionStatusChanged(executionStatus);
    } else {
        dcl::util::Logger << dcl::util::Error
                << "Command listener not found (ID=" << commandId
                << ')' << std::endl;
    }
}

void CLMessageDispatcher::programBuildComplete(
        dcl::dcl_object_id programBuildId,
        const std::vector<dcl::dcl_object_id>& deviceIds,
        const std::vector<cl_build_status>& buildStatus) const {
    dcl::ProgramBuildListener *programBuildListener = _objectRegistry.lookup<dcl::ProgramBuildListener>(programBuildId);
    if (programBuildListener) {
        programBuildListener->onComplete(deviceIds, buildStatus);
    } else {
        dcl::util::Logger << dcl::util::Error
                << "Program build listener not found (ID=" << programBuildId
                << ')' << std::endl;
    }
}

void CLMessageDispatcher::synchronizeEvent(
        dcl::dcl_object_id eventId,
        dcl::Process& process) const {
    dcl::EventListener *eventListener = _objectRegistry.lookup<dcl::EventListener>(eventId);
    if (eventListener) {
        eventListener->onSynchronize(process);
    } else {
        dcl::util::Logger << dcl::util::Error
                << "Event listener not found (ID=" << eventId
                << ')' << std::endl;
    }
}

/*
 * GCF message listener API
 */

void CLMessageDispatcher::dispatch(gcf::Serializable& message,
        const gcf::Process& process) {
    switch (message.getType()) {

    /*
     * Asynchronous notification messages
     * Messages must be copied as they will be deleted by the GCF when dispatch
     * returns.
     */
    case dcl::comm::ContextErrorMessage::TYPE:
    {
        dcl::util::Logger << dcl::util::Debug
                << "Received context error message" << std::endl;

        dcl::comm::ContextErrorMessage& notification =
                static_cast<dcl::comm::ContextErrorMessage&>(message);
        contextError(notification.contextId(), notification.errorInfo(),
                notification.privateInfo());
        break;
    }
    case dcl::comm::ContextPrintfMessage::TYPE:
    {
        dcl::util::Logger << dcl::util::Debug
                << "Received context printf message" << std::endl;

        dcl::comm::ContextPrintfMessage& notification =
                static_cast<dcl::comm::ContextPrintfMessage&>(message);
        contextPrintf(notification.contextId(), notification.printfData());
        break;
    }

    case dcl::comm::CommandExecutionStatusChangedMessage::TYPE:
    {
        dcl::util::Logger << dcl::util::Debug
                << "Received command execution status changed message" << std::endl;

        dcl::comm::CommandExecutionStatusChangedMessage& notification =
                static_cast<dcl::comm::CommandExecutionStatusChangedMessage&>(message);
        executionStatusChanged(notification.commandId(), notification.status());
        break;
    }

    case dcl::comm::ProgramBuildMessage::TYPE:
    {
        std::vector<cl_build_status> buildStatus;

        dcl::util::Logger << dcl::util::Debug
                << "Received program build message" << std::endl;

        dcl::comm::ProgramBuildMessage& notification =
                static_cast<dcl::comm::ProgramBuildMessage&>(message);
        notification.getBuildStatus(buildStatus);
        programBuildComplete(notification.programBuildId(),
                notification.deviceIds(), buildStatus);
        break;
    }

    default: /* invalid message */
        dcl::util::Logger << dcl::util::Error
                << "Received unknown message" << std::endl;
        break;
    }
}

/******************************************************************************
 * Host side message dispatcher
 ******************************************************************************/

CLComputeNodeMessageDispatcher::CLComputeNodeMessageDispatcher(
        const dclgcf::ConnectionManagerImpl& connectionManager,
        dcl::CLObjectRegistry& objectRegistry) :
    CLMessageDispatcher(objectRegistry), _connectionManager(connectionManager) {
    /* Message dispatcher must be initialized before a thread is created */
    _thread = boost::thread(&CLComputeNodeMessageDispatcher::run, this);
}

CLComputeNodeMessageDispatcher::~CLComputeNodeMessageDispatcher() {
    /* interrupt blocking queue to stop thread */
    _executionStatusMessages.interrupt();
    if (_thread.joinable()) _thread.join();
}

void CLComputeNodeMessageDispatcher::run() {
    try {
        /* Wait for and dispatch execution status changes */
        while (true) {
            const std::pair<dcl::dcl_object_id, cl_int>& executionStatus = _executionStatusMessages.front();
            executionStatusChanged(executionStatus.first, executionStatus.second);
            _executionStatusMessages.pop();
        }
    } catch (const boost::thread_interrupted&) {
        /* task queue has been interrupted */
    }
}

void CLComputeNodeMessageDispatcher::dispatch(gcf::Serializable& message,
        const gcf::Process& process) {
    switch (message.getType()) {
    case dcl::comm::CommandExecutionStatusChangedMessage::TYPE:
    {
        dcl::util::Logger << dcl::util::Debug
                << "Received command execution status changed message from compute node" << std::endl;

        dcl::comm::CommandExecutionStatusChangedMessage& notification =
                static_cast<dcl::comm::CommandExecutionStatusChangedMessage&>(message);
        _executionStatusMessages.push(std::make_pair(notification.commandId(), notification.status()));
        break;
    }

    case dcl::comm::EventSynchronizationMessage::TYPE:
    {
        dcl::util::Logger << dcl::util::Debug
                << "Received event synchronization message from compute node" << std::endl;

        dcl::comm::EventSynchronizationMessage& notification =
                static_cast<dcl::comm::EventSynchronizationMessage&>(message);
        ComputeNodeImpl *computeNode = _connectionManager.getComputeNode(process);
        assert(computeNode != NULL);
        synchronizeEvent(notification.eventId(), *computeNode);
        break;
    }

    default:
        CLMessageDispatcher::dispatch(message, process);
    }
}

/******************************************************************************
 * Compute node side message dispatcher
 ******************************************************************************/

CLHostMessageDispatcher::CLHostMessageDispatcher(HostImpl *host) :
    CLMessageDispatcher(host->objectRegistry()), _host(*host) { }

/*
 * GCF message listener API
 */

void CLHostMessageDispatcher::dispatch(gcf::Serializable& message,
        const gcf::Process& process) {
    if (process != _host.process()) return; // discard message
    
    switch (message.getType()) {
    case dcl::comm::EventSynchronizationMessage::TYPE:
    {
        dcl::util::Logger << dcl::util::Debug
                << "Received event synchronization message from host" << std::endl;

        dcl::comm::EventSynchronizationMessage& notification =
                static_cast<dcl::comm::EventSynchronizationMessage&>(message);
        synchronizeEvent(notification.eventId(), _host);
        break;
    }

    default:
        CLMessageDispatcher::dispatch(message, process);
    }
}

} /* namespace comm */

} /* namespace dclgcf */
