/******************************************************************************
 * This file is part of dOpenCL.
 * 
 * dOpenCL is an implementation of the OpenCL application programming
 * interface for distributed systems. See <http://dopencl.uni-muenster.de/>
 * for more information.
 * 
 * Developed by: Research Group Parallel and Distributed Systems
 *               Department of Mathematics and Computer Science
 *               University of Muenster, Germany
 *               <http://pvs.uni-muenster.de/>
 * 
 * Copyright (C) 2013  Philipp Kegel <philipp.kegel@uni-muenster.de>
 *
 * dOpenCL is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * dOpenCL is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with dOpenCL. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Permission to use dOpenCL for scientific, non-commercial work is
 * granted under the terms of the dOpenCL Academic License provided
 * appropriate credit is given. See the dOpenCL Academic License for
 * more details.
 * 
 * You should have received a copy of the dOpenCL Academic License
 * along with dOpenCL. If not, see <http://dopencl.uni-muenster.de/>.
 ******************************************************************************/

/*
 * @file CLMessageDispatcher.h
 *
 * @date 2011-12-16
 * @author Philipp Kegel
 * @author Sebastian Pribnow
 */

#ifndef CLMESSAGEDISPATCHER_H_
#define CLMESSAGEDISPATCHER_H_

#include <string>
#include <utility>
#include <vector>

#include <boost/thread/thread.hpp>

#ifdef __APPLE__
#include <OpenCL/cl.h>
#else
#include <CL/cl.h>
#endif

#include <gcf/CallbackPolicy.h>
#include <gcf/GCFTypedefs.h>
#include <gcf/MessageListener.h>
#include <gcf/Process.h>
#include <gcf/Serializable.h>

#include <dcl/Binary.h>
#include <dcl/BlockingQueue.h>
#include <dcl/CLObjectRegistry.h>
#include <dcl/ContextListener.h>
#include <dcl/DCLTypes.h>
#include <dcl/ProgramBuildListener.h>
#include <dcl/EventListener.h>

#include "../ConnectionManagerImpl.h"

namespace dclgcf {

/* forward declaration */
class HostImpl;

/******************************************************************************/

namespace comm {

/**
 * @brief A dispatch for incoming application level notification messages.
 */
class CLMessageDispatcher: public gcf::MessageListener {
public:
    /**
     * @brief Returns a GCF callback policy comprising the message types processed by a message dispatcher.
     *
     * @return a callback policy
     */
    static const gcf::CallbackPolicy& callbackPolicy();


    CLMessageDispatcher(
            dcl::CLObjectRegistry&  /* objectRegistry */);
	virtual ~CLMessageDispatcher() { }

    /*
     * GCF message listener API
     */
    virtual void dispatch(
            gcf::Serializable&  /* message */,
            const gcf::Process& /* process */);

protected:
    /**
     * @brief A list of dOpenCL message types.
     *
     * This class is used internally for static initialization of a GCF callback
     * policy.
     */
    class MessageTypes : public std::vector<GCFTypeID_t> {
    public:
        MessageTypes();
    };


    void contextError(
            dcl::dcl_object_id  /* contextId */,
            const std::string&  /* errorInfo */,
            const dcl::Binary&  /* privateInfo */) const;

    void contextPrintf(
            dcl::dcl_object_id  /* contextId */,
            const std::string&  /* printfData */) const;

    void executionStatusChanged(
            dcl::dcl_object_id  /* commandId */,
            cl_int              /* executionStatus */) const;

    void programBuildComplete(
            dcl::dcl_object_id                      /* programBuildId */,
            const std::vector<dcl::dcl_object_id>&  /* deviceIds */,
            const std::vector<cl_build_status>&     /* buildStatus */) const;

    /**
     * @brief Callback for an incoming event synchronization request
     *
     * @param[in]  eventId  the ID of the event to synchronize
     * @param[in]  process  the process that requested the synchronization
     */
    void synchronizeEvent(
            dcl::dcl_object_id  eventId,
            dcl::Process&       process) const;

    dcl::CLObjectRegistry& _objectRegistry; /**< Registry for application objects */
};

/******************************************************************************/

/**
 * @brief A dispatch for incoming messages from compute nodes.
 *
 * This dispatcher is used on the host!
 *
 * This class should become the only implementation of a message dispatcher.
 * Due to implementation issues of ID generation (each host uses its own ID
 * range) it currently has to be distinguished from message dispatchers used on
 * the compute node side.
 *
 * @see CLHostMessageDispatcher
 */
class CLComputeNodeMessageDispatcher: public CLMessageDispatcher {
public:
    CLComputeNodeMessageDispatcher(
            const dclgcf::ConnectionManagerImpl& connectionManager,
            dcl::CLObjectRegistry&               objectRegistry);
    virtual ~CLComputeNodeMessageDispatcher();

    void run();

    /*
     * GCF message listener API
     */
    void dispatch(
            gcf::Serializable&  /* message */,
            const gcf::Process& /* process */);

private:
    dcl::BlockingQueue<std::pair<dcl::dcl_object_id, cl_int> > _executionStatusMessages;

    boost::thread _thread;
    const dclgcf::ConnectionManagerImpl& _connectionManager;
};

/******************************************************************************/

/**
 * @brief A dispatch for incoming messages from hosts.
 *
 * This dispatcher is used on compute nodes!
 *
 * Messages from hosts have to be dispatched differently as the ID within these
 * messages refers to a particular host and, thus, are not unique if multiple
 * hosts are connected to a compute nodes.
 * Hence, a different message dispatcher is required for each host that only
 * processes messages it received from this associated host.
 *
 * In future version of dOpenCL UUID could be used to assign globally unique IDs
 * to objects, such that a message dispatcher does not have to consider the
 * message source to resolve these IDs. Then, CLHostMessageDispatcher and
 * CLComputeNodeMessageDispatcher can be unified.
 */
class CLHostMessageDispatcher: public CLMessageDispatcher {
public:
    CLHostMessageDispatcher(
            HostImpl *  /* host */);

    /*
     * GCF message listener API
     */
    void dispatch(
            gcf::Serializable&  /* message */,
            const gcf::Process& /* process */);

private:
    HostImpl& _host;
};

} /* namespace comm */

} /* namespace dclgcf */

#endif /* CLMESSAGEDISPATCHER_H_ */
