/******************************************************************************
 * This file is part of dOpenCL.
 * 
 * dOpenCL is an implementation of the OpenCL application programming
 * interface for distributed systems. See <http://dopencl.uni-muenster.de/>
 * for more information.
 * 
 * Developed by: Research Group Parallel and Distributed Systems
 *               Department of Mathematics and Computer Science
 *               University of Muenster, Germany
 *               <http://pvs.uni-muenster.de/>
 * 
 * Copyright (C) 2013  Philipp Kegel <philipp.kegel@uni-muenster.de>
 *
 * dOpenCL is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * dOpenCL is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with dOpenCL. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Permission to use dOpenCL for scientific, non-commercial work is
 * granted under the terms of the dOpenCL Academic License provided
 * appropriate credit is given. See the dOpenCL Academic License for
 * more details.
 * 
 * You should have received a copy of the dOpenCL Academic License
 * along with dOpenCL. If not, see <http://dopencl.uni-muenster.de/>.
 ******************************************************************************/

/**
 * @file    CLRequestDispatcher.cpp
 *
 * @date    31.10.2011
 * @author  Philipp Kegel
 */

#include <cassert>
#include <cstddef>
#include <cstdlib>
#include <ostream>
#include <utility>
#include <vector>

#include <boost/shared_ptr.hpp>

//#include <boost/thread/xtime.hpp>

#define __CL_ENABLE_EXCEPTIONS
#include <CL/cl.hpp>
#include <CL/cl_wwu_dcl.h>

#include <gcf/CallbackPolicy.h>
#include <gcf/CommunicationManager.h>
#include <gcf/Process.h>
#include <gcf/Serializable.h>

#include <dcl/CLCommandListener.h>
#include <dcl/CommandListener.h>
#include <dcl/DCLTypes.h>
#include <dcl/Event.h>

#include <dcl/comm/BuildProgram.h>
#include <dcl/comm/CreateBuffer.h>
#include <dcl/comm/CreateCommandQueue.h>
#include <dcl/comm/CreateContext.h>
#include <dcl/comm/CreateKernel.h>
#include <dcl/comm/CreateKernelsInProgram.h>
#include <dcl/comm/CreateProgramWithSource.h>
#include <dcl/comm/DeleteMemory.h>
#include <dcl/comm/DeleteCommandQueue.h>
#include <dcl/comm/DeleteContext.h>
#include <dcl/comm/DeleteKernel.h>
#include <dcl/comm/DeleteProgram.h>
#include <dcl/comm/EnqueueCollective.h>
#include <dcl/comm/EnqueueCopy.h>
#include <dcl/comm/EnqueueNDRangeKernel.h>
#include <dcl/comm/EnqueueReadBuffer.h>
#if defined(CL_USE_DEPRECATED_OPENCL_1_1_APIS) || (defined(CL_VERSION_1_1) && !defined(CL_VERSION_1_2))
#include <dcl/comm/EnqueueWaitForEvents.h>
#endif // #if defined(CL_USE_DEPRECATED_OPENCL_1_1_APIS)
#include <dcl/comm/EnqueueWriteBuffer.h>
#include <dcl/comm/ErrorResponse.h>
#include <dcl/comm/EventCommand.h>
#include <dcl/comm/FlushFinishRequest.h>
#include <dcl/comm/GetKernelInfo.h>
#include <dcl/comm/InfoResponse.h>
#include <dcl/comm/Request.h>
#include <dcl/comm/Response.h>
#include <dcl/comm/SetKernelArg.h>

#include <dcl/util/Logger.h>

#include "../ComputeNodeConnectionManagerImpl.h"
#include "../ComputeNodeImpl.h"
#include "../HostImpl.h"

#include "CLRequestDispatcher.h"
#include "DeviceIDsResponse.h"
#include "DeviceInfosResponse.h"
#include "GetDeviceIDs.h"
#include "GetDeviceInfos.h"

/* In future versions all request classes (currently in namespace dcl::comm)
 * should go to dclgcf::comm. To avoid tedious renaming at that time, 'using
 * namespace' is an acceptable solution for now. */
using namespace dcl::comm;

namespace dclgcf {

namespace comm {

CLRequestDispatcher::MessageTypes::MessageTypes() {
    /* add command request message types */
    push_back(BuildProgram::TYPE);
    push_back(CreateBuffer::TYPE);
    push_back(CreateCommandQueue::TYPE);
    push_back(CreateContext::TYPE);
    push_back(CreateEvent::TYPE);
    push_back(CreateKernel::TYPE);
    push_back(CreateKernelsInProgram::TYPE);
    push_back(CreateProgramWithSource::TYPE);
    push_back(CreateUserEvent::TYPE);
    push_back(GetDeviceIDs::TYPE);
    push_back(GetDeviceInfos::TYPE);
    push_back(GetEventProfilingInfo::TYPE);
    push_back(GetKernelInfo::TYPE);
    push_back(GetKernelWorkGroupInfo::TYPE);
    push_back(DeleteMemory::TYPE);
    push_back(DeleteCommandQueue::TYPE);
    push_back(DeleteContext::TYPE);
    push_back(DeleteEvent::TYPE);
    push_back(DeleteKernel::TYPE);
    push_back(DeleteProgram::TYPE);
    push_back(EnqueueBarrier::TYPE);
    push_back(EnqueueBroadcastBuffer::TYPE);
    push_back(EnqueueCopyBuffer::TYPE);
    push_back(EnqueueMarker::TYPE);
    push_back(EnqueueNDRangeKernel::TYPE);
    push_back(EnqueueReadBuffer::TYPE);
    push_back(EnqueueReduceBuffer::TYPE);
#if defined(CL_USE_DEPRECATED_OPENCL_1_1_APIS) || (defined(CL_VERSION_1_1) && !defined(CL_VERSION_1_2))
    push_back(EnqueueWaitForEvents::TYPE);
#endif // #if defined(CL_USE_DEPRECATED_OPENCL_1_1_APIS)
    push_back(EnqueueWriteBuffer::TYPE);
    push_back(FinishRequest::TYPE);
    push_back(FlushRequest::TYPE);
    push_back(SetKernelArg::TYPE);
    push_back(SetKernelArgBinary::TYPE);
    push_back(SetKernelArgMemObject::TYPE);
    push_back(SetUserEventStatusRequest::TYPE);
}

/******************************************************************************/

const gcf::CallbackPolicy& CLRequestDispatcher::callbackPolicy() {
    static gcf::CallbackPolicy policy(
            MessageTypes(),
            false /* do not exclude but include specified types */);
    return policy;
}

CLRequestDispatcher::CLRequestDispatcher(
        ComputeNodeConnectionManagerImpl& connectionManager) :
        _connectionManager(connectionManager) {
}

CLRequestDispatcher::~CLRequestDispatcher() {
}

void CLRequestDispatcher::addCLCommandListener(const gcf::Process& process,
        dcl::CLCommandListener& listener) {
    std::pair<std::map<const gcf::Process *, dcl::CLCommandListener *>::iterator, bool> res;
    res = _clCommandListeners.insert(std::make_pair(&process, &listener));
    assert(res.second);
}

void CLRequestDispatcher::removeCLCommandListener(const gcf::Process& process) {
    std::map<const gcf::Process *, dcl::CLCommandListener *>::size_type numErased;
    numErased = _clCommandListeners.erase(&process);
    assert(numErased == 1);
}

dcl::CLCommandListener& CLRequestDispatcher::lookupCommandListener(
        const gcf::Process *process) {
    std::map<const gcf::Process *, dcl::CLCommandListener *>::const_iterator i;
    dcl::CLCommandListener *clCommandListener;

    i = _clCommandListeners.find(process);
    if (i != _clCommandListeners.end()) {
        clCommandListener = i->second;
    }
    assert(clCommandListener != 0);

    return *clCommandListener;
}

/******************************************************************************/

/* TODO Replace command listeners, i.e. sessions, by object registry(s)
 * A registry is provided by each host as each host assigns its own (and thus
 * possibly identical) IDs to objects. Therefore, lookup host in compute node's
 * connection manager, obtain registry, and lookup object in registry, rather
 * than in the command listener or session, respectively.
 * This also requires to register newly created objects in the registry within
 * this class.
 *
 * Resolve an object ID:
         HostImpl *host = _connectionManager.getHost();
         assert(host != 0);
         host->objectRegistry().lookup...(...Id);
 * Register a new object:
         object = ...(...Id);
         HostImpl *host = _connectionManager.getHost();
         assert(host != 0);
         host->objectRegistry().register...(...Id, object);
 *
 * Object lookup is hard to implement, if the request is not sent by a host but
 * rather another compute node or a device manager (formerly resource manager).
 * This is, e.g., the case if events status changes are sent between compute
 * nodes, directly.
 * In these cases, it is not clear at which registry the objects are registered
 * whose IDs are used by the compute node or device manager.
 *
 * Device IDs should not be resolved by a host's registry, as devices are not
 * directly associated which a host but rather a platform.
 * Solution idea: use connection manager to assign and resolve device IDs
 * 1) add platform (cf. DeviceManager) to ComputeNodeConnectionManager
 * 2) Connection manager obtains list of all devices from platform and assigns
 *    IDs to all devices
 */

#if 0
/**
 * @brief Obtain all platform information and return as compute node information.
 */
template<>
Response * CLRequestDispatcher::execute(const GetPlatformInfos& request) {
    Response *response;

    try {
        dcl::Platform platform = _connectionManager.platform();

        response = new PlatformInfosResponse(request,
                platform.getInfo<CL_PLATFORM_EXTENSIONS>(),
                platform.getInfo<CL_PLATFORM_NAME>(),
                platform.getInfo<CL_PLATFORM_PROFILE>(),
                platform.getInfo<CL_PLATFORM_VENDOR>(),
                platform.getInfo<CL_PLATFORM_VERSION>());
    } catch (const cl::Error& err) {
        response = new ErrorResponse(request, err.err());
    }

    return response;
}
#endif

template<>
Response * CLRequestDispatcher::execute(const GetDeviceIDs& request,
        const gcf::Process& process) {
    Response *response;

    try {
        std::vector<dcl::dcl_object_id> deviceIDs;

        /* FIXME Do not query devices from command listener, or whatever process-dependent object */
#if 0
        std::vector<dcl::Device> devices;

        _connectionManager.platform().getDevices(request.deviceType(), devices);
        dcl::Device::getDeviceIDs(devices, deviceIDs);
#endif
        lookupCommandListener(&process).getDeviceIDs(
                request.deviceType(), deviceIDs);

        response = new DeviceIDsResponse(request, deviceIDs);
    } catch (const cl::Error& err) {
        response = new ErrorResponse(request, err.err());
    }

    return response;
}

template<>
Response * CLRequestDispatcher::execute(const GetDeviceInfos& request,
        const gcf::Process& process) {
    Response *response;

    try {
        void *paramValue;
        size_t paramSize;

        /* FIXME Do not query device info from command listener, or whatever process-dependent object */
#if 0
        cl::Device device;
        cl_int err;

        device = _connectionManager.objectRegistry().lookupDevice(request.device());
        /* Obtain device info using OpenCL C API to avoid unnecessary type
         * conversions. */
        err = ::clGetDeviceInfo(device(), request.paramName(), 0, 0, &paramSize);
        if (err == CL_SUCCESS) {
            paramValue = (void *) malloc(paramSize);
            if (paramValue) {
                err = ::clGetDeviceInfo(device(), request.paramName(), paramSize,
                        paramValue, 0);
                if (err == CL_SUCCESS) {
                    response = new DeviceInfosResponse(request, paramSize, paramValue);
                }

                /* clean up */
                free(paramValue);
            } else {
                err = CL_OUT_OF_HOST_MEMORY;
            }
        }

        if (err != CL_SUCCESS) {
            response = new ErrorResponse(request, err);
        }
#endif
        lookupCommandListener(&process).getDeviceInfo(
                request.device(), request.paramName(), &paramValue, &paramSize);

        response = new DeviceInfosResponse(request, paramSize, paramValue);

        free(paramValue);
    } catch (const cl::Error& err) {
        response = new ErrorResponse(request, err.err());
    }

    return response;
}

template<>
Response * CLRequestDispatcher::execute(const CreateContext& request,
        const gcf::Process& process) {
    std::vector<ComputeNodeImpl *> computeNodes;
    Response *response;

    /* TODO Lookup (or create) compute nodes
     * This operation can introduce new compute nodes to this compute node */

    try {
        /* TODO Replace command listeners, i.e. sessions, by object registry(s) */
#if 0
        std::vector<dcl::Device *> devices;
        HostImpl *host = _connectionManager.getHost(process);
        assert(host != 0);

        _connectionManager.objectRegistry().lookupDevices(request.deviceIds(), devices);
        dcl::Context *context = _connectionManager.platform().createContext(
                request.contextId(),
                host,
                std::vector<dcl::ComputeNode *>(computeNodes.begin(), computeNodes.end()),
                devices);
        // TODO Use dcl::ContextListener to abstract from context ID and host
//        dcl::Context *context = _connectionManager.platform().createContext(
//                ContextListenerImpl(request.contextId(), host),
//                std::vector<dcl::ComputeNode *>(computeNodes.begin(), computeNodes.end()),
//                devices);
        host->objectRegistry().bindContext(request.contextId(), context);
#endif
        lookupCommandListener(&process).createContext(
                request.contextId(),
                std::vector<dcl::ComputeNode *>(computeNodes.begin(), computeNodes.end()),
                request.deviceIds());

        /* TODO Asynchronously connect to created compute nodes
         * Return response when compute nodes have been connected */
//        boost::thread thread(boost::bind(
//                &ComputeNodeConnectionManagerImpl::connectComputeNodes,
//                &_connectionManager, computeNodes));
        response = new Response(request);
    } catch (const cl::Error& err) {
        response = new ErrorResponse(request, err.err());
    }

    return response;
}

template<>
Response * CLRequestDispatcher::execute(const DeleteContext& request,
        const gcf::Process& process) {
    Response *response;

    try {
        lookupCommandListener(&process).releaseContext(request.contextId());
        response = new Response(request);
    } catch (const cl::Error& err) {
        response = new ErrorResponse(request, err.err());
    }

    return response;
}

template<>
Response * CLRequestDispatcher::execute(const CreateBuffer& request,
        const gcf::Process& process) {
    Response *response;

    try {
        lookupCommandListener(&process).createBuffer(
                request.contextId(), request.bufferId(), request.flags(), request.size());
        response = new Response(request);
    } catch (const cl::Error& err) {
        response = new ErrorResponse(request, err.err());
    }

    return response;
}

template<>
Response * CLRequestDispatcher::execute(const DeleteMemory& request,
        const gcf::Process& process) {
    Response *response;

    try {
        lookupCommandListener(&process).releaseMemObject(request.memObjectId());
        response = new Response(request);
    } catch (const cl::Error& err) {
        response = new ErrorResponse(request, err.err());
    }

    return response;
}

template<>
Response * CLRequestDispatcher::execute(const CreateCommandQueue& request,
        const gcf::Process& process) {
    Response *response;

    try {
        lookupCommandListener(&process).createCommandQueue(
                request.contextId(), request.deviceId(),
                request.commandQueueId(), request.properties());
        response = new Response(request);
    } catch (const cl::Error& err) {
        response = new ErrorResponse(request, err.err());
    }

    return response;
}

template<>
Response * CLRequestDispatcher::execute(const DeleteCommandQueue& request,
        const gcf::Process& process) {
    Response *response;

    try {
        lookupCommandListener(&process).releaseCommandQueue(request.commandQueueId());
        response = new Response(request);
    } catch (const cl::Error& err) {
        response = new ErrorResponse(request, err.err());
    }

    return response;
}

template<>
Response * CLRequestDispatcher::execute(const CreateProgramWithSource& request,
        const gcf::Process& process) {
    Response *response;

    try {
        lookupCommandListener(&process).createProgram(
                request.contextId(), request.programId(),
                request.length());

        response = new Response(request);
    } catch (const cl::Error& err) {
        response = new ErrorResponse(request, err.err());
    }

    return response;
}

template<>
Response * CLRequestDispatcher::execute(const DeleteProgram& request,
        const gcf::Process& process) {
    Response *response;

    try {
        lookupCommandListener(&process).releaseProgram(request.programId());
        response = new Response(request);
    } catch (const cl::Error& err) {
        response = new ErrorResponse(request, err.err());
    }

    return response;
}

template<>
Response * CLRequestDispatcher::execute(const BuildProgram& request,
        const gcf::Process& process) {
    Response *response;

    try {
        lookupCommandListener(&process).buildProgram(
                request.programId(),
                (request.deviceIds().empty() ? 0 : &request.deviceIds()),
                request.options().c_str(), request.programBuildId());
        response = new Response(request);
    } catch (const cl::Error& err){
        response = new ErrorResponse(request, err.err());
    }

    return response;
}

template<>
Response * CLRequestDispatcher::execute(const CreateKernel& request,
        const gcf::Process& process) {
    Response *response;

    try{
        lookupCommandListener(&process).createKernel(
                request.programId(), request.kernelId(),
                request.kernelName().c_str());
        response = new Response(request);
    } catch (const cl::Error& err){
        response = new ErrorResponse(request, err.err());
    }

    return response;
}

template<>
Response * CLRequestDispatcher::execute(const CreateKernelsInProgram& request,
        const gcf::Process& process) {
    Response *response;

    try {
        lookupCommandListener(&process).createKernelsInProgram(
                request.programId(),
                request.kernelIds());
        response = new Response(request);
    } catch (const cl::Error& err) {
        response = new ErrorResponse(request, err.err());
    }

    return response;
}

template<>
Response * CLRequestDispatcher::execute(const DeleteKernel& request,
        const gcf::Process& process) {
    Response *response;

    try {
        lookupCommandListener(&process).releaseKernel(request.kernelId());
        response = new Response(request);
    } catch (const cl::Error& err) {
        response = new ErrorResponse(request, err.err());
    }

    return response;
}

template<>
Response * CLRequestDispatcher::execute(const CreateEvent& request,
        const gcf::Process& process) {
    Response *response;

    try {
        HostImpl *host = _connectionManager.getHost(process);
        dcl::Event *event;

        assert(host != 0);

        event = lookupCommandListener(&process).createEvent(
                request.contextId(), request.eventId(),
                request.memObjectIds());
        /* register command listener */
        host->objectRegistry().bind<dcl::CommandListener>(request.eventId(), *event);

        response = new Response(request);
    } catch (const cl::Error& err) {
        response = new ErrorResponse(request, err.err());
    }

    return response;
}

template<>
Response * CLRequestDispatcher::execute(const CreateUserEvent& request,
        const gcf::Process& process) {
    Response *response;

    try {
        lookupCommandListener(&process).createUserEvent(
                request.contextId(), request.eventId());
        response = new Response(request);
    } catch (const cl::Error& err) {
        response = new ErrorResponse(request, err.err());
    }

    return response;
}

template<>
Response * CLRequestDispatcher::execute(const DeleteEvent& request,
        const gcf::Process& process) {
    Response *response;

    try {
        HostImpl *host = _connectionManager.getHost(process);

        assert(host != 0);

        lookupCommandListener(&process).releaseEvent(request.eventId());
        /* deregister command listener */
        host->objectRegistry().unbind<dcl::CommandListener>(request.eventId());

        response = new Response(request);
    } catch (const cl::Error& err) {
        response = new ErrorResponse(request, err.err());
    }

    return response;
}

template<>
Response * CLRequestDispatcher::execute(const GetEventProfilingInfo& request,
        const gcf::Process& process) {
    Response *response;

    try {
        dcl::dcl_object_id event_id = request.eventId();
        cl_ulong received, queued, submit, start, end;

        lookupCommandListener(&process).getEventProfilingInfo(
                event_id, CL_PROFILING_COMMAND_RECEIVED_WWU, received);
        lookupCommandListener(&process).getEventProfilingInfo(
                event_id, CL_PROFILING_COMMAND_QUEUED, queued);
        lookupCommandListener(&process).getEventProfilingInfo(
                event_id, CL_PROFILING_COMMAND_SUBMIT, submit);
        lookupCommandListener(&process).getEventProfilingInfo(
                event_id, CL_PROFILING_COMMAND_START, start);
        lookupCommandListener(&process).getEventProfilingInfo(
                event_id, CL_PROFILING_COMMAND_END, end);

        response = new EventProfilingInfo(
                request, received, queued, submit, start, end);
    } catch (const cl::Error& err) {
        response = new ErrorResponse(request, err.err());
    }

    return response;
}

template<>
Response * CLRequestDispatcher::execute(const GetKernelInfo& request,
        const gcf::Process& process) {
    Response *response;

    try {
        void *paramValue;
        size_t paramSize;

        lookupCommandListener(&process).getKernelInfo(
                request.kernelId(), request.paramName(),
                &paramValue, &paramSize);

        response = new InfoResponse(request, paramSize, paramValue);

        free(paramValue);
    } catch (const cl::Error& err) {
        response = new ErrorResponse(request, err.err());
    }

    return response;
}

template<>
Response * CLRequestDispatcher::execute(const GetKernelWorkGroupInfo& request,
        const gcf::Process& process) {
    Response *response;

    try {
        void *paramValue;
        size_t paramSize;

        lookupCommandListener(&process).getKernelWorkGroupInfo(
                request.kernelId(), request.deviceId(), request.paramName(),
                &paramValue, &paramSize);

        response = new InfoResponse(request, paramSize, paramValue);

        free(paramValue);
    } catch (const cl::Error& err) {
        response = new ErrorResponse(request, err.err());
    }

    return response;
}

template<>
Response * CLRequestDispatcher::execute(const EnqueueWriteBuffer& request,
        const gcf::Process& process) {
    Response *response;

    try {
        lookupCommandListener(&process).enqueueWriteBuffer(
                request.commandQueueId(), request.commandId(),
                request.bufferId(),	request.blocking(), request.offset(),
                request.cb(),
                (request.eventIdWaitList().empty() ? 0 : &request.eventIdWaitList()),
                request.event()
        );

        response = new Response(request);
    } catch (const cl::Error& err) {
        response = new ErrorResponse(request, err.err());
    }

    return response;
}

template<>
Response * CLRequestDispatcher::execute(const EnqueueCopyBuffer& request,
        const gcf::Process& process) {
    Response *response;

    try {
        lookupCommandListener(&process).enqueueCopyBuffer(
                request.commandQueueId(), request.commandId(),
                request.srcBufferId(), request.dstBufferId(),
                request.srcOffset(), request.dstOffset(), request.cb(),
                (request.eventIdWaitList().empty() ? 0 : &request.eventIdWaitList()),
                request.event()
        );

        response = new Response(request);
    } catch (const cl::Error& err) {
        response = new ErrorResponse(request, err.err());
    }

    return response;
}

template<>
Response * CLRequestDispatcher::execute(const EnqueueReadBuffer& request,
        const gcf::Process& process) {
    Response *response;

    try {
        lookupCommandListener(&process).enqueueReadBuffer(
                request.commandQueueId(), request.commandId(),
                request.bufferId(),	request.blocking(), request.offset(),
                request.cb(),
                (request.eventIdWaitList().empty() ? 0 : &request.eventIdWaitList()),
                request.event()
        );

        response = new Response(request);
    } catch (const cl::Error& err) {
        response = new ErrorResponse(request, err.err());
    }

    return response;
}

template<>
Response * CLRequestDispatcher::execute(const EnqueueBroadcastBuffer& request,
        const gcf::Process& process) {
    Response *response;

    try {
        std::vector<size_t> dstOffsets;

        request.dstOffsets(dstOffsets);

        lookupCommandListener(&process).enqueueBroadcastBuffer(
                request.commandQueueIds(), request.commandId(),
                request.srcBufferId(), request.dstBufferIds(),
                request.srcOffset(), dstOffsets,
                request.cb(),
                (request.eventIdWaitList().empty() ? 0 : &request.eventIdWaitList()),
                request.event()
        );

        response = new Response(request);
    } catch (const cl::Error& err) {
        response = new ErrorResponse(request, err.err());
    }

    return response;
}

template<>
Response * CLRequestDispatcher::execute(const EnqueueReduceBuffer& request,
        const gcf::Process& process) {
    Response *response;

    try {
        cl::NDRange offset;
        cl::NDRange global;
        cl::NDRange local;
        std::vector<size_t> global_work_offset;
        std::vector<size_t> global_work_size;
        std::vector<size_t> local_work_size;

        request.offset(global_work_offset);
        switch (global_work_offset.size()) {
        case 0:
            offset = cl::NullRange;
            break;
        case 1:
            offset = cl::NDRange(global_work_offset[0]);
            break;
        case 2:
            offset = cl::NDRange(global_work_offset[0], global_work_offset[1]);
            break;
        case 3:
            offset = cl::NDRange(global_work_offset[0], global_work_offset[1],
                    global_work_offset[2]);
            break;
        }

        request.global(global_work_size);
        switch (global_work_size.size()) {
        case 0:
            global = cl::NullRange;
            break;
        case 1:
            global = cl::NDRange(global_work_size[0]);
            break;
        case 2:
            global = cl::NDRange(global_work_size[0], global_work_size[1]);
            break;
        case 3:
            global = cl::NDRange(global_work_size[0], global_work_size[1],
                    global_work_size[2]);
            break;
        }

        request.local(local_work_size);
        switch (local_work_size.size()) {
        case 0:
            local = cl::NullRange;
            break;
        case 1:
            local = cl::NDRange(local_work_size[0]);
            break;
        case 2:
            local = cl::NDRange(local_work_size[0],local_work_size[1]);
            break;
        case 3:
            local = cl::NDRange(local_work_size[0], local_work_size[1],
                    local_work_size[2]);
            break;
        }

        lookupCommandListener(&process).enqueueReduceBuffer(
                request.commandQueueId(), request.commandId(),
                request.srcIds(), request.dstId(),
                request.kernelId(), offset, global, local,
                (request.eventIdWaitList().empty() ? 0 : &request.eventIdWaitList()),
                request.event()
        );

        response = new Response(request);
    } catch (const cl::Error& err) {
        response = new ErrorResponse(request, err.err());
    }

    return response;
}

template<>
Response * CLRequestDispatcher::execute(const EnqueueNDRangeKernel& request,
        const gcf::Process& process) {
    Response *response;

    try {
        cl::NDRange offset;
        cl::NDRange global;
        cl::NDRange local;
        std::vector<size_t> global_work_offset;
        std::vector<size_t> global_work_size;
        std::vector<size_t> local_work_size;

        request.offset(global_work_offset);
        switch (global_work_offset.size()) {
        case 0:
            offset = cl::NullRange;
            break;
        case 1:
            offset = cl::NDRange(global_work_offset[0]);
            break;
        case 2:
            offset = cl::NDRange(global_work_offset[0], global_work_offset[1]);
            break;
        case 3:
            offset = cl::NDRange(global_work_offset[0], global_work_offset[1],
                    global_work_offset[2]);
            break;
        }

        request.global(global_work_size);
        switch (global_work_size.size()) {
        case 0:
            global = cl::NullRange;
            break;
        case 1:
            global = cl::NDRange(global_work_size[0]);
            break;
        case 2:
            global = cl::NDRange(global_work_size[0], global_work_size[1]);
            break;
        case 3:
            global = cl::NDRange(global_work_size[0], global_work_size[1],
                    global_work_size[2]);
            break;
        }

        request.local(local_work_size);
        switch (local_work_size.size()) {
        case 0:
            local = cl::NullRange;
            break;
        case 1:
            local = cl::NDRange(local_work_size[0]);
            break;
        case 2:
            local = cl::NDRange(local_work_size[0],local_work_size[1]);
            break;
        case 3:
            local = cl::NDRange(local_work_size[0], local_work_size[1],
                    local_work_size[2]);
            break;
        }

        lookupCommandListener(&process).enqueueNDRangeKernel(
                request.commandQueueId(), request.commandId(),
                request.kernelId(), offset, global, local,
                (request.eventIdWaitList().empty() ? 0 : &request.eventIdWaitList()),
                request.event()
        );

        response = new Response(request);
    } catch (const cl::Error& err) {
        response = new ErrorResponse(request, err.err());
    }

    return response;
}

template<>
Response * CLRequestDispatcher::execute(const EnqueueBarrier& request,
        const gcf::Process& process) {
    Response *response;

    try {
        lookupCommandListener(&process).enqueueBarrier(
                request.commandQueueId(),
                request.commandId(),
                (request.eventIdWaitList().empty() ? 0 : &request.eventIdWaitList()),
                request.event()
        );

        response = new Response(request);
    } catch (const cl::Error& err) {
        response = new ErrorResponse(request, err.err());
    }

    return response;
}

#if defined(CL_USE_DEPRECATED_OPENCL_1_1_APIS) || (defined(CL_VERSION_1_1) && !defined(CL_VERSION_1_2))
template<>
Response * CLRequestDispatcher::execute(const EnqueueWaitForEvents& request,
        const gcf::Process& process) {
    Response *response;

    try {
        lookupCommandListener(&process).enqueueWaitForEvents(
                request.commandQueueId(),
                request.eventIdList());

        response = new Response(request);
    } catch (const cl::Error& err) {
        response = new ErrorResponse(request, err.err());
    }

    return response;
}
#endif // #if defined(CL_USE_DEPRECATED_OPENCL_1_1_APIS)

template<>
Response * CLRequestDispatcher::execute(const EnqueueMarker& request,
        const gcf::Process& process) {
    Response *response;

    try {
        lookupCommandListener(&process).enqueueMarker(
                request.commandQueueId(),
                request.commandId(),
                (request.eventIdWaitList().empty() ? 0 : &request.eventIdWaitList()),
                request.event()
        );

        response = new Response(request);
    } catch (const cl::Error& err) {
        response = new ErrorResponse(request, err.err());
    }

    return response;
}

template<>
Response * CLRequestDispatcher::execute(const FinishRequest& request,
        const gcf::Process& process) {
    Response *response;

    try {
        lookupCommandListener(&process).finish(request.commandQueueId());
        response = new Response(request);
    } catch (const cl::Error& err) {
        response = new ErrorResponse(request, err.err());
    }

    return response;
}

template<>
Response * CLRequestDispatcher::execute(const FlushRequest& request,
        const gcf::Process& process) {
    Response *response;

    try {
        lookupCommandListener(&process).flush(request.commandQueueId());
        response = new Response(request);
    } catch (const cl::Error& err) {
        response = new ErrorResponse(request, err.err());
    }

    return response;
}

template<>
Response * CLRequestDispatcher::execute(const SetKernelArgMemObject& request,
        const gcf::Process& process) {
    Response *response;

    try {
        lookupCommandListener(&process).setKernelArg(
                request.kernelId(),
                request.argIndex(),
                dcl::MEMORY,
                request.argSize(),
                request.argValue());
        response = new Response(request);
    } catch (const cl::Error& err) {
        response = new ErrorResponse(request, err.err());
    }

    return response;
}

template<>
Response * CLRequestDispatcher::execute(const SetKernelArgBinary& request,
        const gcf::Process& process) {
    Response *response;

    try {
        lookupCommandListener(&process).setKernelArg(
                request.kernelId(),
                request.argIndex(),
                dcl::BINARY,
                request.argSize(),
                request.argValue());
        response = new Response(request);
    } catch (const cl::Error& err) {
        response = new ErrorResponse(request, err.err());
    }

    return response;
}

template<>
Response * CLRequestDispatcher::execute(const SetUserEventStatusRequest& request,
        const gcf::Process& process) {
    Response *response;

    try {
        lookupCommandListener(&process).setUserEventStatus(
                request.eventId(), request.status());
        response = new Response(request);
    } catch (const cl::Error& err) {
        response = new ErrorResponse(request, err.err());
    }

    return response;
}

/******************************************************************************
 * GCF message listener API
 ******************************************************************************/

void CLRequestDispatcher::dispatch(gcf::Serializable& message,
        const gcf::Process& process) {
    boost::shared_ptr<Response> response;
//    boost::xtime start, end;

//    boost::xtime_get(&start, boost::TIME_UTC);

    /* TODO Discard request from unknown processes
     * Only processes that have identified ('connected') themselves as host
     * or possibly also compute node are allowed to issue requests. */

//    dcl::util::Logger << dcl::util::Debug
//            << "\treceived message (type=" << message.getType() << ')'
//            << std::endl;

    /*
     * Dispatch message
     */
    switch (message.getType()) {

    /* Requests sent by any dOpenCL node */
    case GetDeviceIDs::TYPE:
        response = boost::shared_ptr<Response>(execute<GetDeviceIDs>(
                static_cast<GetDeviceIDs&>(message), process));
        break;
    case GetDeviceInfos::TYPE:
        response = boost::shared_ptr<Response>(execute<GetDeviceInfos>(
                static_cast<GetDeviceInfos&>(message), process));
        break;

    /* Request sent by hosts */
    case CreateContext::TYPE:
        response = boost::shared_ptr<Response>(execute<CreateContext>(
                static_cast<CreateContext&>(message), process));
        break;
    case DeleteContext::TYPE:
        response = boost::shared_ptr<Response>(execute<DeleteContext>(
                static_cast<DeleteContext&>(message), process));
        break;
    case DeleteMemory::TYPE:
        response = boost::shared_ptr<Response>(execute<DeleteMemory>(
                static_cast<DeleteMemory&>(message), process));
        break;
    case CreateBuffer::TYPE:
        response = boost::shared_ptr<Response>(execute<CreateBuffer>(
                static_cast<CreateBuffer&>(message), process));
        break;
    case CreateCommandQueue::TYPE:
        response = boost::shared_ptr<Response>(execute<CreateCommandQueue>(
                static_cast<CreateCommandQueue&>(message), process));
        break;
    case DeleteCommandQueue::TYPE:
        response = boost::shared_ptr<Response>(execute<DeleteCommandQueue>(
                static_cast<DeleteCommandQueue&>(message), process));
        break;
    case EnqueueBarrier::TYPE:
        response = boost::shared_ptr<Response>(execute<EnqueueBarrier>(
                static_cast<EnqueueBarrier&>(message), process));
        break;
    case EnqueueBroadcastBuffer::TYPE:
        response = boost::shared_ptr<Response>(execute<EnqueueBroadcastBuffer>(
                static_cast<EnqueueBroadcastBuffer&>(message), process));
        break;
    case EnqueueCopyBuffer::TYPE:
        response = boost::shared_ptr<Response>(execute<EnqueueCopyBuffer>(
                static_cast<EnqueueCopyBuffer&>(message), process));
        break;
    case EnqueueMarker::TYPE:
        response = boost::shared_ptr<Response>(execute<EnqueueMarker>(
                static_cast<EnqueueMarker&>(message), process));
        break;
    case EnqueueNDRangeKernel::TYPE:
        response = boost::shared_ptr<Response>(execute<EnqueueNDRangeKernel>(
                static_cast<EnqueueNDRangeKernel&>(message), process));
        break;
    case EnqueueWriteBuffer::TYPE:
        response = boost::shared_ptr<Response>(execute<EnqueueWriteBuffer>(
                static_cast<EnqueueWriteBuffer&>(message), process));
        break;
    case EnqueueReadBuffer::TYPE:
        response = boost::shared_ptr<Response>(execute<EnqueueReadBuffer>(
                static_cast<EnqueueReadBuffer&>(message), process));
        break;
    case EnqueueReduceBuffer::TYPE:
        response = boost::shared_ptr<Response>(execute<EnqueueReduceBuffer>(
                static_cast<EnqueueReduceBuffer&>(message), process));
        break;
#if defined(CL_USE_DEPRECATED_OPENCL_1_1_APIS) || (defined(CL_VERSION_1_1) && !defined(CL_VERSION_1_2))
    case EnqueueWaitForEvents::TYPE:
        response = boost::shared_ptr<Response>(execute<EnqueueWaitForEvents>(
                static_cast<EnqueueWaitForEvents&>(message), process));
        break;
#endif // #if defined(CL_USE_DEPRECATED_OPENCL_1_1_APIS)
    case FinishRequest::TYPE:
        response = boost::shared_ptr<Response>(execute<FinishRequest>(
                static_cast<FinishRequest&>(message), process));
        break;
    case FlushRequest::TYPE:
        response = boost::shared_ptr<Response>(execute<FlushRequest>(
                static_cast<FlushRequest&>(message), process));
        break;
    case CreateProgramWithSource::TYPE:
        response = boost::shared_ptr<Response>(execute<CreateProgramWithSource>(
                static_cast<CreateProgramWithSource&>(message), process));
        break;
    case DeleteProgram::TYPE:
        response = boost::shared_ptr<Response>(execute<DeleteProgram>(
                static_cast<DeleteProgram&>(message), process));
        break;
    case BuildProgram::TYPE:
        response = boost::shared_ptr<Response>(execute<BuildProgram>(
                static_cast<BuildProgram&>(message), process));
        break;
    case CreateKernel::TYPE:
        response = boost::shared_ptr<Response>(execute<CreateKernel>(
                static_cast<CreateKernel&>(message), process));
        break;
    case CreateKernelsInProgram::TYPE:
        response = boost::shared_ptr<Response>(execute<CreateKernelsInProgram>(
                static_cast<CreateKernelsInProgram&>(message), process));
        break;
    case DeleteKernel::TYPE:
        response = boost::shared_ptr<Response>(execute<DeleteKernel>(
                static_cast<DeleteKernel&>(message), process));
        break;
    case CreateEvent::TYPE:
        response = boost::shared_ptr<Response>(execute<CreateEvent>(
                static_cast<CreateEvent&>(message), process));
        break;
    case CreateUserEvent::TYPE:
        response = boost::shared_ptr<Response>(execute<CreateUserEvent>(
                static_cast<CreateUserEvent&>(message), process));
        break;
    case SetUserEventStatusRequest::TYPE:
        response = boost::shared_ptr<Response>(execute<SetUserEventStatusRequest>(
                static_cast<SetUserEventStatusRequest&>(message), process));
        break;
    case DeleteEvent::TYPE:
        response = boost::shared_ptr<Response>(execute<DeleteEvent>(
                static_cast<DeleteEvent&>(message), process));
        break;
    case GetKernelInfo::TYPE:
        response = boost::shared_ptr<Response>(execute<GetKernelInfo>(
                static_cast<GetKernelInfo&>(message), process));
        break;
    case GetKernelWorkGroupInfo::TYPE:
        response = boost::shared_ptr<Response>(execute<GetKernelWorkGroupInfo>(
                static_cast<GetKernelWorkGroupInfo&>(message), process));
        break;
    case GetEventProfilingInfo::TYPE:
        response = boost::shared_ptr<Response>(execute<GetEventProfilingInfo>(
                static_cast<GetEventProfilingInfo&>(message), process));
        break;
    case SetKernelArgMemObject::TYPE:
        response = boost::shared_ptr<Response>(execute<SetKernelArgMemObject>(
                static_cast<SetKernelArgMemObject&>(message), process));
        break;
    case SetKernelArgBinary::TYPE:
        response = boost::shared_ptr<Response>(execute<SetKernelArgBinary>(
                static_cast<SetKernelArgBinary&>(message), process));
        break;

    default: /* invalid message */
        dcl::util::Logger << dcl::util::Error
                << "Received message of unknown type" << std::endl;
        break;
    }

//    boost::xtime_get(&end, boost::TIME_UTC);
//    end.sec  -= start.sec;
//    end.nsec -= start.nsec;
//    if (end.nsec < 0) {
//        --end.sec;
//        end.nsec += 1000000000;
//    }
//    dcl::util::Logger << dcl::util::Debug
//            << "Processed request of type " << message.getType()
//            << " in " << end.sec << ':' << std::setw(9) << std::setfill('0') << end.nsec
//            << std::endl;

    /*
     * A response may have been created if a request has been received.
     */
    if (response) {
        _connectionManager.communicationManager().send(process, *response);
//        dcl::util::Logger << dcl::util::Debug
//                << "\tsent response (request ID=" << response->requestId
//                << ", type=" << response->getType() << ')'
//                << std::endl;
    }
}

} /* namespace comm */

} /* namespace dclgcf */
