/******************************************************************************
 * This file is part of dOpenCL.
 * 
 * dOpenCL is an implementation of the OpenCL application programming
 * interface for distributed systems. See <http://dopencl.uni-muenster.de/>
 * for more information.
 * 
 * Developed by: Research Group Parallel and Distributed Systems
 *               Department of Mathematics and Computer Science
 *               University of Muenster, Germany
 *               <http://pvs.uni-muenster.de/>
 * 
 * Copyright (C) 2013  Philipp Kegel <philipp.kegel@uni-muenster.de>
 *
 * dOpenCL is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * dOpenCL is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with dOpenCL. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Permission to use dOpenCL for scientific, non-commercial work is
 * granted under the terms of the dOpenCL Academic License provided
 * appropriate credit is given. See the dOpenCL Academic License for
 * more details.
 * 
 * You should have received a copy of the dOpenCL Academic License
 * along with dOpenCL. If not, see <http://dopencl.uni-muenster.de/>.
 ******************************************************************************/

/**
 * @file    CLRequestDispatcher.h
 *
 * @date    31.10.2011
 * @author  Philipp Kegel
 */

#ifndef CLREQUESTDISPATCHER_H_
#define CLREQUESTDISPATCHER_H_

#include <map>
#include <vector>

#include <gcf/CallbackPolicy.h>
#include <gcf/GCFTypedefs.h>
#include <gcf/MessageListener.h>
#include <gcf/Process.h>
#include <gcf/Serializable.h>

#include <dcl/CLCommandListener.h>

#include <dcl/comm/Response.h>

namespace dclgcf {

/* forward declaration */
class ComputeNodeConnectionManagerImpl;

/******************************************************************************/

namespace comm {

/**
 * @brief A central dispatch for incoming command request messages from GCF.
 *
 * Requests are unmarshalled and forwarded to the application.
 */
class CLRequestDispatcher: public gcf::MessageListener {
public:
    /**
     * @brief Returns a GCF callback policy comprising the command request types processed by a request dispatcher.
     *
     * @return a callback policy
     */
    static const gcf::CallbackPolicy& callbackPolicy();


    CLRequestDispatcher(
            ComputeNodeConnectionManagerImpl&  /* connectionManager */);
    virtual ~CLRequestDispatcher();

    /**
     * @brief Registers an OpenCL command listener for a given process.
     *
     * @param[in]  process
     * @param[in]  listener
     */
    void addCLCommandListener(
            const gcf::Process&     /* process */,
            dcl::CLCommandListener& /* listener */);

    /**
     * @brief Removed an OpenCL command listener for a given process.
     *
     * @param[in]  process
     */
    void removeCLCommandListener(
            const gcf::Process& /* process */);

    /*
     * GCF message listener API
     */
    void dispatch(
            gcf::Serializable&  /* message */,
            const gcf::Process& /* process */);

private:
    /**
     * @brief A list of dOpenCL command request message types.
     *
     * This class is used internally for static initialization of a GCF callback
     * policy.
     */
    class MessageTypes : public std::vector<GCFTypeID_t> {
    public:
        MessageTypes();
    };


    /* TODO Replace CLRequestDispatcher::lookupCommandListener by lookup methods of object registry */
    dcl::CLCommandListener& lookupCommandListener(
            const gcf::Process *    /* process */);

    /**
     * @brief Execute a given request
     *
     * A request is unmarshalled and an appropriate method is called.
     * The method's output is marshalled into a response.
     *
     * @param[in]  request  the request
     * @param[in]  process  the requesting process
     * @return the response
     */
    template<class T>
    dcl::comm::Response * execute(
            const T&            /* request */,
            const gcf::Process& /* process */);

    ComputeNodeConnectionManagerImpl& _connectionManager;

    std::map<const gcf::Process *, dcl::CLCommandListener *> _clCommandListeners;
};

} /* namespace comm */

} /* namespace dclgcf */

#endif /* CLREQUESTDISPATCHER_H_ */
