/******************************************************************************
 * This file is part of dOpenCL.
 * 
 * dOpenCL is an implementation of the OpenCL application programming
 * interface for distributed systems. See <http://dopencl.uni-muenster.de/>
 * for more information.
 * 
 * Developed by: Research Group Parallel and Distributed Systems
 *               Department of Mathematics and Computer Science
 *               University of Muenster, Germany
 *               <http://pvs.uni-muenster.de/>
 * 
 * Copyright (C) 2013  Philipp Kegel <philipp.kegel@uni-muenster.de>
 *
 * dOpenCL is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * dOpenCL is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with dOpenCL. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Permission to use dOpenCL for scientific, non-commercial work is
 * granted under the terms of the dOpenCL Academic License provided
 * appropriate credit is given. See the dOpenCL Academic License for
 * more details.
 * 
 * You should have received a copy of the dOpenCL Academic License
 * along with dOpenCL. If not, see <http://dopencl.uni-muenster.de/>.
 ******************************************************************************/

/**
 * @file    CLResponseDispatcher.cpp
 *
 * @date    14.07.2012
 * @author  Philipp Kegel
 */

#include <cassert>
#include <ostream>

#include <gcf/CallbackPolicy.h>
#include <gcf/Process.h>
#include <gcf/Serializable.h>

#include <dcl/comm/ErrorResponse.h>
#include <dcl/comm/EventCommand.h>
#include <dcl/comm/InfoResponse.h>
#include <dcl/comm/Response.h>

#include <dcl/util/Logger.h>

#include "../ComputeNodeImpl.h"
#include "../ConnectionManagerImpl.h"

#include "CLResponseDispatcher.h"
#include "DeviceIDsResponse.h"
#include "DeviceInfosResponse.h"
#include "ResponseBuffer.h"

/* In future versions all response classes (currently in namespace dcl::comm)
 * should go to dclgcf::comm. To avoid tedious renaming at that time, 'using
 * namespace' is an acceptable solution for now. */
using namespace dcl::comm;

namespace dclgcf {

namespace comm {

CLResponseDispatcher::MessageTypes::MessageTypes() {
    /* add command response message types */
    push_back(Response::TYPE);
    push_back(DeviceIDsResponse::TYPE);
    push_back(DeviceInfosResponse::TYPE);
    push_back(ErrorResponse::TYPE);
    push_back(EventProfilingInfo::TYPE);
    push_back(InfoResponse::TYPE);
}

/******************************************************************************/

const gcf::CallbackPolicy& CLResponseDispatcher::callbackPolicy() {
    static gcf::CallbackPolicy policy(
            MessageTypes(),
            false /* do not exclude but include specified types */);
    return policy;
}

CLResponseDispatcher::CLResponseDispatcher(
        const ConnectionManagerImpl& connectionManager) :
        _connectionManager(connectionManager) {
}

CLResponseDispatcher::~CLResponseDispatcher() {
}

/******************************************************************************
 * GCF message listener API
 ******************************************************************************/

void CLResponseDispatcher::dispatch(gcf::Serializable& message,
		const gcf::Process& process) {
	ComputeNodeImpl *computeNode;
	Response *response;

	/*
	 * Find response buffer
	 */
	computeNode = _connectionManager.getComputeNode(process);
	assert(computeNode != 0);

	switch (message.getType()) {

	/*
	 * Command response messages
	 * Messages must be copied as they will be deleted by the GCF when dispatch
	 * returns.
	 */
	/* TODO Implement move semantics for messages, such that responses will not
	 *      be deeply copied for buffering */
	case DeviceIDsResponse::TYPE:
		response = new DeviceIDsResponse(
				static_cast<DeviceIDsResponse&>(message));
		break;
	case DeviceInfosResponse::TYPE:
		response = new DeviceInfosResponse(
				static_cast<DeviceInfosResponse&>(message));
		break;
	case ErrorResponse::TYPE:
		response = new ErrorResponse(
				static_cast<ErrorResponse&>(message));
		break;
	case EventProfilingInfo::TYPE:
		response = new EventProfilingInfo(
				static_cast<EventProfilingInfo&>(message));
		break;
	case InfoResponse::TYPE:
		response = new InfoResponse(
				static_cast<InfoResponse&>(message));
		break;
	case Response::TYPE:
		response = new Response(
				static_cast<Response&>(message));
		break;

	default: /* invalid message */
		dcl::util::Logger << dcl::util::Error
				<< "Received message of unknown type" << std::endl;
		break;
	}

	if (response) {
	    computeNode->responseBuffer().add(response);
	}
}

} /* namespace comm */

} /* namespace dclgcf */
