/******************************************************************************
 * This file is part of dOpenCL.
 * 
 * dOpenCL is an implementation of the OpenCL application programming
 * interface for distributed systems. See <http://dopencl.uni-muenster.de/>
 * for more information.
 * 
 * Developed by: Research Group Parallel and Distributed Systems
 *               Department of Mathematics and Computer Science
 *               University of Muenster, Germany
 *               <http://pvs.uni-muenster.de/>
 * 
 * Copyright (C) 2013  Philipp Kegel <philipp.kegel@uni-muenster.de>
 *
 * dOpenCL is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * dOpenCL is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with dOpenCL. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Permission to use dOpenCL for scientific, non-commercial work is
 * granted under the terms of the dOpenCL Academic License provided
 * appropriate credit is given. See the dOpenCL Academic License for
 * more details.
 * 
 * You should have received a copy of the dOpenCL Academic License
 * along with dOpenCL. If not, see <http://dopencl.uni-muenster.de/>.
 ******************************************************************************/

/*
 * ContextMessage.cpp
 *
 *  Created on: 27.03.2012
 *      Author: Sebastian Pribnow
 */

#include <string>

#ifdef __APPLE__
#include <OpenCL/cl.h>
#else
#include <CL/cl.h>
#endif

#include <gcf/GCFTypedefs.h>
#include <gcf/GenericSerializerImplNew.cpp>
#include <gcf/Serializable.h>

#include <dcl/Binary.h>
#include <dcl/DCLTypes.h>

#include <dcl/comm/ContextMessage.h>

namespace gcf {

template<>
size_t writeCustomSTR<dcl::comm::ContextErrorMessage> (
		dcl::comm::ContextErrorMessage *object,
		const gcf::GenericSerializeTemplateRecord& gstr,
		char *buffer) {
	assert(gstr.getType() == gcf::ST_Custom1);

	gcf::ByteBuffer buf(buffer, gstr.getLength());
	dcl::Binary& argValue =
			*gcf::GenericSerializeTemplateRecord::getAttribute<
					dcl::comm::ContextErrorMessage, dcl::Binary,
					gcf::ST_Custom1>(object, gstr);

	buf << argValue;

	return buf.getLength();
}

template<>
size_t readCustomSTR<dcl::comm::ContextErrorMessage> (
		dcl::comm::ContextErrorMessage *object,
		const gcf::GenericSerializeTemplateRecord& gstr,
		const char *buffer) {
	assert(gstr.getType() == gcf::ST_Custom1);

	gcf::ByteBuffer buf(buffer, gstr.getLength());
	dcl::Binary& argValue =
			*gcf::GenericSerializeTemplateRecord::getAttribute<
					dcl::comm::ContextErrorMessage, dcl::Binary,
					gcf::ST_Custom1>(object, gstr);

	buf >> argValue;

	return buf.getLength();
}

} /* namespace gcf */

/******************************************************************************/

namespace dcl {

namespace comm {

ContextErrorMessage::ContextErrorMessage() {
	initSerializable();
}

ContextErrorMessage::ContextErrorMessage(dcl::dcl_object_id contextId,
		const std::string& errInfo, const dcl::Binary& privateInfo) :
	_contextId(contextId), _errorInfo(errInfo), _privateInfo(privateInfo)
{
	initSerializable();
}

ContextErrorMessage::ContextErrorMessage(const ContextErrorMessage& rhs) :
	_contextId(rhs._contextId), _errorInfo(rhs._errorInfo),
	_privateInfo(rhs._privateInfo)
{
	initSerializable();
}

ContextErrorMessage::~ContextErrorMessage() { }

dcl_object_id ContextErrorMessage::contextId() const {
	return _contextId;
}

const std::string& ContextErrorMessage::errorInfo() const {
	return _errorInfo;
}

const dcl::Binary& ContextErrorMessage::privateInfo() const {
	return _privateInfo;
}

IMPLEMENT_SERIALIZABLE_EVENT(
		dcl::comm::ContextErrorMessage,
		ADD_ATTRIBUTE_DEFAULT(dcl::comm::ContextErrorMessage, _contextId);
		ADD_ATTRIBUTE_DEFAULT(dcl::comm::ContextErrorMessage, _errorInfo);
		ADD_ATTRIBUTE_CUSTOM(dcl::comm::ContextErrorMessage, _privateInfo,
						gcf::ST_Custom1, dcl::Binary, 256, gcf::Reliable,
						gcf::Public);
		)

/******************************************************************************/

ContextPrintfMessage::ContextPrintfMessage() {
    initSerializable();
}

ContextPrintfMessage::ContextPrintfMessage(dcl::dcl_object_id contextId,
        const std::string& printfData) :
    _contextId(contextId), _printfData(printfData)
{
    initSerializable();
}

ContextPrintfMessage::ContextPrintfMessage(const ContextPrintfMessage& rhs) :
    _contextId(rhs._contextId), _printfData(rhs._printfData)
{
    initSerializable();
}

ContextPrintfMessage::~ContextPrintfMessage() { }

dcl_object_id ContextPrintfMessage::contextId() const {
    return _contextId;
}

const std::string& ContextPrintfMessage::printfData() const {
    return _printfData;
}

IMPLEMENT_SERIALIZABLE_EVENT(
        dcl::comm::ContextPrintfMessage,
        ADD_ATTRIBUTE_DEFAULT(dcl::comm::ContextPrintfMessage, _contextId);
        ADD_ATTRIBUTE_DEFAULT(dcl::comm::ContextPrintfMessage, _printfData);
        )

} /* namespace comm */

} /* namespace dcl */




