/******************************************************************************
 * This file is part of dOpenCL.
 * 
 * dOpenCL is an implementation of the OpenCL application programming
 * interface for distributed systems. See <http://dopencl.uni-muenster.de/>
 * for more information.
 * 
 * Developed by: Research Group Parallel and Distributed Systems
 *               Department of Mathematics and Computer Science
 *               University of Muenster, Germany
 *               <http://pvs.uni-muenster.de/>
 * 
 * Copyright (C) 2013  Philipp Kegel <philipp.kegel@uni-muenster.de>
 *
 * dOpenCL is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * dOpenCL is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with dOpenCL. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Permission to use dOpenCL for scientific, non-commercial work is
 * granted under the terms of the dOpenCL Academic License provided
 * appropriate credit is given. See the dOpenCL Academic License for
 * more details.
 * 
 * You should have received a copy of the dOpenCL Academic License
 * along with dOpenCL. If not, see <http://dopencl.uni-muenster.de/>.
 ******************************************************************************/

/**
 * @file    CreateContext.cpp
 *
 * @date    10.05.2011
 * @author  Karsten Jeschkies
 */

#include <cassert>
#include <cstddef>
#include <vector>

#ifdef __APPLE__
#include <OpenCL/cl.h>
#else
#include <CL/cl.h>
#endif

#include <gcf/GCFTypedefs.h>
#include <gcf/GenericSerializerImplNew.cpp>
#include <gcf/Process.h>
#include <gcf/ProcessID.h>
#include <gcf/Serializable.h>

#include <gcf/serialization/ByteBuffer.h>

#include <dcl/DCLTypes.h>

#include <dcl/comm/CreateContext.h>
#include <dcl/comm/Request.h>

namespace gcf {

template<>
size_t writeCustomSTR<dcl::comm::CreateContext> (
        dcl::comm::CreateContext *object,
        const gcf::GenericSerializeTemplateRecord& gstr,
        char *buffer) {
    assert((gstr.getType() == gcf::ST_Custom1)
            || (gstr.getType() == gcf::ST_Custom2)
            || (gstr.getType() == gcf::ST_Custom3));

    gcf::ByteBuffer buf(buffer, gstr.getLength());

    switch (gstr.getType()) {
    case gcf::ST_Custom1:
    {
        std::vector<gcf::ProcessID>& computeNodeIds =
                *gcf::GenericSerializeTemplateRecord::getAttribute<
                        dcl::comm::CreateContext, std::vector<gcf::ProcessID>,
                        gcf::ST_Custom1>(object, gstr);

        buf << static_cast<gcf_uint32>(computeNodeIds.size());
        for (std::vector<gcf::ProcessID>::const_iterator i = computeNodeIds.begin();
                i != computeNodeIds.end(); ++i) {
            buf << *i;
        }
        break;
    }
    case gcf::ST_Custom2:
    {
        cl_device_type& deviceType =
                *gcf::GenericSerializeTemplateRecord::getAttribute<
                        dcl::comm::CreateContext, cl_device_type,
                        gcf::ST_Custom2>(object, gstr);

        buf << deviceType;
        break;
    }
    case gcf::ST_Custom3:
    {
        std::vector<dcl::dcl_object_id>& deviceIds =
                *gcf::GenericSerializeTemplateRecord::getAttribute<
                        dcl::comm::CreateContext, std::vector<dcl::dcl_object_id>,
                        gcf::ST_Custom3>(object, gstr);

        buf << deviceIds;
        break;
    }
    default:
        /* invalid value */
        ;
    }

    return buf.getLength();
}

template<>
size_t readCustomSTR<dcl::comm::CreateContext> (
        dcl::comm::CreateContext *object,
        const gcf::GenericSerializeTemplateRecord& gstr,
        const char *buffer) {
    assert((gstr.getType() == gcf::ST_Custom1)
            || (gstr.getType() == gcf::ST_Custom2)
            || (gstr.getType() == gcf::ST_Custom3));

    gcf::ByteBuffer buf(buffer, gstr.getLength());

    switch (gstr.getType()) {
    case gcf::ST_Custom1:
    {
        std::vector<gcf::ProcessID>& computeNodeIds =
                *gcf::GenericSerializeTemplateRecord::getAttribute<
                        dcl::comm::CreateContext, std::vector<gcf::ProcessID>,
                        gcf::ST_Custom1>(object, gstr);
        gcf_uint32 size;

        buf >> size;

        computeNodeIds.resize(size);
        for (gcf_uint32 i = 0; i < size; ++i) {
            buf >> computeNodeIds[i];
        }
        break;
    }
    case gcf::ST_Custom2:
    {
        cl_device_type& deviceType =
                *gcf::GenericSerializeTemplateRecord::getAttribute<
                        dcl::comm::CreateContext, cl_device_type,
                        gcf::ST_Custom2>(object, gstr);

        buf >> deviceType;
        break;
    }
    case gcf::ST_Custom3:
    {
        std::vector<dcl::dcl_object_id>& deviceIds =
                *gcf::GenericSerializeTemplateRecord::getAttribute<
                        dcl::comm::CreateContext, std::vector<dcl::dcl_object_id>,
                        gcf::ST_Custom3>(object, gstr);

        deviceIds.clear();
        buf >> deviceIds;
        break;
    }
    default:
        /* invalid value */
        ;
    }

    return buf.getLength();
}

} /* namespace gcf */

/******************************************************************************/

namespace dcl {

namespace comm {

CreateContext::CreateContext() :
		 _contextId(0) {
	initSerializable();
}

CreateContext::CreateContext(
        dcl_object_id contextId,
      const std::vector<gcf::ProcessID>& computeNodeIds) :
    _contextId(contextId), _computeNodeIds(computeNodeIds),
    _deviceType(CL_DEVICE_TYPE_ALL)
{
    initSerializable();
}

CreateContext::CreateContext(
        dcl_object_id contextId,
      const std::vector<gcf::ProcessID>& computeNodeIds,
        cl_device_type deviceType) :
    _contextId(contextId), _computeNodeIds(computeNodeIds),
    _deviceType(deviceType)
{
    initSerializable();
}

CreateContext::CreateContext(
		dcl_object_id contextId,
		const std::vector<gcf::ProcessID>& computeNodeIds,
		const std::vector<dcl_object_id>& deviceIds) :
	_contextId(contextId), _computeNodeIds(computeNodeIds),
    _deviceType(CL_DEVICE_TYPE_ALL), _deviceIds(deviceIds)
{
    initSerializable();
}

CreateContext::CreateContext(const CreateContext& rhs) :
    Request(rhs), _contextId(rhs._contextId),
    _computeNodeIds(rhs._computeNodeIds),
    _deviceType(rhs._deviceType), _deviceIds(rhs._deviceIds)
{
    initSerializable();
}

dcl_object_id CreateContext::contextId() const {
	return _contextId;
}

const std::vector<gcf::ProcessID>& CreateContext::computeNodeIds() const {
    return _computeNodeIds;
}

cl_device_type CreateContext::deviceType() const {
    return _deviceType;
}

const std::vector<dcl_object_id>& CreateContext::deviceIds() const {
	return _deviceIds;
}

IMPLEMENT_SERIALIZABLE_DERIVED_EVENT(
		dcl::comm::CreateContext,
		dcl::comm::Request,
		ADD_ATTRIBUTE_DEFAULT(dcl::comm::CreateContext, _contextId);
        ADD_ATTRIBUTE_CUSTOM(dcl::comm::CreateContext, _computeNodeIds,
                gcf::ST_Custom1, std::vector<gcf::ProcessID>, 256,
                gcf::Reliable, gcf::Public);
        ADD_ATTRIBUTE_CUSTOM(dcl::comm::CreateContext, _deviceType,
                gcf::ST_Custom2, cl_device_type, 8, gcf::Reliable, gcf::Public);
        ADD_ATTRIBUTE_CUSTOM(dcl::comm::CreateContext, _deviceIds,
                gcf::ST_Custom3, std::vector<dcl::dcl_object_id>, 256,
                gcf::Reliable, gcf::Public);
        )

} /* namespace comm */

} /* namespace dcl */
