/******************************************************************************
 * This file is part of dOpenCL.
 * 
 * dOpenCL is an implementation of the OpenCL application programming
 * interface for distributed systems. See <http://dopencl.uni-muenster.de/>
 * for more information.
 * 
 * Developed by: Research Group Parallel and Distributed Systems
 *               Department of Mathematics and Computer Science
 *               University of Muenster, Germany
 *               <http://pvs.uni-muenster.de/>
 * 
 * Copyright (C) 2013  Philipp Kegel <philipp.kegel@uni-muenster.de>
 *
 * dOpenCL is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * dOpenCL is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with dOpenCL. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Permission to use dOpenCL for scientific, non-commercial work is
 * granted under the terms of the dOpenCL Academic License provided
 * appropriate credit is given. See the dOpenCL Academic License for
 * more details.
 * 
 * You should have received a copy of the dOpenCL Academic License
 * along with dOpenCL. If not, see <http://dopencl.uni-muenster.de/>.
 ******************************************************************************/

/**
 * @file    DataDispatcher.cpp
 *
 * @author  Philipp Kegel
 * @date    16.02.2011
 */

#include <cstddef>

#include <boost/make_shared.hpp>
#include <boost/shared_ptr.hpp>

#include <boost/thread/thread.hpp>

#include <gcf/transport/IODataStream.h>

#include <dcl/DCLTypes.h>

#include "../DataTransferImpl.h"

#include "DataDispatcher.h"

namespace dclgcf {

namespace comm {

DataDispatcher::DataDispatcher() {
    /* Data dispatcher must be initialized before a thread is created */
    _thread = boost::thread(&DataDispatcher::run, this);
}

DataDispatcher::~DataDispatcher() {
    /* interrupt blocking queue to stop thread */
    _dataTransfers.interrupt();
    if (_thread.joinable()) _thread.join();
}

boost::shared_ptr<DataTransferImpl> DataDispatcher::receiveData(
		gcf::IODataStream& dataStream, size_t size, void *ptr) {
	boost::shared_ptr<DataReceipt> receipt(
	        boost::make_shared<DataReceipt>(dataStream.recv(ptr, size)));
	_dataTransfers.push(receipt);
	return receipt;
}

boost::shared_ptr<DataTransferImpl> DataDispatcher::sendData(
		gcf::IODataStream& dataStream, size_t size, const void *ptr) {
	boost::shared_ptr<DataSending> sending(
	        boost::make_shared<DataSending>(dataStream.send(ptr, size)));
	_dataTransfers.push(sending);
	return sending;
}

void DataDispatcher::run() {
	try {
		/* Wait for, fetch, and execute tasks from task queue */
		while (true) {
			boost::shared_ptr<DataTransferImpl> dataTransfer = _dataTransfers.front();
			_dataTransfers.pop();
			dataTransfer->execute();
		}
	} catch (const boost::thread_interrupted&) {
		/* task queue has been interrupted */
	}
}

} /* namespace comm */

} /* namespace dclgcf */
