/******************************************************************************
 * This file is part of dOpenCL.
 * 
 * dOpenCL is an implementation of the OpenCL application programming
 * interface for distributed systems. See <http://dopencl.uni-muenster.de/>
 * for more information.
 * 
 * Developed by: Research Group Parallel and Distributed Systems
 *               Department of Mathematics and Computer Science
 *               University of Muenster, Germany
 *               <http://pvs.uni-muenster.de/>
 * 
 * Copyright (C) 2013  Philipp Kegel <philipp.kegel@uni-muenster.de>
 *
 * dOpenCL is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * dOpenCL is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with dOpenCL. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Permission to use dOpenCL for scientific, non-commercial work is
 * granted under the terms of the dOpenCL Academic License provided
 * appropriate credit is given. See the dOpenCL Academic License for
 * more details.
 * 
 * You should have received a copy of the dOpenCL Academic License
 * along with dOpenCL. If not, see <http://dopencl.uni-muenster.de/>.
 ******************************************************************************/

/*
 * EnqueueCollective.cpp
 *
 *  Created on: 02.07.2012
 *      Author: Philipp Kegel
 */

#include <cstddef>
#include <vector>

#include <gcf/GCFTypedefs.h>
#include <gcf/GenericSerializerImplNew.cpp>
#include <gcf/Serializable.h>
#include <gcf/serialization/ByteBuffer.h>

#include <dcl/DCLTypes.h>
#include <dcl/comm/EnqueueCollective.h>
#include <dcl/comm/Request.h>

namespace gcf {

template<>
size_t readCustomSTR<dcl::comm::EnqueueReduceBuffer> (
		dcl::comm::EnqueueReduceBuffer *object,
		const gcf::GenericSerializeTemplateRecord& gstr,
		const char *buffer) {
	assert(gstr.getType() == gcf::ST_Custom1);

	gcf::ByteBuffer buf(buffer, gstr.getLength());
	std::vector<gcf_uint32>& ndRange =
			*gcf::GenericSerializeTemplateRecord::getAttribute<
					dcl::comm::EnqueueReduceBuffer, std::vector<gcf_uint32>,
					gcf::ST_Custom1>(object, gstr);

	ndRange.clear();
	buf >> ndRange;

	return buf.getLength();
}

template<>
size_t writeCustomSTR<dcl::comm::EnqueueReduceBuffer> (
		dcl::comm::EnqueueReduceBuffer *object,
		const gcf::GenericSerializeTemplateRecord& gstr,
		char *buffer) {
	assert(gstr.getType() == gcf::ST_Custom1);

	gcf::ByteBuffer buf(buffer, gstr.getLength());
	std::vector<gcf_uint32>& ndRange =
			*gcf::GenericSerializeTemplateRecord::getAttribute<
					dcl::comm::EnqueueReduceBuffer, std::vector<gcf_uint32>,
					gcf::ST_Custom1>(object, gstr);

	buf << ndRange;

	return buf.getLength();
}

} /* namespace gcf */

/******************************************************************************/

namespace dcl {

namespace comm {

static void workDimensionHtoN(
		const std::vector<size_t>& in,
		std::vector<gcf_uint32>& out) {
	std::vector<size_t>::const_iterator src;
	std::vector<gcf_uint32>::iterator dst;

	out.resize(in.size());
	src = in.begin();
	dst = out.begin();

	while (src != in.end()) {
		*dst++ = static_cast<gcf_uint32>(*src++);
	}
}

static void workDimensionNtoH(
		const std::vector<gcf_uint32>& in,
		std::vector<size_t>& out) {
	std::vector<gcf_uint32>::const_iterator src;
	std::vector<size_t>::iterator dst;

	out.resize(in.size());
	src = in.begin();
	dst = out.begin();

	while (src != in.end()) {
		*dst++ = static_cast<size_t>(*src++);
	}
}

/******************************************************************************/

EnqueueBroadcastBuffer::EnqueueBroadcastBuffer() :
	_commandId(0), _srcBufferId(0),	_event(false) {
	initSerializable();
}

EnqueueBroadcastBuffer::EnqueueBroadcastBuffer(
		const std::vector<dcl_object_id>& commandQueueIds,
		dcl_object_id commandId,
		dcl_object_id srcBufferId,
		const std::vector<dcl_object_id>& dstBufferIds,
		size_t srcOffset,
		const std::vector<size_t>& dstOffsets,
		size_t cb,
		const std::vector<dcl_object_id> *eventIdWaitList,
		bool event) :
	_commandQueueIds(commandQueueIds), _commandId(commandId),
			_srcBufferId(srcBufferId), _dstBufferIds(dstBufferIds),
			_srcOffset(static_cast<gcf_uint32> (srcOffset)),
			_cb(static_cast<gcf_uint32> (cb)),
			_event(static_cast<gcf_bool>(event))
{
	initSerializable();

	/* convert destination buffer offsets */
	workDimensionHtoN(dstOffsets, _dstOffsets);

	if (eventIdWaitList) {
		_eventIdWaitList = *eventIdWaitList;
	}
}

EnqueueBroadcastBuffer::EnqueueBroadcastBuffer(const EnqueueBroadcastBuffer& rhs) :
	Request(rhs), _commandQueueIds(rhs._commandQueueIds), _commandId(
			rhs._commandId), _srcBufferId(rhs._srcBufferId), _dstBufferIds(
			rhs._dstBufferIds), _srcOffset(rhs._srcOffset), _dstOffsets(
			rhs._dstOffsets), _cb(rhs._cb), _eventIdWaitList(
			rhs._eventIdWaitList), _event(rhs._event)
{
	initSerializable();
}

const std::vector<dcl_object_id>& EnqueueBroadcastBuffer::commandQueueIds() const {
	return _commandQueueIds;
}

dcl_object_id EnqueueBroadcastBuffer::commandId() const {
	return _commandId;
}

dcl_object_id EnqueueBroadcastBuffer::srcBufferId() const {
	return _srcBufferId;
}

const std::vector<dcl_object_id>& EnqueueBroadcastBuffer::dstBufferIds() const {
	return _dstBufferIds;
}

size_t EnqueueBroadcastBuffer::srcOffset() const {
	return static_cast<size_t> (_srcOffset);
}

void EnqueueBroadcastBuffer::dstOffsets(std::vector<size_t>& dstOffsets) const {
	workDimensionNtoH(_dstOffsets, dstOffsets);
}

size_t EnqueueBroadcastBuffer::cb() const {
	return static_cast<size_t> (_cb);
}

const std::vector<dcl_object_id>& EnqueueBroadcastBuffer::eventIdWaitList() const {
	return _eventIdWaitList;
}

bool EnqueueBroadcastBuffer::event() const {
	return static_cast<bool>(_event);
}

IMPLEMENT_SERIALIZABLE_DERIVED_EVENT(
		dcl::comm::EnqueueBroadcastBuffer,
		dcl::comm::Request,
		ADD_ATTRIBUTE_DEFAULT(dcl::comm::EnqueueBroadcastBuffer, _commandQueueIds);
		ADD_ATTRIBUTE_DEFAULT(dcl::comm::EnqueueBroadcastBuffer, _commandId);
		ADD_ATTRIBUTE_DEFAULT(dcl::comm::EnqueueBroadcastBuffer, _srcBufferId);
		ADD_ATTRIBUTE_DEFAULT(dcl::comm::EnqueueBroadcastBuffer, _dstBufferIds);
		ADD_ATTRIBUTE_DEFAULT(dcl::comm::EnqueueBroadcastBuffer, _srcOffset);
		ADD_ATTRIBUTE_DEFAULT(dcl::comm::EnqueueBroadcastBuffer, _dstOffsets);
		ADD_ATTRIBUTE_DEFAULT(dcl::comm::EnqueueBroadcastBuffer, _cb);
		ADD_ATTRIBUTE_DEFAULT(dcl::comm::EnqueueBroadcastBuffer, _eventIdWaitList);
		ADD_ATTRIBUTE_DEFAULT(dcl::comm::EnqueueBroadcastBuffer, _event);
	)

/******************************************************************************/

EnqueueReduceBuffer::EnqueueReduceBuffer() :
	_commandId(0), _dstId(0), _kernelId(0), _event(false) {
	initSerializable();
}

EnqueueReduceBuffer::EnqueueReduceBuffer(
		dcl_object_id commandQueueId,
		dcl_object_id commandId,
		const std::vector<dcl_object_id>& srcIds,
		dcl_object_id dstId,
		dcl_object_id kernelId,
		const std::vector<size_t>& offset,
		const std::vector<size_t>& global,
		const std::vector<size_t>& local,
		const std::vector<dcl_object_id> *eventIdWaitList,
		bool event) :
	_commandQueueId(commandQueueId), _commandId(commandId),
			_srcIds(srcIds), _dstId(dstId), _kernelId(kernelId),
			_offset(offset.size()), _global(global.size()),
			_local(local.size()), _event(event) {
	initSerializable();

	/* convert work dimensions from host to network representation */
	workDimensionHtoN(offset, _offset);
	workDimensionHtoN(global, _global);
	workDimensionHtoN(local, _local);

	if (eventIdWaitList) {
		_eventIdWaitList = *eventIdWaitList;
	}
}

EnqueueReduceBuffer::EnqueueReduceBuffer(
		const EnqueueReduceBuffer& rhs) :
	Request(rhs), _commandQueueId(rhs._commandQueueId), _commandId(rhs._commandId),
			_srcIds(rhs._srcIds), _dstId(rhs._dstId),
			_kernelId(rhs._kernelId), _global(rhs._global), _local(rhs._local),
			_eventIdWaitList(rhs._eventIdWaitList), _event(rhs._event) {
	initSerializable();
}

dcl_object_id EnqueueReduceBuffer::commandQueueId() const {
	return _commandQueueId;
}

dcl_object_id EnqueueReduceBuffer::commandId() const {
	return _commandId;
}

const std::vector<dcl_object_id>& EnqueueReduceBuffer::srcIds() const {
	return _srcIds;
}

dcl_object_id EnqueueReduceBuffer::dstId() const {
	return _dstId;
}

dcl_object_id EnqueueReduceBuffer::kernelId() const {
	return _kernelId;
}

void EnqueueReduceBuffer::offset(std::vector<size_t>& offset) const {
	workDimensionNtoH(_offset, offset);
}

void EnqueueReduceBuffer::global(std::vector<size_t>& global) const {
	workDimensionNtoH(_global, global);
}

void EnqueueReduceBuffer::local(std::vector<size_t>& local) const {
	workDimensionNtoH(_local, local);
}

const std::vector<dcl_object_id>& EnqueueReduceBuffer::eventIdWaitList() const {
	return _eventIdWaitList;
}

bool EnqueueReduceBuffer::event() const {
	return static_cast<bool>(_event);
}

IMPLEMENT_SERIALIZABLE_DERIVED_EVENT(
		dcl::comm::EnqueueReduceBuffer,
		dcl::comm::Request,
		ADD_ATTRIBUTE_DEFAULT(dcl::comm::EnqueueReduceBuffer, _commandQueueId);
		ADD_ATTRIBUTE_DEFAULT(dcl::comm::EnqueueReduceBuffer, _commandId);
		ADD_ATTRIBUTE_DEFAULT(dcl::comm::EnqueueReduceBuffer, _srcIds);
		ADD_ATTRIBUTE_DEFAULT(dcl::comm::EnqueueReduceBuffer, _dstId);
		ADD_ATTRIBUTE_DEFAULT(dcl::comm::EnqueueReduceBuffer, _kernelId);
		ADD_ATTRIBUTE_CUSTOM(
				dcl::comm::EnqueueReduceBuffer,
				_offset,
				gcf::ST_Custom1,
				std::vector<gcf_uint32>,
				256,
				gcf::Reliable,
				gcf::Public);
		ADD_ATTRIBUTE_CUSTOM(
				dcl::comm::EnqueueReduceBuffer,
				_global,
				gcf::ST_Custom1,
				std::vector<gcf_uint32>,
				256,
				gcf::Reliable,
				gcf::Public);
		ADD_ATTRIBUTE_CUSTOM(
				dcl::comm::EnqueueReduceBuffer,
				_local,
				gcf::ST_Custom1,
				std::vector<gcf_uint32>,
				256,
				gcf::Reliable,
				gcf::Public);
		ADD_ATTRIBUTE_DEFAULT(dcl::comm::EnqueueReduceBuffer, _eventIdWaitList);
		ADD_ATTRIBUTE_DEFAULT(dcl::comm::EnqueueReduceBuffer, _event);
)

} /* namespace comm */

} /* namespace dcl */
