/******************************************************************************
 * This file is part of dOpenCL.
 * 
 * dOpenCL is an implementation of the OpenCL application programming
 * interface for distributed systems. See <http://dopencl.uni-muenster.de/>
 * for more information.
 * 
 * Developed by: Research Group Parallel and Distributed Systems
 *               Department of Mathematics and Computer Science
 *               University of Muenster, Germany
 *               <http://pvs.uni-muenster.de/>
 * 
 * Copyright (C) 2013  Philipp Kegel <philipp.kegel@uni-muenster.de>
 *
 * dOpenCL is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * dOpenCL is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with dOpenCL. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Permission to use dOpenCL for scientific, non-commercial work is
 * granted under the terms of the dOpenCL Academic License provided
 * appropriate credit is given. See the dOpenCL Academic License for
 * more details.
 * 
 * You should have received a copy of the dOpenCL Academic License
 * along with dOpenCL. If not, see <http://dopencl.uni-muenster.de/>.
 ******************************************************************************/

/*
 * @file    EnqueueMapMemObject.cpp
 *
 * @date    2012-08-08
 * @author  Philipp Kegel
 */

#include <cstddef>
#include <vector>

#ifdef __APPLE__
#include <OpenCL/cl.h>
#else
#include <CL/cl.h>
#endif

#include <gcf/GCFTypedefs.h>
#include <gcf/GenericSerializerImplNew.cpp>
#include <gcf/Serializable.h>

#include <dcl/DCLTypes.h>
#include <dcl/comm/EnqueueMapMemObject.h>
#include <dcl/comm/Request.h>

namespace gcf {

template<>
size_t writeCustomSTR<dcl::comm::EnqueueMapBuffer> (
        dcl::comm::EnqueueMapBuffer *object,
        const gcf::GenericSerializeTemplateRecord& gstr,
        char *buffer) {
    assert(gstr.getType() == gcf::ST_Custom1);

    gcf::ByteBuffer buf(buffer, gstr.getLength());
    gcf_uint64& properties =
            *gcf::GenericSerializeTemplateRecord::getAttribute<
                    dcl::comm::EnqueueMapBuffer, gcf_uint64,
                    gcf::ST_Custom1>(object, gstr);

    buf << properties;

    return buf.getLength();
}

template<>
size_t readCustomSTR<dcl::comm::EnqueueMapBuffer> (
        dcl::comm::EnqueueMapBuffer *object,
        const gcf::GenericSerializeTemplateRecord& gstr,
        const char *buffer) {
    assert(gstr.getType() == gcf::ST_Custom1);

    gcf::ByteBuffer buf(buffer, gstr.getLength());
    gcf_uint64& properties =
            *gcf::GenericSerializeTemplateRecord::getAttribute<
                    dcl::comm::EnqueueMapBuffer, gcf_uint64,
                    gcf::ST_Custom1>(object, gstr);

    buf >> properties;

    return buf.getLength();
}

} /* namespace gcf */

/******************************************************************************/

namespace dcl {

namespace comm {

EnqueueMapBuffer::EnqueueMapBuffer() {
	initSerializable();
}

EnqueueMapBuffer::EnqueueMapBuffer(
		dcl_object_id commandQueueId,
		dcl_object_id commandId,
		dcl_object_id bufferId,
		bool blocking_read,
		cl_map_flags map_flags,
		size_t offset,
		size_t cb,
		const std::vector<dcl_object_id> *eventIdWaitList,
		bool event) :
	_commandQueueId(commandQueueId), _commandId(commandId), _bufferId(bufferId),
			_blocking(static_cast<gcf_bool>(blocking_read)),
            _mapFlags(static_cast<gcf_uint64>(map_flags)),
			_offset(static_cast<gcf_uint32>(offset)),
			_cb(static_cast<gcf_uint32>(cb)),
			_event(static_cast<gcf_bool>(event)) {
	initSerializable();

	if (eventIdWaitList) {
		_eventIdWaitList = *eventIdWaitList;
	}
}

EnqueueMapBuffer::EnqueueMapBuffer(
		const EnqueueMapBuffer& rhs) :
	Request(rhs), _commandQueueId(rhs._commandQueueId),
			_commandId(rhs._commandId),	_bufferId(rhs._bufferId),
			_blocking(rhs._blocking), _mapFlags(rhs._mapFlags),
			_offset(rhs._offset), _cb(rhs._cb),
			_eventIdWaitList(rhs._eventIdWaitList), _event(rhs._event) {
	initSerializable();
}

dcl_object_id EnqueueMapBuffer::commandQueueId() const {
	return _commandQueueId;
}

dcl_object_id EnqueueMapBuffer::commandId() const {
	return _commandId;
}

dcl_object_id EnqueueMapBuffer::bufferId() const {
	return _bufferId;
}

bool EnqueueMapBuffer::blocking() const {
	return static_cast<bool>(_blocking);
}

cl_mem_flags EnqueueMapBuffer::mapFlags() const {
    return static_cast<cl_mem_flags> (_mapFlags);
}

size_t EnqueueMapBuffer::offset() const {
	return static_cast<size_t> (_offset);
}

size_t EnqueueMapBuffer::cb() const {
	return static_cast<size_t> (_cb);
}

const std::vector<dcl_object_id>& EnqueueMapBuffer::eventIdWaitList(void) const {
	return _eventIdWaitList;
}

bool EnqueueMapBuffer::event() const {
	return static_cast<bool>(_event);
}

IMPLEMENT_SERIALIZABLE_DERIVED_EVENT(
		dcl::comm::EnqueueMapBuffer,
		dcl::comm::Request,
		ADD_ATTRIBUTE_DEFAULT(dcl::comm::EnqueueMapBuffer, _commandQueueId);
		ADD_ATTRIBUTE_DEFAULT(dcl::comm::EnqueueMapBuffer, _commandId);
		ADD_ATTRIBUTE_DEFAULT(dcl::comm::EnqueueMapBuffer, _bufferId);
		ADD_ATTRIBUTE_DEFAULT(dcl::comm::EnqueueMapBuffer, _blocking);
        ADD_ATTRIBUTE_CUSTOM(dcl::comm::EnqueueMapBuffer, _mapFlags,
                gcf::ST_Custom1, gcf_uint64, 8, gcf::Reliable, gcf::Public);
		ADD_ATTRIBUTE_DEFAULT(dcl::comm::EnqueueMapBuffer, _offset);
		ADD_ATTRIBUTE_DEFAULT(dcl::comm::EnqueueMapBuffer, _cb);
		ADD_ATTRIBUTE_DEFAULT(dcl::comm::EnqueueMapBuffer, _eventIdWaitList);
		ADD_ATTRIBUTE_DEFAULT(dcl::comm::EnqueueMapBuffer, _event);
	)

/******************************************************************************/

EnqueueUnmapMemObject::EnqueueUnmapMemObject() {
    initSerializable();
}

EnqueueUnmapMemObject::EnqueueUnmapMemObject(
        dcl_object_id commandQueueId,
        dcl_object_id commandId,
        dcl_object_id bufferId,
        const std::vector<dcl_object_id> *eventIdWaitList,
        bool event) :
    _commandQueueId(commandQueueId), _commandId(commandId), _bufferId(bufferId),
            _event(static_cast<gcf_bool>(event)) {
    initSerializable();

    if (eventIdWaitList) {
        _eventIdWaitList = *eventIdWaitList;
    }
}

EnqueueUnmapMemObject::EnqueueUnmapMemObject(
        const EnqueueUnmapMemObject& rhs) :
    Request(rhs), _commandQueueId(rhs._commandQueueId),
            _commandId(rhs._commandId), _bufferId(rhs._bufferId),
            _eventIdWaitList(rhs._eventIdWaitList), _event(rhs._event) {
    initSerializable();
}

dcl_object_id EnqueueUnmapMemObject::commandQueueId() const {
    return _commandQueueId;
}

dcl_object_id EnqueueUnmapMemObject::commandId() const {
    return _commandId;
}

dcl_object_id EnqueueUnmapMemObject::bufferId() const {
    return _bufferId;
}

const std::vector<dcl_object_id>& EnqueueUnmapMemObject::eventIdWaitList(void) const {
    return _eventIdWaitList;
}

bool EnqueueUnmapMemObject::event() const {
    return static_cast<bool>(_event);
}

IMPLEMENT_SERIALIZABLE_DERIVED_EVENT(
        dcl::comm::EnqueueUnmapMemObject,
        dcl::comm::Request,
        ADD_ATTRIBUTE_DEFAULT(dcl::comm::EnqueueUnmapMemObject, _commandQueueId);
        ADD_ATTRIBUTE_DEFAULT(dcl::comm::EnqueueUnmapMemObject, _commandId);
        ADD_ATTRIBUTE_DEFAULT(dcl::comm::EnqueueUnmapMemObject, _bufferId);
        ADD_ATTRIBUTE_DEFAULT(dcl::comm::EnqueueUnmapMemObject, _eventIdWaitList);
        ADD_ATTRIBUTE_DEFAULT(dcl::comm::EnqueueUnmapMemObject, _event);
    )

} /* namespace comm */

} /* namespace dcl */
