/******************************************************************************
 * This file is part of dOpenCL.
 * 
 * dOpenCL is an implementation of the OpenCL application programming
 * interface for distributed systems. See <http://dopencl.uni-muenster.de/>
 * for more information.
 * 
 * Developed by: Research Group Parallel and Distributed Systems
 *               Department of Mathematics and Computer Science
 *               University of Muenster, Germany
 *               <http://pvs.uni-muenster.de/>
 * 
 * Copyright (C) 2013  Philipp Kegel <philipp.kegel@uni-muenster.de>
 *
 * dOpenCL is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * dOpenCL is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with dOpenCL. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Permission to use dOpenCL for scientific, non-commercial work is
 * granted under the terms of the dOpenCL Academic License provided
 * appropriate credit is given. See the dOpenCL Academic License for
 * more details.
 * 
 * You should have received a copy of the dOpenCL Academic License
 * along with dOpenCL. If not, see <http://dopencl.uni-muenster.de/>.
 ******************************************************************************/

/*
 * EnqueueNDRangeKernel.cpp
 *
 *  Created on: Jul 13, 2011
 *      Author: taylan
 */

#include <cassert>
#include <cstddef>
#include <vector>

#include <gcf/GCFTypedefs.h>
#include <gcf/GenericSerializerImplNew.cpp>
#include <gcf/Serializable.h>
#include <gcf/serialization/ByteBuffer.h>

#include <dcl/DCLTypes.h>
#include <dcl/comm/EnqueueNDRangeKernel.h>
#include <dcl/comm/Request.h>

namespace gcf {

template<>
size_t readCustomSTR<dcl::comm::EnqueueNDRangeKernel> (
		dcl::comm::EnqueueNDRangeKernel *object,
		const gcf::GenericSerializeTemplateRecord& gstr,
		const char *buffer) {
	assert(gstr.getType() == gcf::ST_Custom1);

	gcf::ByteBuffer buf(buffer, gstr.getLength());
	std::vector<gcf_uint32>& ndRange =
			*gcf::GenericSerializeTemplateRecord::getAttribute<
					dcl::comm::EnqueueNDRangeKernel, std::vector<gcf_uint32>,
					gcf::ST_Custom1>(object, gstr);

	ndRange.clear();
	buf >> ndRange;

	return buf.getLength();
}

template<>
size_t writeCustomSTR<dcl::comm::EnqueueNDRangeKernel> (
		dcl::comm::EnqueueNDRangeKernel *object,
		const gcf::GenericSerializeTemplateRecord& gstr,
		char *buffer) {
	assert(gstr.getType() == gcf::ST_Custom1);

	gcf::ByteBuffer buf(buffer, gstr.getLength());
	std::vector<gcf_uint32>& ndRange =
			*gcf::GenericSerializeTemplateRecord::getAttribute<
					dcl::comm::EnqueueNDRangeKernel, std::vector<gcf_uint32>,
					gcf::ST_Custom1>(object, gstr);

	buf << ndRange;

	return buf.getLength();
}

} /* namespace gcf */

/******************************************************************************/

namespace dcl {

namespace comm {

static void workDimensionHtoN(
		const std::vector<size_t>& in,
		std::vector<gcf_uint32>& out) {
	std::vector<size_t>::const_iterator src;
	std::vector<gcf_uint32>::iterator dst;

	out.resize(in.size());
	src = in.begin();
	dst = out.begin();

	while (src != in.end()) {
		*dst++ = static_cast<gcf_uint32>(*src++);
	}
}

static void workDimensionNtoH(
		const std::vector<gcf_uint32>& in,
		std::vector<size_t>& out) {
	std::vector<gcf_uint32>::const_iterator src;
	std::vector<size_t>::iterator dst;

	out.resize(in.size());
	src = in.begin();
	dst = out.begin();

	while (src != in.end()) {
		*dst++ = static_cast<size_t>(*src++);
	}
}

/******************************************************************************/

EnqueueNDRangeKernel::EnqueueNDRangeKernel() :
	_commandQueueId(0), _commandId(0), _kernelId(0), _event(false) {
	initSerializable();
}

EnqueueNDRangeKernel::EnqueueNDRangeKernel(
		dcl_object_id commandQueueId,
		dcl_object_id commandId,
		dcl_object_id kernelId,
		const std::vector<size_t>& offset,
		const std::vector<size_t>& global,
		const std::vector<size_t>& local,
		const std::vector<dcl_object_id> *eventIdWaitList,
		bool event) :
	_commandQueueId(commandQueueId), _commandId(commandId), _kernelId(kernelId),
			_offset(offset.size()), _global(global.size()),
			_local(local.size()), _event(event) {
	initSerializable();

	/* convert work dimensions from host to network representation */
	workDimensionHtoN(offset, _offset);
	workDimensionHtoN(global, _global);
	workDimensionHtoN(local, _local);

	if (eventIdWaitList) {
		_eventIdWaitList = *eventIdWaitList;
	}
}

EnqueueNDRangeKernel::EnqueueNDRangeKernel(
		const EnqueueNDRangeKernel& rhs) :
	Request(rhs), _commandQueueId(rhs._commandQueueId), _commandId(rhs._commandId),
			_kernelId(rhs._kernelId), _global(rhs._global), _local(rhs._local),
			_eventIdWaitList(rhs._eventIdWaitList), _event(rhs._event) {
	initSerializable();
}

dcl_object_id EnqueueNDRangeKernel::commandQueueId() const {
	return _commandQueueId;
}

dcl_object_id EnqueueNDRangeKernel::commandId() const {
	return _commandId;
}

dcl_object_id EnqueueNDRangeKernel::kernelId() const {
	return _kernelId;
}

void EnqueueNDRangeKernel::offset(std::vector<size_t>& offset) const {
	workDimensionNtoH(_offset, offset);
}

void EnqueueNDRangeKernel::global(std::vector<size_t>& global) const {
	workDimensionNtoH(_global, global);
}

void EnqueueNDRangeKernel::local(std::vector<size_t>& local) const {
	workDimensionNtoH(_local, local);
}

const std::vector<dcl_object_id>& EnqueueNDRangeKernel::eventIdWaitList() const {
	return _eventIdWaitList;
}

bool EnqueueNDRangeKernel::event() const {
	return static_cast<bool>(_event);
}

IMPLEMENT_SERIALIZABLE_DERIVED_EVENT(
		dcl::comm::EnqueueNDRangeKernel,
		dcl::comm::Request,
		ADD_ATTRIBUTE_DEFAULT(dcl::comm::EnqueueNDRangeKernel, _commandQueueId);
		ADD_ATTRIBUTE_DEFAULT(dcl::comm::EnqueueNDRangeKernel, _commandId);
		ADD_ATTRIBUTE_DEFAULT(dcl::comm::EnqueueNDRangeKernel, _kernelId);
		ADD_ATTRIBUTE_CUSTOM(
				dcl::comm::EnqueueNDRangeKernel,
				_offset,
				gcf::ST_Custom1,
				std::vector<gcf_uint32>,
				256,
				gcf::Reliable,
				gcf::Public);
		ADD_ATTRIBUTE_CUSTOM(
				dcl::comm::EnqueueNDRangeKernel,
				_global,
				gcf::ST_Custom1,
				std::vector<gcf_uint32>,
				256,
				gcf::Reliable,
				gcf::Public);
		ADD_ATTRIBUTE_CUSTOM(
				dcl::comm::EnqueueNDRangeKernel,
				_local,
				gcf::ST_Custom1,
				std::vector<gcf_uint32>,
				256,
				gcf::Reliable,
				gcf::Public);
		ADD_ATTRIBUTE_DEFAULT(dcl::comm::EnqueueNDRangeKernel, _eventIdWaitList);
		ADD_ATTRIBUTE_DEFAULT(dcl::comm::EnqueueNDRangeKernel, _event);
)

} /* namespace comm */

} /* namespace dcl */
