/******************************************************************************
 * This file is part of dOpenCL.
 * 
 * dOpenCL is an implementation of the OpenCL application programming
 * interface for distributed systems. See <http://dopencl.uni-muenster.de/>
 * for more information.
 * 
 * Developed by: Research Group Parallel and Distributed Systems
 *               Department of Mathematics and Computer Science
 *               University of Muenster, Germany
 *               <http://pvs.uni-muenster.de/>
 * 
 * Copyright (C) 2013  Philipp Kegel <philipp.kegel@uni-muenster.de>
 *
 * dOpenCL is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * dOpenCL is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with dOpenCL. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Permission to use dOpenCL for scientific, non-commercial work is
 * granted under the terms of the dOpenCL Academic License provided
 * appropriate credit is given. See the dOpenCL Academic License for
 * more details.
 * 
 * You should have received a copy of the dOpenCL Academic License
 * along with dOpenCL. If not, see <http://dopencl.uni-muenster.de/>.
 ******************************************************************************/

/*
 * EnqueueWaitForEvents.cpp
 *
 *  Created on: 18.12.2011
 *      Author: Philipp Kegel
 */

#include <cassert>
#include <cstddef>
#include <vector>

#include <gcf/GCFTypedefs.h>
#include <gcf/GenericSerializerImplNew.cpp>
#include <gcf/Serializable.h>

#include <dcl/DCLTypes.h>
#include <dcl/comm/Request.h>
#include <dcl/comm/EnqueueWaitForEvents.h>

namespace gcf {

template<>
size_t writeCustomSTR<dcl::comm::EnqueueWaitForEvents> (
		dcl::comm::EnqueueWaitForEvents *object,
		const gcf::GenericSerializeTemplateRecord& gstr,
		char *buffer) {
	assert(gstr.getType() == gcf::ST_Custom1);

	gcf::ByteBuffer buf(buffer, gstr.getLength());
	std::vector<dcl::dcl_object_id>& eventIds =
			*gcf::GenericSerializeTemplateRecord::getAttribute<
					dcl::comm::EnqueueWaitForEvents, std::vector<dcl::dcl_object_id>,
					gcf::ST_Custom1>(object, gstr);

	buf << eventIds;

	return buf.getLength();
}

template<>
size_t readCustomSTR<dcl::comm::EnqueueWaitForEvents> (
		dcl::comm::EnqueueWaitForEvents *object,
		const gcf::GenericSerializeTemplateRecord& gstr,
		const char *buffer) {
	assert(gstr.getType() == gcf::ST_Custom1);

	gcf::ByteBuffer buf(buffer, gstr.getLength());
	std::vector<dcl::dcl_object_id>& eventIds =
			*gcf::GenericSerializeTemplateRecord::getAttribute<
					dcl::comm::EnqueueWaitForEvents, std::vector<dcl::dcl_object_id>,
					gcf::ST_Custom1>(object, gstr);

	eventIds.clear();
	buf >> eventIds;

	return buf.getLength();
}

} /* namespace gcf */

/******************************************************************************/

namespace dcl {

namespace comm {

EnqueueMarker::EnqueueMarker() {
	initSerializable();
}

EnqueueMarker::EnqueueMarker(
        dcl_object_id commandQueueId,
        dcl_object_id commandId,
        const std::vector<dcl_object_id> *eventIdWaitList,
        bool event) :
    _commandQueueId(commandQueueId), _commandId(commandId),
            _event(static_cast<gcf_bool>(event)) {
    initSerializable();

    if (eventIdWaitList) {
        _eventIdWaitList = *eventIdWaitList;
    }
}

EnqueueMarker::EnqueueMarker(
        const EnqueueMarker& rhs) :
    Request(rhs), _commandQueueId(rhs._commandQueueId),
            _commandId(rhs._commandId), _eventIdWaitList(rhs._eventIdWaitList),
            _event(rhs._event) {
    initSerializable();
}

dcl::dcl_object_id EnqueueMarker::commandQueueId() const {
	return _commandQueueId;
}

dcl_object_id EnqueueMarker::commandId() const {
    return _commandId;
}

const std::vector<dcl_object_id>& EnqueueMarker::eventIdWaitList(void) const {
    return _eventIdWaitList;
}

bool EnqueueMarker::event() const {
    return static_cast<bool>(_event);
}

IMPLEMENT_SERIALIZABLE_DERIVED_EVENT(
		dcl::comm::EnqueueMarker,
		dcl::comm::Request,
		ADD_ATTRIBUTE_DEFAULT(dcl::comm::EnqueueMarker, _commandQueueId);
        ADD_ATTRIBUTE_DEFAULT(dcl::comm::EnqueueMarker, _commandId);
        ADD_ATTRIBUTE_DEFAULT(dcl::comm::EnqueueMarker, _eventIdWaitList);
        ADD_ATTRIBUTE_DEFAULT(dcl::comm::EnqueueMarker, _event);
		)

/******************************************************************************/

EnqueueWaitForEvents::EnqueueWaitForEvents() {
	initSerializable();
}

EnqueueWaitForEvents::EnqueueWaitForEvents(
		dcl::dcl_object_id commandQueueId,
		const std::vector<dcl::dcl_object_id>& eventIdList):
		_commandQueueId(commandQueueId), _eventIdList(eventIdList) {
	initSerializable();
}

EnqueueWaitForEvents::EnqueueWaitForEvents(const EnqueueWaitForEvents& rhs) :
	Request(rhs), _commandQueueId(rhs._commandQueueId),
	_eventIdList(rhs._eventIdList) {
	initSerializable();
}

dcl::dcl_object_id EnqueueWaitForEvents::commandQueueId() const {
	return _commandQueueId;
}

const std::vector<dcl::dcl_object_id>& EnqueueWaitForEvents::eventIdList() const {
	return _eventIdList;
}

IMPLEMENT_SERIALIZABLE_DERIVED_EVENT(
		dcl::comm::EnqueueWaitForEvents,
		dcl::comm::Request,
		ADD_ATTRIBUTE_DEFAULT(dcl::comm::EnqueueWaitForEvents, _commandQueueId);
		ADD_ATTRIBUTE_CUSTOM(
				dcl::comm::EnqueueWaitForEvents, _eventIdList, gcf::ST_Custom1,
				std::vector<dcl::dcl_object_id>, 256, gcf::Reliable,
				gcf::Public);
		)

/******************************************************************************/

EnqueueBarrier::EnqueueBarrier() {
	initSerializable();
}

EnqueueBarrier::EnqueueBarrier(
        dcl_object_id commandQueueId,
        dcl_object_id commandId,
        const std::vector<dcl_object_id> *eventIdWaitList,
        bool event) :
    _commandQueueId(commandQueueId), _commandId(commandId),
            _event(static_cast<gcf_bool>(event)) {
    initSerializable();

    if (eventIdWaitList) {
        _eventIdWaitList = *eventIdWaitList;
    }
}

EnqueueBarrier::EnqueueBarrier(
        const EnqueueBarrier& rhs) :
    Request(rhs), _commandQueueId(rhs._commandQueueId),
            _commandId(rhs._commandId), _eventIdWaitList(rhs._eventIdWaitList),
            _event(rhs._event) {
    initSerializable();
}

dcl::dcl_object_id EnqueueBarrier::commandQueueId() const {
    return _commandQueueId;
}

dcl_object_id EnqueueBarrier::commandId() const {
    return _commandId;
}

const std::vector<dcl_object_id>& EnqueueBarrier::eventIdWaitList(void) const {
    return _eventIdWaitList;
}

bool EnqueueBarrier::event() const {
    return static_cast<bool>(_event);
}

IMPLEMENT_SERIALIZABLE_DERIVED_EVENT(
        dcl::comm::EnqueueBarrier,
        dcl::comm::Request,
        ADD_ATTRIBUTE_DEFAULT(dcl::comm::EnqueueBarrier, _commandQueueId);
        ADD_ATTRIBUTE_DEFAULT(dcl::comm::EnqueueBarrier, _commandId);
        ADD_ATTRIBUTE_DEFAULT(dcl::comm::EnqueueBarrier, _eventIdWaitList);
        ADD_ATTRIBUTE_DEFAULT(dcl::comm::EnqueueBarrier, _event);
        )

} /* namespace comm */

} /* namespace dcl */
