/******************************************************************************
 * This file is part of dOpenCL.
 * 
 * dOpenCL is an implementation of the OpenCL application programming
 * interface for distributed systems. See <http://dopencl.uni-muenster.de/>
 * for more information.
 * 
 * Developed by: Research Group Parallel and Distributed Systems
 *               Department of Mathematics and Computer Science
 *               University of Muenster, Germany
 *               <http://pvs.uni-muenster.de/>
 * 
 * Copyright (C) 2013  Philipp Kegel <philipp.kegel@uni-muenster.de>
 *
 * dOpenCL is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * dOpenCL is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with dOpenCL. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Permission to use dOpenCL for scientific, non-commercial work is
 * granted under the terms of the dOpenCL Academic License provided
 * appropriate credit is given. See the dOpenCL Academic License for
 * more details.
 * 
 * You should have received a copy of the dOpenCL Academic License
 * along with dOpenCL. If not, see <http://dopencl.uni-muenster.de/>.
 ******************************************************************************/

/*!
 * \file EventCommand.cpp
 *
 * Request/response messages for event objects.
 *
 * \date 2011-07-11
 * \author Philipp Kegel
 */

#include <cassert>
#include <cstddef>
#include <vector>

#ifdef __APPLE__
#include <OpenCL/cl.h>
#else
#include <CL/cl.h>
#endif

#include <gcf/GCFTypedefs.h>
#include <gcf/GenericSerializerImplNew.cpp>
#include <gcf/Serializable.h>

#include <dcl/DCLTypes.h>
#include <dcl/comm/EventCommand.h>
#include <dcl/comm/Request.h>
#include <dcl/comm/Response.h>

namespace gcf {

template<>
size_t writeCustomSTR<dcl::comm::CreateEvent> (
		dcl::comm::CreateEvent *object,
		const gcf::GenericSerializeTemplateRecord& gstr,
		char *buffer) {
	assert(gstr.getType() == gcf::ST_Custom1);

	gcf::ByteBuffer buf(buffer, gstr.getLength());
	std::vector<dcl::dcl_object_id>& memObjectIds =
			*gcf::GenericSerializeTemplateRecord::getAttribute<
					dcl::comm::CreateEvent, std::vector<dcl::dcl_object_id>,
					gcf::ST_Custom1>(object, gstr);

	buf << memObjectIds;

	return buf.getLength();
}

template<>
size_t readCustomSTR<dcl::comm::CreateEvent> (
		dcl::comm::CreateEvent *object,
		const gcf::GenericSerializeTemplateRecord& gstr,
		const char *buffer) {
	assert(gstr.getType() == gcf::ST_Custom1);

	gcf::ByteBuffer buf(buffer, gstr.getLength());
	std::vector<dcl::dcl_object_id>& memObjectIds =
			*gcf::GenericSerializeTemplateRecord::getAttribute<
					dcl::comm::CreateEvent, std::vector<dcl::dcl_object_id>,
					gcf::ST_Custom1>(object, gstr);

	memObjectIds.clear();
	buf >> memObjectIds;

	return buf.getLength();
}

/******************************************************************************/

template<>
size_t writeCustomSTR<dcl::comm::EventProfilingInfo> (
		dcl::comm::EventProfilingInfo *object,
		const gcf::GenericSerializeTemplateRecord& gstr,
		char *buffer) {
	assert(gstr.getType() == gcf::ST_Custom1);

	gcf::ByteBuffer buf(buffer, gstr.getLength());
	gcf_uint64& profilingInfo =
			*gcf::GenericSerializeTemplateRecord::getAttribute<
					dcl::comm::EventProfilingInfo, gcf_uint64,
					gcf::ST_Custom1>(object, gstr);

	buf << profilingInfo;

	return buf.getLength();
}

template<>
size_t readCustomSTR<dcl::comm::EventProfilingInfo> (
		dcl::comm::EventProfilingInfo *object,
		const gcf::GenericSerializeTemplateRecord& gstr,
		const char *buffer) {
	assert(gstr.getType() == gcf::ST_Custom1);

	gcf::ByteBuffer buf(buffer, gstr.getLength());
	gcf_uint64& profilingInfo =
			*gcf::GenericSerializeTemplateRecord::getAttribute<
					dcl::comm::EventProfilingInfo, gcf_uint64,
					gcf::ST_Custom1>(object, gstr);

	buf >> profilingInfo;

	return buf.getLength();
}

} /* namespace gcf */

/******************************************************************************/

namespace dcl {

namespace comm {

/******************************************************************************
 * Create user event
 ******************************************************************************/

CreateEvent::CreateEvent() {
	initSerializable();
}

CreateEvent::CreateEvent(dcl::dcl_object_id contextId,
		dcl::dcl_object_id eventId,
		const std::vector<dcl_object_id>& memObjectIds) :
	_contextId(contextId), _eventId(eventId), _memObjectIds(memObjectIds) {
	initSerializable();
}

CreateEvent::CreateEvent(
		const CreateEvent& rhs) :
	Request(rhs), _contextId(rhs._contextId), _eventId(rhs._eventId),
	_memObjectIds(rhs._memObjectIds)  {
	initSerializable();
}

CreateEvent::~CreateEvent() { }

dcl_object_id CreateEvent::contextId() const {
	return _contextId;
}

dcl_object_id CreateEvent::eventId() const {
	return _eventId;
}

const std::vector<dcl_object_id>& CreateEvent::memObjectIds() const {
	return _memObjectIds;
}

IMPLEMENT_SERIALIZABLE_DERIVED_EVENT(
		dcl::comm::CreateEvent,
		dcl::comm::Request,
		ADD_ATTRIBUTE_DEFAULT(dcl::comm::CreateEvent, _contextId);
		ADD_ATTRIBUTE_DEFAULT(dcl::comm::CreateEvent, _eventId);
		ADD_ATTRIBUTE_CUSTOM(dcl::comm::CreateEvent, _memObjectIds, gcf::ST_Custom1,
				std::vector<dcl::dcl_object_id>, 256, gcf::Reliable, gcf::Public);
		)

/******************************************************************************
 * Create user event
 ******************************************************************************/

CreateUserEvent::CreateUserEvent() {
	initSerializable();
}

CreateUserEvent::CreateUserEvent(
		dcl::dcl_object_id contextId, dcl::dcl_object_id eventId):
	_contextId(contextId), _eventId(eventId) {
	initSerializable();
}

CreateUserEvent::CreateUserEvent(
		const CreateUserEvent& rhs) :
	Request(rhs), _contextId(rhs._contextId), _eventId(rhs._eventId) {
	initSerializable();
}

CreateUserEvent::~CreateUserEvent() { }

dcl_object_id CreateUserEvent::contextId() const {
	return _contextId;
}

dcl_object_id CreateUserEvent::eventId() const {
	return _eventId;
}

IMPLEMENT_SERIALIZABLE_DERIVED_EVENT(
		dcl::comm::CreateUserEvent,
		dcl::comm::Request,
		ADD_ATTRIBUTE_DEFAULT(dcl::comm::CreateUserEvent, _contextId);
		ADD_ATTRIBUTE_DEFAULT(dcl::comm::CreateUserEvent, _eventId);
		)

/******************************************************************************
 * Delete event
 ******************************************************************************/

DeleteEvent::DeleteEvent() :
	_eventId(0) {
	initSerializable();
}

DeleteEvent::DeleteEvent(
		dcl_object_id eventId):
	_eventId(eventId) {
	initSerializable();
}

DeleteEvent::DeleteEvent(
		const DeleteEvent& rhs) :
	Request(rhs), _eventId(rhs._eventId) {
	initSerializable();
}

DeleteEvent::~DeleteEvent() { }

dcl_object_id DeleteEvent::eventId() const {
	return _eventId;
}

IMPLEMENT_SERIALIZABLE_DERIVED_EVENT(
		dcl::comm::DeleteEvent,
		dcl::comm::Request,
		ADD_ATTRIBUTE_DEFAULT(dcl::comm::DeleteEvent, _eventId);
		)

/******************************************************************************
 * Set user event status
 ******************************************************************************/

SetUserEventStatusRequest::SetUserEventStatusRequest() {
	initSerializable();
}

SetUserEventStatusRequest::SetUserEventStatusRequest(
		dcl::dcl_object_id eventId, cl_int status):
	_eventId(eventId), _status(static_cast<gcf_uint32>(status)) {
	initSerializable();
}

SetUserEventStatusRequest::SetUserEventStatusRequest(
		const SetUserEventStatusRequest& rhs) :
	Request(rhs), _eventId(rhs._eventId), _status(rhs._status) {
	initSerializable();
}

SetUserEventStatusRequest::~SetUserEventStatusRequest() { }

dcl_object_id SetUserEventStatusRequest::eventId() const {
	return _eventId;
}

cl_int SetUserEventStatusRequest::status() const {
	return static_cast<cl_int>(_status);
}

IMPLEMENT_SERIALIZABLE_DERIVED_EVENT(
		dcl::comm::SetUserEventStatusRequest,
		dcl::comm::Request,
		ADD_ATTRIBUTE_DEFAULT(dcl::comm::SetUserEventStatusRequest, _eventId);
		ADD_ATTRIBUTE_DEFAULT(dcl::comm::SetUserEventStatusRequest, _status);
		)

/******************************************************************************
 * Get event profiling info
 ******************************************************************************/

GetEventProfilingInfo::GetEventProfilingInfo() {
	initSerializable();
}

GetEventProfilingInfo::GetEventProfilingInfo(
		dcl_object_id eventID) : _eventId(eventID) {
	initSerializable();
}

GetEventProfilingInfo::GetEventProfilingInfo(const GetEventProfilingInfo& rhs) :
	_eventId(rhs._eventId) {
	initSerializable();
}

GetEventProfilingInfo::~GetEventProfilingInfo() { }

dcl_object_id GetEventProfilingInfo::eventId() const {
	return _eventId;
}

IMPLEMENT_SERIALIZABLE_DERIVED_EVENT(
		dcl::comm::GetEventProfilingInfo,
		dcl::comm::Request,
		ADD_ATTRIBUTE_DEFAULT(dcl::comm::GetEventProfilingInfo, _eventId);
		)

/******************************************************************************/

EventProfilingInfo::EventProfilingInfo() {
	initSerializable();
}

EventProfilingInfo::EventProfilingInfo(const Request& request,
        cl_ulong received, cl_ulong queued, cl_ulong submit, cl_ulong start,
        cl_ulong end) :
	Response(request),
    _received(static_cast<gcf_uint64>(received)),
	_queued(static_cast<gcf_uint64>(queued)),
	_submit(static_cast<gcf_uint64>(submit)),
	_start(static_cast<gcf_uint64>(start)),
	_end(static_cast<gcf_uint64>(end))
{
	initSerializable();
}

EventProfilingInfo::EventProfilingInfo(const EventProfilingInfo& rhs) :
	Response(rhs), _received(rhs._received), _queued(rhs._queued),
	_submit(rhs._submit), _start(rhs._start), _end(rhs._end)
{
	initSerializable();
}

EventProfilingInfo::~EventProfilingInfo() { }

cl_ulong EventProfilingInfo::received() const {
    return static_cast<cl_ulong>(_received);
}

cl_ulong EventProfilingInfo::queued() const {
	return static_cast<cl_ulong>(_queued);
}

cl_ulong EventProfilingInfo::submit() const {
	return static_cast<cl_ulong>(_submit);
}

cl_ulong EventProfilingInfo::start() const {
	return static_cast<cl_ulong>(_start);
}

cl_ulong EventProfilingInfo::end() const {
	return static_cast<cl_ulong>(_end);
}

/*
 * GCF does not provide automatic serialization for gcf_uint64.
 * Hence, values of this type have to be serialized explicitly.
 */
IMPLEMENT_SERIALIZABLE_DERIVED_EVENT(
		dcl::comm::EventProfilingInfo,
		dcl::comm::Response,
        ADD_ATTRIBUTE_CUSTOM(dcl::comm::EventProfilingInfo, _received,
                gcf::ST_Custom1, gcf_uint64, 8, gcf::Reliable, gcf::Public);
		ADD_ATTRIBUTE_CUSTOM(dcl::comm::EventProfilingInfo, _queued,
				gcf::ST_Custom1, gcf_uint64, 8, gcf::Reliable, gcf::Public);
		ADD_ATTRIBUTE_CUSTOM(dcl::comm::EventProfilingInfo, _submit,
				gcf::ST_Custom1, gcf_uint64, 8, gcf::Reliable, gcf::Public);
		ADD_ATTRIBUTE_CUSTOM(dcl::comm::EventProfilingInfo, _start,
				gcf::ST_Custom1, gcf_uint64, 8, gcf::Reliable, gcf::Public);
		ADD_ATTRIBUTE_CUSTOM(dcl::comm::EventProfilingInfo, _end,
				gcf::ST_Custom1, gcf_uint64, 8, gcf::Reliable, gcf::Public);
		)

} /* namespace comm */

} /* namespace dcl */
