/******************************************************************************
 * This file is part of dOpenCL.
 * 
 * dOpenCL is an implementation of the OpenCL application programming
 * interface for distributed systems. See <http://dopencl.uni-muenster.de/>
 * for more information.
 * 
 * Developed by: Research Group Parallel and Distributed Systems
 *               Department of Mathematics and Computer Science
 *               University of Muenster, Germany
 *               <http://pvs.uni-muenster.de/>
 * 
 * Copyright (C) 2013  Philipp Kegel <philipp.kegel@uni-muenster.de>
 *
 * dOpenCL is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * dOpenCL is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with dOpenCL. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Permission to use dOpenCL for scientific, non-commercial work is
 * granted under the terms of the dOpenCL Academic License provided
 * appropriate credit is given. See the dOpenCL Academic License for
 * more details.
 * 
 * You should have received a copy of the dOpenCL Academic License
 * along with dOpenCL. If not, see <http://dopencl.uni-muenster.de/>.
 ******************************************************************************/

/*
 * ResponseBuffer.cpp
 *
 *  Created on: 18.04.2011
 *      Author: Philipp Kegel
 */

#include <cstddef>
#include <vector>

#include <boost/thread/condition_variable.hpp>
#include <boost/thread/exceptions.hpp>
#include <boost/thread/locks.hpp>
#include <boost/thread/mutex.hpp>

#include <dcl/comm/Request.h>
#include <dcl/comm/Response.h>

#include "ResponseBuffer.h"

namespace dcl {

namespace comm {

ResponseBuffer::ResponseBuffer(size_t size) :
	_responses(size, 0), _head(_responses.begin()), _tail(_responses.begin()), _interrupt(false) {
}

ResponseBuffer::~ResponseBuffer() {
}

void ResponseBuffer::add(Response *response) {
	boost::lock_guard<boost::mutex> lock(_mutex);

	while (!_interrupt && !insert(response)) {
		_responseRemoved.wait(_mutex);
	}
	if (_interrupt) throw boost::thread_interrupted();
}

Response * ResponseBuffer::tryGet(const Request& request) {
	boost::lock_guard<boost::mutex> lock(_mutex);
	return remove(request);
}

Response * ResponseBuffer::get(const Request& request) {
	boost::lock_guard<boost::mutex> lock(_mutex);
    Response *response;

	while (!_interrupt && !(response = remove(request))) {
		_responseAdded.wait(_mutex);
	};
	if (_interrupt) throw boost::thread_interrupted();

	return response;
}

void ResponseBuffer::interrupt() {
	boost::lock_guard<boost::mutex> lock(_mutex);
	_interrupt = true;
	_responseAdded.notify_all();
	_responseRemoved.notify_all();
}

void ResponseBuffer::clear() {
	boost::lock_guard<boost::mutex> lock(_mutex);

	for (std::vector<Response *>::iterator i = _responses.begin();
			i != _responses.end(); ++i) {
		delete (*i);
	}
	_responses.clear();
	_tail = _head = _responses.begin();

    _responseRemoved.notify_all();
}

bool ResponseBuffer::insert(Response *response) {
	bool inserted = false; /* buffer full */
    std::vector<Response *>::iterator i = _tail;

    do {
    	std::vector<Response *>::iterator entry = i;

        /* next buffer position */
    	if (++i == _responses.end()) {
    		i = _responses.begin();
    	}

        if (!*entry) {
            *entry = response; /* insert response into buffer */
            inserted = true;   /* response inserted successfully */
            _tail = i;         /* remember next buffer position for subsequent insert operation */
            _responseAdded.notify_all();
            break;
        }
    } while (i != _tail); /* only run through buffer once */

    return inserted;
}

Response * ResponseBuffer::remove(const Request& request) {
	Response *responseRet = 0;
    std::vector<Response *>::iterator i = _head;

    do {
    	std::vector<Response *>::iterator entry = i;
    	Response *response = *entry;

        /* next buffer position */
        if (++i == _responses.end()) {
            i = _responses.begin();
        }
        if (response && response->requestId == request.id) {
            *entry = 0;          	/* remove response from buffer */
            responseRet = response; /* response removed successfully */
            _head = i;           	/* remember next buffer position for subsequent search */
            _responseRemoved.notify_one();
            break;
        }

    } while (i != _head); /* only run through buffer once */

    return responseRet;
}

} /* namespace comm */

} /* namespace dcl */
