/******************************************************************************
 * This file is part of dOpenCL.
 * 
 * dOpenCL is an implementation of the OpenCL application programming
 * interface for distributed systems. See <http://dopencl.uni-muenster.de/>
 * for more information.
 * 
 * Developed by: Research Group Parallel and Distributed Systems
 *               Department of Mathematics and Computer Science
 *               University of Muenster, Germany
 *               <http://pvs.uni-muenster.de/>
 * 
 * Copyright (C) 2013  Philipp Kegel <philipp.kegel@uni-muenster.de>
 *
 * dOpenCL is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * dOpenCL is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with dOpenCL. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Permission to use dOpenCL for scientific, non-commercial work is
 * granted under the terms of the dOpenCL Academic License provided
 * appropriate credit is given. See the dOpenCL Academic License for
 * more details.
 * 
 * You should have received a copy of the dOpenCL Academic License
 * along with dOpenCL. If not, see <http://dopencl.uni-muenster.de/>.
 ******************************************************************************/

/*
 * ResponseBuffer.h
 *
 *  Created on: 18.04.2011
 *      Author: Philipp Kegel
 *
 * A ring buffer for saving responses from compute nodes.
 */

#ifndef RESPONSEBUFFER_H_
#define RESPONSEBUFFER_H_

#include <cstddef>
#include <vector>

#include <boost/thread/condition_variable.hpp>
#include <boost/thread/exceptions.hpp>
#include <boost/thread/locks.hpp>
#include <boost/thread/mutex.hpp>

#include <dcl/comm/Request.h>
#include <dcl/comm/Response.h>

namespace dcl {

namespace comm {

class ResponseBuffer {
public:
	static const size_t DEFAULT_SIZE = 64;

	ResponseBuffer(size_t size = DEFAULT_SIZE);
	virtual ~ResponseBuffer();

	void add(Response *response);
    /* TODO Add timed method for adding response */

	Response * tryGet(const Request& request);
	/**
	 * @brief Waits for a response indefinitely.
	 */
	Response * get(const Request& request);

	/**
	 * @brief Waits for a response.
	 *
	 * @param request   the request which is associated with the response
	 * @param timeout	a timeout
	 * @return the response, or NULL if @c timeout has been reached
	 */
	template<typename duration_type>
	Response * get(
	        const Request& request,
	        const duration_type& timeout) {
	    boost::lock_guard<boost::mutex> lock(_mutex);
	    Response *response;

	    while (!_interrupt && !(response = remove(request))) {
	        if (!_responseAdded.timed_wait(_mutex, timeout)) {
	            /* false if time timeout is reached, otherwise true */
	            response = 0;
	            break;
	        }
	    };
	    if (_interrupt) throw boost::thread_interrupted();

	    return response;
	}

	void interrupt();
	void clear();

private:
	bool insert(Response *response);
	Response * remove(const Request& request);

	std::vector<Response *> _responses;
	std::vector<Response *>::iterator _head, _tail;

	boost::mutex _mutex;                            /**< buffer mutex */
	boost::condition_variable_any _responseAdded;   /**< condition: added response to buffer */
	boost::condition_variable_any _responseRemoved; /**< condition: removed response from buffer */
	bool _interrupt;
};

} /* namespace comm */

} /* namespace dcl */

#endif /* RESPONSEBUFFER_H_ */
