/******************************************************************************
 * This file is part of dOpenCL.
 * 
 * dOpenCL is an implementation of the OpenCL application programming
 * interface for distributed systems. See <http://dopencl.uni-muenster.de/>
 * for more information.
 * 
 * Developed by: Research Group Parallel and Distributed Systems
 *               Department of Mathematics and Computer Science
 *               University of Muenster, Germany
 *               <http://pvs.uni-muenster.de/>
 * 
 * Copyright (C) 2013  Philipp Kegel <philipp.kegel@uni-muenster.de>
 *
 * dOpenCL is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * dOpenCL is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with dOpenCL. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Permission to use dOpenCL for scientific, non-commercial work is
 * granted under the terms of the dOpenCL Academic License provided
 * appropriate credit is given. See the dOpenCL Academic License for
 * more details.
 * 
 * You should have received a copy of the dOpenCL Academic License
 * along with dOpenCL. If not, see <http://dopencl.uni-muenster.de/>.
 ******************************************************************************/

/**
 * @file    Event.cpp
 *
 * @date    2011-06-21
 * @author  Philipp Kegel
 * @author  Louay Hashim
 */

#include <algorithm>
#include <cassert>
#include <cstdlib>
#include <iostream>
#include <map>
#include <set>
#include <stdexcept>
#include <utility>
#include <vector>

#include <boost/bind.hpp>
#include <boost/shared_ptr.hpp>

#include <boost/thread/condition_variable.hpp>
#include <boost/thread/locks.hpp>
#include <boost/thread/recursive_mutex.hpp>

#ifdef __APPLE__
#include <OpenCL/cl.h>
#include <OpenCL/cl_wwu_dcl.h>
#else
#include <CL/cl.h>
#include <CL/cl_wwu_dcl.h>
#endif

#include <dcl/CLError.h>
#include <dcl/CLObjectRegistry.h>
#include <dcl/ComputeNode.h>
#include <dcl/DataTransfer.h>
#include <dcl/DCLException.h>
#include <dcl/DCLTypes.h>
#include "dcl/Remote.h"

#include <dcl/comm/CommandMessage.h>
#include <dcl/comm/EventCommand.h>
#include <dcl/comm/EventSynchronizationMessage.h>

#include <dcl/util/Clock.h>
#include <dcl/util/Logger.h>

#include "Context.h"
#include "CommandQueue.h"
#include "Error.h"
#include "Event.h"
#include "Memory.h"
#include "Platform.h"
#include "Retainable.h"
#include "utility.h"


_cl_event::_cl_event(cl_context context, cl_int status) :
	_context(context), _status(status) {
    if (!context) { throw dclicd::Error(CL_INVALID_CONTEXT); }
}

_cl_event::~_cl_event() { }

void _cl_event::retain() {
    boost::lock_guard<boost::recursive_mutex> lock(_statusMutex);
    ++_ref_count;
}

bool _cl_event::release() {
    boost::lock_guard<boost::recursive_mutex> lock(_statusMutex);
    cl_uint refCount;

    refCount = _ref_count--;
    assert(refCount > 0); // _ref_count must be greater than zero when calling release

    if (refCount == 1 && isComplete()) {
        destroy();
        return true;
    } else {
        return false;
    }
}

void _cl_event::destroy() {
    /* Events must only be deleted if their reference count is 0 *and* their
     * associated command is completed (or terminated).
     * The event is required to forward its execution status to remote events.
     *
     * Note that the event on the host does not have to be retained for
     * enqueued commands that require a wait for this event as the remote
     * events are implicitly retained by the compute nodes' OpenCL
     * implementations. */
	assert(_ref_count == 0);
	assert(isComplete());

	try {
		dcl::comm::DeleteEvent request(remoteId());
		dcl::executeCommand(_context->computeNodes(), request);
		dcl::util::Logger << dcl::util::Info
				<< "Event deleted (ID=" << remoteId() << ')' << std::endl;
	} catch (const dcl::CLError& err) {
		throw dclicd::Error(err);
	} catch (const dcl::IOException& err) {
		throw dclicd::Error(err);
	} catch (const dcl::ProtocolException& err) {
		throw dclicd::Error(err);
	}
}

void _cl_event::waitForEvents(const std::vector<cl_event>& event_list) {
	cl_event event;
	cl_context context;
	cl_command_queue queue;
	std::set<cl_command_queue> queues;

	if (event_list.empty()) throw dclicd::Error(CL_INVALID_VALUE);

	/*
	 * Validate first event in list
	 */
	event = event_list.front();
	if (!event) throw dclicd::Error(CL_INVALID_EVENT);

	context = event->_context;

	queue = event->commandQueue();
	if (queue) {
		queues.insert(queue);
	}

	/*
	 * Validate remaining events
	 */
	for (std::vector<cl_event>::const_iterator i = event_list.begin();
			i != event_list.end(); ++i)	{
		event = *i;
		if (!event) throw dclicd::Error(CL_INVALID_EVENT);
		if (event->_context != context) throw dclicd::Error(CL_INVALID_CONTEXT);

		queue = event->commandQueue();
		if (queue) {
			queues.insert(queue);
		}
	}

	/* Flush command queues */
	std::for_each(queues.begin(), queues.end(),
			boost::bind(&_cl_command_queue::flush, _1));

	/*
	 * Wait for events
	 */
	try {
		/* Here, @c waitNoFlush is used to avoid redundant flushing of command queues. */
		std::for_each(event_list.begin(), event_list.end(),
				boost::bind(&_cl_event::waitNoFlush, _1));
	} catch (const dclicd::Error&) {
		/* the execution status of any of the events in event_list is a negative
		 * integer value */
		throw dclicd::Error(CL_EXEC_STATUS_ERROR_FOR_EVENTS_IN_WAIT_LIST);
	}
}

void _cl_event::setCallback(
		cl_int command_exec_callback_type,
		void (CL_CALLBACK *pfn_event_notify)(
				cl_event event,
				cl_int event_command_exec_status,
				void *user_data),
		void *user_data) {
	if (command_exec_callback_type != CL_COMPLETE) throw dclicd::Error(CL_INVALID_VALUE);
	if (!pfn_event_notify) throw dclicd::Error(CL_INVALID_VALUE);

	{
		boost::lock_guard<boost::recursive_mutex> lock(_statusMutex);

		/*
		 * The following mechanism is the reason an event requires its own
		 * command execution status rather than querying it from its associated
		 * command.
		 * When a callback is set, it must be called immediately if the event's
		 * command execution status already is equal to (or lower than) the
		 * callback's command execution status. However, this status must not
		 * change between adding the callback to the list and checking if the
		 * callback should be triggered immediately.
		 * Therefore, adding the callback and checking the event's command
		 * execution status must be performed atomically.
		 */
		if (_status <= command_exec_callback_type) {
			pfn_event_notify(this, CL_COMPLETE, user_data);
		}

		/* Add callback to list */
		_callbacks[command_exec_callback_type].push_back(
				std::make_pair(pfn_event_notify, user_data));
	}
}

void _cl_event::getInfo(
		cl_event_info param_name,
		size_t param_value_size,
		void *param_value,
		size_t *param_value_size_ret) {
	switch (param_name) {
	case CL_EVENT_COMMAND_QUEUE:
		dclicd::copy_info(commandQueue(), param_value_size,
				param_value, param_value_size_ret);
		break;
	case CL_EVENT_CONTEXT:
		dclicd::copy_info(_context, param_value_size,
				param_value, param_value_size_ret);
		break;
	case CL_EVENT_COMMAND_EXECUTION_STATUS:
    {
        boost::lock_guard<boost::recursive_mutex> lock(_statusMutex);
		dclicd::copy_info(_status, param_value_size,
			param_value, param_value_size_ret);
		break;
    }
	case CL_EVENT_REFERENCE_COUNT:
	{
        boost::lock_guard<boost::recursive_mutex> lock(_statusMutex);
		dclicd::copy_info(_ref_count, param_value_size,
				param_value, param_value_size_ret);
		break;
	}
	case CL_EVENT_COMMAND_TYPE:
		dclicd::copy_info(commandType(), param_value_size,
				param_value, param_value_size_ret);
		break;
	default:
		throw dclicd::Error(CL_INVALID_VALUE);
	}
}

bool _cl_event::isComplete() const {
	return (_status < 0 /* terminated */
			|| _status == CL_COMPLETE);
}

void _cl_event::waitNoFlush() const {
	boost::lock_guard<boost::recursive_mutex> lock(_statusMutex);
	while (_status > CL_COMPLETE) _statusChanged.wait(_statusMutex);
}

bool _cl_event::setCommandExecutionStatus(cl_int status) {
    boost::lock_guard<boost::recursive_mutex> lock(_statusMutex);

    /* Trigger callback's *before* setting the event complete. Otherwise, an
     * application thread waiting for the event to complete may be resumed and
     * then releases *and* deletes the event concurrently.
     *
     * In order to allow callbacks to call methods of the event, the event's
     * lock has to be reentrant. */
    triggerCallbacks(status);

    _status = status;
    _statusChanged.notify_all();

    if (_ref_count == 0 && isComplete()) {
        destroy();
        return true;
    } else {
        return false;
    }
}

void _cl_event::triggerCallbacks(cl_int status) {
	std::map<cl_int, std::vector<std::pair<void(*)(cl_event, cl_int, void *),
			void *> > >::const_iterator i;

	/* FIXME Event callbacks should be able to release their event
	 * Though, an event must not be deleted (by the implementation) before all
	 * callbacks have been called, as the event is passed as an argument. */

	/*
	 * Call callback functions registered for CL_COMPLETE if the event's
	 * command has completed successfully or is abnormally terminated (i.e.,
	 * status < 0).
	 */
	i = _callbacks.find((status <= CL_COMPLETE) ? CL_COMPLETE : status);
	if (i != _callbacks.end()) {
		const std::vector<std::pair<void(*)(cl_event, cl_int, void *),
				void *> > & callbacks = i->second;

		for (std::vector<std::pair<void(*)(cl_event, cl_int, void *), void *> >::const_iterator j = callbacks.begin();
				j != callbacks.end(); ++j) {
			j->first(this, status, j->second);
		}
	}
}

/******************************************************************************
 * Event
 ******************************************************************************/

namespace dclicd {

Event::Event(cl_context context, 
	const boost::shared_ptr<dclicd::command::Command>& command,
	const std::vector<cl_mem>& memoryObjects) :
	_cl_event(context, CL_QUEUED), dcl::Remote(command->remoteId()),
	_command(command), _commandQueued(dcl::util::clock.getTime()),
	_profilingInfo(0), _memoryObjects(memoryObjects)
{
    assert(command != 0); // command must not be NULL
#ifndef NDEBUG
    /* Context must be context of command's command queue */
    command->commandQueue()->getInfo(CL_QUEUE_CONTEXT, sizeof(context), &context, 0);
    assert(context == _context);
#endif

    /* register event (required for consistency protocol) */
    _context->getPlatform()->remote().objectRegistry().bind<dcl::EventListener>(_id, *this);

	try {
		std::vector<dcl::dcl_object_id> memoryObjectIds;
		
		for (std::vector<cl_mem>::const_iterator i = memoryObjects.begin(); i != memoryObjects.end(); ++i) {
		    memoryObjectIds.push_back((*i)->remoteId());
		}
		
		dcl::comm::CreateEvent createEvent(_context->remoteId(),
				_command->remoteId(), memoryObjectIds);
		std::vector<dcl::ComputeNode *> computeNodes(_context->computeNodes());

		/* Create list of 'other' compute nodes.
		 * Other compute nodes are compute nodes where no command has been
		 * enqueued but which belong to the same context as the command queue
		 * where the associated command has been enqueued. */
		computeNodes.erase(std::find(computeNodes.begin(), computeNodes.end(),
				&_command->commandQueue()->computeNode()));

		/* Create substitute events on other compute nodes */
		dcl::executeCommand(computeNodes, createEvent);
		dcl::util::Logger << dcl::util::Info
				<< "Event created (ID=" << _command->remoteId() << ')'
				<< std::endl;
	} catch (const dcl::CLError& err) {
		throw dclicd::Error(err);
	} catch (const dcl::IOException& err) {
		throw dclicd::Error(err);
	} catch (const dcl::ProtocolException& err) {
		throw dclicd::Error(err);
	}

    _command->setEvent(*this); // attach event to local command
	_command->commandQueue()->retain();
}

Event::~Event() {
    dclicd::release(_command->commandQueue());

	/* deregister event */
    _context->getPlatform()->remote().objectRegistry().unbind<dcl::EventListener>(_id);

	delete _profilingInfo;
}

dcl::dcl_object_id Event::remoteId() const {
	return _command->remoteId();
}

void Event::wait() const {
	/* Perform an implicit flush to ensure that the associated command will be
	 * executed eventually */
	_command->commandQueue()->flush();

	/* Do not wait for the associated command's execution status but on the
	 * event's one as the event status also includes the completion of
	 * operations associated with the event, e.g., triggering the callbacks. */
	waitNoFlush();
}

bool Event::onCommandExecutionStatusChanged(cl_int status) {
    /* TODO Send event status update to other compute nodes directly from the
     *      compute node hosting the original event.
     * The following is a work-around for missing node-to-node communication.
     * However, commands that are finished by the host, e.g., 'read buffer',
     * still have to send the event status update. */
	if (status < 0 || status == CL_COMPLETE) { /* command failed or has been completed */
		std::vector<dcl::ComputeNode *> computeNodes = _context->computeNodes();

		/*
		 * Forward event status change to other compute nodes in context
		 * The event status has been set to an error code or to 'complete' on
		 * the compute node owning the event. The statuses of the corresponding
		 * substitute events on other compute nodes of the context now have to
		 * be updated accordingly.
		 */
		computeNodes.erase(std::find(computeNodes.begin(), computeNodes.end(),
				&_command->commandQueue()->computeNode()));
		if (!computeNodes.empty()) {
            try {
                dcl::comm::CommandExecutionStatusChangedMessage message(_command->remoteId(), status);

                dcl::sendMessage(computeNodes, message);
                dcl::util::Logger << dcl::util::Debug
                        << "Forwarded update of command execution status to compute nodes (ID=" << remoteId()
                        << ", status=" << status
                        << ')' << std::endl;
            } catch (const dcl::DCLException& err) {
                /* Application state has become inconsistent, abort */
                std::cerr << "ERROR: event status update failed" << std::endl;
                abort();
            }
		}
	}

	return setCommandExecutionStatus(status);
}

void Event::getProfilingInfo(
		cl_kernel_info param_name,
		size_t param_value_size,
		void *param_value,
		size_t *param_value_size_ret) {
    /*
     * Do NOT check locally if event is complete!
     * Remote event may be complete after _cl_command_queue::finish, but local
     * event still can be incomplete because of message latency
	if (!isComplete()) throw dclicd::Error(CL_PROFILING_INFO_NOT_AVAILABLE);
     */

	if (!_profilingInfo) {
		/* Query profiling info from compute node */
		try {
			dcl::comm::GetEventProfilingInfo request(_command->remoteId());
			boost::shared_ptr<dcl::comm::EventProfilingInfo> response(
					dynamic_cast<dcl::comm::EventProfilingInfo *>(
							_command->commandQueue()->computeNode().executeCommand(
									request, dcl::comm::EventProfilingInfo::TYPE)));

			_profilingInfo = new detail::EventProfilingInfo(
			        response->received(),response->queued(), response->submit(),
			        response->start(), response->end());
		} catch (const std::bad_alloc&) {
			throw dclicd::Error(CL_OUT_OF_HOST_MEMORY);
		} catch (const dcl::CLError& err) {
			throw dclicd::Error(err);
		} catch (const dcl::IOException& err) {
			throw dclicd::Error(err);
		} catch (const dcl::ProtocolException& err) {
			throw dclicd::Error(err);
		}
	}

    switch (param_name) {
    case CL_PROFILING_COMMAND_QUEUED:
        /* time of queuing the command on the host */
        dclicd::copy_info(_commandQueued, param_value_size, param_value,
                param_value_size_ret);
        break;
    case CL_PROFILING_COMMAND_RECEIVED_WWU:
        /* time of receiving/queuing the command on the compute node */
        dclicd::copy_info(_profilingInfo->received(), param_value_size,
                param_value, param_value_size_ret);
        break;
    case CL_PROFILING_COMMAND_SUBMIT:
        dclicd::copy_info(_profilingInfo->submit(), param_value_size,
                param_value, param_value_size_ret);
        break;
    case CL_PROFILING_COMMAND_START:
        dclicd::copy_info(_profilingInfo->start(), param_value_size,
                param_value, param_value_size_ret);
        break;
    case CL_PROFILING_COMMAND_END:
        dclicd::copy_info(_profilingInfo->end(), param_value_size,
                param_value, param_value_size_ret);
        break;
    default:
        throw dclicd::Error(CL_INVALID_VALUE);
    }
}

cl_command_type Event::commandType() const {
	return _command->type();
}

cl_command_queue Event::commandQueue() const {
	return _command->commandQueue();
}

void Event::synchronize() {
    /* TODO Implement dclicd::Event::synchronize
     * 1. Send synchronization request to compute node
     *    Presumably, this step can be omitted as a synchronization request is
     *    implied by clWaitForEvents and clFinish, which are the only APIs (?)
     *    that require event synchronization on the host.
     * 2. Acquire associated memory object changes */
}

void Event::onSynchronize(dcl::Process& process) {
    dcl::util::Logger << dcl::util::Debug
            << "(MEM) Event synchronization (ID=" << _id
            << ") requested by compute node '" << process.url() << '\''
            << std::endl;

    if (_memoryObjects.empty()) return;

    /* forward synchronization request to event's compute node */
    dcl::comm::EventSynchronizationMessage msg(_id);
    _command->commandQueue()->computeNode().sendMessage(msg);
    dcl::util::Logger << dcl::util::Debug
            << "(MEM) Forwarded event synchronization request (ID=" << remoteId()
            << ") to compute node '" << _command->commandQueue()->computeNode().url() << '\''
            << std::endl;

    /* TODO Implement forwarding of acquire request in this method rather than using a memory object's mechanisms
     * The acquire/release operations will be implemented on the host to
     * acquire/release the associated changes of the memory objects from a
     * compute node to the memory objects' host pointers. These operations must
     * not interfere with synchronization operations between compute nodes that
     * are currently mediated by the host  (work-around for missing
     * node-to-node communication). */

    /* acquire and release memory objects from event's compute node */
    for (std::vector<cl_mem>::const_iterator it = _memoryObjects.begin();
            it != _memoryObjects.end(); ++it) {
        (*it)->onAcquire(process, _command->commandQueue()->computeNode());
    }
}

/******************************************************************************
 * User event
 ******************************************************************************/

UserEvent::UserEvent(cl_context context) :
	dcl::Remote(dcl::Remote::generateId()),
	_cl_event(context, CL_SUBMITTED)
{
	try {
		dcl::comm::CreateUserEvent request(context->remoteId(), _id);
		dcl::executeCommand(_context->computeNodes(), request);
		dcl::util::Logger << dcl::util::Info
				<< "User event created (ID=" << _id << ')' << std::endl;
	} catch (const dcl::CLError& err) {
		throw dclicd::Error(err);
	} catch (const dcl::IOException& err) {
		throw dclicd::Error(err);
	} catch (const dcl::ProtocolException& err) {
		throw dclicd::Error(err);
	}

	_context->retain();
}

UserEvent::~UserEvent() {
    dclicd::release(_context);
}

dcl::dcl_object_id UserEvent::remoteId() const {
	return _id;
}

void UserEvent::wait() const {
	/*
	 * User events are not associated with a command queue. Therefore, no
	 * command queue is flushed when waiting for a user event.
	 */
	waitNoFlush();
}

void UserEvent::setStatus(cl_int status) {
	if (status > 0) {
		/* Event status to set must be either CL_COMPLETE or a negative integer
		 * value */
		throw dclicd::Error(CL_INVALID_VALUE);
	}

	{
		boost::lock_guard<boost::recursive_mutex> lock(_statusMutex);
		if (_status < 0	|| _status == CL_COMPLETE) {
			/* Event status has already been set */
			throw dclicd::Error(CL_INVALID_OPERATION);
		}
	}

	/*
	 * Broadcast user event status update
	 */
	try {
		dcl::comm::SetUserEventStatusRequest request(_id, status);
		dcl::executeCommand(_context->computeNodes(), request);
		/* FIXME Ensure that user event status has been set on all compute nodes
		 * Otherwise, the application's status has become inconsistent, i.e.,
		 * the user event's status is not the same on all compute nodes. */
		dcl::util::Logger << dcl::util::Info
				<< "User event status set (ID=" << remoteId()
				<< ", status=" << status
				<< ')' << std::endl;
	} catch (const dcl::CLError& err) {
		throw dclicd::Error(err);
	} catch (const dcl::IOException& err) {
		throw dclicd::Error(err);
	} catch (const dcl::ProtocolException& err) {
		throw dclicd::Error(err);
	}

	setCommandExecutionStatus(status);
}

void UserEvent::getProfilingInfo(
		cl_profiling_info /* param_name */,
		size_t            /* param_value_size */,
		void *            /* param_value */,
		size_t *          /* param_value_size_ret */) {
	/* Profiling is never available for user events */
	throw dclicd::Error(CL_PROFILING_INFO_NOT_AVAILABLE);
}

cl_command_type UserEvent::commandType() const {
	/* User events always are associated with a user command */
	return CL_COMMAND_USER;
}

cl_command_queue UserEvent::commandQueue() const {
	/* User events are not associated with a command queue */
	return 0;
}

} /* namespace dclicd */
