/******************************************************************************
 * This file is part of dOpenCL.
 * 
 * dOpenCL is an implementation of the OpenCL application programming
 * interface for distributed systems. See <http://dopencl.uni-muenster.de/>
 * for more information.
 * 
 * Developed by: Research Group Parallel and Distributed Systems
 *               Department of Mathematics and Computer Science
 *               University of Muenster, Germany
 *               <http://pvs.uni-muenster.de/>
 * 
 * Copyright (C) 2013  Philipp Kegel <philipp.kegel@uni-muenster.de>
 *
 * dOpenCL is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * dOpenCL is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with dOpenCL. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Permission to use dOpenCL for scientific, non-commercial work is
 * granted under the terms of the dOpenCL Academic License provided
 * appropriate credit is given. See the dOpenCL Academic License for
 * more details.
 * 
 * You should have received a copy of the dOpenCL Academic License
 * along with dOpenCL. If not, see <http://dopencl.uni-muenster.de/>.
 ******************************************************************************/

/**
 * @file    Event.h
 *
 * @date    21.06.2011
 * @author  Louay Hashim
 * @author  Philipp Kegel
 */

#ifndef EVENT_H_
#define EVENT_H_

#include <map>
#include <utility>
#include <vector>

#include <boost/shared_ptr.hpp>

#include <boost/thread/condition_variable.hpp>
#include <boost/thread/recursive_mutex.hpp>

#ifdef __APPLE__
#include <OpenCL/cl.h>
#else
#include <CL/cl.h>
#endif

#include <dcl/DataTransfer.h>
#include <dcl/DCLTypes.h>
#include <dcl/EventListener.h>
#include <dcl/Process.h>
#include <dcl/Remote.h>

#include "Memory.h"
#include "Retainable.h"

#include "detail/EventProfilingInfo.h"


class _cl_event:
public _cl_retainable {
public:
    /**
     * @brief Waits until all events in event_list are complete.
     * This method implements clWaitForEvents.
     * As by the OpenCL specification this is a blocking operation and thus
     * implicitly calls @c _cl_command_queue::flush on the events' command
     * queues.
     *
     * @param[in] event_list    the events to wait for
     */
    static void waitForEvents(const std::vector<cl_event>& event_list);


    virtual ~_cl_event();

    void retain();
    bool release();

    virtual dcl::dcl_object_id remoteId() const = 0;

    /**
     * @brief Wait for the event to be completed.
     *
     * This method is a convenience method for @c _cl_event::waitForEvents.
     */
    virtual void wait() const = 0;

    void setCallback(
            cl_int              /* command_exec_callback_type */,
            void (CL_CALLBACK * /* pfn_event_notify */)(
                    cl_event event,
                    cl_int event_command_exec_status,
                    void *user_data),
            void *              /* user_data */);

    void getInfo(
            cl_event_info   /* param_name */,
            size_t          /* param_value_size */,
            void *          /* param_value */,
            size_t *        /* param_value_size_ret */);

    /**
     * @brief Returns profiling information on this event.
     *
     * @param param_name
     * @param param_size
     * @param param_value
     * @param param_size_ret
     */
    virtual void getProfilingInfo(
            cl_profiling_info   /* param_name */,
            size_t              /* param_value_size */,
            void *              /* param_value */,
            size_t *            /* param_value_size_ret */) = 0;

protected:
    /**
     * @brief Creates a (user) event.
     *
     * @param[in]  context
     * @param[in]  status   initial event status
     */
    _cl_event(
            cl_context  /* context */,
            cl_int      /* status */);

    /**
     * @brief Query command type.
     *
     * This method is reserved for internal use by @c _cl_event::getInfo.
     *
     * @return the type of the command that is associated with this event
     */
    virtual cl_command_type commandType() const = 0;

    /**
     * @brief Query the command queue associated with this event.
     *
     * This method is reserved for internal use by @c _cl_event::getInfo.
     *
     * @return the command queue associated with this event
     */
    virtual cl_command_queue commandQueue() const = 0;

    /**
     * @brief Checks, if this event is complete
     *
     * This method is reserved for internal use.
     *
     * @return true, if this event is complete, otherwise false
     */
    bool isComplete() const;

    /**
     * @brief Waits for the event to be completed.
     *
     * This method is reserved for internal use.
     * Unlike @c wait it does *not* perform an implicit flush.
     */
    void waitNoFlush() const;

    /**
     * @brief Sets the event's command execution status.
     *
     * This event is destroyed, if its reference count is zero and its
     * associated command has finished. The caller is responsible for deleting
     * this event.
     *
     * This method is reserved for internal use.
     *
     * @param[in]  status   the command execution status this events should switch to
     * @return @c true, if this event has been destroyed, otherwise @c false
     */
    bool setCommandExecutionStatus(
            cl_int  /* status */);

    void destroy();

    cl_context _context;

    cl_int _status; /**< Event status */
    mutable boost::recursive_mutex _statusMutex; /**< Mutex for event status */
    mutable boost::condition_variable_any _statusChanged; /**< Condition: event status changed */

private:
    /**
     * @brief Triggers callbacks that have been registered for this event using @c setCallback.
     *
     * @param[in]  status   the command execution status for which the callbacks
     *                      should be triggered for
     */
    void triggerCallbacks(cl_int /* status */);

    /**
     * Saves a list of callbacks for each command execution status
     */
    std::map<cl_int, std::vector<std::pair<void (*)(cl_event, cl_int, void *), void *> > > _callbacks;
};

/******************************************************************************/

namespace dclicd {

namespace command {

/* forward declarations */
class Command;

} /* namespace command */

/******************************************************************************/

class Event:
	public _cl_event,           // extends _cl_event
	public dcl::Remote,         // extends Remote
	public dcl::EventListener { // implements EventListener
public:
    /**
     * @brief Creates an event associated with a specified command.
     *
     * On compute nodes other than the compute node where the associated command
     * has been enqueued, a substitute event is created to replace the native
     * OpenCL event which resides on the compute node where the associated
     * command has been enqueued.
     * Internally, these substitute event are implemented as user events, but
     * they hold additional information for synchronization purposes.
     *
     * @param[in]  context          the context that is associated with this event
     * @param[in]  command          the command that this event is associated with
     * @param[in]  memoryObjects    the memory objects associated with this event
     */
    Event(
            cl_context                                         context,
            const boost::shared_ptr<dclicd::command::Command>& command,
            const std::vector<cl_mem>&                         memoryObjects = std::vector<cl_mem>());
    virtual ~Event();

    dcl::dcl_object_id remoteId() const;

    /**
     * @brief Wait for the event to be completed.
     *
     * This method is a convenience method for @c _cl_event::waitForEvents.
     * Unlike @_cl_event::wait it performs an implicit flush on the event's
     * associated command queue.
     */
    void wait() const;

    /**
     * @brief Sets the event's command execution status.
     *
     * This method should only be called by the event's associated command to
     * synchronize the event's command execution status with the actual
     * execution status of its associated command
     *
     * @param[in]  status   the new command execution status
     * @return @c true, if this event has been destroyed, otherwise @c false
     *
     * @see _cl_event::setCommandExecutionStatus
     */
    bool onCommandExecutionStatusChanged(
            cl_int  /* status */);

    void getProfilingInfo(
            cl_profiling_info   /* param_name */,
            size_t              /* param_value_size */,
            void *              /* param_value */,
            size_t *            /* param_value_size_ret */);

    /**
     * @brief Synchronizes (acquires) the changes associated with this event
     *
     * Usually, this method is implicitly called when waiting for an event.
     * However, @c clFinish also is a synchronization point and, therefore,
     * requires the implementation to call this method.
     */
    void synchronize();

    /*
     * EventListener API
     *
     * This method is only required to work-around the missing node-to-node
     * communication. The host will never own an event which a compute node can
     * synchronize with. However, the host can synchronize with a compute node's
     * event.
     */

    void onSynchronize(
            dcl::Process& process);

protected:
    cl_command_type commandType() const;
    cl_command_queue commandQueue() const;

private:
    boost::shared_ptr<dclicd::command::Command> _command; /**< Command that is associated with this event */
    cl_ulong _commandQueued; /**< Queuing time of command on host */
    detail::EventProfilingInfo *_profilingInfo; /**< Profiling info (optional) */
    std::vector<cl_mem> _memoryObjects; /**< Memory objects associated with this event */
};

/******************************************************************************/

#if 0
class ReadMemoryEvent : public Event { // extends Event
public:
    /**
     * @brief Creates an event associated with a specified read memory object command.
     *
     * Unlike other events, this event is responsible for broadcasting a
     * 'command complete' message to other compute nodes, as the command is
     * finished on the host.
     * Currently, the same is performed by Event but as a work-around for
     * missing node-to-node communication.
     *
     * @param[in]  context  the context that is associated with this event
     * @param[in]  command  the read memory command that this event is associated with
     */
    ReadMemoryEvent(
            cl_context                          /* context */,
            dclicd::command::ReadMemoryCommand& /* command */);
    virtual ~Event();

    void onCommandExecutionStatusChanged(
            cl_int  /* status */);
};
#endif

/******************************************************************************/

class UserEvent: public dcl::Remote, public _cl_event {
public:
    UserEvent(
            cl_context  /* context */);
    virtual ~UserEvent();

    dcl::dcl_object_id remoteId() const;

    void wait() const;

    void setStatus(
            cl_int /* status */);

    void getProfilingInfo(
            cl_profiling_info   /* param_name */,
            size_t              /* param_value_size */,
            void *              /* param_value */,
            size_t *            /* param_value_size_ret */);

protected:
    cl_command_type commandType() const;
    cl_command_queue commandQueue() const;
};

} /* namespace dclicd */

#endif /* EVENT_H_ */
