/******************************************************************************
 * This file is part of dOpenCL.
 * 
 * dOpenCL is an implementation of the OpenCL application programming
 * interface for distributed systems. See <http://dopencl.uni-muenster.de/>
 * for more information.
 * 
 * Developed by: Research Group Parallel and Distributed Systems
 *               Department of Mathematics and Computer Science
 *               University of Muenster, Germany
 *               <http://pvs.uni-muenster.de/>
 * 
 * Copyright (C) 2013  Philipp Kegel <philipp.kegel@uni-muenster.de>
 *
 * dOpenCL is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * dOpenCL is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with dOpenCL. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Permission to use dOpenCL for scientific, non-commercial work is
 * granted under the terms of the dOpenCL Academic License provided
 * appropriate credit is given. See the dOpenCL Academic License for
 * more details.
 * 
 * You should have received a copy of the dOpenCL Academic License
 * along with dOpenCL. If not, see <http://dopencl.uni-muenster.de/>.
 ******************************************************************************/

/**
 * @file    Retainable.h
 *
 * @date    2011-05-18
 * @author  Louay Hashim
 * @author  Philipp Kegel
 */

#ifndef RETAINABLE_H_
#define RETAINABLE_H_

#include <cassert>

#ifdef __APPLE__
#include <OpenCL/cl.h>
#include <OpenCL/cl_wwu_dcl.h>
#else
#include <CL/cl.h>
#include <CL/cl_wwu_dcl.h>
#endif

class _cl_retainable {
public:
    /**
     * @brief Creates a retainable object.
     *
     * The object is retained implicitly, i.e., the reference counter is set to 1.
     */
    _cl_retainable();
    virtual ~_cl_retainable();

    /**
     * @brief Increases this object's reference count by one.
     */
    virtual void retain();

    /**
     * @brief Decreases this object's reference count by one.
     *
     * This object is destroyed, if its reference count becomes zero.
     *
     * @return @c true, if this object has been destroyed, otherwise @c false
     */
    virtual bool release();

protected:
    virtual void destroy() = 0;

    cl_uint _ref_count;
};

/******************************************************************************/

namespace dclicd {

/**
 * @brief Releases and possibly deletes a retainable object.
 *
 * The object must have been created using operator new.
 *
 * @param[in]  object   the object to release. object must not be NULL.
 */
template<typename T>
void release(T *object) {
    assert(object != 0); // caller should return appropriate CL_INVALID_... error code
    /* FIXME Ensure that object has been created with new */
    if (object->release()) delete object;
}

/*
 * Compute nodes are deleted in _cl_platform::destroyComputeNode, such that
 * there must not be a redundant delete in dclicd::release.
 */
template<>
void release(cl_compute_node_WWU compute_node);

/******************************************************************************/

namespace detail {

/* This class is performs implicit reference counting to alleviate handling of
 * shared objects */
#if 0
template<typename T>
class Retainable {
    Retainable() {
    }

    Retainable(T *object) : _object(object) {
        /* FIXME Never(!) call a virtual function within a constructor */
//		validate();
        retain();
    }

    Retainable(const Retainable& rhs) : _object(rhs._object) {
        retain();
    }

    virtual ~Retainable() {
        if (_object == 0) return;
        release();
    }

    Retainable& operator=(const Retainable& rhs) {
        if (_object != rhs._object) {
            release();
            _object = rhs._object;
            retain();
        }

        return *this;
    }

    operator T *() {
        retain();
        return _object;
    }

    void retain() {
        if (_object == 0) return;
        ++_object->ref_count;
    }

    void release() {
        if (_object == 0) return;
        assert(_object->ref_count > 0);
        if (--_object->ref_count == 0) {
            destroy();
            delete _object;
        }
    }

protected:
    virtual void destroy() = 0;

    T *_object;
};
#endif

} /* namespace detail */

} /* namespace dclicd */

#endif /* RETAINABLE_H_ */
