/******************************************************************************
 * This file is part of dOpenCL.
 * 
 * dOpenCL is an implementation of the OpenCL application programming
 * interface for distributed systems. See <http://dopencl.uni-muenster.de/>
 * for more information.
 * 
 * Developed by: Research Group Parallel and Distributed Systems
 *               Department of Mathematics and Computer Science
 *               University of Muenster, Germany
 *               <http://pvs.uni-muenster.de/>
 * 
 * Copyright (C) 2013  Philipp Kegel <philipp.kegel@uni-muenster.de>
 *
 * dOpenCL is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * dOpenCL is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with dOpenCL. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Permission to use dOpenCL for scientific, non-commercial work is
 * granted under the terms of the dOpenCL Academic License provided
 * appropriate credit is given. See the dOpenCL Academic License for
 * more details.
 * 
 * You should have received a copy of the dOpenCL Academic License
 * along with dOpenCL. If not, see <http://dopencl.uni-muenster.de/>.
 ******************************************************************************/

/*
 * Command.h
 *
 *  Created on: 24.03.2012
 *      Author: Philipp Kegel
 */

#ifndef COMMAND_H_
#define COMMAND_H_

#include <boost/thread/condition_variable.hpp>
#include <boost/thread/recursive_mutex.hpp>

#ifdef __APPLE__
#include <OpenCL/cl.h>
#else
#include <CL/cl.h>
#endif

#include <dcl/CommandListener.h>
#include <dcl/Remote.h>

namespace dclicd {

/* forward declarations */
class Event;

/******************************************************************************/

namespace command {

class Command :
public dcl::Remote,           // extends Remote
public dcl::CommandListener { // implements CommandListener
public:
    /**
     * @brief Creates a command
     *
     * @param[in]  command_type     the type of this command
     * @param[in]  command_queue    the command queue that is associated with this command
     */
    Command(
            cl_command_type     /* type */,
            cl_command_queue    /* commandQueue */);
    virtual ~Command();

    cl_command_type type() const;
    cl_command_queue commandQueue() const;

    /**
     * @brief Attaches an event to this command.
     *
     * This operation must be called only once, as the event shall not be
     * detached or changed.
     *
     * @param[in]  event    the event to attach to this command
     */
    void setEvent(
            dclicd::Event&  /* event */);

    /**
     * @brief Checks, if this command is complete
     *
     * @return true, if this command is complete, otherwise false
     */
    bool isComplete() const;

    /**
     * @brief Wait for the command to be completed.
     */
    void wait() const;

    /*
     * Command listener API
     */

    void onExecutionStatusChanged(
            cl_int  /* executionStatus */);

protected:
    /**
     * @brief Executes this command when its execution status changes to CL_SUBMITTED.
     *
     * @return a new execution status of the command
     * The returned execution status depends on whether the host can complete
     * the command directly (CL_COMPLETE is returned), or the command will be
     * completed later (by the host or compute node, CL_RUNNING will be
     * returned).
     */
    virtual cl_int execute();

    cl_command_type _type; /**< Command type */
    cl_command_queue _commandQueue; /**< Command queue that the command has been enqueued to */
    cl_int _executionStatus; /**< Execution status */
    mutable boost::recursive_mutex _executionStatusMutex; /**< Mutex for command execution status */
    mutable boost::condition_variable_any _executionStatusChanged; /**< Condition: command execution status changed */

    dclicd::Event *_event; /**< Associated event; can be NULL */
};

} /* namespace command */

} /* namespace dclicd */

#endif /* COMMAND_H_ */
