/******************************************************************************
 * This file is part of dOpenCL.
 * 
 * dOpenCL is an implementation of the OpenCL application programming
 * interface for distributed systems. See <http://dopencl.uni-muenster.de/>
 * for more information.
 * 
 * Developed by: Research Group Parallel and Distributed Systems
 *               Department of Mathematics and Computer Science
 *               University of Muenster, Germany
 *               <http://pvs.uni-muenster.de/>
 * 
 * Copyright (C) 2013  Philipp Kegel <philipp.kegel@uni-muenster.de>
 *
 * dOpenCL is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * dOpenCL is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with dOpenCL. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Permission to use dOpenCL for scientific, non-commercial work is
 * granted under the terms of the dOpenCL Academic License provided
 * appropriate credit is given. See the dOpenCL Academic License for
 * more details.
 * 
 * You should have received a copy of the dOpenCL Academic License
 * along with dOpenCL. If not, see <http://dopencl.uni-muenster.de/>.
 ******************************************************************************/

/*
 * @file    MappingCommand.cpp
 *
 * @date    2012-03-24
 * @author  Philipp Kegel
 */

#include <cassert>
#include <cstddef>

#include <boost/bind.hpp>
#include <boost/shared_ptr.hpp>

#ifdef __APPLE__
#include <OpenCL/cl.h>
#else
#include <CL/cl.h>
#endif

#include <dcl/ComputeNode.h>
#include <dcl/DataTransfer.h>
#include <dcl/DCLTypes.h>

#include "../CommandQueue.h"
#include "../Memory.h"
#include "../Retainable.h"

#include "../detail/MappedMemory.h"

#include "Command.h"
#include "MappingCommand.h"

namespace dclicd {

namespace command {

MapBufferCommand::MapBufferCommand(
        cl_command_queue commandQueue,
        Buffer *buffer,
        void *ptr) :
    Command(CL_COMMAND_MAP_BUFFER, commandQueue), _buffer(buffer), _ptr(ptr) {
    assert(_buffer != 0); // buffer must not be NULL
    _buffer->retain();
}

MapBufferCommand::~MapBufferCommand() {
    dclicd::release(_buffer);
}

cl_int MapBufferCommand::execute() {
    const detail::MappedBufferRegion *mapping;

    /* Obtain mapping data */
    mapping = _buffer->findMapping(_ptr);
    assert(mapping != 0); // _ptr must point to a mapped region of _buffer

    if ((mapping->flags() & CL_MAP_READ)) {
        /*
         * The mapped memory region has to be synchronized, i.e., it has to be
         * downloaded to the mapped pointer.
         */
        /* Start data receipt */
        boost::shared_ptr<dcl::DataTransfer> receipt(
                _commandQueue->computeNode().receiveData(mapping->cb(), _ptr));
        /* Register callback to complete MapBufferCommand */
        receipt->setCallback(boost::bind(
                &MapBufferCommand::onExecutionStatusChanged, this, _1));

        return CL_RUNNING;
    } else {
        return CL_COMPLETE;
    }
}

/******************************************************************************/

UnmapMemoryCommand::UnmapMemoryCommand(
        cl_command_queue commandQueue,
        cl_mem memobj,
        void *ptr) :
    Command(CL_COMMAND_UNMAP_MEM_OBJECT, commandQueue), _memobj(memobj),
    _ptr(ptr) {
    assert(_memobj != 0); // memory object must not be NULL
    _memobj->retain();
}

UnmapMemoryCommand::~UnmapMemoryCommand() {
    dclicd::release(_memobj);
}

void UnmapMemoryCommand::unmap(cl_int errcode) {
    switch (errcode) {
    case CL_SUCCESS:
        _memobj->unmap(_ptr); // release pointer to mapped memory
        break;
    default:
        ;
        /* no break */
    }
}

cl_int UnmapMemoryCommand::execute() {
//    const detail::MappedMemory *mapping;
    const detail::MappedBufferRegion *mapping;

    /* Obtain mapping data */
    /* TODO Generalize dclicd::command::UnmapMemoryCommand for memory objects */
//    mapping = _memobj.findMapping(_ptr);
    Buffer *buffer = dynamic_cast<Buffer *>(_memobj);
    mapping = buffer->findMapping(_ptr);
    assert(mapping != 0); // _ptr must point to a mapped region of _memobj

    if (mapping->flags() & CL_MAP_WRITE) {
        /*
         * The mapped memory region has to be synchronized, i.e., its data has
         * to be uploaded to the command queue's compute node.
         */
        /* Start data sending */
        boost::shared_ptr<dcl::DataTransfer> sending(
                _commandQueue->computeNode().sendData(mapping->cb(), _ptr));
        /* Register callback to release pointer to mapped memory.
         * The status of the pointer is undefined if the unmap operations fails. */
        sending->setCallback(boost::bind(&UnmapMemoryCommand::unmap, this, _1));

        /* UnmapMemoryCommand will be completed by compute node */

        return CL_RUNNING;
    } else {
        _memobj->unmap(_ptr); // release pointer to mapped memory

        return CL_COMPLETE;
    }
}

} /* namespace command */

} /* namespace dclicd */
