/******************************************************************************
 * This file is part of dOpenCL.
 * 
 * dOpenCL is an implementation of the OpenCL application programming
 * interface for distributed systems. See <http://dopencl.uni-muenster.de/>
 * for more information.
 * 
 * Developed by: Research Group Parallel and Distributed Systems
 *               Department of Mathematics and Computer Science
 *               University of Muenster, Germany
 *               <http://pvs.uni-muenster.de/>
 * 
 * Copyright (C) 2013  Philipp Kegel <philipp.kegel@uni-muenster.de>
 *
 * dOpenCL is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * dOpenCL is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with dOpenCL. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Permission to use dOpenCL for scientific, non-commercial work is
 * granted under the terms of the dOpenCL Academic License provided
 * appropriate credit is given. See the dOpenCL Academic License for
 * more details.
 * 
 * You should have received a copy of the dOpenCL Academic License
 * along with dOpenCL. If not, see <http://dopencl.uni-muenster.de/>.
 ******************************************************************************/

/*!
 * \file ProgramBuild.cpp
 *
 * \date 2012-08-04
 * \author Philipp Kegel
 */

#include <cassert>
#include <map>
#include <ostream>
#include <string>
#include <utility>
#include <vector>

#include <boost/thread/condition_variable.hpp>
#include <boost/thread/locks.hpp>
#include <boost/thread/mutex.hpp>

#ifdef __APPLE__
#include <OpenCL/cl.h>
#include <OpenCL/cl_wwu_dcl.h>
#else
#include <CL/cl.h>
#include <CL/cl_wwu_dcl.h>
#endif

#include <dcl/CLError.h>
#include <dcl/CLObjectRegistry.h>
#include <dcl/ComputeNode.h>
#include <dcl/DCLTypes.h>
#include <dcl/DCLException.h>
#include <dcl/ProgramBuildListener.h>
#include <dcl/Remote.h>

#include <dcl/comm/BuildProgram.h>

#include <dcl/util/Logger.h>

#include "../Context.h"
#include "../Device.h"
#include "../Error.h"
#include "../Platform.h"
#include "../Program.h"

#include "ProgramBuild.h"

namespace dclicd {

namespace detail {

ProgramBuild::ProgramBuild(
        cl_program program,
        const std::vector<cl_device_id>& devices,
        const std::string& options,
        void (*pfn_notify)(cl_program, void *),
        void *user_data) :
    dcl::Remote(dcl::Remote::generateId()),
    _program(program), _devices(devices), _options(options),
    _pfnNotify(pfn_notify), _userData(user_data), _buildStatus(CL_BUILD_NONE)
{
    assert(program != 0); // constructor is only called internally
    assert(!devices.empty()); // program must be build for at least one device
    if (pfn_notify == 0 && user_data != 0) {
        throw dclicd::Error(CL_INVALID_VALUE);
    }

    /* register build *before* sending request;
     * otherwise a build completion message might be lost */
    _program->context()->getPlatform()->remote().objectRegistry().bind<dcl::ProgramBuildListener>(_id, *this);

    submit();

    /* TODO Trigger callback asynchronously */
    if (_buildStatus == CL_BUILD_SUCCESS && _pfnNotify) {
        _pfnNotify(_program, _userData);
    }
}

ProgramBuild::~ProgramBuild() {
    /* Deregister program build */
    _program->context()->getPlatform()->remote().objectRegistry().unbind<dcl::ProgramBuildListener>(_id);
}

bool ProgramBuild::isComplete() {
    boost::lock_guard<boost::mutex> lock(_buildStatusMutex);
    return testComplete();
}

bool ProgramBuild::hasFailed() {
    boost::lock_guard<boost::mutex> lock(_buildStatusMutex);
    return _buildStatus == CL_BUILD_PROGRAM_FAILURE;
}

void ProgramBuild::wait() {
    boost::lock_guard<boost::mutex> lock(_buildStatusMutex);
    while (!testComplete()) {
        _buildCompleted.wait(_buildStatusMutex);
    }
}

void ProgramBuild::onComplete(
        const std::vector<dcl::dcl_object_id>& deviceIds,
        const std::vector<cl_build_status>& buildStatus) {
    assert(deviceIds.size() == buildStatus.size());

    /* TODO Update build status and save build info (status and options) for
     * devices of completed (!) build operation
    for (unsigned int i = 0; i < deviceIds->size(); ++i) {
        _program->onBuildCompleted(deviceIds[i], buildStatus[i], _options);
    }
     */

    {
        boost::lock_guard<boost::mutex> lock(_buildStatusMutex);
        if (testComplete()) {

            /* Trigger callback if all compute nodes reported completion of program
             * build */
            if (_pfnNotify) {
                _pfnNotify(_program, _userData);
            }
        }
        _buildCompleted.notify_all();
    }
}

void ProgramBuild::submit() {
    std::map<dcl::ComputeNode *, std::vector<dcl::dcl_object_id> > nodeDeviceIds;

    /*
     * Create compute nodes' lists of device IDs
     */
    for (std::vector<cl_device_id>::const_iterator i = _devices.begin();
            i < _devices.end(); ++i) {
        cl_device_id device = *i;
        assert(device != 0); // devices must not be NULL
        nodeDeviceIds[&device->remote().getComputeNode()].push_back(device->remote().getId());
    }

    {
        boost::lock_guard<boost::mutex> lock(_buildStatusMutex);
        std::vector<std::pair<dcl::ComputeNode *, dcl::comm::BuildProgram> > requests;
        cl_int err = CL_SUCCESS; // assume successful command submission

        /*
         * Create and send requests
         */
        for (std::map<dcl::ComputeNode *, std::vector<dcl::dcl_object_id> >::const_iterator
                i = nodeDeviceIds.begin(); i != nodeDeviceIds.end(); ++i) {
            dcl::comm::BuildProgram request(_program->remoteId(), i->second, _options, _id);
            dcl::ComputeNode *computeNode = i->first;

            /* Send requests to *all* compute nodes, i.e. do not stop on failure */
            try {
                computeNode->sendRequest(request);
                requests.push_back(std::make_pair(computeNode, request)); // save pending request
            } catch (const dcl::IOException&) {
                err = CL_IO_ERROR_WWU;
            } catch (const dcl::ProtocolException&) {
                err = CL_PROTOCOL_ERROR_WWU;
            }

            if (err != CL_SUCCESS) {
                /* Do not attempt to build program on remaining compute nodes,
                 * but finish build process on other compute nodes. */
                break;
            }
        }

        /*
         * Await responses from all compute nodes
         * Also await responses if request failed on some compute nodes.
         */
        for (std::vector<std::pair<dcl::ComputeNode *, dcl::comm::BuildProgram> >::const_iterator
                i = requests.begin(); i != requests.end(); ++i) {
            dcl::ComputeNode *computeNode = i->first;
            cl_int nodeErr = CL_SUCCESS; // assume successful response from compute node

            /* Receive responses from *all* compute nodes, i.e. do not stop on failure */
            try {
                computeNode->awaitResponse(i->second);
                _computeNodes.push_back(computeNode); // compute node now executes program build
            } catch (const dcl::CLError& err) {
                nodeErr = err.err();
            } catch (const dcl::IOException&) {
                nodeErr = CL_IO_ERROR_WWU;
            } catch (const dcl::ProtocolException&) {
                nodeErr = CL_PROTOCOL_ERROR_WWU;
            }

            if (nodeErr != CL_SUCCESS) {
                if (err == CL_SUCCESS) { // return error from first failure
                    err = nodeErr;
                }
            }
        }

        if (err == CL_SUCCESS) { // program build submitted to all compute nodes
            /* TODO Update build info (status and options) for devices on compute node
            for (std::vector<cl_device_id>::const_iterator i = devices->begin();
                    i != devices->end(); ++i) {
                _program->onBuildStatusChanged(*i, CL_BUILD_IN_PROGRESS, _options));
            }
             */
            /* TODO Implement asynchronous program build */
//            _buildStatus = CL_BUILD_IN_PROGRESS;
            _buildStatus = CL_BUILD_SUCCESS;

            dcl::util::Logger << dcl::util::Info
                    << "Program build submitted (program ID=" << _program->remoteId()
                    << ", build ID=" << _id
                    << ')' << std::endl;
        } else {
            _buildStatus = CL_BUILD_PROGRAM_FAILURE;
        }
    }
}

bool ProgramBuild::testComplete() {
    /* TODO Implement asynchronous program build.
     * Currently a program is always build synchronously, such that a program
     * build always is complete. */
    return (_buildStatus == CL_BUILD_SUCCESS ||
            _buildStatus == CL_BUILD_PROGRAM_FAILURE);
}

} /* namespace detail */

} /* namespace dclicd */
