/******************************************************************************
 * This file is part of dOpenCL.
 * 
 * dOpenCL is an implementation of the OpenCL application programming
 * interface for distributed systems. See <http://dopencl.uni-muenster.de/>
 * for more information.
 * 
 * Developed by: Research Group Parallel and Distributed Systems
 *               Department of Mathematics and Computer Science
 *               University of Muenster, Germany
 *               <http://pvs.uni-muenster.de/>
 * 
 * Copyright (C) 2013  Philipp Kegel <philipp.kegel@uni-muenster.de>
 *
 * dOpenCL is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * dOpenCL is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with dOpenCL. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Permission to use dOpenCL for scientific, non-commercial work is
 * granted under the terms of the dOpenCL Academic License provided
 * appropriate credit is given. See the dOpenCL Academic License for
 * more details.
 * 
 * You should have received a copy of the dOpenCL Academic License
 * along with dOpenCL. If not, see <http://dopencl.uni-muenster.de/>.
 ******************************************************************************/

/**
 * @file Buffer.cpp
 *
 * @date 2013-04-07
 * @author Philipp Kegel
 */

#include <cstddef>
#include <cstring>
#include <iostream>
#include <vector>

#ifdef __APPLE__
#include <OpenCL/cl.h>
#else
#include <CL/cl.h>
#endif

#define BOOST_TEST_DYN_LINK
#define BOOST_TEST_MODULE Memory
#include <boost/test/unit_test.hpp>

/**
 * @brief Returns the first platform available
 *
 * @return a platform
 */
cl_platform_id getPlatform() {
    cl_platform_id platform;
    cl_uint num_platforms;
    size_t param_value_size;
    cl_int err;

    err = clGetPlatformIDs(1, &platform, &num_platforms);
    BOOST_REQUIRE_EQUAL(err, CL_SUCCESS);
    BOOST_REQUIRE_MESSAGE(num_platforms >= 1, "No platform found");

    /* platform should be 'dOpenCL' */
    err = clGetPlatformInfo(platform, CL_PLATFORM_NAME, 0, NULL,
            &param_value_size);
    BOOST_REQUIRE_EQUAL(err, CL_SUCCESS);
    char platform_name[param_value_size];
    err = clGetPlatformInfo(platform, CL_PLATFORM_NAME, param_value_size,
            platform_name, NULL);
    BOOST_REQUIRE_EQUAL(err, CL_SUCCESS);
    BOOST_CHECK_MESSAGE(strcmp(platform_name, "dOpenCL") == 0, "First platform must be 'dOpenCL'");

    return platform;
}

/**
 * @brief Returns the platform's first device.
 *
 * @param[in]  platform the platform
 * @return the platform's first device
 */
cl_device_id getDevice(cl_platform_id platform) {
    cl_device_id device = NULL;
    cl_uint num_devices;
    cl_int err;

    err = clGetDeviceIDs(platform, CL_DEVICE_TYPE_ALL, 1, &device,
            &num_devices);
    BOOST_REQUIRE_EQUAL(err, CL_SUCCESS);
    BOOST_REQUIRE_GE(num_devices, 1);

    return device;
}

cl_context createContext(cl_uint num_devices, cl_device_id *devices) {
    cl_context context;
    cl_int err;

    BOOST_REQUIRE_GE(num_devices, 1);
    BOOST_REQUIRE(devices != NULL);

    context = clCreateContext(0, num_devices, devices, 0, 0, &err);
    BOOST_REQUIRE_EQUAL(err, CL_SUCCESS);

    return context;
}

cl_command_queue createCommandQueue(cl_context context, cl_device_id device) {
    cl_command_queue commandQueue;
    cl_int err;

    commandQueue = clCreateCommandQueue(context, device, 0, &err);
    BOOST_REQUIRE_EQUAL(err, CL_SUCCESS);

    return commandQueue;
}

/******************************************************************************
 * Shared test suite fixture
 ******************************************************************************/

struct Context {
    cl_context context;
    cl_command_queue commandQueue;

    Context() {
        cl_platform_id platform = getPlatform();
        cl_device_id device = getDevice(platform);

        context = createContext(1, &device);
        commandQueue = createCommandQueue(context, device);

        BOOST_TEST_MESSAGE("Set up fixture");
    }

    ~Context() {
        cl_int err;

        /* clean up */
        err = clReleaseCommandQueue(commandQueue);
        BOOST_CHECK_EQUAL(err, CL_SUCCESS);
        err = clReleaseContext(context);
        BOOST_CHECK_EQUAL(err, CL_SUCCESS);

        BOOST_TEST_MESSAGE("Teared down fixture");
    }
};

BOOST_FIXTURE_TEST_SUITE( Buffer, Context )

/******************************************************************************
 * Test cases
 ******************************************************************************/

template<typename T>
void fillVector(std::vector<T>& vec, T first, T step) {
    T value = first;
    for (typename std::vector<T>::iterator i = vec.begin(); i != vec.end(); ++i) {
        *i = value;
        value += step;
    }
}

BOOST_AUTO_TEST_CASE( CreateBuffer )
{
    size_t size = 1024;
    cl_mem buffer;
    cl_int err;

    /* create buffer */
    buffer = clCreateBuffer(context, CL_MEM_READ_WRITE, size, NULL, &err);
    BOOST_REQUIRE_EQUAL(err, CL_SUCCESS);

    /* clean up */
    err = clReleaseMemObject(buffer);
    BOOST_CHECK_EQUAL(err, CL_SUCCESS);
}

BOOST_AUTO_TEST_CASE( CreateBufferCopyHostPtr )
{
    size_t vecSize = 1024 * 1024;
    cl_mem buffer;
    std::vector<cl_int> vec1(vecSize, 0), vec2(vecSize, 1);
    cl_int err;

    fillVector(vec1, 1, 1); // initialize input data

    /* create buffer from first host pointer */
    buffer = clCreateBuffer(context, CL_MEM_READ_WRITE | CL_MEM_COPY_HOST_PTR, vecSize * sizeof(cl_int), &vec1.front(), &err);
    BOOST_REQUIRE_EQUAL(err, CL_SUCCESS);

    /* download buffer to second host pointer */
    err = clEnqueueReadBuffer(
            commandQueue,
            buffer,
            CL_TRUE,
            0, vecSize * sizeof(cl_int), &vec2.front(),
            0, NULL, NULL);
    BOOST_REQUIRE_EQUAL(err, CL_SUCCESS);

    BOOST_CHECK_MESSAGE(vec1 == vec2, "Input and output buffers differ"); // compare input and output data

    /* clean up */
    err = clReleaseMemObject(buffer);
    BOOST_CHECK_EQUAL(err, CL_SUCCESS);
}

BOOST_AUTO_TEST_SUITE_END()
